/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

require([
    "app/common",
    "app/messages",
    "app/tags",
    "app/wizardData",
    "aps/load",
    "aps/xhr",
    "aps/Button",
    "aps/Gauge",
    "aps/ResourceStore",
    "dojo/when",
    "dojo/promise/all",
    "dijit/registry"
], function (
    common,
    messages,
    tags,
    wizardData,
    load,
    xhr,
    Button,
    Gauge,
    Store,
    when,
    all,
    registry
) {
    var globalId = aps.context.vars.globals.aps.id,
        tenantId = aps.context.vars.tenant.aps.id,
        accountStore = new Store({
            apsType: common.TYPES.USERS,
            target: "/aps/2/resources"
        }),
        subscriptionAccountStore = new Store({
            target: "/aps/2/resources/" + tenantId + "/users",
            idProperty: "aps.id"
        }),
        subscriptionStore = new Store({
            target: "/aps/2/resources/" + tenantId + "/subscription",
            idProperty: "aps.id"
        }),
        getStatisticsDeffer = xhr.get("/aps/2/resources/" + tenantId + "/updateStatistics").then(
            function (statistics) {
                return statistics;
            }, messages.displayError),
        deleteInProgress = false,
        userViewViewId = common.getRelatedViewId('view'),
        userNewSelectViewId = common.getRelatedViewId('new-select'),
        userNewConfigureViewId = common.getRelatedViewId('new-configure'),
        userActivateViewId = common.getRelatedViewId('activate');

    function output(html) {
        return ["aps/Output", {id: common.genId('output'), value: html, escapeHTML: false}];
    }

    function image(src) {
        return "<img src=" + src + ">";
    }

    function span(text, cssClass, title) {
        return "<span" + (title ? " title='" + title + "'" : "") + (cssClass ? " class=" + cssClass : "") + ">" +
            text + "</span>";
    }

    function formatStorageUsage(usageObject) {
        var result = (usageObject.usage == null) ? '0' : common.formatStorageSize(usageObject.usage);
        if (usageObject.limit != null && usageObject.limit > 0) {
            result = result + ' / ' + common.formatStorageSize(usageObject.limit);
        }
        return result;
    }

    function formatUnitUsage(usageObject) {
        var result = (usageObject.usage == null) ? '0' : usageObject.usage;
        if (usageObject.limit != null && usageObject.limit > 0) {
            result = result + ' / ' + usageObject.limit;
        }
        return result;
    }

    function prepareUsageStats(value, limit) {
        value = Number(value);
        limit = Number(limit);
        var usageStats = (value > 0) ? value : 0;
        if (limit >= 0) {
            return usageStats + ' / ' + limit;
        }
        return usageStats.toString();
    }

    function checkButtonState() {
        var selectedRecordsCount = registry.byId("account-grid").get("selectionArray").length;
        registry.byId("newuser").set('disabled', deleteInProgress || selectedRecordsCount > 0);
        registry.byId("deleteuser").set('disabled', deleteInProgress || selectedRecordsCount === 0);
    }

    function onNewAccountRequestFail(err) {
        messages.displayError(err);
        registry.byId("newuser").cancel();
    }

    xhr.get("/aps/2/resources/" + globalId + "/settings").then(function (settings) {
        var userStore = new Store({
            apsType: !settings.ccpv1_show_staff_members ?
                "http://aps-standard.org/types/core/service-user/1.0" :
                "http://parallels.com/aps/types/pa/user/1.0",
            target: "/aps/2/resources"
        });

        all({
            statistics: getStatisticsDeffer,
            users: userStore.query()
        }).then(function (response) {
            var statistics = response.statistics,
                users = response.users,
                needRefresh = false, // if any account is in-progress we need to refresh the grid
                createBackupAccount = function () {
                    when(subscriptionStore.query(), function (subscriptions) {
                        var subscription = subscriptions[0],
                            resourcesStore = new Store({
                                target: "/aps/2/resources/" + subscription.aps.id + "/resources",
                                idProperty: "aps.id"
                            });

                        when(resourcesStore.query(), function (resources) {
                            for (var i = 0; i < resources.length; i++) {
                                if (resources[i].apsType == common.TYPES.USERS) {
                                    if (resources[i].usage >= resources[i].limit) {
                                        messages.displayWarning(_("Limit for backup users has been reached."));
                                        registry.byId("newuser").cancel();
                                        return;
                                    }
                                }
                            }

                            when(accountStore.query(), function (accounts) {
                                for (var i = 0; i < users.length; i++) {
                                    var found = false;
                                    for (var j = 0; j < accounts.length; j++) {
                                        if (accounts[j].serviceUserId == users[i].aps.id) {
                                            found = true;
                                        }
                                    }
                                }

                                wizardData.resetWizardData(userNewSelectViewId);
                                wizardData.resetWizardData(userNewConfigureViewId);
                                aps.apsc.gotoView(userNewSelectViewId);
                            }, onNewAccountRequestFail);
                        }, onNewAccountRequestFail);
                    }, onNewAccountRequestFail);
                },
                deleteBackupAccount = function () {
                    if (!confirm(_("Are you sure that you want to unassign this service from the user(s)? All user backups will be deleted."))) {
                        this.cancel();
                        return;
                    }

                    var accountGrid = registry.byId("account-grid"),
                        selectionArray = accountGrid.get("selectionArray"),
                        deffers = [],
                        deletedIds = [],
                        deleteBtn = this;

                    messages.removeAll();

                    selectionArray.forEach(function (id) {
                        var deffer = subscriptionAccountStore.remove(id).then(function () {
                            deletedIds.push(id);
                        }, messages.displayErrorAndGo);
                        deffers.push(deffer);
                    });

                    deleteInProgress = true;
                    checkButtonState();

                    all(deffers).then(function () {
                        accountGrid.refresh();

                        deletedIds.forEach(function (id) {
                            selectionArray.splice(selectionArray.indexOf(id), 1);
                        });

                        deleteInProgress = false;
                        deleteBtn.cancel();
                        checkButtonState();
                    });
                },
                totalPanel = ['aps/Container', {'id': common.genId('container')}, [
                    ['aps/Container', {'id': common.genId('container')}, [
                        output(tags.div([
                            statistics.tenant.StorageUsage && statistics.tenant.StorageUsage.limit ? span(
                                tags.icon('i-cloud-storage-o--16 acronis-gray-icon', 16) +
                                span(formatStorageUsage(statistics.tenant.StorageUsage)),
                                'acronis-total-usage',
                                _('Cloud Storage usage')
                            ) : '',

                            statistics.tenant.LocalStorageUsage && statistics.tenant.LocalStorageUsage.limit ?
                                span(
                                    tags.icon('i-local-backup-o--16 acronis-gray-icon', 16) +
                                    span(formatStorageUsage(statistics.tenant.LocalStorageUsage)),
                                    'acronis-total-usage',
                                    _('Local Storage usage')
                                ) : ''
                            ,

                            statistics.tenant.WorkstationCount && statistics.tenant.WorkstationCount.limit ?
                                span(
                                    tags.icon('i-workstations-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.WorkstationCount)),
                                    'acronis-total-usage',
                                    _('Number of workstations')
                                ) : ''
                            ,

                            statistics.tenant.ServerCount && statistics.tenant.ServerCount.limit ?
                                span(
                                    tags.icon('i-server-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.ServerCount)),
                                    'acronis-total-usage',
                                    _('Number of servers')
                                ) : ''
                            ,

                            statistics.tenant.VmCount && statistics.tenant.VmCount.limit ?
                                span(
                                    tags.icon('i-virtual-machines-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.VmCount)),
                                    'acronis-total-usage',
                                    _('Number of virtual machines')
                                ) : ''
                            ,

                            statistics.tenant.MobileDeviceCount && statistics.tenant.MobileDeviceCount.limit ?
                                span(
                                    tags.icon('i-mobile-devices-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.MobileDeviceCount)),
                                    'acronis-total-usage',
                                    _('Number of mobile devices')
                                ) : ''
                            ,

                            statistics.tenant.WebsiteCount && statistics.tenant.WebsiteCount.limit ?
                                span(
                                    tags.icon('i-website-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.WebsiteCount)),
                                    'acronis-total-usage',
                                    _('Number of websites')
                                ) : ''
                            ,

                            statistics.tenant.MailboxCount && statistics.tenant.MailboxCount.limit ?
                                span(
                                    tags.icon('i-mailboxes-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.MailboxCount)),
                                    'acronis-total-usage',
                                    _('Number of office 365 seats')
                                ) : ''
                            ,

                            statistics.tenant.O365Mailboxes && statistics.tenant.O365Mailboxes.limit ?
                                span(
                                    tags.icon('i-mail-o--24 acronis-gray-icon', 16) +
                                    span(_('Enabled')),
                                    'acronis-total-usage',
                                    _('Office 365 Mailboxes')
                                ) : ''
                            ,

                            statistics.tenant.O365OneDrive && statistics.tenant.O365OneDrive.limit ?
                                span(
                                    tags.icon('i-one-drive--16 acronis-gray-icon', 16) +
                                    span(_('Enabled')),
                                    'acronis-total-usage',
                                    _('Office 365 OneDrive')
                                ) : ''
                            ,

                            statistics.tenant.O365SharePointSitesCount && statistics.tenant.O365SharePointSitesCount.limit ?
                                span(
                                    tags.icon('i-o-365-sharepoint-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.O365SharePointSitesCount)),
                                    'acronis-total-usage',
                                    _('Number of office 365 SharePoint sites')
                                ) : ''
                            ,

                            statistics.tenant.WebHostingServerCount && statistics.tenant.WebHostingServerCount.limit ?
                                span(
                                    tags.icon('i-web-hosting-server-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.WebHostingServerCount)),
                                    'acronis-total-usage',
                                    _('Number of web hosting servers')
                                ) : ''
                            ,

                            statistics.tenant.DrStorageUsage && statistics.tenant.DrStorageUsage.limit ?
                                span(
                                    tags.icon('i-dr-storage-o--16 acronis-gray-icon', 16) +
                                    span(formatStorageUsage(statistics.tenant.DrStorageUsage)),
                                    'acronis-total-usage',
                                    _('Disaster recovery storage usage')
                                ) : ''
                            ,

                            statistics.tenant.ComputePointsCount && statistics.tenant.ComputePointsCount.limit ?
                                span(
                                    tags.icon('i-cp-used-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.ComputePointsCount)),
                                    'acronis-total-usage',
                                    _('Number of compute points')
                                ) : ''
                            ,

                            statistics.tenant.PublicIPsCount && statistics.tenant.PublicIPsCount.limit ?
                                span(
                                    tags.icon('i-public-ips-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.PublicIPsCount)),
                                    'acronis-total-usage',
                                    _('Number of public IP addresses')
                                ) : ''
                            ,

                            statistics.tenant.CloudServersCount && statistics.tenant.CloudServersCount.limit ?
                                span(
                                    tags.icon('i-cloud-server-o--16 acronis-gray-icon', 16) +
                                    span(formatUnitUsage(statistics.tenant.CloudServersCount)),
                                    'acronis-total-usage',
                                    _('Number of cloud servers')
                                ) : ''
                            ,

                            statistics.tenant.InternetAccess && statistics.tenant.InternetAccess.limit ?
                                span(
                                    tags.icon('i-internet-access-o--16 acronis-gray-icon', 16) +
                                    span(_('Enabled')),
                                    'acronis-total-usage',
                                    _('Internet access')
                                ) : ''

                        ].join(''), {'class': 'acronis-total-wrapper'}))
                    ]]
                ]],
                toolbar = ["aps/Toolbar", [
                    ["aps/ToolbarButton", {
                        id: "newuser",
                        label: _("Add New Backup User"),
                        iconClass: "sb-add-new",
                        onClick: createBackupAccount
                    }],
                    ["aps/ToolbarButton", {
                        id: "deleteuser",
                        label: _("Delete"),
                        iconClass: "sb-delete",
                        onClick: deleteBackupAccount
                    }]
                ]],
                gridColumns = [
                    {
                        name: image("images/user_statuses/ico_state_on_16px.png"),
                        field: "userstatus",
                        escapeHTML: false,
                        renderCell: function (row, data) {
                            switch (data) {
                                case "1":
                                    return image("images/user_statuses/ico_state_on_16px.png");
                                case "2":
                                    needRefresh = true;
                                    return image("images/user_statuses/wheel-progress.gif");
                                default:
                                    return image("images/user_statuses/ico_state_off_16px.png");
                            }
                        }
                    },
                    {
                        name: _("Name"),
                        field: "displayName",
                        renderCell: function (row) {
                            for (var i = 0; i < users.length; i++) {
                                if (row.serviceUserId == users[i].aps.id) {
                                    return common.escapeAngularBrackets(users[i].displayName);
                                }
                            }
                            return "";
                        }
                    },
                    {name: _("Login"), field: "endUserLogin", type: "resourceName"},
                    {
                        name: span(tags.icon('i-cloud-storage-o--16 acronis-gray-icon', 16) + _("Cloud Storage usage"), 'aconis-ccp1-storage-usage-th-title',
                            _("Cloud Storage usage")),
                        escapeHTML: false,
                        field: "userStorageUsage.usage",
                        renderCell: function (row) {
                            var usage = statistics.userUsage[row.mspUserId];
                            if (!usage) {
                                return _("N/A");
                            }

                            var data = usage.userStorageUsage;
                            if (data.limit == common.QUOTA_UNLIMITED_VALUE) {
                                if (data.usage == null) {
                                    return _("N/A");
                                }
                                return common.formatStorageSize(data.usage);
                            }

                            var classesMap = {};
                            classesMap[Math.round(data.limit * 0.5)] = "warn";
                            classesMap[Math.round(data.limit * 0.75)] = "over";
                            var currentUsageForGauge = (data.usage == null) ? _("N/A") : common.formatStorageSize(
                                data.usage);
                            return new Gauge({
                                label: _("Cloud Storage usage"),
                                legend: currentUsageForGauge + " " + _("used of") + " " + common.formatStorageSize(
                                    data.limit),
                                minimum: 0,
                                maximum: data.limit,
                                value: data.usage == null ? 0 : data.usage,
                                classesMap: classesMap
                            });
                        }
                    }
                ],
                userGrid = ["aps/Grid", {
                    id: "account-grid",
                    showPaging: false,
                    recordsTotalText: _("item(s) total"),
                    columns: gridColumns,
                    store: subscriptionAccountStore,
                    selectionMode: 'multiple',
                    apsResourceViewId: userViewViewId
                }];

            if (statistics.tenant.WorkstationCount && statistics.tenant.WorkstationCount.limit) {
                gridColumns.push({
                    name: span(tags.icon('i-workstations-o--16 acronis-gray-icon', 16), null, _("Number of workstations")),
                    escapeHTML: false,
                    field: "UserWorkstationCount.usage",
                    renderCell: function (row) {
                        var usage = statistics.userUsage[row.mspUserId];
                        return usage ? prepareUsageStats(usage.UserWorkstationCount.usage,
                            usage.UserWorkstationCount.limit) : _("N/A");
                    }
                });
            }
            if (statistics.tenant.ServerCount && statistics.tenant.ServerCount.limit) {
                gridColumns.push({
                    name: span(tags.icon('i-server-o--16 acronis-gray-icon', 16), null, _("Number of servers")),
                    escapeHTML: false,
                    field: "userServerCount.usage",
                    renderCell: function (row) {
                        var usage = statistics.userUsage[row.mspUserId];
                        return usage ? prepareUsageStats(usage.userServerCount.usage, usage.userServerCount.limit) : _("N/A");
                    }
                });
            }
            if (statistics.tenant.VmCount && statistics.tenant.VmCount.limit) {
                gridColumns.push({
                    name: span(tags.icon('i-virtual-machines-o--16 acronis-gray-icon', 16), null, _("Number of virtual machines")),
                    escapeHTML: false,
                    field: "UserVmCount.usage",
                    renderCell: function (row) {
                        var usage = statistics.userUsage[row.mspUserId];
                        return usage ? prepareUsageStats(usage.UserVmCount.usage, usage.UserVmCount.limit) : _("N/A");
                    }
                });
            }
            if (statistics.tenant.MobileDeviceCount && statistics.tenant.MobileDeviceCount.limit) {
                gridColumns.push({
                    name: span(tags.icon('i-mobile-devices-o--16 acronis-gray-icon', 16), null, _("Number of mobile devices")),
                    escapeHTML: false,
                    field: "UserMobileDeviceCount.usage",
                    renderCell: function (row) {
                        var usage = statistics.userUsage[row.mspUserId];
                        return usage ? prepareUsageStats(usage.UserMobileDeviceCount.usage, usage.UserMobileDeviceCount.limit) : _("N/A");
                    }
                });
            }
            if (statistics.tenant.WebHostingServerCount && statistics.tenant.WebHostingServerCount.limit) {
                gridColumns.push({
                    name: span(tags.icon('i-web-hosting-server-o--16 acronis-gray-icon', 16), null, _("Number of web hosting servers")),
                    escapeHTML: false,
                    field: "userWebHostingServerCount.usage",
                    renderCell: function (row) {
                        var usage = statistics.userUsage[row.mspUserId];
                        return usage ? prepareUsageStats(usage.userWebHostingServerCount.usage, usage.userWebHostingServerCount.limit) : _("N/A");
                    }
                });
            }
            if (statistics.tenant.WebsiteCount && statistics.tenant.WebsiteCount.limit) {
                gridColumns.push({
                    name: span(tags.icon('i-website-o--16 acronis-gray-icon', 16), null, _("Number of websites")),
                    escapeHTML: false,
                    field: "UserWebsiteCount.usage",
                    renderCell: function (row) {
                        var usage = statistics.userUsage[row.mspUserId];
                        return usage ? prepareUsageStats(usage.UserWebsiteCount.usage, usage.UserWebsiteCount.limit) : _("N/A");
                    }
                });
            }
            gridColumns.push({
                name: "",
                renderCell: function (user) {
                    switch (user.userstatus) {
                        case common.USER_STATUS.ACTIVE:
                            return new Button({
                                id: common.genId("login_button"),
                                label: _("Log into Backup Console"),
                                'class': 'acronis-ccp1-button log-into-backup-console',
                                iconClass: "sb-export",
                                onClick: function () {
                                    var self = this,
                                        w = window.open('', '_blank');

                                    common.loginToConsole(user).then(function (uri) {
                                        w.location.href = uri;
                                        self.cancel();
                                    }, function (error) {
                                        w.close();
                                        console.error("ERR: \n" + JSON.stringify(error));
                                        messages.displayError(error);
                                        self.cancel();
                                    });

                                }
                            });
                        case common.USER_STATUS.NOT_ACTIVATED:
                            return new Button({
                                id: common.genId("activate_button"),
                                label: _("Activate"),
                                'class': 'acronis-ccp1-button log-into-backup-console',
                                onClick: function () {
                                    aps.apsc.gotoView(userActivateViewId, user.aps.id);
                                }
                            });
                    }
                    return "";
                }
            });

            load(["aps/PageContainer", {id: "apsPageContainer", 'class': "acronis-tenant"}, [
                ["aps/Container", {
                    id: common.genId('usage'),
                    title: _("Total usage")
                }, [
                    totalPanel
                ]],
                ["aps/Container", {
                    id: common.genId('container'),
                    title: _("Backup users"),
                    description: _("A backup user can access the backup management console and manage backups of his/her own data.")
                }, [
                    toolbar, userGrid
                ]]
            ]]).then(function () {
                if (statistics.error) {
                    messages.displayMessage(statistics.error, 'error');
                }
                (function setRefreshTimeout() {
                    setTimeout(function () {
                        if (needRefresh) {
                            getStatisticsDeffer.then(function(newStatistics) {
                                statistics = newStatistics;
                                registry.byId("account-grid").refresh();
                                needRefresh = false;
                            });
                        }
                    }, 10000);
                })();

                checkButtonState();
                registry.byId("account-grid").get("selectionArray").watchElements(checkButtonState);
            });
        });
    }, messages.displayError);
});
