/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

require([
    "aps/load",
    "dojo/promise/all",
    "aps/Store",
    "aps/Button",
    "aps/Memory",
    "aps/xhr",
    "app/agentsHelper",
    "app/common",
    "app/messages",
    "app/tags"
], function (
    load,
    all,
    Store,
    Button,
    Memory,
    xhr,
    agentsHelper,
    common,
    messages,
    tags
) {
    var language = common.getLanguage(aps.context.locale),
        enduser = aps.context.vars.enduser,
        userId = enduser.aps.id,
        userEmail = enduser.endUserEmail,
        isUserDisabled = (enduser.userstatus == "0"),
        clientUrlStore = new Store({target: "/aps/2/resources/" + userId + "/getclient?language=" + language}),
        usageStore = new Store({target: "/aps/2/resources/" + userId + "/machineUsage"});

    if (isUserDisabled) {
        load(["aps/PageContainer", {id: "apsPageContainer"}]);
        messages.displayWarning(_("The subscription to the service is currently disabled. Contact the system administrator."));
        return;
    }

    function output(html) {
        return ["aps/Output", {content: html}];
    }

    function link(href, text) {
        return '<a href=' + href + ' target="_blank">' + text + '</a>';
    }

    function image(src) {
        return '<img src=' + src + '>';
    }

    function ol(items) {
        return '<ol><li>' + items.join('</li><li>') + '</li></ol>';
    }

    function span(text, cssClass, title) {
        return "<span" + (title ? " title='" + title + "'" : "") + (cssClass ? " class=" + cssClass : "") + ">" +
            text + "</span>";
    }

    function subtitle(text) {
        return ["aps/Output", {value: text, "class": "acronis-usercp-header"}];
    }

    function agentLink(agent) {
        return output(span(link(agent.url, image(agent.icon)), null, agent.title));
    }

    function formatStorageUsage(usageObject) {
        var result = (usageObject.usage == null) ? '0' : common.formatStorageSize(usageObject.usage);
        if (usageObject.limit != null && usageObject.limit > 0) {
            result = result + ' / ' + common.formatStorageSize(usageObject.limit);
        }
        return result;
    }

    function formatUnitUsage(usageObject) {
        var result = (usageObject.usage == null) ? '0' : usageObject.usage;
        if (usageObject.limit != null && usageObject.limit > 0) {
            result = result + ' / ' + usageObject.limit;
        }
        return result;
    }

    function convertList(array, type) {
        var mailboxes = [],
            machines = [];
        array.forEach(function (machine) {
            if (machine.type === "msexchange_mailbox") {
                mailboxes.push(machine);
            } else if (machine.type === "machine" || machine.type === "virtual_machine") {
                machines.push(machine);
            }
        });

        return (type === 'mailboxes') ? mailboxes : machines;
    }

    /**
     * Returns right status text.
     * @param value
     * @returns string
     */
    function getStatusText(value) {
        switch (value) {
            case "warning":
                return _("Warnings");
            case "error":
                return _("Error");
            case "paused":
                return _("Backup paused");
            case "canceled":
                return _("Backup canceled");
            case "none":
            case "unknown":
            case "not_run":
                return _("Not protected");
            default:
                return _("OK");
        }
    }

    /**
     * Returns right status text.
     * @param value
     * @returns string
     */
    function getStatusIcon(value) {
        switch (value) {
            case "warning":
            case "error":
                return value;
            case "paused":
            case "canceled":
                return "notify";
            case "none":
            case "unknown":
            case "not_run":
                return 'unprotected';
            default:
                return 'ok';
        }
    }

    all({
        tenant: xhr.get('/aps/2/resources/' + enduser.aps.id + '/getTenant'),
        userData: clientUrlStore.query(),
        machineUsage: usageStore.query()
    }).then(function (response) {
        var clients = response.userData.clients,
            statistics = response.userData.statistics,
            mailboxList = convertList(response.machineUsage, 'mailboxes'),
            machineList = convertList(response.machineUsage, 'machines'),
            manageBackupsButton = ["aps/Button", {
                id: "simpleButton",
                label: _("Manage Backups"),
                iconClass: "sb-export",
                autoBusy: false,
                onClick: function () {
                    var w = window.open('', '_blank');
                    xhr.get("/aps/2/resources/" + userId + "/loginByToken").then(function (url) {
                        w.location.href = url;
                    }, function (error) {
                        w.close();
                        messages.displayError(error);
                    });
                }
            }],
            totalPanel = ['aps/Container', {'class': 'total-panel'}, [
                output([
                    span(_('Total usage:'), 'total-usage-title'),
                    span(tags.icon('i-cloud-storage-o--16 acronis-gray-icon', 16) +
                        span(formatStorageUsage(statistics.userStorageUsage)), 'total-usage', _('Cloud Storage usage')),
                    !statistics.UserWorkstationCount.limit ? '' : span(tags.icon('i-workstations-o--16 acronis-gray-icon', 16) +
                        span(formatUnitUsage(statistics.UserWorkstationCount)), 'total-usage', _('Number of workstations')),
                    !statistics.userServerCount.limit ? '' : span(tags.icon('i-server-o--16 acronis-gray-icon', 16) +
                        span(formatUnitUsage(statistics.userServerCount)), 'total-usage', _('Number of servers')),
                    !statistics.UserVmCount.limit ? '' : span(tags.icon('i-virtual-machines-o--16 acronis-gray-icon', 16) +
                        span(formatUnitUsage(statistics.UserVmCount)), 'total-usage', _('Number of virtual machines')),
                    !statistics.UserMobileDeviceCount.limit ? '' : span(tags.icon('i-mobile-devices-o--16 acronis-gray-icon', 16) +
                        span(formatUnitUsage(statistics.UserMobileDeviceCount)), 'total-usage', _('Number of mobile devices')),
                    !statistics.userWebHostingServerCount.limit ? '' : span(tags.icon('i-web-hosting-server-o--16 acronis-gray-icon', 16) +
                        span(formatUnitUsage(statistics.userWebHostingServerCount)), 'total-usage', _('Number of web hosting servers')),
                    !statistics.UserWebsiteCount.limit ? '' : span(tags.icon('i-website-o--16 acronis-gray-icon', 16) +
                        span(formatUnitUsage(statistics.UserWebsiteCount)), 'total-usage', _('Number of websites'))
                ].join(''))
            ]],
            gridColumns = [
                {
                    "name": _("Devices"),
                    "escapeHTML": false,
                    "field": "name",
                    "class": "acronis-width-25-percent",
                    renderCell: function (row, data) {
                        data = common.escapeAngularBrackets(data);
                        if (row.type === 'virtual_machine') {
                            return '<img src="images/statuses/virtual16.png" class="machine-icon">' + data;
                        }
                        if (row.type === 'machine') {
                            return '<img src="images/statuses/machine16.png" class="machine-icon">' + data;
                        }
                        return data;
                    }
                },
                {
                    "name": _("Status"),
                    "escapeHTML": false,
                    "field": "status",
                    "class": "acronis-width-25-percent",
                    renderCell: function (row, data) {
                        var text = getStatusText(data),
                            icon = getStatusIcon(data);
                        return '<img src="images/statuses/' + icon + '24.png"> ' + text;
                    }
                },
                {
                    "name": _("Last backup"),
                    "field": "last_backup",
                    "class": "acronis-width-25-percent",
                    renderCell: function (row, data) {
                        if (!data) {
                            return _("Never");
                        }
                        var date = new Date(0);
                        date.setUTCSeconds(data);
                        return common.formatLocalDateString(date, aps.context.locale);
                    }
                },
                {
                    "name": _("Next backup"),
                    "field": "next_backup",
                    "class": "acronis-width-25-percent",
                    renderCell: function (row, data) {
                        if (!data) {
                            return _("Not scheduled");
                        }
                        var date = new Date(0);
                        date.setUTCSeconds(data);
                        return common.formatLocalDateString(date, aps.context.locale);
                    }
                }
            ],
            mailboxesGridColumns = [
                {
                    "name": _("Office 365 seats"),
                    "escapeHTML": false,
                    "field": "name",
                    "class": "acronis-width-25-percent",
                    renderCell: function (row, data) {
                        data = common.escapeAngularBrackets(data);
                        if (row.type === 'virtual_machine') {
                            return '<img src="images/statuses/virtual16.png" class="machine-icon">' + data;
                        }
                        if (row.type === 'machine') {
                            return '<img src="images/statuses/machine16.png" class="machine-icon">' + data;
                        }
                        return data;
                    }
                },
                {
                    "name": _("Status"),
                    "escapeHTML": false,
                    "field": "status",
                    "class": "acronis-width-25-percent",
                    renderCell: function (row, data) {
                        var text = getStatusText(data),
                            icon = getStatusIcon(data);
                        return '<img src="images/statuses/' + icon + '24.png"> ' + text;
                    }
                },
                {
                    "name": _("Last backup"),
                    "field": "last_backup",
                    "class": "acronis-width-25-percent",
                    renderCell: function (row, data) {
                        if (!data) {
                            return _("Never");
                        }
                        var date = new Date(0);
                        date.setUTCSeconds(data);
                        return common.formatLocalDateString(date, aps.context.locale);
                    }
                },
                {
                    "name": _("Next backup"),
                    "field": "next_backup",
                    "class": "acronis-width-25-percent",
                    renderCell: function (row, data) {
                        if (!data) {
                            return _("Not scheduled");
                        }
                        var date = new Date(0);
                        date.setUTCSeconds(data);
                        return common.formatLocalDateString(date, aps.context.locale);
                    }
                }
            ],
            machinesGrid = ["aps/Grid", {
                showPaging: false,
                recordsTotalText: _("item(s) total"),
                columns: gridColumns,
                store: new Memory({data: machineList})
            }],
            mailboxesGrid = ["aps/Grid", {
                showPaging: false,
                sort: {
                    attribute: "name"
                },
                recordsTotalText: _("item(s) total"),
                columns: mailboxesGridColumns,
                store: new Memory({data: mailboxList})
            }],
            agentsList = agentsHelper.buildBackupAgentsList(clients),
            rightColumn = [
                subtitle(_("Download"))
            ],
            leftColumn = [
                subtitle(_("Getting Started")),
                output(ol([
                    _("Select the backup agent for your device."),
                    _("Download the backup agent installer to your device."),
                    _("Run the installer and following installation instructions. After that the device appears in the web interface."),
                    _("Select the device and create new backup plan: select the items that you want to protect, modify backup settings and click Apply to protect your device.")
                ]))
            ];

        agentsList.forEach(function(agentCategory) {
            rightColumn.push([
                "aps/FieldSet", {showLabels: false}, [
                    ["aps/Output", {value: agentCategory.title, "class": "title"}],
                    ["aps/Container", {"class": "agent-box"}, agentCategory.agents.map(agentLink)]
                ]]
            );
        });

        rightColumn.push(["aps/Output", {
            value: _('To download other agents, click "Manage Backups" and then click "Add".')
        }]);

        load(["aps/PageContainer", {id: "apsPageContainer", "class": "acronis-usercp"}, [
            manageBackupsButton,
            totalPanel,
            ["aps/Container", {cols: 2}, [
                ["aps/Container", {"class": "left-box"}, leftColumn],
                ["aps/Container", {"class": "right-box"}, rightColumn]
            ]],
            ["aps/Container", {
                title: _("Machines"),
                "class": "machine-grid",
                visible: machineList.length
            }, [machinesGrid]],
            ["aps/Container", {
                title: _("Microsoft Office 365 seats"),
                "class": "mailboxes-grid",
                visible: mailboxList.length
            }, [mailboxesGrid]
            ]]
        ]);
    }, function (err) {
        aps.apsc.hideLoading();
        console.log(err);
    });
});
