/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/Deferred",
    "dojox/mvc/at",
    "dojox/mvc/getStateful",
    "dojox/mvc/getPlainValue",
    "aps/_View",
    "aps/Memory",
    "aps/Container",
    "aps/Output",
    "aps/load",
    "./../components/getUserResourcesGrid",
    "./../../messages",
    "./../../common"
], function (
    declare,
    lang,
    Deferred,
    at,
    getStateful,
    getPlainValue,
    _View,
    Memory,
    Container,
    Output,
    load,
    getUserResourcesGrid,
    messages,
    common
) {
    return declare(_View, {
        _container: null,
        _tenant: null,
        _model: null,
        _userQuota: null,
        _deltas: [],
        _users: null,

        _submit: function () {
            var self = this,
                actions = {
                    provisionResource: [],
                    configureResource: []
                },
                data = getPlainValue(self._model),
                users = self._users;

            users.forEach(function (user) {
                var backupAccount = lang.clone(data);
                if (user.backupUser) {
                    backupAccount.aps = {id: user.backupUser.aps.id};
                    actions.configureResource.push(backupAccount);
                } else {
                    backupAccount.user = {aps: {id: user.aps.id}};
                    backupAccount.aps = {
                        type: common.TYPES.USERS,
                        subscription: self._tenant.aps.subscription
                    };
                    backupAccount.tenant = [{aps: {id: self._tenant.aps.id}}];
                    actions.provisionResource.push(backupAccount);
                }
            });

            var operation = {
                "call": {
                    "resourceId": self._tenant.aps.id,
                    "operation": "setUsersLicensesAfterUpsell",
                    "parameters": {"data": actions}
                }
            };

            return aps.biz.requestUsageChange({
                deltas: self._deltas,
                operations: [operation],
                subscriptionId: self._tenant.aps.subscription
            }).then(function() {
                return aps.biz.commit();
            });
        },

        init: function () {
            this._container = new Container();
            this._container.placeAt(this.domNode);
        },
        onHide: function () {
            messages.removeAll();
        },
        onCancel: function () {
            aps.apsc.prev();
        },
        onSubmit: function () {
            var self = this;
            if (!self._container.validate()) {
                aps.apsc.cancelProcessing();
                return;
            }

            self._submit().then(function () {
                aps.apsc.prev();
            }, messages.displayError);
        },
        onContext: function (context) {
            var self = this,
                usersIds = aps.context.params.usersIds,
                users = [],
                allUsers = context.vars.users;

            allUsers.forEach(function (user) {
                if (0 <= usersIds.indexOf(user.aps.id))
                    users.push(user);
            });

            self._container.removeAll();
            self.set("label", _("Assign Backup Service for __count__ user(s)", {count: users.length}));
            self._container.addChild(new Output({
                value: _("To assign backup service you need to define resource limits. These limits will be applied for each user you previously selected.")
            }));

            self._tenant = context.vars.tenant;
            self._users = users;

            common.getBackupProfile(self._tenant).then(function (profile) {
                self._userQuota = common.getUserQuota(profile);
                self._model = getStateful({
                    notifyBackup: common.getBackupNotification(profile),
                    notificationsManagement: common.getManagementNotification(profile),

                    privileged: common.USER_PRIVILEGED.NO,

                    activateDuringCreation: false
                });
                if (self._userQuota) {
                    self._model.set({
                        userStorageUsage: getStateful({limit: 0}),
                        userServerCount: getStateful({limit: 0}),
                        UserVmCount: getStateful({limit: 0}),
                        UserWorkstationCount: getStateful({limit: 0}),
                        UserMobileDeviceCount: getStateful({limit: 0}),
                        userWebHostingServerCount: getStateful({limit: 0}),
                        UserWebsiteCount: getStateful({limit: 0})
                    });
                }
                else {
                    self._model.set({
                        userStorageUsage: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                        userServerCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                        UserVmCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                        UserWorkstationCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                        UserMobileDeviceCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                        userWebHostingServerCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE}),
                        UserWebsiteCount: getStateful({limit: common.QUOTA_UNLIMITED_VALUE})
                    });
                }

                var form = ["aps/Panel", {title: _("Backup user settings")}, [
                    ["aps/FieldSet", {gridSize: "md-12 xs-12"}, [
                        ["aps/Select", {
                            id: self.genId('backup_notifications'),
                            gridSize: "md-6 xs-12",
                            label: _("Backup Notifications"),
                            value: at(self._model, "notifyBackup").transform({
                                parse: function (value) {
                                    return Number(value);
                                }
                            }),
                            required: true,
                            store: new Memory({data: common.getBackupNotificationOptions(), idProperty: "value"})
                        }],
                        ["aps/Select", {
                            id: self.genId('business_notifications'),
                            gridSize: "md-6 xs-12",
                            label: _("Business Notifications"),
                            description: _("Enable notifications about quota excess for this user."),
                            value: at(self._model, "notificationsManagement").transform({
                                format: function (value) {
                                    return Number(value);
                                },
                                parse: function (value) {
                                    return !!Number(value);
                                }
                            }),
                            required: true,
                            store: new Memory({data: common.getEnabledDisabledOptions(), idProperty: "value"})
                        }],
                        ["aps/Select", {
                            id: self.genId('administrator_privileges'),
                            gridSize: "md-6 xs-12",
                            label: _("Administrator Privileges"),
                            description: _("An administrator can manage backups of other users and back up Office 365 seats."),
                            value: at(self._model, "privileged").transform({
                                parse: function(value) {
                                    return Number(value);
                                },
                                format: function(value) {
                                    return String(value);
                                }
                            }),
                            required: true,
                            store: new Memory({idProperty: "value", data: common.getEnabledDisabledOptions()})
                        }]
                    ]]
                ]];

                load(form, self._container).then(function () {
                    if (self._userQuota) {
                        getUserResourcesGrid(self._tenant, self._model, users.length).then(function (resources) {
                            self._container.addChild(resources.grid);
                            self._deltas = resources.deltas;
                            aps.apsc.hideLoading();
                        }, messages.displayError);
                    }
                    else {
                        aps.apsc.hideLoading();
                    }
                }, messages.displayError);

            }, messages.displayError);
        }
    });
});

