/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/_base/declare",
    "aps/Container",
    "aps/Grid",
    "aps/FieldSet",
    "aps/Memory",
    "./base/Output",
    "./base/Status",
    "./../../uomConverter"
], function (
    declare,
    Container,
    Grid,
    FieldSet,
    Memory,
    Output,
    Status,
    uomConverter
) {

    var statusInfo = {
        updating: {"label": _("UPDATING"), "type": "default", isLoad: true},
        deleting: {"label": _("DELETING"), "type": "danger", isLoad: true},
        notAssigned: {"label": _("NOT ASSIGNED"), "type": "default", isLoad: false},
        notActivated: {"label": _("NOT ACTIVATED"), "type": "warning", isLoad: false},
        activating: {"label": _("ACTIVATING"), "type": "warning", isLoad: true, useIcon: false},
        activated: {"label": _("ACTIVATED"), "type": "success", isLoad: false},
        disabled: {"label": _("DISABLED"), "type": "default", isLoad: true}
    };

    function createStatus(statusName) {
        return new Status({
            useIcon: false,
            gridSize: 'xs-12',
            status: statusName,
            statusInfo: statusInfo
        });
    }

    function createResourceUsageOutput(title, value) {
        return new Output({
            content: '<span title="${title}">${value}</span>',
            title: title,
            value: value
        });
    }

    function createResourceUsage(activated, backupStatistic) {
        var fieldSet = new FieldSet();
        if (backupStatistic.userStorageUsage && backupStatistic.userStorageUsage.limit) {
            fieldSet.addChild(createResourceUsageOutput(
                _("Cloud Storage usage"),
                activated ? prepareStorageUsageStats(
                    backupStatistic.userStorageUsage.usage,
                    backupStatistic.userStorageUsage.limit
                ) : _('N/A')
            ));
        }

        if (backupStatistic.UserWorkstationCount && backupStatistic.UserWorkstationCount.limit) {
            fieldSet.addChild(createResourceUsageOutput(
                _("Workstations"),
                activated ? prepareUsageStats(
                    backupStatistic.UserWorkstationCount.usage,
                    backupStatistic.UserWorkstationCount.limit
                ) : _('N/A')
            ));
        }

        if (backupStatistic.userServerCount && backupStatistic.userServerCount.limit) {
            fieldSet.addChild(createResourceUsageOutput(
                _("Servers"),
                activated ? prepareUsageStats(
                    backupStatistic.userServerCount.usage,
                    backupStatistic.userServerCount.limit
                ) : _('N/A')
            ));
        }

        if (backupStatistic.UserVmCount && backupStatistic.UserVmCount.limit) {
            fieldSet.addChild(createResourceUsageOutput(
                _("Virtual Machines"),
                activated ? prepareUsageStats(
                    backupStatistic.UserVmCount.usage,
                    backupStatistic.UserVmCount.limit
                ) : _('N/A')
            ));
        }

        if (backupStatistic.UserMobileDeviceCount && backupStatistic.UserMobileDeviceCount.limit) {
            fieldSet.addChild(createResourceUsageOutput(
                _("Mobile Devices"),
                activated ? prepareUsageStats(
                    backupStatistic.UserMobileDeviceCount.usage,
                    backupStatistic.UserMobileDeviceCount.limit
                ) : _('N/A')
            ));
        }

        if (backupStatistic.UserWebsiteCount && backupStatistic.UserWebsiteCount.limit) {
            fieldSet.addChild(createResourceUsageOutput(
                _("Web Sites"),
                activated ? prepareUsageStats(
                    backupStatistic.UserWebsiteCount.usage,
                    backupStatistic.UserWebsiteCount.limit
                ) : _('N/A')
            ));
        }

        if (backupStatistic.userWebHostingServerCount && backupStatistic.userWebHostingServerCount.limit) {
            fieldSet.addChild(createResourceUsageOutput(
                _("Web Hosting Servers"),
                activated ? prepareUsageStats(
                    backupStatistic.userWebHostingServerCount.usage,
                    backupStatistic.userWebHostingServerCount.limit
                ) : _('N/A')
            ));
        }

        return fieldSet;
    }

    function prepareUsageStats(value, limit) {
        var usageStats = (value > 0) ? value : 0;
        if (limit >= 0) {
            return usageStats + ' / ' + limit;
        }
        return usageStats;
    }

    function prepareStorageUsageStats(usage, limit) {
        if (limit === 0) {
            return _('None');
        }
        var usageSize = uomConverter.convertByteToMajorityUnit(usage);
        if (limit < 0) {
            return usageSize.value + ' ' + uomConverter.getUnitName(usageSize.unit);
        }
        var limitSize = uomConverter.convertByteToMajorityUnit(limit);
        return usageSize.value + ' ' + uomConverter.getUnitName(usageSize.unit) + ' / ' + limitSize.value + ' ' + uomConverter.getUnitName(limitSize.unit);
    }

    function isUserActivated(user) {
        return user.backupUserStatus === "activated";
    }

    return declare(Container, {
        users: [],
        gridSize: 'xs-12',
        style: "padding: 0px;",

        _setUsersAttr: function (val) {
            var userStore = new Memory({data: val, idProperty: "aps.id"});
            var layout = [
                {
                    "field": "backupUserStatus",
                    "escapeHTML": false,
                    renderCell: function (row, data) {
                        return createStatus(data);
                    }
                },
                {
                    "type": "resourceName",
                    "field": "displayName"
                },
                {
                    "field": "login"
                },
                {
                    "escapeHTML": false,
                    "field": "backupUserStatistic",
                    renderCell: function (row, data) {
                        return createResourceUsage(isUserActivated(row) && data, data);
                    }
                }
            ];

            var grid = new Grid({
                gridSize: 'xs-12',
                style: "padding: 0px;",
                columns: layout,
                store: userStore,
                apsResourceViewId: 'http://www.parallels.com/ccp-users#viewUser'
            });

            var container = new Container({
                title: _("Backup Users"),
                gridSize: 'xs-12',
                style: "padding: 0px;"
            });
            container.addChild(grid);

            this.removeAll();
            this.addChild(container);

            this.users = val;
        }

    });
});

