/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

define([
    "dojo/_base/declare",
    "aps/Button",
    "aps/Container",
    "aps/DropDownButton",
    "aps/Grid",
    "aps/Memory",
    "aps/Toolbar",
    "./base/Status",
    "./../components/getStorageOutput",
    "./../components/BackupAdminActions",
    "./../../common",
    "./../../messages"
], function (
    declare,
    Button,
    Container,
    DropDownButton,
    Grid,
    Memory,
    Toolbar,
    Status,
    getStorageOutput,
    BackupAdminActions,
    common,
    messages
) {

    var statusInfo = {
        updating: {"label": _("UPDATING"), "type": "default", isLoad: true},
        deleting: {"label": _("DELETING"), "type": "danger", isLoad: true},
        notAssigned: {"label": _("NOT ASSIGNED"), "type": "default", isLoad: false},
        notActivated: {"label": _("NOT ACTIVATED"), "type": "warning", isLoad: false},
        activating: {"label": _("ACTIVATING"), "type": "warning", isLoad: true, useIcon: false},
        activated: {"label": _("ACTIVATED"), "type": "success", isLoad: false},
        disabled: {"label": _("DISABLED"), "type": "default", isLoad: true}
    };

    function createStatus(statusName) {
        return new Status({
            useIcon: false,
            gridSize: 'md-12',
            status: statusName,
            statusInfo: statusInfo
        });
    }

    function prepareAllowedDevicesLimit(statistic, tenant) {

        function getCounterLimit(counter) {
            if (!(counter && counter.limit)) {
                return null;
            }
            if (counter.limit < 0) {
                return _("Unlimited");
            }
            return String(counter.limit);
        }

        function checkUnlimited(counter, tenantCounter) {
            return !tenantCounter || counter.limit < 0;
        }

        if (checkUnlimited(statistic.userServerCount, tenant.ServerCount) &&
            checkUnlimited(statistic.UserVmCount, tenant.VmCount) &&
            checkUnlimited(statistic.UserWorkstationCount, tenant.WorkstationCount) &&
            checkUnlimited(statistic.UserMobileDeviceCount, tenant.MobileDeviceCount) &&
            checkUnlimited(statistic.userWebHostingServerCount, tenant.WebHostingServerCount) &&
            checkUnlimited(statistic.UserWebsiteCount, tenant.WebsiteCount)) {
                return _("Unlimited");
        }

        var limit, result = [];
        limit = getCounterLimit(statistic.userServerCount);
        if (limit) {
            result.push(_("{0} Server(s)").replace("{0}", limit));
        }
        limit = getCounterLimit(statistic.UserVmCount);
        if (limit) {
            result.push(_("{0} Virtual Machine(s)").replace("{0}", limit));
        }
        limit = getCounterLimit(statistic.UserWorkstationCount);
        if (limit) {
            result.push(_("{0} Workstation(s)").replace("{0}", limit));
        }
        limit = getCounterLimit(statistic.UserMobileDeviceCount);
        if (limit) {
            result.push(_("{0} Mobile Device(s)").replace("{0}", limit));
        }
        limit = getCounterLimit(statistic.userWebHostingServerCount);
        if (limit) {
            result.push(_("{0} Web Hosting Server(s)").replace("{0}", limit));
        }
        limit = getCounterLimit(statistic.UserWebsiteCount);
        if (limit) {
            result.push(_("{0} Website(s)").replace("{0}", limit));
        }

        if (!result.length) {
            return _('None');
        }
        return result.join(', ');
    }

    return declare(Container, {
        title: _("Backup Users"),
        users: [],
        gridSize: "md-12 xs-12",

        _setUsersAttr: function (val) {
            var tenant = aps.context.vars.tenant,
                userQuota = common.getUserQuota(tenant.profile),
                userStore = new Memory({data: val, idProperty: "aps.id"}),
                layout = [
                    {
                        name: _('User'),
                        field: "displayName",
                        type: "resourceName",
                        filter: true
                    },
                    {
                        name: _("Service Status"),
                        field: "backupUserStatus",
                        sortable: false,
                        renderCell: function (row, data) {
                            return createStatus(data);
                        }
                    },
                    {
                        name: _("Cloud Storage usage"),
                        field: "backupUserStatistic.userStorageUsage",
                        sortable: false,
                        renderCell: function (row, data) {
                            if (!data || data.usage === null) {
                                return _('N/A');
                            }

                            return getStorageOutput({
                                limitedLegend: _("${usage} ${usageUnit} of ${limit} ${limitUnit} used"),
                                unlimitedLegend: _("${usage} ${usageUnit} used"),
                                usage: data.usage || 0,
                                limit: data.limit || 0
                            });
                        }
                    },
                    {
                        name: _("Allowed Devices Limit"),
                        field: "backupUserStatistic",
                        sortable: false,
                        renderCell: function (row, data) {
                            if (!data) {
                                return _('N/A');
                            }
                            if (userQuota) {
                                return prepareAllowedDevicesLimit(data, tenant);
                            }
                            return _("Unlimited");
                        }
                    },
                    {
                        name: '',
                        field: "backupUser",
                        sortable: false,
                        renderCell: function (row) {

                            var backupAccountActions = new BackupAdminActions(row, tenant),
                                separatorMenuItem = {
                                    type: "separator"
                                },
                                viewMenuItem = {
                                    iconClass: "fa-user",
                                    label: _("View User"),
                                    onClick: function () {
                                        aps.apsc.gotoView("http://www.parallels.com/ccp-users#viewUser", row.aps.id);
                                    }
                                },
                                assignMenuItem = {
                                    iconClass: "fa-check",
                                    label: _("Assign Service"),
                                    onClick: function () {
                                        backupAccountActions.assign().then(null, messages.displayError);
                                    }
                                },
                                activateMenuItem = {
                                    iconClass: "fa-magic",
                                    type: "warning",
                                    label: _("Activate"),
                                    onClick: function () {
                                        backupAccountActions.activate().then(null, messages.displayError);
                                    }
                                },
                                configureMenuItem = {
                                    iconClass: "fa-pencil",
                                    label: _("Change Settings"),
                                    onClick: function () {
                                        backupAccountActions.configure().then(null, messages.displayError);
                                    }
                                },
                                unassignMenuItem = {
                                    iconClass: "fa-times",
                                    label: _("Unassign Service"),
                                    onClick: function () {
                                        backupAccountActions.unassign().then(null, messages.displayError);
                                    }
                                },
                                enableMenuItem = {
                                    iconClass: "fa-times",
                                    label: _("Enable"),
                                    onClick: function () {
                                        backupAccountActions.enable().then(null, messages.displayError);
                                    }
                                },
                                disableMenuItem = {
                                    iconClass: "fa-times",
                                    label: _("Disable"),
                                    onClick: function () {
                                        backupAccountActions.disable().then(null, messages.displayError);
                                    }
                                },
                                loginToImpersonatedBackupConsole = {
                                    iconClass: "fa-external-link",
                                    label: _("Log into Backup Console"),
                                    onClick: function () {
                                        var w = window.open('', '_blank');
                                        backupAccountActions.loginToConsole().then(function (uri) {
                                            w.location.href = uri;
                                        }, function (error) {
                                            w.close();
                                            messages.displayError(error);
                                        });

                                    }
                                },
                                menuItems = [];

                            switch (row.backupUserStatus) {
                                case 'notAssigned':
                                    menuItems = [assignMenuItem, viewMenuItem];
                                    break;

                                case 'notActivated':
                                    menuItems = [activateMenuItem, viewMenuItem, separatorMenuItem, unassignMenuItem];
                                    break;

                                case 'activated':
                                    menuItems = [loginToImpersonatedBackupConsole, separatorMenuItem, configureMenuItem, viewMenuItem];
                                    if (row.backupUser.servicestatus == common.STATUS.ENABLED) {
                                        menuItems.push(separatorMenuItem);
                                        menuItems.push(disableMenuItem);
                                    }
                                    menuItems.push(unassignMenuItem);
                                    break;
                                case 'disabled':
                                    configureMenuItem.disabled = true;
                                    menuItems = [configureMenuItem, viewMenuItem, separatorMenuItem, enableMenuItem, unassignMenuItem];
                                    break;
                                default:
                                    configureMenuItem.disabled = true;
                                    unassignMenuItem.disabled = true;
                                    menuItems = [configureMenuItem, viewMenuItem, separatorMenuItem, unassignMenuItem];
                            }

                            return new DropDownButton({
                                iconClass: "fa-cog",
                                label: _("Actions"),
                                items: menuItems
                            });
                        }
                    }
                ];

            var toolbar = new Toolbar(),
                grid = new Grid({
                    selectionMode: "multiple",
                    style: "padding: 0px;",
                    columns: layout,
                    store: userStore,
                    showPaging: true,
                    rowsPerPage: 20,
                    pageSizeOptions: [20],
                    apsResourceViewId: 'http://www.parallels.com/ccp-users#viewUser'
                }),
                addButton = new Button({
                    autoBusy: false,
                    label: _("Add New Users"),
                    iconClass: "fa-plus",
                    isDefault: true,
                    onClick: function () {
                        aps.apsc.gotoView("http://www.parallels.com/ccp-users#addUser");
                    }
                }),
                assignButton = new Button({
                    autoBusy: false,
                    label: _("Assign Service"),
                    iconClass: "fa-pencil",
                    isDefault: true,
                    disabled: true,
                    onClick: function () {
                        var usersIds = grid.selectionArray,
                            selectedUsersIds = [];
                        usersIds.forEach(function (userId) {
                            var user = grid.store.get(userId);
                            if (!user.backupUser) {
                                selectedUsersIds.push(userId);
                            }
                        });
                        aps.apsc.gotoView("nextcp-ccp-multiple-setquotas", null, {
                            usersIds: selectedUsersIds
                        });
                    }
                });

            grid.selectionArray.watchElements(function () {
                var users = grid.selectionArray.filter(function (userId) {
                    var user = grid.store.get(userId);
                    return !user.backupUser;
                });
                assignButton.set("disabled", !users.length);
            });

            toolbar.addChild(addButton);
            toolbar.addChild(assignButton);
            grid.addChild(toolbar);

            this.removeAll();
            this.addChild(grid);

            this.users = val;
        }

    });
});

