/*
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

require([
    "aps/Button",
    "aps/Container",
    "aps/load",
    "aps/Memory",
    "aps/Output",
    "aps/xhr",
    "dijit/registry",
    "dojo/promise/all",
    "./js/common.js",
    "./js/messages.js",
    "./js/wizardData.js"
], function (
    Button,
    Container,
    load,
    Memory,
    Output,
    xhr,
    registry,
    all,
    common,
    messages,
    wizardData
) {
    var tenant = aps.context.vars.tenant,
        administratorNewViewId = common.getRelatedViewId('new'),
        administratorViewViewId = common.getRelatedViewId('view');

    function getUserStatus(status) {
        var content;
        switch (status) {
            case -1:
                content = '<img src=images/user_statuses/ico_state_off_16px.png> ' + _("Not activated");
                break;
            case 0:
                content = '<img src=images/user_statuses/ico_state_off_16px.png> ' + _("Disabled");
                break;
            case 1:
                content = '<img src=images/user_statuses/ico_state_on_16px.png> ' + _("Enabled");
                break;
        }
        return content;
    }

    function output(html) {
        return ["aps/Output", {id: common.genId('output'), value: html, escapeHTML: false}];
    }

    function ul(items) {
        return '<ul><li>' + items.join('</li><li>') + '</li></ul>';
    }

    function div(html, cssClass) {
        return '<div class=' + cssClass + '>' + html + '</div>';
    }

    function onAddNew() {
        aps.apsc.gotoView(administratorNewViewId);
    }

    function onRemove() {
        var self = this;
        if (!confirm(_("Are you sure that you want to remove the administrator account?"))) {
            self.cancel();
            return;
        }

        var grid = registry.byId("adminGrid"),
            selectionArray = grid.get("selectionArray"),
            defers = [],
            deletedIds = [];

        messages.removeAll();

        selectionArray.forEach(function (id) {
            var admin = grid.store.get(id),
                deffer = xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + admin.id, {
                    method: "DELETE",
                    query: {
                        datacenterId: admin.datacenterId
                    }
                }).then(function () {
                    deletedIds.push(id);
                }, messages.displayErrorAndGo);
            defers.push(deffer);
        });

        all(defers).then(function () {
            deletedIds.forEach(function (id) {
                grid.store.remove(id);
                selectionArray.splice(selectionArray.indexOf(id), 1);
            });
            grid.refresh();
            self.cancel();
        });
    }

    function getUserActions(row) {
        var widget;
        switch (row.status) {
            case +common.STATUS.INACTIVE:
                widget = new Button({
                    iconClass: "sb-export",
                    label: _("Resend activation email"),
                    onClick: function () {
                        var self = this;
                        messages.removeAll();
                        xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + row.id + "/resend", {
                            method: "PUT",
                            query: {datacenterId: row.datacenterId}
                        }).then(
                            function () {
                                messages.displayMessage(_("The activation message has been successfully sent to {0}.").replace("{0}", row.email), 'info');
                                self.cancel();
                            },
                            function () {
                                messages.displayMessage(_("Failed to send the activation message to {0}.").replace("{0}", row.login), 'error');
                                self.cancel();
                            }
                        );
                    }
                });
                break;
            case +common.STATUS.DISABLED:
                widget = new Button({
                    iconClass: "sb-export",
                    label: _("Enable"),
                    onClick: function () {
                        var self = this;
                        messages.removeAll();
                        xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + row.id + "/enable", {
                            method: "PUT",
                            query: {datacenterId: row.datacenterId}
                        }).then(
                            function () {
                                messages.displayMessage(_("Administrator {0} has been successfully enabled.").replace("{0}", row.login), 'info');
                                xhr.get("/aps/2/resources/" + tenant.aps.id + "/administrators").then(function (administrators) {
                                    var grid = registry.byId("adminGrid");
                                    grid.store.setData(administrators);
                                    grid.refresh();
                                }, messages.displayError);
                            },
                            function () {
                                messages.displayMessage(_("Failed to enable adminisrtator {0}.").replace("{0}", row.login), 'error');
                                self.cancel();
                            }
                        );
                    }
                });
                break;
            case +common.STATUS.ENABLED:
                widget = new Container();
                widget.addChild(new Button({
                    iconClass: "sb-export",
                    label: _("Disable"),
                    onClick: function () {
                        var self = this;
                        messages.removeAll();
                        xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + row.id + "/disable", {
                            method: "PUT",
                            query: {datacenterId: row.datacenterId}
                        }).then(
                            function () {
                                messages.displayMessage(_("Administrator {0} has been successfully disabled.").replace("{0}", row.login), 'info');
                                xhr.get("/aps/2/resources/" + tenant.aps.id + "/administrators").then(function (administrators) {
                                    var grid = registry.byId("adminGrid");
                                    grid.store.setData(administrators);
                                    grid.refresh();
                                }, messages.displayError);
                            },
                            function () {
                                messages.displayMessage(_("Failed to disable adminisrtator {0}.").replace("{0}", row.login), 'error');
                                self.cancel();
                            }
                        );
                    }
                }));
                widget.addChild(new Button({
                    iconClass: "sb-export",
                    label: _("Manage"),
                    onClick: function () {
                        var self = this,
                            newWindow = window.open("", "_blank");
                        messages.removeAll();
                        xhr("/aps/2/resources/" + tenant.aps.id + "/administrators/" + row.id + "/console", {
                            method: "GET",
                            query: {datacenterId: row.datacenterId}
                        }).then(function (response) {
                            newWindow.location.href = response.url;
                            self.cancel();
                        }, function (error) {
                            newWindow.close();
                            messages.displayError(error);
                            self.cancel();
                        });
                    }
                }));
                break;
        }
        return widget;
    }

    all({
        administrators: xhr.get("/aps/2/resources/" + tenant.aps.id + "/administrators"),
        datacenters: xhr.get("/aps/2/resources/" + tenant.aps.id + "/getDatacenters")
    }).then(function (response) {
        var administrators = response.administrators,
            showDatacenters = response.datacenters.length != 1;

        function getOutput(label, row) {
            return new Output({
                content: '<a href=#>${value}</a>',
                value: label,
                escapeHTML: false,
                onClick: function () {
                    wizardData.setWizardData({
                        adminId: row.id,
                        datacenterId: row.datacenterId,
                        showDatacenters: showDatacenters
                    }, administratorViewViewId);
                    aps.apsc.gotoView(administratorViewViewId);
                }
            });
        }

         var columns = [
                {
                    field: "id", name: "ID", visible: false
                },
                {
                    field: "firstName", name: _("First name"),
                    renderCell: function (row) {
                        return getOutput(row.firstName, row);
                    }
                },
                {
                    field: "lastName", name: _("Last name"),
                    renderCell: function (row) {
                        return getOutput(row.lastName, row);
                    }
                },
                {
                    field: "login", name: _("Login"),
                    renderCell: function (row) {
                        return getOutput(row.login, row);
                    }
                },
                {
                    field: "email", name: _("Email"),
                    renderCell: function (row) {
                        return getOutput(row.email, row);
                    }
                },
                {
                    field: "datacenterName", name: _("Datacenter"), visible: showDatacenters
                },
                {
                    field: "status", name: _("Status"), escapeHTML: false,
                    renderCell: function (row, data) {
                        return getUserStatus(data);
                    }
                },
                {
                    field: "status", name: "", sortable: false,
                    renderCell: function (row) {
                        return getUserActions(row);
                    }
                }
            ];

        load(["aps/PageContainer", {id: "apsPageContainer", "class": "acronis-administrator"}, [
            ["aps/Container", {"id": common.genId('container'), title: _("Information")}, [
                output([
                    div(_("Administrators of the backup service can access the backup management console to:"), "info"),
                    ul([
                        _("manage all customers from the centralized interface"),
                        _("enable branding"),
                        _("register own cloud storage to keep the data of your customers in it")
                    ]),
                    div(_("Administrators can receive backup notifications and notifications about exceeded quotas."), "info")
                ].join(""))
            ]],
            ["aps/Container", {"id": common.genId('container'),title: _("Backup Administrators")}, [
                ["aps/Grid", {
                    id: "adminGrid",
                    selectionMode: "multiple",
                    apsResourceViewId: administratorViewViewId,
                    showPaging: false,
                    pageSizeOptions: "GRID_UNLIMITED",
                    columns: columns,
                    store: new Memory({data: administrators, idProperty: "id"})
                }, [
                    ["aps/Toolbar", {id: "adminToolbar"}, [
                        ["aps/ToolbarButton", {
                            id: "adminNew",
                            iconClass: "sb-add-new",
                            label: _("Add"),
                            disabled: !response.datacenters.length,
                            onClick: onAddNew
                        }],
                        ["aps/ToolbarSeparator"],
                        ["aps/ToolbarButton", {
                            id: "adminDelete",
                            iconClass: "sb-delete",
                            label: _("Remove"),
                            requireItems: true,
                            disabled: !response.datacenters.length,
                            onClick: onRemove
                        }]
                    ]]
                ]]
            ]]
        ]]).then(function () {
            messages.removeAll();
            if (!response.datacenters.length) {
                messages.displayMessage(_("The provider does not have configured datacenters."), "warning");
            }
            var data = wizardData.getWizardData(administratorNewViewId);
            if (data && data.newAdminEmail) {
                messages.displayMessage(_("The administrator account has been created. The activation message has been sent to email '{0}'. To activate the account, please follow the instructions from this message.").replace("{0}", data.newAdminEmail), "info");
                wizardData.resetWizardData(administratorNewViewId);
            }
        });
    }, messages.displayError);
});
