/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

/*
 * Methods used for calculating quotas and render quota editing controls.
 */
define([
    "aps/TextBox",
    "./js/common.js"
],
function (
    TextBox,
    common
) {
    function getQuotaControls(user, tenant, quotasStatistics, setToMax, userQuotaMode) {
        var controlsData = {};

        if (tenant.StorageUsage && tenant.StorageUsage.limit) {
            controlsData.storage = {
                id: "userStorageUsage",
                label: _("Cloud Storage (GB)"),
                quotaStatistic: quotasStatistics['storageUsage'],
                allocatedToUser: user && user.userStorageUsage && user.userStorageUsage.limit ? user.userStorageUsage.limit : 0,
                isStorage: true
            };
        }

        if (tenant.WorkstationCount && tenant.WorkstationCount.limit) {
            controlsData.workstations = {
                id: "userWorkstationCount",
                label: _("Physical workstations"),
                quotaStatistic: quotasStatistics['workstationCount'],
                allocatedToUser: user && user.UserWorkstationCount && user.UserWorkstationCount.limit ? user.UserWorkstationCount.limit : 0
            };
        }

        if (tenant.ServerCount && tenant.ServerCount.limit) {
            controlsData.servers = {
                id: "userServerCount",
                label: _("Physical servers"),
                quotaStatistic: quotasStatistics['serverCount'],
                allocatedToUser: user && user.userServerCount && user.userServerCount.limit ? user.userServerCount.limit : 0
            };
        }

        if (tenant.VmCount && tenant.VmCount.limit) {
            controlsData.vms = {
                id: "userVmCount",
                label: _("Virtual machines"),
                quotaStatistic: quotasStatistics['vmCount'],
                allocatedToUser: user && user.UserVmCount && user.UserVmCount.limit ? user.UserVmCount.limit : 0
            };
        }

        if (tenant.MobileDeviceCount && tenant.MobileDeviceCount.limit) {
            controlsData.mobiles = {
                id: "userMobileDeviceCount",
                label: _("Mobile devices"),
                quotaStatistic: quotasStatistics['mobileDeviceCount'],
                allocatedToUser: user && user.UserMobileDeviceCount && user.UserMobileDeviceCount.limit? user.UserMobileDeviceCount.limit : 0
            };
        }

        if (tenant.WebHostingServerCount && tenant.WebHostingServerCount.limit) {
            controlsData.webHostingServers = {
                id: "userWebHostingServerCount",
                label: _("Web Hosting Servers"),
                quotaStatistic: quotasStatistics['webHostingServerCount'],
                allocatedToUser: user && user.userWebHostingServerCount && user.userWebHostingServerCount.limit ? user.userWebHostingServerCount.limit : 0
            };
        }

        if (tenant.WebsiteCount && tenant.WebsiteCount.limit) {
            controlsData.websites = {
                id: "userWebsiteCount",
                label: _("Websites"),
                quotaStatistic: quotasStatistics['websiteCount'],
                allocatedToUser: user && user.UserWebsiteCount && user.UserWebsiteCount.limit ? user.UserWebsiteCount.limit : 0
            };
        }

        var controls = {};
        for (var controlType in controlsData) {
            if (!controlsData.hasOwnProperty(controlType)) {
                continue;
            }
            var config = controlsData[controlType];
            config.setToMax = setToMax;
            config.userQuotaMode = userQuotaMode;
            config.isStorage = config.isStorage || false;
            config.isForNewUser = !user;
            controls[controlType] = createTextBox(config);
        }
        return controls;
    }

    function createTextBox(config) {
        var allocatedTotal = config.quotaStatistic.allocatedTotal == common.QUOTA_UNLIMITED_VALUE ?
            common.QUOTA_UNLIMITED_VALUE : (
                config.isStorage ?
                    Math.ceil(config.quotaStatistic.allocatedTotal / common.BYTES_IN_GB) :
                    config.quotaStatistic.allocatedTotal
            );

        var allocatedToUser = config.allocatedToUser == common.QUOTA_UNLIMITED_VALUE ?
            common.QUOTA_UNLIMITED_VALUE : (
                config.isStorage ?
                    Math.ceil(config.allocatedToUser / common.BYTES_IN_GB) :
                    config.allocatedToUser
            );

        var availableToAllocate = common.QUOTA_UNLIMITED_VALUE,
            limit = common.QUOTA_UNLIMITED_VALUE;

        if (config.quotaStatistic.limit != common.QUOTA_UNLIMITED_VALUE &&
            config.userQuotaMode == common.USER_QUOTA.STRICT
        ) {
            limit = config.isStorage ?
                Math.floor(config.quotaStatistic.limit / common.BYTES_IN_GB) :
                config.quotaStatistic.limit;

            availableToAllocate = allocatedTotal != common.QUOTA_UNLIMITED_VALUE ?
                Math.max(limit - (allocatedTotal - allocatedToUser), 0) : 0;
        }

        var isAvailableToAllocateUnlimited = availableToAllocate == common.QUOTA_UNLIMITED_VALUE;

        function getQuotaHint() {
            var dataForText = {
                user_alloc: allocatedToUser < 0 ? _("Unlimited") : allocatedToUser,
                total_alloc: allocatedTotal < 0 ? _("Unlimited") : allocatedTotal,
                limit: limit < 0 ? _("Unlimited") : limit,
                available_alloc: availableToAllocate < 0 ? _("Unlimited") : availableToAllocate
            };

            if (isAvailableToAllocateUnlimited) {
                if (config.isForNewUser || allocatedToUser == common.QUOTA_UNLIMITED_VALUE) {
                    return _("Resource is not limited.");
                }
                return config.isStorage ?
                    _("__user_alloc__ GB currently allocated to this user.", dataForText) :
                    _("__user_alloc__ currently allocated to this user.", dataForText);
            }

            if (config.isForNewUser || !allocatedToUser) {
                return config.isStorage ?
                    _("__total_alloc__ GB of __limit__ GB allocated in total, __available_alloc__ GB available to the user.", dataForText) :
                    _("__total_alloc__ of __limit__ allocated in total, __available_alloc__ available to the user.", dataForText);
            }

            return config.isStorage ?
                _("__user_alloc__ GB currently allocated to this user. __total_alloc__ GB of __limit__ GB allocated in total, __available_alloc__ GB available to the user.", dataForText) :
                _("__user_alloc__ currently allocated to this user. __total_alloc__ of __limit__ allocated in total, __available_alloc__ available to the user.", dataForText);
        }

        var constraints = {
                min: 0
            },
            invalidMessage = _("The value must be between __min__ and __max__.", constraints);

        if (availableToAllocate !== common.QUOTA_UNLIMITED_VALUE) {
            constraints.max = availableToAllocate;
            invalidMessage = _("The value must not exceed {0}").replace("{0}", constraints.max);
        }

        return new TextBox({
            id: config.id,
            label: config.label,
            placeHolder: isAvailableToAllocateUnlimited ? _('Unlimited') : "",
            required: !isAvailableToAllocateUnlimited,
            value: !isAvailableToAllocateUnlimited && config.setToMax ? availableToAllocate : "",
            pattern: config.pattern || "^[0-9]+$",
            constraints: constraints,
            hint: getQuotaHint() + (config.extraHint ? ' ' + config.extraHint : ''),
            optionalText: "",
            missingMessage: _("This value is required."),
            invalidMessage: invalidMessage,
            validator: function(value, constraints){
                if (this.disabled || !this.required && value === "") {
                    return true;
                }
                if (this.required && value === "")  {
                    this.invalidMessage = _("This value is required.");
                    return false;
                }
                if (isNaN(value)) {
                    this.invalidMessage = _("Please enter a numeric value.");
                    return false;
                }
                if (this.required && (value < constraints.min || value > constraints.max)) {
                    this.invalidMessage = _("The value must not exceed {0}.").replace("{0}", constraints.max);
                    return false;
                }
                if (!value.match(this.pattern)) {
                    this.invalidMessage = _("Please enter a numeric value.");
                    return false;
                }
                return true;
            }
        });
    }

    function canShowQuotas(userQuota, account) {
        return userQuota != common.USER_QUOTA.OFF ||
            (account.userStorageUsage.limit != -1 && account.userStorageUsage.limit != null) ||
            (account.UserWorkstationCount.limit != -1 && account.UserWorkstationCount.limit != null) ||
            (account.userServerCount.limit != -1 && account.userServerCount.limit != null) ||
            (account.UserVmCount.limit != -1 && account.UserVmCount.limit != null) ||
            (account.UserMobileDeviceCount.limit  != -1 && account.UserMobileDeviceCount.limit != null) ||
            (account.userWebHostingServerCount.limit  != -1 && account.userWebHostingServerCount.limit != null) ||
            (account.UserWebsiteCount.limit  != -1 && account.UserWebsiteCount.limit != null);
    }

    function getQuotaDescription(userQuota) {
        switch (Number(userQuota)) {
            case common.USER_QUOTA.OFF:
                return _("Quota management for backup users is disabled.");
            case common.USER_QUOTA.STRICT:
            case common.USER_QUOTA.SOFT:
                return _("Specify the quotas for the backup user.");
            default:
                return "";
        }
    }

    return {
        getQuotaDescription: getQuotaDescription,
        canShowQuotas: canShowQuotas,
        getQuotaControlsForExistingUser: getQuotaControls,
        getQuotaControlsForNewUser: function () {
            var args = Array.prototype.slice.call(arguments);
            args.unshift(null);
            return getQuotaControls.apply(null, args);
        }
    };
});
