<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

require_once(__DIR__ . "/app/bootstrap.php");
require_once(__DIR__ . "/migration/ApsMigration.php");
require_once(__DIR__ . "/migration/CsvMigration.php");

chdir(__DIR__);
define('APS_RUNTIME_NO_PROCESS', true);

const RUNNING_MIGRATION_MESSAGE = "Running migration: group name %s in application with ID %s\n";
const COMMAND_LINE_PARAMETERS_ARE_NOT_SPECIFIED = "Error: command line parameters are not specified\n";
const APPLICATION_ID_IS_NOT_INTEGER = "Error: applicationId is not integer\n";
const APPLICATION_ID_SHOULD_BE_LARGER_THAN_ZERO = "Error: applicationId should be a number larger than zero\n";
const SETTINGS_ARE_NOT_VALID_MESSAGE = "Settings are not valid please edit ./app/config.php\n";
const PROPERTY = "[PROPERTY]";
const PROPERTY_IS_NOT_SET_MESSAGE = "MigrationSettings [PROPERTY] is not set\n";
const NOT_SET_VALUE = "NOT_SET";
const MIGRATE_COMMAND_LINE_EXAMPLE = "Example: sudo -H -u apache bash -c 'php %s -i 17 -d 1 -g testGroup -g testGroup2 -g testGroup3 -r 100001 >> /tmp/migration.log &' \n";
const BA_SERVER_DNS = 'BACKUPAGENT_SERVER_DNS';
const BA_SERVER_ADMIN_LOGIN = 'BACKUPAGENT_SERVER_ADMIN_LOGIN';
const BA_SERVER_ADMIN_PASSWORD = 'BACKUPAGENT_SERVER_ADMIN_PASSWORD';
const PARAMETERS_OFF = 'off';
const PARAMETERS_ON = 'on';
const OUTPUT_INCORRECT_OPTION_VALUE_ACTIVATE_USER = "Option activateUser must be equal to on/off \n";
const OUTPUT_OPTION_UNDEFINED = "Option '%s' undefined \n";
const OUTPUT_OPTION_DATACENTER_ID_NOT_EXIST = "Option Datacenter ID must be exist \n";

const OPTION_PARSE_CHAR = ':';
const OPTION_FILE = 'f';
const OPTION_APPLICATION_ID = 'i';
const OPTION_ACTIVATE_USER = 'a';
const OPTION_GROUP_NAME = 'g';
const OPTION_TERMINATE_STATUS = 't';
const OPTION_CONTINUE_STATUS = 'c';
const OPTION_COMPOSITE_RESOURCE_ID = 'r';
const OPTION_HELP = 'h';
const OPTION_DATACENTER_ID = 'd';

$opt = getopt(
    OPTION_FILE . OPTION_PARSE_CHAR .
    OPTION_APPLICATION_ID . OPTION_PARSE_CHAR .
    OPTION_ACTIVATE_USER . OPTION_PARSE_CHAR .
    OPTION_GROUP_NAME . OPTION_PARSE_CHAR .
    OPTION_TERMINATE_STATUS . OPTION_PARSE_CHAR .
    OPTION_CONTINUE_STATUS . OPTION_PARSE_CHAR .
    OPTION_DATACENTER_ID . OPTION_PARSE_CHAR .
    OPTION_COMPOSITE_RESOURCE_ID . OPTION_PARSE_CHAR .
    OPTION_HELP
);

if (isset($opt[OPTION_HELP])) {
    showUsage();
    return;
}

$file = getOptions(OPTION_FILE, $opt);
$terminateStatuses = getOptions(OPTION_TERMINATE_STATUS, $opt);
$continueStatuses = getOptions(OPTION_CONTINUE_STATUS, $opt);

try {
    $apiFactory = new ApiFactory();
    if (!$file) {
        if (!validateCommandLineParameters($opt) || !validateSettings()) {
            return;
        }

        $applicationId = getOptions(OPTION_APPLICATION_ID, $opt);
        $activateUser = getOptions(OPTION_ACTIVATE_USER, $opt);
        $groupNames = getOptions(OPTION_GROUP_NAME, $opt);
        $dataCenterID = getOptions(OPTION_DATACENTER_ID, $opt);
        $compositeResourceIDs = getOptions(OPTION_COMPOSITE_RESOURCE_ID, $opt);

        $apsMigration = new ApsMigration($apiFactory, $continueStatuses, $terminateStatuses);
        $apsMigration->migrate($groupNames, $applicationId, $activateUser, $compositeResourceIDs, $dataCenterID);
    } else {
        $apsMigration = new CsvMigration($apiFactory, $continueStatuses, $terminateStatuses);
        $apsMigration->migrate($file);
    }
} catch (Exception $e ) {
    echo sprintf('Error: %s', $e->getMessage()) . PHP_EOL;
}

/**
 *  Validate command line parameters
 * @param array $options
 * @return boolean
 */
function validateCommandLineParameters($options)
{
    if (empty($options[OPTION_APPLICATION_ID]) || empty($options[OPTION_GROUP_NAME])) {
        echo COMMAND_LINE_PARAMETERS_ARE_NOT_SPECIFIED;
        showUsage();
        return false;
    }

    if (!is_numeric($options[OPTION_APPLICATION_ID])) {
        echo APPLICATION_ID_IS_NOT_INTEGER;
        showUsage();
        return false;
    }

    if (intval($options[OPTION_APPLICATION_ID]) <= 0) {
        echo APPLICATION_ID_SHOULD_BE_LARGER_THAN_ZERO;
        showUsage();
        return false;
    }

    if (isset($options[OPTION_ACTIVATE_USER]) && $options[OPTION_ACTIVATE_USER] != PARAMETERS_OFF && $options[OPTION_ACTIVATE_USER] != PARAMETERS_ON) {
        echo OUTPUT_INCORRECT_OPTION_VALUE_ACTIVATE_USER;
        showUsage();
        return false;
    }

    if (!isset($options[OPTION_DATACENTER_ID])) {
        echo OUTPUT_OPTION_DATACENTER_ID_NOT_EXIST;
        showUsage();
        return false;
    }

    return true;
}

/**
 *  Validate settings in config.php file
 * @return boolean
 */
function validateSettings()
{
    $errorMessage = SETTINGS_ARE_NOT_VALID_MESSAGE;

    if (!validateStringSetting(BACKUPAGENT_SERVER_DNS, BA_SERVER_DNS, $errorMessage)) {
        return false;
    }
    if (!validateStringSetting(BACKUPAGENT_SERVER_ADMIN_LOGIN, BA_SERVER_ADMIN_LOGIN, $errorMessage)) {
        return false;
    }

    if (!validateStringSetting(BACKUPAGENT_SERVER_ADMIN_PASSWORD, BA_SERVER_ADMIN_PASSWORD, $errorMessage)) {
        return false;
    }

    return true;
}

/**
 *  Validate string setting from config file
 * @param string $value
 * @param string $name
 * @param string $pleaseEditConfigFileMessage
 * @return boolean
 */
function validateStringSetting($value, $name, $pleaseEditConfigFileMessage)
{
    $valueIsNotSetMessage = PROPERTY_IS_NOT_SET_MESSAGE;
    if (!isset($value) || $value == NOT_SET_VALUE) {
        echo $pleaseEditConfigFileMessage;
        echo str_replace(PROPERTY, $name, $valueIsNotSetMessage);

        return false;
    }

    return true;
}

/**
 *  Prints usage to the screen
 */
function showUsage()
{
    $fileName = basename($_SERVER['PHP_SELF']);

    echo "Usage: sudo -H -u apache bash -c 'php {$fileName} -i [APPLICATION_ID] -d [DATACENTER_ID] -g [group1] -g [group2] -g [groupN] -r [COMPOSITE_RESOURCE_ID] >> /tmp/migration.log &'

Options:
-f, CSV file
-h, Show this help message and exit
-i, Backup Agent Application ID
-d, Datacenter ID
-c, optional, continue status of upgrade resource order
-t, optional, terminate status of upgrade resource order
-r, optional, ID of composite resource
-g, group name in Backup Agent server \n";
    echo sprintf(MIGRATE_COMMAND_LINE_EXAMPLE, $fileName);
    echo PHP_EOL;
}

/**
 * @param $optionName
 * @param $options
 * @return null
 */
function getOptions($optionName, $options)
{
    $option = null;

    switch ($optionName) {
        case OPTION_FILE:
            $option = isset($options[OPTION_FILE]) ? trim($options[OPTION_FILE]) : false;
            break;
        case OPTION_APPLICATION_ID:
            $option = intval($options[OPTION_APPLICATION_ID]);
            break;
        case OPTION_ACTIVATE_USER:
            $option = (isset($options[OPTION_ACTIVATE_USER]) && $options[OPTION_ACTIVATE_USER] == PARAMETERS_OFF) ? false : true;
            break;
        case OPTION_TERMINATE_STATUS:
            $value = (isset($options[OPTION_TERMINATE_STATUS])) ? $options[OPTION_TERMINATE_STATUS] : ApsMigration::DEFAULT_TERMINATE_ORDER_STATUS;
            $option = (is_array($value)) ? $value : array($value);
            break;
        case OPTION_CONTINUE_STATUS:
            $value = (isset($options[OPTION_CONTINUE_STATUS])) ? $options[OPTION_CONTINUE_STATUS] : ApsMigration::DEFAULT_CONTINUE_ORDER_STATUS;
            $option = (is_array($value)) ? $value : array($value);
            break;
        case OPTION_GROUP_NAME:
            $option = (is_array($options[OPTION_GROUP_NAME])) ? $options[OPTION_GROUP_NAME] : array($options[OPTION_GROUP_NAME]);
            break;
        case OPTION_DATACENTER_ID:
            $option = $options[OPTION_DATACENTER_ID];
            break;
        case OPTION_COMPOSITE_RESOURCE_ID:
            $value =  (isset($options[OPTION_COMPOSITE_RESOURCE_ID])) ? $options[OPTION_COMPOSITE_RESOURCE_ID] : array();
            $option = (is_array($value)) ? $value : array($options[OPTION_COMPOSITE_RESOURCE_ID]);
            break;
    }

    if (is_null($option)) {
        echo sprintf(OUTPUT_OPTION_UNDEFINED, $optionName);
        showUsage();
        return false;
    }
    return $option;
}
