<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\CloudApi;

use Acronis\App\Utils\CacheableTrait;
use Acronis\App\Utils\MemoizeTrait;
use Arr;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Exception\GuzzleException;
use Str;
use Swagger\Client\ApiException;
use Swagger\Client\HttpException;
use Swagger\Client\IOException;
use Swagger\Client\Model\Applications\Application;
use Swagger\Client\Model\Infra\Infra;
use Swagger\Client\Model\OfferingItems\OfferingItemOutput;
use Swagger\Client\Model\OfferingItems\Quota;
use Swagger\Client\Model\Reports\Report;
use Swagger\Client\Model\Reports\ReportPost;
use Swagger\Client\Model\Reports\ReportPostKind;
use Swagger\Client\Model\Reports\ReportPostLevel;
use Swagger\Client\Model\Reports\ReportPostParameters;
use Swagger\Client\Model\Reports\ReportPostResultAction;
use Swagger\Client\Model\Reports\ReportPostSchedule;
use Swagger\Client\Model\Reports\ReportPostType;
use Swagger\Client\Model\Reports\Stored\StoredReportParamsItems;
use Swagger\Client\Model\Tenants\OfferingItem;
use Swagger\Client\Model\Tenants\OfferingItemsPut;
use Swagger\Client\Model\Tenants\Tenant;
use Swagger\Client\Model\Usages\UsageOutput;
use Swagger\Client\Model\Users\User;

class Api extends AuthorizedApi implements ApiInterface
{
    use CacheableTrait;
    use MemoizeTrait;

    const CACHE_TTL = 7200;
    const URL_PATH_API_V1 = '/api/1';
    const URL_PATH_API_V2 = '/api/2';

    /** @var string */
    private $url;

    /** @var string */
    private $serverUrl;

    /** @var string */
    private $login;

    /** @var string */
    private $password;

    public function __construct($url, $login, $password)
    {
        $this->url = $url;
        $this->login = $login;
        $this->password = $password;
    }

    /**
     * @param UsageOutput|OfferingItem|OfferingItemOutput $offeringItem
     * @return string
     */
    public static function getOfferingItemHash($offeringItem)
    {
        return Str::format(
            '%s:%s',
            $offeringItem->getName(), $offeringItem->hasInfraId() ? $offeringItem->getInfraId() : ''
        );
    }

    /**
     * @param UsageOutput[]|OfferingItem[]|OfferingItemOutput[] $offeringItems
     * @return array
     */
    public static function getOfferingItemsHashTable(array $offeringItems)
    {
        return Arr::map(
            $offeringItems,
            function ($offeringItem) {
                return static::getOfferingItemHash($offeringItem);
            },
            function ($offeringItem) {
                return $offeringItem;
            }
        );
    }

    /**
     * @return string
     */
    public function getUrl()
    {
        return $this->url;
    }

    /**
     * @return string
     */
    public function getLogin()
    {
        return $this->login;
    }

    /**
     * @return string
     */
    public function getPassword()
    {
        return $this->password;
    }

    /* ====== Users API ====== */

    /**
     * @return User
     * @throws ApiException
     * @throws HttpException
     */
    public function getMe()
    {
        return $this->memoize(function () {
            return $this->authorizedCall(function () {
                return $this->getUsersApi()->getUsersMe();
            });
        });
    }

    /**
     * Returns True if the login exists otherwise False
     *
     * @param string $login
     * @return bool
     * @throws ApiException
     * @throws HttpException
     */
    public function checkLogin($login)
    {
        return $this->authorizedCall(function () use ($login) {
            try {
                $this->getUsersApi()->getUsersCheckLogin($login);

                return false;
            } catch (HttpException $e) {
                if ($e->getCode() != 409) {
                    throw $e;
                }

                return true;
            }
        });
    }

    /* ====== Methods to manage tenants ====== */

    /**
     * @param string $tenantId
     * @return Tenant
     * @throws HttpException
     * @throws ApiException
     */
    public function getTenant($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsByTenantId($tenantId);
        });
    }

    /**
     * @param $rootTenantId
     * @param $leafTenantId
     * @return string[] a list of UUIDs
     * @throws ApiException
     * @throws HttpException
     */
    public function resolveTenantPath($rootTenantId, $leafTenantId)
    {
        $tenantId = $leafTenantId;
        $tenantPath = [$tenantId];
        while ($tenantId !== $rootTenantId) {
            $tenantData = $this->getTenant($tenantId);
            $tenantId = $tenantData->getParentId();

            $tenantPath[] = $tenantId;
        }

        return $tenantPath;
    }

    /**
     * @param string $tenantId
     * @return UsageOutput[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantUsages($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsUsagesByTenantId($tenantId);
        })->getItems();
    }

    /* ====== Methods to manage tenant's applications ====== */

    /**
     * @return Application[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getApplications()
    {
        return $this->memoize(function () {
            return $this->authorizedCall(function () {
                return $this->getApplicationsApi()->getApplications();
            })->getItems();
        });
    }

    /**
     * @param string $type
     * @return Application|null
     * @throws ApiException
     * @throws HttpException
     */
    public function getApplicationByType($type)
    {
        $applications = array_filter($this->getApplications(), function ($application) use ($type) {
            /** @var Application $application */
            return $application->getType() === $type;
        });

        return reset($applications) ?: null;
    }

    /**
     * @param string $tenantId
     * @return string[] a list of UUIDs
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantApplicationsUuids($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsApplicationsByTenantId($tenantId);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @return bool
     * @throws ApiException
     * @throws HttpException
     */
    public function isApplicationEnabledForTenant($tenantId, $applicationId)
    {
        return in_array($applicationId, $this->getTenantApplicationsUuids($tenantId));
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @throws ApiException
     * @throws HttpException
     */
    public function enableApplicationForTenant($tenantId, $applicationId)
    {
        $this->authorizedCall(function () use ($applicationId, $tenantId) {
            return $this->getApplicationsApi()->postApplicationsBindingsTenants($applicationId, $tenantId);
        });
    }

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @throws ApiException
     * @throws HttpException
     */
    public function disableApplicationForTenant($tenantId, $applicationId)
    {
        $this->authorizedCall(function () use ($applicationId, $tenantId) {
            return $this->getApplicationsApi()->deleteApplicationsBindingsTenants($applicationId, $tenantId);
        });
    }

    /* ====== Methods to manage offering items ====== */

    /**
     * @param string $tenantId
     * @param string $applicationId
     * @return OfferingItemOutput[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantApplicationOfferingItems($tenantId, $applicationId)
    {
        return $this->authorizedCall(function () use ($tenantId, $applicationId) {
            return $this->getTenantsApi()->getTenantsApplicationsOfferingItems($tenantId, $applicationId);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @param OfferingItem[] $offeringItems
     * @return OfferingItemOutput[]
     * @throws ApiException
     * @throws HttpException
     */
    public function updateTenantApplicationOfferingItems($tenantId, $offeringItems)
    {
        $body = new OfferingItemsPut();
        $body->setOfferingItems($offeringItems);

        return $this->authorizedCall(function () use ($tenantId, $body) {
            return $this->getTenantsApi()->putTenantsOfferingItemsByTenantId($tenantId, $body);
        })->getItems();
    }

    /**
     * Recursive enable an application for a hierarchy of tenants except root tenant.
     *
     * @param string $rootTenantId
     * @param string $leafTenantId
     * @param string $applicationId
     * @throws ApiException
     * @throws HttpException
     */
    public function enableApplicationForTenantsChain($rootTenantId, $leafTenantId, $applicationId)
    {
        if ($leafTenantId === $rootTenantId) {
            return;
        }

        if ($this->isApplicationEnabledForTenant($leafTenantId, $applicationId)) {
            return;
        }

        // enable for the parent
        $tenant = $this->getTenant($leafTenantId);
        $this->enableApplicationForTenantsChain($rootTenantId, $tenant->getParentId(), $applicationId);
        // enable for the tenant
        $this->enableApplicationForTenant($leafTenantId, $applicationId);
    }

    /**
     * Recursive enable specified offering items with unlimited quotas for a hierarchy of tenants except root tenant.
     * Note: Enabled or not specified offering items won't be affected
     *
     * @param string $rootTenantId
     * @param string $leafTenantId
     * @param string $applicationId
     * @param OfferingItem[] $offeringItems
     * @throws ApiException
     * @throws HttpException
     */
    public function enableApplicationOfferingItemsForTenantsChain(
        $rootTenantId,
        $leafTenantId,
        $applicationId,
        array $offeringItems
    ) {
        if ($leafTenantId === $rootTenantId) {
            return;
        }

        // enable only OIs with status 1, other will be ignored
        $neededOfferingItems = static::getOfferingItemsHashTable(
            $this->getActiveOfferingItems($offeringItems)
        );
        $availableOfferingItems = static::getOfferingItemsHashTable(
            $this->getTenantApplicationOfferingItems($leafTenantId, $applicationId)
        );

        // Enable not available offering items for the parent
        $notAvailableOfferingItems = array_diff_key($neededOfferingItems, $availableOfferingItems);
        if (!empty($notAvailableOfferingItems)) {
            $tenant = $this->getTenant($leafTenantId);
            $this->enableApplicationOfferingItemsForTenantsChain(
                $rootTenantId, $tenant->getParentId(), $applicationId, $notAvailableOfferingItems
            );
            $availableOfferingItems = static::getOfferingItemsHashTable(
                $this->getTenantApplicationOfferingItems($leafTenantId, $applicationId)
            );
        }

        // Exist if all needed offering items are enabled
        $enabledOfferingItems = array_filter($availableOfferingItems, function ($offeringItem) {
            /** @var OfferingItem $offeringItem */
            return $offeringItem->getStatus();
        });
        $disabledOfferingItems = array_diff_key($neededOfferingItems, $enabledOfferingItems);
        if (empty($disabledOfferingItems)) {
            return;
        }

        // Enable disabled offering items
        $offeringItemsToUpdate = [];
        foreach ($disabledOfferingItems as $offeringItemHash => $neededOfferingItem) {
            /** @var OfferingItem $neededOfferingItem */

            $oi = new OfferingItem();
            $oi->setApplicationId($applicationId);
            $oi->setName($neededOfferingItem->getName());
            $oi->setStatus(static::OFFERING_ITEM_STATUS_ACTIVE);

            if ($neededOfferingItem->hasInfraId()) {
                $oi->setInfraId($neededOfferingItem->getInfraId());
            }

            $offeringItemsToUpdate [] = $oi;
        }

        // Sent request to update offering items
        $this->updateTenantApplicationOfferingItems($leafTenantId, $offeringItemsToUpdate);
    }

    /**
     * Update specified only offering items for $leafTenantId.
     * Recursive enable offering items for tenant's parent (see method enableUnlimitedOfferingItemsForTenantsChain)
     *
     * @param string $rootTenantId
     * @param string $leafTenantId
     * @param string $applicationId
     * @param OfferingItem[] $offeringItems
     * @throws ApiException
     * @throws HttpException
     */
    public function updateApplicationOfferingItemsForTenantsChain(
        $rootTenantId,
        $leafTenantId,
        $applicationId,
        array $offeringItems
    ) {
        if ($leafTenantId === $rootTenantId) {
            throw new HttpException(Str::format(
                'Unable to update offers for yourself. Root tenant ID equals leaf tenant ID %s.',
                $leafTenantId
            ), 403);
        }

        $neededOfferingItems = static::getOfferingItemsHashTable($offeringItems);
        $availableOfferingItems = static::getOfferingItemsHashTable(
            $this->getTenantApplicationOfferingItems($leafTenantId, $applicationId)
        );

        // enable not available offering items for tenant's parent if there are such
        $notAvailableOfferingItems = array_diff_key(
            $this->getActiveOfferingItems($neededOfferingItems),
            $availableOfferingItems
        );
        if (!empty($notAvailableOfferingItems)) {
            $tenant = $this->getTenant($leafTenantId);
            $this->enableApplicationOfferingItemsForTenantsChain(
                $rootTenantId, $tenant->getParentId(), $applicationId, $notAvailableOfferingItems
            );

            // refresh a list of available offering items
            $availableOfferingItems = static::getOfferingItemsHashTable(
                $this->getTenantApplicationOfferingItems($leafTenantId, $applicationId)
            );
        }

        $offeringItemsToUpdate = [];
        foreach ($neededOfferingItems as $offeringItemHash => $offeringItem) {
            /** @var OfferingItem $offeringItem */
            $status = $offeringItem->hasStatus()
                ? $offeringItem->getStatus()
                : static::OFFERING_ITEM_STATUS_ACTIVE;

            $oi = new OfferingItem();
            $oi->setStatus($status);
            $oi->setApplicationId($applicationId);
            $oi->setName($offeringItem->getName());

            if ($offeringItem->hasInfraId()) {
                $oi->setInfraId($offeringItem->getInfraId());
            }

            if ($status !== static::OFFERING_ITEM_STATUS_INACTIVE && $offeringItem->hasQuota()) {
                $offeringItemQuota = $offeringItem->getQuota();

                $quota = new Quota();
                $quota->setValue($offeringItemQuota->getValue());
                $quota->setOverage($offeringItemQuota->getOverage());

                $quota->setVersion(0);

                // set current version of quota
                /** @var OfferingItem $availableOfferingItem */
                $availableOfferingItem = Arr::get($availableOfferingItems, $offeringItemHash);

                if ($availableOfferingItem && $availableOfferingItem->hasQuota()) {
                    $availableOfferingItemQuota = $availableOfferingItem->getQuota();
                    if ($availableOfferingItemQuota->hasVersion()) {
                        $quota->setVersion($availableOfferingItemQuota->getVersion());
                    }
                }

                $oi->setQuota($quota);
            }

            $offeringItemsToUpdate[] = $oi;
        }

        // sent request to update offering items
        $this->updateTenantApplicationOfferingItems($leafTenantId, $offeringItemsToUpdate);
    }

    /**
     * @param string $offeringItemName
     * @return bool
     */
    public function isOfferingItemForCustomer($offeringItemName)
    {
        static $hashtable = null;
        if (is_null($hashtable)) {
            $hashtable = array_flip(static::CUSTOMER_OFFERING_ITEMS);
        }

        return isset($hashtable[$offeringItemName]);
    }

    /**
     * @param string $offeringItemName
     * @return bool
     */
    public function isOfferingItemForPartner($offeringItemName)
    {
        static $hashtable = null;
        if (is_null($hashtable)) {
            $hashtable = array_flip(static::PARTNER_OFFERING_ITEMS);
        }

        return isset($hashtable[$offeringItemName]);
    }

    /**
     * @param string $offeringItemName
     * @return bool
     */
    public function isOfferingItemForUser($offeringItemName)
    {
        static $hashtable = null;
        if (is_null($hashtable)) {
            $hashtable = array_flip(static::USER_OFFERING_ITEMS);
        }

        return isset($hashtable[$offeringItemName]);
    }

    /* ====== Methods to manage reports ====== */

    /**
     * @param string $reportId
     * @return StoredReportParamsItems[]
     * @throws ApiException
     * @throws HttpException
     */
    public function getStoredUsageReports($reportId)
    {
        return $this->authorizedCall(function () use ($reportId) {
            return $this->getReportsApi()->getReportsStoredByReportId($reportId);
        })->getItems();
    }

    /**
     * @param string $reportId
     * @param string $storedReportId
     * @param string $destinationFile
     * @throws ApiException
     * @throws HttpException
     */
    public function downloadStoredUsageReportArchive($reportId, $storedReportId, $destinationFile)
    {
        $this->authorizedCall(function () use ($reportId, $storedReportId, $destinationFile) {
            $resourcePath = Str::format('/reports/%s/stored/%s', $reportId, $storedReportId);
            $apiClientConfig = $this->getApiClient()->getConfig();
            $url = $apiClientConfig->getHost() . $resourcePath;
            $headers = [
                'Authorization' => 'Bearer ' . $apiClientConfig->getAccessToken(),
                'User-Agent' => $apiClientConfig->getUserAgent(),
            ];
            $this->getLogger()->info(Str::format(
                'Download file "%s" to "%s".',
                $url, $destinationFile
            ));
            $this->downloadAndSaveFile($url, $destinationFile, $headers);
        });
    }

    /**
     * @param string $tenantId
     * @return Report
     * @throws ApiException
     * @throws HttpException
     */
    public function orderCurrentUsageReportForAccounts($tenantId)
    {
        $params = new ReportPostParameters();
        $params->setTenantId($tenantId);
        $params->setKind(ReportPostKind::CURRENT);
        $params->setLevel(ReportPostLevel::ACCOUNTS);

        $schedule = new ReportPostSchedule();
        $schedule->setType(ReportPostType::ONCE);

        $body = new ReportPost();
        $body->setParameters($params);
        $body->setSchedule($schedule);
        $body->setRecipients([]);
        $body->setResultAction(ReportPostResultAction::SAVE);

        return $this->authorizedCall(function () use ($body) {
            return $this->getReportsApi()->postReports($body);
        });
    }

    /* ====== Methods to manage infra objects ====== */

    /**
     * @param string $tenantId
     * @return string[] a list of UUIDs
     * @throws ApiException
     * @throws HttpException
     */
    public function getTenantInfrasUuids($tenantId)
    {
        return $this->authorizedCall(function () use ($tenantId) {
            return $this->getTenantsApi()->getTenantsInfraByTenantId($tenantId);
        })->getInfras();
    }

    /**
     * @param string[] $uuids
     * @return Infra[]
     * @throws ApiException
     * @throws HttpException
     */
    public function fetchInfras($uuids)
    {
        return $this->authorizedCall(function () use ($uuids) {
            return $this->getInfraApi()->getInfra($uuids);
        })->getItems();
    }

    /**
     * @param string $tenantId
     * @return Infra[]
     * @throws ApiException
     * @throws HttpException
     */
    public function fetchTenantInfras($tenantId)
    {
        $uuids = $this->getTenantInfrasUuids($tenantId);
        if (empty($uuids)) {
            return [];
        }

        return $this->fetchInfras($uuids);
    }

    /**
     * @param $rootTenantId
     * @param $leafTenantId
     * @return string[] a list of UUIDs
     * @throws ApiException
     * @throws HttpException
     */
    public function getInfrasUuidsForTenantsChain($rootTenantId, $leafTenantId)
    {
        $tenantPath = $this->resolveTenantPath($rootTenantId, $leafTenantId);

        $infras = [];
        foreach ($tenantPath as $tenantId) {
            $infras = array_merge($infras, $this->getTenantInfrasUuids($tenantId));
        }

        return array_unique($infras);
    }

    /**
     * @param $rootTenantId
     * @param $leafTenantId
     * @return Infra[]
     * @throws ApiException
     * @throws HttpException
     */
    public function fetchInfrasForTenantsChain($rootTenantId, $leafTenantId)
    {
        $uuids = $this->getInfrasUuidsForTenantsChain($rootTenantId, $leafTenantId);

        return $this->fetchInfras($uuids);
    }

    /* ====== Protected utility methods ====== */

    /**
     * @return string
     * @throws HttpException
     * @throws IOException
     */
    protected function getServerUrl()
    {
        if (!$this->serverUrl) {
            $this->serverUrl = $this->resolveServerUrlForLogin();
            $this->serverUrl .= static::URL_PATH_API_V2;
        }

        return $this->serverUrl;
    }

    /* ====== Private methods ====== */

    /**
     * @return string
     * @throws HttpException
     * @throws IOException
     */
    private function resolveServerUrlForLogin()
    {
        return $this->fromCache(function () {
            $accountInfo = $this->fetchAccountInfo($this->login);

            return rtrim($accountInfo->server_url, '/');
        }, static::CACHE_TTL, false, $this->getCredentialsHash());
    }

    /**
     * @param string $login
     * @return object
     * @throws HttpException
     * @throws IOException
     */
    private function fetchAccountInfo($login)
    {
        $url = rtrim($this->url, '/') . static::URL_PATH_API_V1;

        $apiClient = $this->createApiClient($url);

        $path = '/accounts';
        $method = 'GET';
        $queryParameters = ['login' => $login];
        $body = null;
        $headers = [];

        list($response, $statusCode, $httpHeader) = $apiClient->callApi(
            $path, $method, $queryParameters, $body, $headers
        );

        return $response;
    }

    /**
     * @param string $url
     * @param string $destinationFile
     * @param array $headers
     * @throws IOException
     * @throws HttpException
     */
    private function downloadAndSaveFile($url, $destinationFile, $headers)
    {
        $client = new Client();

        $fp = fopen($destinationFile, 'wb');
        try {
            $client->request('GET', $url, [
                'sink' => $fp,
                'headers' => $headers,
                'verify' => false,
                'decode_content' => false
            ]);
        } catch (BadResponseException $e) {
            throw new HttpException($e->getMessage(), $e->getCode());
        } catch (GuzzleException $e) {
            throw new IOException($e->getMessage(), $e->getCode());
        }
    }

    protected function getCredentialsHash()
    {
        return md5($this->getUrl() . $this->getLogin() . $this->getPassword());
    }

    /**
     * @param OfferingItem[] $offeringItems
     * @return array
     */
    private function getActiveOfferingItems(array $offeringItems)
    {
        return array_filter($offeringItems, function ($offeringItem) {
            /** @var OfferingItem $offeringItem */
            // if status is not specified the offering item will be enabled
            if (!$offeringItem->hasStatus()) {
                return true;
            }

            return $offeringItem->getStatus() === static::OFFERING_ITEM_STATUS_ACTIVE;
        });
    }
}