<?php
/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Console\Command\UsageReport;

use Acronis\App\Console\Command\ApsInstanceCommand;
use Acronis\App\Console\Command\LockableTrait;
use Acronis\App\Database\Model\ReportEntry;
use Acronis\App\Database\Model\ReportEntryHelper;
use Acronis\App\Services\Logger\AppLoggerAwareTrait;
use Acronis\App\Services\UsageReportManager\UsageReportManagerAwareTrait;
use Acronis\App\UsageReport\Exception\ManagerException;
use Acronis\App\UsageReport\Manager;
use datacenter;
use Exception;
use Rest\RestException;
use Str;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Time;

/**
 * @package Acronis\App\Console\Command\UsageReport
 */
final class Process extends ApsInstanceCommand
{
    const COMMAND_NAME = 'usage-report:process';

    use UsageReportManagerAwareTrait;
    use AppLoggerAwareTrait;
    use LockableTrait;

    const ARGUMENT_DATACENTER = 'datacenterApsId';
    const ARGUMENT_DATACENTER_DESCRIPTION = 'APS ID (UUID) of datacenter resource.';

    /** @var datacenter */
    private $datacenter;

    /** @var Manager */
    private $reportManager;

    protected function configure()
    {
        parent::configure();

        $this->addArgument(
            static::ARGUMENT_DATACENTER,
            InputArgument::REQUIRED,
            static::ARGUMENT_DATACENTER_DESCRIPTION
        );
    }

    protected function getCommandName()
    {
        return static::COMMAND_NAME;
    }

    protected function getCommandDescription()
    {
        return 'Command for usage report processing.';
    }

    protected function getCommandHelp()
    {
        return <<<TEXT
<info>
This command orders an usage report, downloads and aggregates it.
The results of the aggregation are stored in the key-value storage.
</info>
TEXT;
    }

    /* ====== Main business logic of this command ====== */

    /**
     * @param InputInterface $input
     * @param OutputInterface $output
     * @return int|null|void
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * @throws Exception
     */
    protected function execute(InputInterface $input, OutputInterface $output)
    {
        parent::execute($input, $output);

        $datacenterApsId = $this->getArgumentDatacenterApsId($input);

        $this->executeWithLock(function () use ($datacenterApsId) {
            $this->processDatacenterReport($datacenterApsId);
        }, [$this->getApsInstanceId(), $datacenterApsId]);
    }

    private function processDatacenterReport($datacenterApsId)
    {
        $reportEntry = $this->getOrCreateCurrentReportEntry($datacenterApsId);

        if (!$this->reportStatusIsOrdered($reportEntry)) {
            $this->orderReport($reportEntry);
        } else {
            $this->getLog()->notice('Skipping report ordering.');
        }

        if (!$this->reportStatusIsReady($reportEntry) && !$this->waitReportIsReady($reportEntry)) {
            if ($this->reportStatusIsError($reportEntry)) {
                $this->getLog()->error('Report generation failed or something went wrong.');
                return;
            }

            $this->getLog()->notice('Report is not ready yet.');
            return;
        }

        if (!$this->reportStatusIsDownloaded($reportEntry)) {
            $this->downloadReport($reportEntry);
        }

        if (!$this->reportStatusIsDownloaded($reportEntry)) {
            $this->getLog()->notice('Report was not downloaded.');
            return;
        }

        if (!$this->reportStatusIsAggregated($reportEntry)) {
            $this->aggregateReport($reportEntry);
        }
    }

    private function getArgumentDatacenterApsId(InputInterface $input)
    {
        $datacenterApsId = $input->getArgument(static::ARGUMENT_DATACENTER);

        //todo validate and throw exception
        return $datacenterApsId;
    }

    /**
     * @param ReportEntry $reportEntry
     * @throws Exception
     * @return string
     */
    private function waitReportIsReady($reportEntry)
    {
        $this->getLog()->notice('Waiting for stored report.');
        $numRetries = $this->getRetriesLimit();
        $timeout = $this->getRetryTimeout();

        for ($retries = 0; $retries < $numRetries; $retries++) {
            $this->checkForStoredReport($reportEntry, $retries);

            if ($this->reportStatusIsError($reportEntry)) {
                return false;
            }

            if ($this->reportStatusIsReady($reportEntry)) {
                $this->getLog()->notice('Found stored report.');
                return true;
            }

            $this->getLog()->notice('Waiting %s seconds before next retry.', $timeout);
            sleep($timeout);
        }

        return false;
    }

    /* ====== Status check functions ====== */

    /**
     * Checks if report is already downloaded (status is equals or greater than REPORT_STATUS_DOWNLOADED)
     * @param ReportEntry $reportEntry
     * @return bool
     */
    private function reportStatusIsDownloaded($reportEntry)
    {
        return $reportEntry->status >= ReportEntry::REPORT_STATUS_DOWNLOADED;
    }

    /**
     * @param ReportEntry $reportEntry
     * @return bool
     */
    private function reportStatusIsAggregated($reportEntry)
    {
        return $reportEntry->status == ReportEntry::REPORT_STATUS_AGGREGATED;
    }

    /**
     * Checks if report is ready for downloading (status is equals or greater than REPORT_STATUS_READY)
     * @param ReportEntry $reportEntry
     * @return bool
     */
    private function reportStatusIsReady($reportEntry)
    {
        return $reportEntry->status >= ReportEntry::REPORT_STATUS_READY;
    }

    /**
     * Checks if report generation failed or something went wrong
     * @param ReportEntry $reportEntry
     * @return bool
     */
    private function reportStatusIsError($reportEntry)
    {
        return $reportEntry->status == ReportEntry::REPORT_STATUS_ERROR;
    }

    /**
     * Checks if report is already ordered (status is equals or greater than REPORT_STATUS_ORDERED)
     * @param ReportEntry $reportEntry
     * @return bool
     */
    private function reportStatusIsOrdered($reportEntry)
    {
        return $reportEntry->status >= ReportEntry::REPORT_STATUS_ORDERED;
    }

    /* ====== Wrappers of main business logic methods ====== */

    /**
     * @param $reportEntry
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * @throws ManagerException
     */
    private function orderReport($reportEntry)
    {
        $this->getLog()->notice('Ordering report.');
        $this->getManager()->orderReport($reportEntry);
    }

    /**
     * @param ReportEntry $reportEntry
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * @throws ManagerException
     */
    private function downloadReport($reportEntry)
    {
        $this->getLog()->notice('Started report downloading.');
        $this->getManager()->downloadReport($reportEntry);
        $this->getLog()->notice(
            'Finished report downloading. Saved to "%s".',
            $reportEntry->file_path
        );
    }

    /**
     * @param ReportEntry $reportEntry
     * @param int $retries
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     * @throws ManagerException
     */
    private function checkForStoredReport($reportEntry, $retries = 0)
    {
        $retriesStr = !$retries ? '' : Str::format(' (%s retry)', $retries);
        $this->getLog()->notice(
            'Getting list of stored reports from Acronis Data Cloud%s.',
            $retriesStr
        );
        $this->getManager()->checkForStoredReport($reportEntry);
    }

    /**
     * @param ReportEntry $reportEntry
     * @throws \Exception
     */
    private function aggregateReport($reportEntry)
    {
        $this->getLog()->notice('Started usage report aggregation.');
        $aggregator = $this->getManager()->getAggregator($reportEntry, $this->getApsInstanceId());
        $this->getManager()->aggregateReport($aggregator, $reportEntry);
        $this->getLog()->notice('Finished usage report aggregation.');
    }

    /* ====== Utility functions ======= */

    private function getOrCreateCurrentReportEntry($datacenterApsId)
    {
        $date = Time::getCurentDate();
        $reportHelper = new ReportEntryHelper();

        return $reportHelper->getOrCreateReport($datacenterApsId, $date, $this->getApsInstanceId());
    }

    /**
     * Call after parent::execute()
     * @return Manager
     */
    private function getManager()
    {
        if (is_null($this->reportManager)) {
            $this->reportManager = $this->getUsageReportManager();
            $this->reportManager->setApsController($this->getApsc());
        }

        return $this->reportManager;
    }

    /* ====== Config parameters accessors ====== */

    private function getRetryTimeout()
    {
        return $this->getManager()->getRetryTimeout();
    }

    private function getRetriesLimit()
    {
        return $this->getManager()->getRetriesLimit();
    }
}