<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Model_MspGroupManager implements Model_MspGroupManagerInterface
{
    const LOG_EVENT_METHOD_WAS_CALLED = 'Method "%s" was called.';
    const LOG_EVENT_MIGRATE_GROUP = 'Migrate group with ID :group_id was updated. Details: group_old_name=":group_old_name", group_new_name=":group_new_name", group_new_version=:group_new_version, old_parent_id=:old_parent_id, new_parent_id=:new_parent_id';
    const LOG_EVENT_NORMALIZE_GROUP_NAME_AFTER_MIGRATE = 'Name of group with ID :group_id was normalized after migration. Details: group_old_name=":group_old_name", group_new_name=":group_new_name", group_new_version=:group_new_version';

    const RESELLER_GROUP_NAME_TEMPLATE = '{companyName} (Reseller {accountId})';
    const CUSTOMER_GROUP_NAME_TEMPLATE = '{companyName} /{suffix}';
    const CUSTOMER_GROUP_NAME_PATTERN = '/^%s(?: \/(\d+))?$/i';
    const PATTERN_DELIMITER = '/';

    const EXCEPTION_CANNOT_CREATE_GROUP_MESSAGE = "Cannot create group for account \"%s\"";
    const WARNING_INTERNAL_TAG_WAS_NOT_UPDATED = 'Internal tag was not updated';

    const HASH = '#';

    const UNLIMITED_QUOTA_VALUE = -1;

    const OSA_ACCOUNT_ID = 'osa_account_id';
    const OSA_SUBSCRIPTION_ID = 'osa_subscription_id';
    const CI_APS_ABC_VERSION = 'ci-aps-abc-version';
    const TIER_LEVEL = 'aps-abc-tier';
    const START_PROD_DATE = 'aps-abc-start-prod-date';

    /** @var Service_MspApiClient_ApiInterface */
    private $_mspApi;
    /** @var Service_APSApi_ApiInterface */
    private $_apsApi;

    /**
     * Model_MspGroupManager constructor.
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param Service_APSApi_ApiInterface $apsApi
     */
    public function __construct(
        Service_MspApiClient_ApiInterface $mspApi,
        Service_APSApi_ApiInterface $apsApi
    ) {
        $this->_mspApi = $mspApi;
        $this->_apsApi = $apsApi;
    }

    /**
     * @param tenant $tenant
     * @param int $parentGroupId
     * @param int $datacenterId
     * @return int
     * @throws Exception
     */
    public function createGroupForReseller($tenant, $parentGroupId, $datacenterId = null)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupName = self::formatResellerGroupName(
            $tenant->account->companyName,
            $tenant->account->id
        );

        $groupId = $this->findGroupByName($parentGroupId, $groupName);
        if (!is_null($groupId)) {
            $this->_apsApi->checkGroupCanBeAssignedOrEx($groupId, $datacenterId);
            $this->updateGroupForReseller($groupId, $tenant);
            return $groupId;
        }

        $groupInfo = new Service_MspApiClient_Request_Struct_CreateGroup();
        $groupInfo->setName($groupName);
        $groupInfo->setKind(Service_MspApiClient_Request_Struct_CreateGroupInterface::RESELLER_GROUP_KIND_ID);

        $pricing = $groupInfo->getPricing();
        $pricing->setMode(Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $groupInfo->setCustomerId(Arr::encode(array(
            self::OSA_ACCOUNT_ID => $tenant->account->id,
            self::OSA_SUBSCRIPTION_ID => $tenant->subscription->subscriptionId,
        )));

        $this->initGroupInfoContact($tenant, $groupInfo);

        $this->initGroupInfoUnlimitedPrivileges($groupInfo);

        $createResult = $this->createGroup($parentGroupId, $groupInfo);
        $version = $createResult->getVersion();

        if (!empty($tenant->startProductionTime) || !$tenant->minimalCommitmentTier->isEmpty()) {
            $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
            $minimalCommitmentTier = $tenant->minimalCommitmentTier;
            $internalTag = array();
            if (!$tenant->minimalCommitmentTier->isEmpty()) {
                $internalTag[self::TIER_LEVEL] = $minimalCommitmentTier->level;
            }
            if (!empty($tenant->startProductionTime)) {
                $internalTag[self::START_PROD_DATE] = $tenant->startProductionTime;
            }
            $updateGroupInfo->setInternalTag(Arr::encode($internalTag));
            $version = $this->updateGroup($createResult->getId(), $updateGroupInfo, $version);
        }

        $this->enableCloudToCloudBackup($groupId, $version);

        return $createResult->getId();
    }

    /**
     * @param int $groupId
     * @param tenant $tenant
     * @return int
     * @throws Exception
     */
    public function updateGroupForReseller($groupId, $tenant)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();

        $groupName = self::formatResellerGroupName(
            $tenant->account->companyName,
            $tenant->account->id
        );
        $updateGroupInfo->setName($groupName);

        $pricing = $updateGroupInfo->getPricing();
        $pricing->setMode(Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $customerIdParts = Arr::decode($groupInfo->getCustomerId());
        $customerIdParts[self::OSA_ACCOUNT_ID] = $tenant->account->id;
        $customerIdParts[self::OSA_SUBSCRIPTION_ID] = $tenant->subscription->subscriptionId;
        $updateGroupInfo->setCustomerId(Arr::encode($customerIdParts));

        $minimalCommitmentTier = $tenant->minimalCommitmentTier;
        $internalTag = Arr::decode($groupInfo->getInternalTag());
        if (!$minimalCommitmentTier->isEmpty()) {
                $internalTag[self::TIER_LEVEL] = $minimalCommitmentTier->level;
        }
        if (!empty($tenant->startProductionTime)) {
            $internalTag[self::START_PROD_DATE] = $tenant->startProductionTime;
        }
        $updateGroupInfo->setInternalTag(Arr::encode($internalTag));

        $this->initGroupInfoContact($tenant, $updateGroupInfo);
        $this->initGroupInfoUnlimitedPrivileges($updateGroupInfo);

        $version = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());

        $version = $this->enableCloudToCloudBackup($groupId, $version);

        return $version;
    }

    /**
     * @param tenant $tenant
     * @param int $parentGroupId
     * @param int $storageId
     * @param int $groupKind
     * @param string $language
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function createGroupForCustomer(
        $tenant,
        $parentGroupId,
        $storageId = null,
        $groupKind = Service_MspApiClient_Request_Struct_CreateGroup::EUC_GROUP_KIND_ID,
        $language = null
    ) {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $companyName = $tenant->account->companyName;
        $groupSuffix = $this->getLastGroupNameSuffix($companyName, $parentGroupId);
        $groupName = is_null($groupSuffix) ? $companyName : self::formatCustomerGroupName($companyName, ++$groupSuffix);

        $groupInfo = new Service_MspApiClient_Request_Struct_CreateGroup();
        $groupInfo->setName($groupName);
        $groupInfo->setKind($groupKind);
        $groupInfo->setStorage($storageId);

        $pricing = $groupInfo->getPricing();
        $mode = $groupKind !== Service_MspApiClient_Request_Struct_CreateGroupInterface::EUC_GROUP_KIND_ID ?
            Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE :
            $this->getGroupMode($tenant);
        $pricing->setMode($mode);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $groupInfo->setCustomerId(Arr::encode(array(
            self::OSA_ACCOUNT_ID => $tenant->account->id,
            self::OSA_SUBSCRIPTION_ID => $tenant->subscription->subscriptionId,
        )));

        $this->initGroupInfoContact($tenant, $groupInfo);
        $this->initGroupInfoPrivileges($tenant, $groupInfo);
        $this->initGroupInfoLocalBackup($tenant, $groupInfo);

        if (!is_null($language)) {
            $groupInfo->setLanguage($language);
        }
        $createResult = $this->createGroup($parentGroupId, $groupInfo);

        $mailboxes = $groupInfo->getPrivileges()->getMailboxCount();
        if ($mailboxes) {
            $this->enableCloudToCloudBackup($createResult->getId(), $createResult->getVersion());
        }

        return $createResult->getId();
    }

    /**
     * @param int $groupId
     * @param tenant $tenant
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function updateGroupForCustomer($groupId, $tenant) {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        $companyName = $tenant->account->companyName;

        $groupName = $this->fetchNameForGroup($companyName, $groupInfo);

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $updateGroupInfo->setName($groupName);

        $pricing = $updateGroupInfo->getPricing();
        $mode = $groupInfo->getKind() !== Service_MspApiClient_Request_Struct_CreateGroupInterface::EUC_GROUP_KIND_ID ?
            Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE :
            $this->getGroupMode($tenant);
        $pricing->setMode($mode);
        $pricing->setEnabled(Service_MspApiClient_Request_Struct_GroupPricing::ENABLED_OFF);

        $customerIdParts = Arr::decode($groupInfo->getCustomerId());
        $customerIdParts[self::OSA_ACCOUNT_ID] = $tenant->account->id;
        $customerIdParts[self::OSA_SUBSCRIPTION_ID] = $tenant->subscription->subscriptionId;
        $updateGroupInfo->setCustomerId(Arr::encode($customerIdParts));

        $this->initGroupInfoContact($tenant, $updateGroupInfo);
        $this->initGroupInfoPrivileges($tenant, $updateGroupInfo);

        $version = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());
        $mailboxes = $groupInfo->getPrivileges()->getMailboxCount();
        if ($mailboxes) {
            $version = $this->enableCloudToCloudBackup($groupId, $version);
        }
        return $version;
    }

    /**
     * @param int $parentGroupId
     * @param string $companyName
     * @return int
     * @throws Exception
     */
    private function findGroupByName($parentGroupId, $companyName)
    {
        /**
         * @var Service_MspApiClient_Response_Struct_GroupItem[] $groups
         */
        $groups = $this->_mspApi->getGroups($parentGroupId);
        foreach ($groups as $group) {
            if ($group->getName() === $companyName) {
                return $group->getId();
            }
        }
        return null;
    }

    /**
     * @param int $parentGroupId
     * @param Service_MspApiClient_Request_Struct_CreateGroupInterface $groupInfo
     * @return Service_MspApiClient_Response_Struct_CreateGroup
     * @throws Exception
     */
    private function createGroup($parentGroupId, Service_MspApiClient_Request_Struct_CreateGroupInterface $groupInfo)
    {
        $creationResult = $this->_mspApi->createGroup($groupInfo, $parentGroupId);
        if (!$creationResult) {
            throw new Exception(sprintf(self::EXCEPTION_CANNOT_CREATE_GROUP_MESSAGE, $groupInfo->getName()));
        }

        return $creationResult;
    }

    /**
     * @param $groupId
     * @param $updateGroupInfo
     * @param $version
     * @return int
     */
    private function updateGroup($groupId, $updateGroupInfo, $version)
    {
        $tenantManager = new Model_TenantManager($this->_mspApi);
        return $tenantManager->updateGroup($groupId, $updateGroupInfo, $version);
    }

    /**
     * @param tenant $tenant
     * @return int
     */
    private function getGroupMode($tenant)
    {
        return $tenant->subscription->trial
            ? Service_MspApiClient_Request_Struct_GroupPricing::TRIAL_MODE
            : Service_MspApiClient_Request_Struct_GroupPricing::PRODUCTION_MODE;
    }

    /**
     * @param tenant $tenant
     * @param Service_MspApiClient_Request_Struct_GroupInterface $groupInfo
     */
    private function initGroupInfoContact($tenant, Service_MspApiClient_Request_Struct_GroupInterface $groupInfo)
    {
        $admin = $this->_apsApi->getAdmin($tenant);
        if (is_null($admin)) {
            return;
        }

        $contact = $groupInfo->getContact();
        $contact->setEmail($admin->email);
        $contact->setFirstName($admin->givenName);
        $contact->setLastName($admin->familyName);
        $contact->setPhone(str_replace(self::HASH, '', $admin->telWork));

        $addressPostal = $admin->addressPostal;
        if (is_null($addressPostal)) {
            return;
        }
        $contact->setCountry(strtoupper($addressPostal->countryName));
        $contact->setZipCode($addressPostal->postalCode);
        $contact->setState($addressPostal->region);
        $contact->setCity($addressPostal->locality);
        $contact->setAddress1($addressPostal->streetAddress);
        $contact->setAddress2($addressPostal->extendedAddress);
    }

    /**
     * @param tenant $tenant
     * @param Service_MspApiClient_Request_Struct_GroupInterface $groupInfo
     */
    private function initGroupInfoPrivileges($tenant, Service_MspApiClient_Request_Struct_GroupInterface $groupInfo)
    {
        $privileges = $groupInfo->getPrivileges();
        $privileges->setStorageSize($this->getCounterLimit($tenant->StorageUsage));
        $privileges->setServerCount($this->getCounterLimit($tenant->ServerCount));
        $privileges->setVmCount($this->getCounterLimit($tenant->VmCount));
        $privileges->setWorkstationCount($this->getCounterLimit($tenant->WorkstationCount));
        $privileges->setMobileCount($this->getCounterLimit($tenant->MobileDeviceCount));
        $privileges->setMailboxCount($this->getCounterLimit($tenant->MailboxCount));
        $privileges->setWebsiteCount($this->getCounterLimit($tenant->WebsiteCount));

        if (!$tenant->profile->isEmpty() && $tenant->profile->freemiumMailboxStorage) {
            $privileges->setMailboxInstanceStorageSize(
                $this->computeMailboxInstanceStorageSizeForTenant($tenant)
            );
        }
    }

    /**
     * @param object | NULL $counter
     * @return int
     */
    private function getCounterLimit($counter)
    {
        if (is_null($counter)) {
            return 0;
        }

        return isset($counter->limit) ? $counter->limit : 0;
    }

    /**
     * @param string $companyName
     * @param int $id
     * @return string
     */
    static public function formatResellerGroupName($companyName, $id)
    {
        return strtr(self::RESELLER_GROUP_NAME_TEMPLATE, array(
            '{companyName}' => $companyName,
            '{accountId}' => $id
        ));
    }

    /**
     * @param string $companyName
     * @param int $suffix
     * @return string
     */
    static public function formatCustomerGroupName($companyName, $suffix)
    {
        return strtr(self::CUSTOMER_GROUP_NAME_TEMPLATE, array(
            '{companyName}' => $companyName,
            '{suffix}' => $suffix
        ));
    }

    /**
     * @param string $companyName
     * @param string $groupName
     * @return bool|int|null NULL - no suffix, int - group suffix, false - names don't match
     */
    private function getGroupNameSuffix($companyName, $groupName)
    {
        $pattern = sprintf(self::CUSTOMER_GROUP_NAME_PATTERN, preg_quote($companyName, self::PATTERN_DELIMITER));
        if (preg_match($pattern, $groupName, $matches)) {
            return isset($matches[1]) ? intval($matches[1]) : null;
        }
        return false;
    }

    /**
     * @param string $companyName
     * @param int $parentGroupId
     * @return int|null
     */
    private function getLastGroupNameSuffix($companyName, $parentGroupId)
    {
        $children = $this->_mspApi->getGroups($parentGroupId);
        $groupSuffix = null;
        foreach ($children as $child) {
            $childSuffix = $this->getGroupNameSuffix($companyName, $child->getName());
            if ($childSuffix === false) {
                continue;
            }

            if (is_null($childSuffix)) {
                $childSuffix = 0;
            }

            $groupSuffix = is_null($groupSuffix) ? $childSuffix : max($groupSuffix, $childSuffix);
        }
        return $groupSuffix;
    }

    /**
     * @param int $groupId
     * @param tenant $tenant
     * @param int $newParentGroupId
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function migrateGroupForCustomer($groupId, $tenant, $newParentGroupId)
    {
        Log::getInstance()->event(sprintf(self::LOG_EVENT_METHOD_WAS_CALLED, __METHOD__));

        $groupInfo = $this->_mspApi->getGroupById($groupId);

        if ($groupInfo->getParentId() == $newParentGroupId) {
            return $groupInfo->getVersion();
        }

        $companyName = $tenant->account->companyName;

        // Actually, we split moving group in 2 step (2 http requests)

        // 1) Set name that will be unique for old and new parents and move group (first http request)

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $tmpGroupName = $companyName." ({$groupId})";
        $updateGroupInfo->setName($tmpGroupName);
        $updateGroupInfo->setParentId($newParentGroupId);
        $groupNewVersion = $this->updateGroup($groupId, $updateGroupInfo, $groupInfo->getVersion());

        Log::getInstance()->event(
            self::LOG_EVENT_MIGRATE_GROUP,
            array(
                ':group_id' => $groupInfo->getId(),
                ':group_old_name' => $groupInfo->getName(),
                ':group_new_name' => $tmpGroupName,
                ':old_parent_id' => $groupInfo->getParentId(),
                ':new_parent_id' => $newParentGroupId,
                ':group_new_version' => $groupNewVersion,
            )
        );

        // 2) Rename moved group deleting unique part (second http request) and enable group if it status is inactive

        $groupSuffix = $this->getLastGroupNameSuffix($companyName, $newParentGroupId);
        $finalGroupName = is_null($groupSuffix) ? $companyName : self::formatCustomerGroupName($companyName, ++$groupSuffix);

        $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
        $updateGroupInfo->setName($finalGroupName);
        if ($groupInfo->getStatus() == Model_TenantManager::STATUS_INACTIVE) {
            $updateGroupInfo->setStatus(Model_TenantManager::STATUS_ACTIVE);
        }
        $groupNewVersion = $this->updateGroup($groupId, $updateGroupInfo, $groupNewVersion);

        Log::getInstance()->event(
            self::LOG_EVENT_NORMALIZE_GROUP_NAME_AFTER_MIGRATE,
            array(
                ':group_id' => $groupInfo->getId(),
                ':group_old_name' => $tmpGroupName,
                ':group_new_name' => $finalGroupName,
            )
        );

        return $groupNewVersion;
    }

    /**
     * @param Service_MspApiClient_Request_Struct_GroupInterface $groupInfo
     */
    private function initGroupInfoUnlimitedPrivileges(Service_MspApiClient_Request_Struct_GroupInterface $groupInfo)
    {
        $privileges = $groupInfo->getPrivileges();

        $privileges->setMailboxCount(self::UNLIMITED_QUOTA_VALUE);
        $privileges->setMobileCount(self::UNLIMITED_QUOTA_VALUE);
        $privileges->setVmCount(self::UNLIMITED_QUOTA_VALUE);
        $privileges->setWorkstationCount(self::UNLIMITED_QUOTA_VALUE);
        $privileges->setStorageSize(self::UNLIMITED_QUOTA_VALUE);
        $privileges->setServerCount(self::UNLIMITED_QUOTA_VALUE);
    }

    private function initGroupInfoLocalBackup($tenant, Service_MspApiClient_Request_Struct_GroupInterface $groupInfo)
    {
        $privileges = $groupInfo->getPrivileges();
        $localBackup = $privileges->getLocalBackup();
        $localBackup->setPermitted($tenant->profile->storageLocation);
    }

    public function updatePackageVersionInfo($groupId)
    {
        $mspApi = $this->_mspApi;
        $group = $mspApi->getGroupById($groupId);
        $buildInfo = new BuildInfo();
        $apsPackageVersion = $buildInfo->getPackageVersion();
        try {
            $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();

            $internalTag = Arr::decode($group->getInternalTag());
            if (isset($internalTag[self::CI_APS_ABC_VERSION])
                && $internalTag[self::CI_APS_ABC_VERSION] == $apsPackageVersion
            ) {
                return;
            }
            $internalTag[self::CI_APS_ABC_VERSION] = $apsPackageVersion;
            $updateGroupInfo->setInternalTag(Arr::encode($internalTag));

            $mspApi->updateGroup($group->getId(), $updateGroupInfo, $group->getVersion());
        } catch (Exception $e) {
            if ($e->getCode() != 403) {
                throw $e;
            }
            Log::getInstance()->warning(self::WARNING_INTERNAL_TAG_WAS_NOT_UPDATED);
        }
    }

    /**
     * @param string $companyName
     * @param Service_MspApiClient_Response_Struct_GroupInfo $groupInfo
     * @return string
     */
    public function fetchNameForGroup($companyName, Service_MspApiClient_Response_Struct_GroupInfo $groupInfo)
    {
        $groupSuffix = $this->getGroupNameSuffix($companyName, $groupInfo->getName());
        if (is_null($groupSuffix)) {
            $groupName = $companyName;
            return $groupName;
        } elseif ($groupSuffix === false) {
            $groupSuffix = $this->getLastGroupNameSuffix($companyName, $groupInfo->getParentId());
            $groupName = is_null($groupSuffix) ? $companyName : self::formatCustomerGroupName($companyName, ++$groupSuffix);
            return $groupName;
        } else {
            $groupName = self::formatCustomerGroupName($companyName, $groupSuffix);
            return $groupName;
        }
    }

    /**
     * @param tenant $tenant
     * @return int
     */
    private function computeMailboxInstanceStorageSizeForTenant($tenant)
    {
        $storageSizeQuota = $this->getCounterLimit($tenant->StorageUsage);
        $mailboxCountQuota = $this->getCounterLimit($tenant->MailboxCount);

        if ($tenant->profile->freemiumMailboxStorage == self::UNLIMITED_QUOTA_VALUE
           || $mailboxCountQuota == self::UNLIMITED_QUOTA_VALUE
        ) {
            return self::UNLIMITED_QUOTA_VALUE;
        }

        return (
            $storageSizeQuota == self::UNLIMITED_QUOTA_VALUE
            || $storageSizeQuota > $mailboxCountQuota * $tenant->profile->freemiumMailboxStorage
        )
            ? self::UNLIMITED_QUOTA_VALUE
            : $tenant->profile->freemiumMailboxStorage;
    }

    /**
     * @param $groupId
     * @param $version
     * @return int
     */
    private function enableCloudToCloudBackup($groupId, $version)
    {
        try {
            $cloudToCloudGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
            $privileges = $cloudToCloudGroupInfo->getPrivileges();
            $privileges->setCloudToCloudBackup(1);
            return $this->updateGroup($groupId, $cloudToCloudGroupInfo, $version);
        } catch (Exception $e) {
            // Skip any errors
            Log::getInstance()->warning($e->getMessage());
        }
        return $version;
    }
}
