<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

use Acronis\App\Base\BackupApplication;
use Acronis\App\CloudApi\Api as CloudApi;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Services\Logger\AppLoggerAwareTrait;
use Acronis\App\Services\UsageReportManager\UsageReportManagerAwareTrait;
use Acronis\App\Utils\MemoizeTrait;
use Rest\RestException;

/**
 * Class Model_Resource_Tenant
 */
abstract class Model_Resource_Tenant implements Model_Resource_TenantInterface
{
    use UsageReportManagerAwareTrait;
    use CountersManagerAwareTrait;
    use MemoizeTrait;
    use AppLoggerAwareTrait;

    const LOG_METHOD_CALL = 'Resource: %s Account: %s Subscription: %s Call: %s';

    const LOG_MIGRATION_WAS_STARTED = 'Tenant migration was started.';
    const LOG_FORMAT_DELETE_MSP_GROUP = 'Delete msp group for tenant (%s)';
    const LOG_UNPROVISIONING_TENANT_IS_MIGRATION = 'The tenant was created during migration. Deleting a group will be skipped';
    const LOG_CALLED_UPDATE_COUNTER_LIMITS = 'Update counter limits for tenant was called.';
    const LOG_CALLED_GET_PARTNER_CREDENTIALS = 'Get partner credentials for tenant was called.';
    const LOG_CALLED_SET_STORAGE_FOR_PLAN = 'Set storage %s for plan %s was called.';
    const LOG_CALLED_LINK_DATACENTER = 'Tenant (%s) link datacenter was called';
    const LOG_WARNING_UNPROVISION_TENANT_DOES_NOT_HAVE_GROUP = "The tenant does not have a group in the Acronis Backup Cloud. Deleting a group will be skipped";
    const LOG_WARNING_UNPROVISION_TENANT_IS_RESELLER = "The tenant is a reseller. Deleting a group will be skipped";
    const LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP = "Unable to find a group by ID %s. Deleting a group will be skipped";
    const LOG_WARNING_BACKUP_PROFILE_HAS_BEEN_CHANGED = 'Backup Profile has been changed from %s to %s in the Service Template. But old Backup Profile will be still linked with tenant (%s)';
    const LOG_WARNING_BACKUP_PROFILE_HAS_BEEN_REMOVED = 'Backup Profile (%s) has been removed from the Service Template. But it will be still linked with tenant (%s)';
    const STAFF_MEMBER_AVAILABLE = 'Staff member available %s';

    const ERROR_ADMINISTRATOR_NOT_FOUND = "Error. Administrator not found.";
    const ERROR_FAILED_LOCK_GROUP = 'Failed to lock group id: %s';
    const EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE = 'Can\'t receive storages for group';
    const EXCEPTION_STORAGE_ID_MUST_BE_PRESENT_MESSAGE = 'Parameter "storageId" must be present.';
    const EXCEPTION_DATACENTER_ID_MUST_BE_PRESENT_MESSAGE = 'Parameter "datacenterId" must be present.';
    const EXCEPTION_RESELLER_HAS_TO_SET_STORAGE_FOR_PLAN_MESSAGE = 'The reseller has to set a storage for the plan';
    const EXCEPTION_NEED_TO_SET_CREDENTIALS_MESSAGE = "Need to set credentials";
    const EXCEPTION_DATACENTER_DOES_NOT_SET_MESSAGE = 'Please set parameters "datacenterId" and "storageId"';
    const EXCEPTION_SELECT_STORAGE_IN_BACKUP_PROFILE_MESSAGE = "Please select a storage in the Backup Profile";
    const EXCEPTION_MORE_THAN_1_COMMITMENT_MESSAGE = 'Only one minimum commitment resource can be included into a service plan.';
    const EXCEPTION_FIELD_MUST_BE_PRESENT_MESSAGE = 'The "%s" must be present.';
    const EXCEPTION_UNKNOWN_STORAGE_MESSAGE = 'Unknown storage. You  have to specify the same option id.';
    const EXCEPTION_RESOURCE_TYPE_MUST_BE_REFERENCE_MESSAGE = 'Resource type "%s" must be a reference.';
    const EXCEPTION_RESOURCE_TYPE_CANNOT_BE_UNLIMITED_MESSAGE = 'Resource type "%s" cannot be unlimited.';
    const EXCEPTION_RESOURCE_TYPE_CANNOT_HAVE_LIMIT_MORE_THEN_ONE_MESSAGE = 'Resource type "%s" cannot have a limit of more than 1.';
    const EXCEPTION_CANNOT_FETCH_ALL_AVAILABLE_TIERS_MESSAGE = 'Cannot fetch all available tiers.';
    const EXCEPTION_INVALID_SEQUENCE_OF_MINIMUM_COMMITMENT_TIER_LEVELS_MESSAGE = 'The service template has an invalid sequence of minimum commitment tier levels. A sequence of levels must start from level "No commitment", contain only unique values, and must not have gaps.';
    const EXCEPTION_TIERS_WITHOUT_ACCOUNT_CURRENCY_MESSAGE = "The service template contains tiers that do not have the minimum commitment in the account's currency.";
    const EXCEPTION_ACCOUNT_ALREADY_HAS_RESELLER_SUBSCRIPTION = "The account already has a reseller subscription.";
    const EXCEPTION_NO_DISCOUNT_FOR_ACCOUNT = 'There is no discount for the account "%s".';
    const EXCEPTION_SEVERAL_DISCOUNTS_FOR_ACCOUNT = 'There are several discounts for the account "%s".';
    const EXCEPTION_CAN_NOT_FIND_DATACENTER_MESSAGE = 'Can\'t find Datacenter by ID "%s"';
    const EXCEPTION_ST_CONTAINS_SEVERAL_BACKUP_PROFILES = 'Only one Backup Profile resource can be included into the service template.';
    const EXCEPTION_ST_CONTAINS_NO_BACKUP_PROFILES = 'The service template doesn\'t contain Backup Profile.';
    const EXCEPTION_LIMIT_OF_RESOURCE_DOES_NOT_ONE = 'The limit of the Backup Profile resource in the service template must be equal to 1.';

    const USAGE = 'usage';
    const LIMIT = 'limit';
    const TENANT = 'tenant';
    const USER_USAGE = 'userUsage';
    const ERROR = 'error';

    const WIN_URL = 'winurl';
    const LINUX_URL = 'linuxurl';
    const MAC_URL = 'macurl';
    const ANDROID_URL = 'androidurl';
    const IOS_URL = 'iosurl';
    const VMWARE_URL = 'vmwareurl';
    const HYPERV_URL = 'hypervurl';
    const VIRTUOZZO_URL = 'virtuozzourl';
    const MICROSOFT_SQL_SERVER_URL = 'microsoftSqlServerUrl';
    const MICROSOFT_EXCHANGE_SERVER_URL = 'microsoftExchangeServerUrl';
    const MICROSOFT_ACTIVE_DIRECTORY_URL = 'microsoftActiveDirectoryUrl';
    const MICROSOFT_O365_URL = 'microsoftO365Url';

    const FIELD_URL = 'url';
    const FIELD_STORAGE_ID = 'id';
    const FIELD_STORAGE_NAME = 'name';
    const FIELD_STORAGE_DATACENTER_ID = 'datacenterId';
    const FIELD_STORAGE_DATACENTER_NAME = 'datacenterName';
    const FIELD_MOBILE_COUNTER = 'mobile';
    const FIELD_MAILBOX_COUNTER = 'mailbox';
    const FIELD_WEBSITE_COUNTER = 'website';

    const URL_REDIRECT_MASK = '%s?jwt=%s';

    const UNLIMITED_VALUE = -1;

    const STORAGE_DEFAULT = "default";
    const STORAGE_PROVIDER = "provider";
    const INHERITED_STORAGE = 1;

    const ACCOUNT_PROCESS_TYPE_ONBOARDING = 'onboarding';
    const ACCOUNT_PROCESS_TYPE_CANCELLATION = 'cancellation';

    const REDIRECT_URL = "%s/#groups/%s";

    //create administrator
    //default permissions for administrator account in baas
    const DEFAULT_GROUP_ADMIN_ACCOUNT_PERMS = 15;
    const RESELLER_GROUP_ADMIN_ACCOUNT_PERMS = 11;

    public static $userQuotasPropertiesMap = array(
        tenant::STORAGE_USAGE => users::STATISTICS_USER_STORAGE_USAGE,
        tenant::SERVER_COUNT => users::STATISTICS_USER_SERVER_COUNT,
        tenant::VM_COUNT => users::STATISTICS_USER_VM_COUNT,
        tenant::WORKSTATION_COUNT => users::STATISTICS_USER_WORKSTATION_COUNT,
        tenant::MOBILE_DEVICE_COUNT => users::STATISTICS_USER_MOBILE_DEVICE_COUNT,
        tenant::WEB_HOSTING_SERVER_COUNT => users::STATISTICS_USER_WEB_HOSTING_SERVER_COUNT,
        tenant::WEBSITE_COUNT => users::STATISTICS_USER_WEBSITE_COUNT,
    );

    /**
     * @var FacadeInterface
     */
    private $facade;

    /**
     * @var Model_ManagerFactoryInterface
     */
    private $managerFactory;

    /**
     * @var \APS\ControllerProxy
     */
    private $apsc;

    /**
     * @var Service_MspApiClient_ApiInterface
     */
    private $mspApi;

    /**
     * @var Service_APSApi_ApiInterface
     */
    private $apsApi;

    /**
     * @var Service_OpenApi_ApiInterface
     */
    private $openApi;

    /**
     * @var Service_PbaApi_ApiInterface
     */
    private $pbaApi;

    private $groupManager;

    private $tenantManager;

    private $userManager;

    private $subscriptionParametersManager;

    private $accountAttributesManager;

    /**
     * @var tenant
     */
    protected $tenant;
    /**
     * @var array
     */
    private $_subscriptionResources;

    /**
     * Model_Resource_Tenant constructor.
     * @param $tenant
     * @param \APS\ControllerProxy $apsc
     * @param FacadeInterface $facade
     * @param Model_ManagerFactoryInterface $managerFactory
     */
    public function __construct(
        $tenant,
        \APS\ControllerProxy $apsc,
        FacadeInterface $facade,
        Model_ManagerFactoryInterface $managerFactory
    )
    {
        $this->apsc = $apsc;

        $this->facade = $facade;
        $this->managerFactory = $managerFactory;

        $this->tenant = $tenant;
    }

    protected function updateTenantId()
    {
        if (isset($this->tenant->tenantId)) {
            return;
        }
        $this->tenant->tenantId = $this->obtainTenantId();
    }

    private function obtainTenantId()
    {
        $group = $this->getTenantGroup($this->tenant->mspGroupId);

        return $group->getUuid();
    }

    private function getTenantGroup($mspGroupId, $datacenter = null)
    {
        $api = is_null($datacenter) ? $this->getMspApi() : $this->createMspApiInstance($datacenter);

        return $api->getGroupById($mspGroupId);
    }

    /**
     * @param tenant $tenant
     * @return string
     */
    private function getCreationDate($tenant, $datacenter = null, $groupId = null)
    {
        $group = $this->getTenantGroup(is_null($groupId) ? $tenant->mspGroupId : $groupId, $datacenter);

        return Time::formatDate(
            (new DateTime())->setTimestamp($group->getCreationDate())
        );
    }

    /**
     * @param $counters
     * @param $datacenter
     * @param $date
     * @throws RestException
     * @throws \Acronis\App\UsageReport\Exception\ManagerException
     */
    protected function updateCountersUsage($datacenter, $groupId = null)
    {
        $date = Time::getCurentDate();
        $dateTime = Time::getCurentDateTime();

        $manager = $this->getUsageReportManager();
        $counters = $manager->getUsagesForSubscription($this->tenant->subscriptionId, $date, $datacenter->aps->id);

        if (is_null($counters)) {

            if ($this->isTenantCreatedAtCurrentDate($dateTime, $datacenter, $groupId)) {
                $this->getLog()->error(
                    'There are no usages for tenant %s.',
                    $this->tenant->tenantId
                );

                return;
            }

            $this->throwRetrieveProcessException($datacenter, $date);
        }

        foreach ($counters as $key => $value) {
            if (property_exists($this->tenant, $key)) {
                $this->tenant->$key->usage += $value;
            }
        }
    }

    /**
     * @param \datacenter|object $datacenter
     * @param string $date
     * @throws RestException
     */
    protected function throwRetrieveProcessException($datacenter, $date)
    {
        throw new RestException(404, Str::format(
            'Aggregated usages for subscription "%s" on datacenter "%s" are not found for date "%s".',
            $this->tenant->subscriptionId, $datacenter->id, $date
        ));
    }

    /**
     * @param string $localDateTime
     * @return bool
     */
    protected function isTenantCreatedAtCurrentDate($localDateTime, $datacenter = null, $groupId = null)
    {
        return $this->getCreationDate($this->tenant, $datacenter, $groupId) == Time::formatDate(new DateTime($localDateTime));
    }

    /**
     * @return string|NULL
     */
    protected static function _getApsIdentityId()
    {
        return isset($_SERVER['HTTP_APS_IDENTITY_ID']) ? $_SERVER['HTTP_APS_IDENTITY_ID'] : NULL;
    }

    protected function throwNotApplicableMethodException()
    {
        throw new Model_Resource_NotApplicableMethodException();
    }

    /**
     * @return \APS\ControllerProxy
     */
    protected function getApsc()
    {
        return $this->apsc;
    }

    /**
     * @param bool $force
     * @return Service_MspApiClient_ApiInterface
     */
    protected function getMspApi($force = false)
    {
        $datacenter = $this->tenant;
        if (is_null($this->mspApi)) {
            $this->mspApi = $this->createMspApiInstance($this->tenant->datacenter, $force);
        } elseif ($force) {
            $this->mspApi->setCredentials(
                $datacenter->login,
                $datacenter->password,
                $datacenter->url,
                $force
            );
            $this->mspApi->login();
        }
        return $this->mspApi;
    }

    /**
     * @param $datacenter
     * @param bool $force
     * @return Service_MspApiClient_ApiInterface
     */
    protected function createMspApiInstance($datacenter, $force = false)
    {
        $instanceId = sprintf(
            '%s_%s',
            $this->getApsc()->getInstanceId(),
            md5($datacenter->login . $datacenter->url)
        );
        $api = $this->facade->getMspApi($datacenter->url, $instanceId);
        $api->setCredentials(
            $datacenter->login,
            $datacenter->password,
            $datacenter->url,
            $force
        );
        if ($force) {
            $api->login();
        }
        return $api;
    }

    /**
     * @param $datacenter
     * @return CloudApi
     */
    protected function createCloudApiInstance($datacenter)
    {
        return new CloudApi(
            $datacenter->url,
            $datacenter->login,
            $datacenter->password
        );
    }

    /**
     * @param string $amsUrl
     * @return Service_AmsApiClient_ApiInterface
     */
    protected function createAmsApiInstance($amsUrl)
    {
        return $this->facade->getAmsApi($amsUrl);
    }

    /**
     * @return Service_PbaApi_ApiInterface
     * @throws Exception
     */
    protected function getPbaApi()
    {
        if (is_null($this->pbaApi)) {
            $config = Settings::getInstance()->getBillingConfig();
            $this->pbaApi = $this->createPbaApiInstance($config);
        }
        return $this->pbaApi;
    }

    /**
     * @param $config
     * @return Service_PbaApi_Api
     */
    protected function createPbaApiInstance($config)
    {
        return $this->facade->getPbaApi($config);
    }

    /**
     * @return Service_APSApi_ApiInterface
     */
    protected function getApsApi()
    {
        if (is_null($this->apsApi)) {
            $this->apsApi = $this->createApsApiInstance($this->getApsc());
        }
        return $this->apsApi;
    }

    /**
     * @param $apsc
     * @return Service_APSApi_ApiInterface
     */
    protected function createApsApiInstance($apsc)
    {
        return $this->facade->getApsApi($apsc);
    }

    /**
     * @return Service_OpenApi_ApiInterface
     */
    protected function getOpenApi()
    {
        if (is_null($this->openApi)) {
            $host = defined('OPEN_API_URI') ? OPEN_API_URI : parse_url(\APS\Proto::getControllerURI(), PHP_URL_HOST);
            $this->openApi = $this->createOpenApiInstance($host);
        }
        return $this->openApi;
    }

    /**
     * @param $host
     * @return Service_OpenApi_ApiInterface
     */
    protected function createOpenApiInstance($host)
    {
        return $this->facade->getOpenApi($host);
    }

    /**
     * @return Model_MspGroupManager
     */
    protected function getMspGroupManager()
    {
        if (is_null($this->groupManager)) {
            $this->groupManager = $this->createMspGroupManagerInstance($this->getMspApi(), $this->getApsApi());
        }
        return $this->groupManager;
    }

    /**
     * @param $mspApi
     * @param $apsApi
     * @return Model_MspGroupManager
     */
    protected function createMspGroupManagerInstance($mspApi, $apsApi)
    {
        return $this->managerFactory->getMspGroupManager($mspApi, $apsApi);
    }

    /**
     * @return Model_TenantManager
     */
    protected function getTenantManager()
    {
        if (is_null($this->tenantManager)) {
            $this->tenantManager = $this->createTenantManagerInstance($this->getMspApi());
        }
        return $this->tenantManager;
    }

    /**
     * @param $mspApi
     * @return Model_TenantManager
     */
    protected function createTenantManagerInstance($mspApi)
    {
        return $this->managerFactory->getTenantManager($mspApi);
    }

    /**
     * @return Model_UserManagerInterface
     */
    protected function getUserManager()
    {
        if (is_null($this->userManager)) {
            $this->userManager = $this->createUserManagerInstance($this->getMspApi());
        }
        return $this->userManager;
    }

    /**
     * @param $mspApi
     * @return Model_UserManagerInterface
     */
    protected function createUserManagerInstance($mspApi)
    {
        return $this->managerFactory->getUserManager($mspApi);
    }

    /**
     * @return Model_OSA_SubscriptionParametersManagerInterface
     */
    protected function getSubscriptionParametersManager()
    {
        if (is_null($this->subscriptionParametersManager)) {
            $this->subscriptionParametersManager = $this->createSubscriptionParametersManagerInstance($this->getPbaApi());
        }
        return $this->subscriptionParametersManager;
    }

    /**
     * @param $pbaApi
     * @return Model_OSA_SubscriptionParametersManagerInterface
     */
    protected function createSubscriptionParametersManagerInstance($pbaApi)
    {
        return $this->managerFactory->getSubscriptionParametersManager($pbaApi);
    }

    /**
     * @return Model_OSA_AccountAttributesManagerInterface
     */
    protected function getAccountAttributesManager()
    {
        if (is_null($this->accountAttributesManager)) {
            $this->accountAttributesManager = $this->createAccountAttributesManagerInstance($this->getPbaApi());
        }
        return $this->accountAttributesManager;
    }

    /**
     * @param $pbaApi
     * @return Model_OSA_AccountAttributesManagerInterface
     */
    protected function createAccountAttributesManagerInstance($pbaApi)
    {
        return $this->managerFactory->getAccountAttributesManager($pbaApi);
    }

    protected function update()
    {
        $this->apsc->updateResource($this->tenant);
    }

    /**
     * @param \APS\EventSubscription $event
     * @param string $resourceId
     */
    protected function subscribeOnEventById($event, $resourceId)
    {
        if (!is_object($event->source)) {
            $event->source = new stdClass();
        }
        $event->source->id = $resourceId;
        $this->getApsc()->impersonate($this->tenant)->subscribe($this->tenant, $event);
    }

    /**
     * @param \APS\EventSubscription $event
     * @param string $resourceType
     */
    protected function subscribeOnEventByType($event, $resourceType)
    {
        if (!is_object($event->source)) {
            $event->source = new stdClass();
        }
        $event->source->type = $resourceType;
        $this->getApsc()->impersonate($this->tenant)->subscribe($this->tenant, $event);
    }

    protected function _subscribeToOnSubscriptionChangeEvent()
    {
        $tenant = $this->tenant;
        $eventSubscription = new \APS\EventSubscription(\APS\EventSubscription::Changed, tenant::ON_SUBSCRIPTION_CHANGE);
        $this->subscribeOnEventById($eventSubscription, $tenant->subscription->aps->id);
    }

    protected function _subscribeToOnAccountChangeEvent()
    {
        $tenant = $this->tenant;
        $eventSubscription = new \APS\EventSubscription(\APS\EventSubscription::Changed, tenant::ON_ACCOUNT_CHANGE);
        $this->subscribeOnEventById($eventSubscription, $tenant->account->aps->id);
    }

    protected function _subscribeToOnUserChangeEvent()
    {
        $adminUser = $this->getAdminUser();
        if (is_null($adminUser)) {
            return;
        }
        $eventSubscription = new \APS\EventSubscription(\APS\EventSubscription::Changed, tenant::ON_USER_CHANGED);
        $this->subscribeOnEventById($eventSubscription, $adminUser->aps->id);
    }

    /**
     * @param bool $tryLink If set TRUE then function will try also to link resource via APSC
     * @throws Exception
     */
    protected function linkProfile($tryLink = true)
    {
        $tenant = $this->tenant;
        if (!$tenant->profile->isEmpty()) {
            return;
        }

        $apsApi = $this->getApsApi();
        $profile = $apsApi->getProfile($this->_getSubscriptionResources());
        if (is_null($profile)) {
            return;
        }

        $tenant->profile->set($profile);
        if (!$tryLink) {
            return;
        }
        try {
            $this->getApsc()->linkResource($tenant, tenant::LINK_NAME_PROFILE, $tenant->profile);
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
            throw $e;
        }
    }

    /**
     * @throws Exception
     */
    protected function checkProfile()
    {
        $profileApsType = \APS\TypeLibrary::getTypeByServiceId('profile');
        $profileApsType = substr($profileApsType, 0, strrpos($profileApsType, '/'));
        $availableProfiles = array();
        foreach ($this->_getSubscriptionResources() as $subscriptionResource) {
            if (!isset($subscriptionResource->apsType)) {
                continue;
            }
            if (strstr($subscriptionResource->apsType, $profileApsType) === false) {
                continue;
            }
            if (!isset($subscriptionResource->apsId)) {
                throw new Exception(Str::format(self::EXCEPTION_RESOURCE_TYPE_MUST_BE_REFERENCE_MESSAGE, $subscriptionResource->id));
            }
            if (!isset($subscriptionResource->limit)) {
                throw new Exception(Str::format(self::EXCEPTION_RESOURCE_TYPE_CANNOT_BE_UNLIMITED_MESSAGE, $subscriptionResource->id));
            }
            if ($subscriptionResource->limit != 1) {
                throw new Exception(Str::format(self::EXCEPTION_LIMIT_OF_RESOURCE_DOES_NOT_ONE, $subscriptionResource->id));
            }
            $availableProfiles[] = $subscriptionResource->apsId;
        }
        $availableProfilesCount = count($availableProfiles);
        if ($availableProfilesCount > 1) {
            throw new Exception(self::EXCEPTION_ST_CONTAINS_SEVERAL_BACKUP_PROFILES);
        }
        if ($availableProfilesCount < 1) {
            throw new Exception(self::EXCEPTION_ST_CONTAINS_NO_BACKUP_PROFILES);
        }
    }

    /**
     * @param bool $tryLink If set TRUE then function will try also to link resource via APSC
     * @throws Exception
     */
    protected function linkDatacenter($tryLink = true)
    {
        $tenant = $this->tenant;
        Log::getInstance()->event(sprintf(self::LOG_CALLED_LINK_DATACENTER, $tenant->aps->id));

        if (!$tenant->datacenter->isEmpty()) {
            return;
        }

        $datacenter = $this->_getDatacenterById($tenant->global, $tenant->datacenterId);
        $tenant->datacenter->set($datacenter);
        if (!$tryLink) {
            return;
        }
        try {
            $this->getApsc()->linkResource($tenant, tenant::LINK_NAME_DATACENTER, $tenant->datacenter);
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
            throw $e;
        }
    }

    /**
     * @param bool $tryLink If set TRUE then function will try also to link resource via APSC
     * @throws Exception
     */
    protected function linkMinimalCommitmentTier($tryLink = true)
    {
        $minimalCommitmentTierType = \APS\TypeLibrary::getTypeByServiceId('minimalCommitmentTier');
        $activatedTiers = array();
        foreach ($this->_getSubscriptionResources() as $subscriptionResource) {
            if (!isset($subscriptionResource->apsType)) {
                continue;
            }
            if ($subscriptionResource->apsType !== $minimalCommitmentTierType) {
                continue;
            }
            if (isset($subscriptionResource->limit) && $subscriptionResource->limit == 0) {
                continue;
            }
            if (!isset($subscriptionResource->apsId)) {
                throw new Exception(sprintf(self::EXCEPTION_RESOURCE_TYPE_MUST_BE_REFERENCE_MESSAGE, $subscriptionResource->id));
            }
            $activatedTiers[] = $subscriptionResource->apsId;
        }
        $tenant = $this->tenant;
        if (empty($activatedTiers)) {
            if (!$tenant->minimalCommitmentTier->isEmpty()) {
                try {
                    $this->getApsc()->unlinkResource($tenant->aps->id, 'minimalCommitmentTier', $tenant->minimalCommitmentTier);
                    $tenant->minimalCommitmentTier->set(null);
                } catch (Exception $e) {
                    Log::getInstance()->error($e->getMessage());
                    throw $e;
                }
            }
            return;
        }
        if (count($activatedTiers) > 1) {
            throw new Exception(self::EXCEPTION_MORE_THAN_1_COMMITMENT_MESSAGE);
        }
        $activatedTierApsId = reset($activatedTiers);
        if (!$tenant->minimalCommitmentTier->isEmpty() && $tenant->minimalCommitmentTier->aps->id === $activatedTierApsId) {
            return;
        }
        $tenant->minimalCommitmentTier->set($this->getApsc()->getResource($activatedTierApsId));
        if (!$tryLink) {
            return;
        }
        try {
            $this->getApsc()->linkResource($tenant->aps->id, 'minimalCommitmentTier', $activatedTierApsId);
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
            throw $e;
        }
    }

    /**
     * @throws Exception
     */
    protected function checkMinimalCommitmentTiers()
    {
        $minimalCommitmentTierType = \APS\TypeLibrary::getTypeByServiceId('minimalCommitmentTier');
        $availableTiers = array();
        $activatedTiers = array();
        foreach ($this->_getSubscriptionResources() as $subscriptionResource) {
            if (!isset($subscriptionResource->apsType)) {
                continue;
            }
            if ($subscriptionResource->apsType !== $minimalCommitmentTierType) {
                continue;
            }
            if (!isset($subscriptionResource->apsId)) {
                throw new Exception(sprintf(self::EXCEPTION_RESOURCE_TYPE_MUST_BE_REFERENCE_MESSAGE, $subscriptionResource->id));
            }
            if (!isset($subscriptionResource->limit)) {
                throw new Exception(sprintf(self::EXCEPTION_RESOURCE_TYPE_CANNOT_BE_UNLIMITED_MESSAGE, $subscriptionResource->id));
            }
            if ($subscriptionResource->limit > 1) {
                throw new Exception(sprintf(self::EXCEPTION_RESOURCE_TYPE_CANNOT_HAVE_LIMIT_MORE_THEN_ONE_MESSAGE, $subscriptionResource->id));
            }
            if ($subscriptionResource->limit == 1) {
                $activatedTiers[] = $subscriptionResource->apsId;
            }
            $availableTiers[] = $subscriptionResource->apsId;
        }
        if (count($activatedTiers) > 1) {
            throw new Exception(self::EXCEPTION_MORE_THAN_1_COMMITMENT_MESSAGE);
        }
        if (empty($availableTiers)) {
            return;
        }
        $tiers = $this->getApsc()->getResources(
            RqlBuilder::createInstance()
                ->in('aps.id', $availableTiers)
                ->getQuery()
        );
        if (count($availableTiers) !== count($tiers)) {
            throw new Exception(self::EXCEPTION_CANNOT_FETCH_ALL_AVAILABLE_TIERS_MESSAGE);
        }

        // Verify levels.
        $levels = array_map(function ($tier) {
            return $tier->level;
        }, $tiers);
        // Exclude custom levels
        $levels = array_filter($levels, function ($level) {
            return $level >= minimalCommitmentTier::LEVEL_NO_COMMITMENT;
        });
        // A sequence of levels must start from level "No commitment", contain only unique values, and must not have gaps.
        sort($levels);
        if ($levels != range(minimalCommitmentTier::LEVEL_NO_COMMITMENT, (count($levels) + minimalCommitmentTier::LEVEL_NO_COMMITMENT) - 1)) {
            throw new Exception(self::EXCEPTION_INVALID_SEQUENCE_OF_MINIMUM_COMMITMENT_TIER_LEVELS_MESSAGE);
        }

        // Verify currencies
        $tenant = $this->tenant;
        $accountId = $tenant->account->id;
        $accountCurrencyInfo = $this->getPbaApi()->getVendorCurrency($accountId);
        foreach ($tiers as $tier) {
            if ($tier->level < minimalCommitmentTier::LEVEL_MINIMUM) {
                continue;
            }
            $tierCurrencies = array_map(function ($minimalCommitment) {
                return $minimalCommitment->currency;
            }, $tier->minimalCommitments);
            if (!in_array($accountCurrencyInfo->getCurrencyID(), $tierCurrencies, true)) {
                throw new Exception(self::EXCEPTION_TIERS_WITHOUT_ACCOUNT_CURRENCY_MESSAGE);
            }
        }
    }

    /**
     * @return array
     */
    protected function _getSubscriptionResources()
    {
        $tenant = $this->tenant;
        if (is_null($this->_subscriptionResources)) {
            $this->_subscriptionResources = $tenant->subscription->resources();
        }
        return $this->_subscriptionResources;
    }

    /**
     * @param globals $global
     * @param string $id
     * @return object
     * @throws Exception
     */
    protected function _getDatacenterById($global, $id)
    {
        $datacenters = $global->datacenters;
        foreach ($datacenters as $datacenter) {
            if ($datacenter->id == $id) {
                return $datacenter;
            }
        }
        $message = sprintf(self::EXCEPTION_CAN_NOT_FIND_DATACENTER_MESSAGE, $id);
        Log::getInstance()->error($message);
        throw new Exception($message);
    }

    /**
     * @return \APS\ResourceBase|stdClass
     */
    protected function getAdminUser()
    {
        $tenant = $this->tenant;
        return $this->getApsApi()->getAdmin($tenant);
    }

    /**
     * @return bool
     */
    protected function updateCounterLimits()
    {
        Log::getInstance()->event(self::LOG_CALLED_UPDATE_COUNTER_LIMITS);
        $propertyList = [];
        foreach ($this->getCountersManager()->getApplicationCounters(CloudApi::APPLICATION_TYPE_BACKUP) as $counter) {
            if ($counter->isFake()) {
                continue;
            }

            $propertyList[$counter->getName()] = null;
        }
        $resources = $this->_getSubscriptionResources();

        if (!is_array($resources)) {
            return false;
        }
        foreach ($resources as $resource) {
            if (!isset($resource->property)) {
                continue;
            }
            $propertyName = $resource->property;
            if (array_key_exists($propertyName, $propertyList)) {
                $propertyList[$propertyName] = $resource;
            }
        }

        $update = false;
        $tenant = $this->tenant;
        foreach ($propertyList as $propertyName => $resource) {
            if (!isset($tenant->{$propertyName})) {
                $tenant->{$propertyName} = \APS\TypeLibrary::newResourceByTypeId('http://aps-standard.org/types/core/resource/1.0#Counter');
                $update = true;
            }

            if (!isset($tenant->{$propertyName}->limit)) {
                $tenant->{$propertyName}->limit = 0;
                $update = true;
            }
            if (!isset($tenant->{$propertyName}->usage)) {
                $tenant->{$propertyName}->usage = 0;
                $update = true;
            }

            if (is_null($resource)) {
                if (!is_null($tenant->{$propertyName}->limit)) {
                    $tenant->{$propertyName}->limit = 0;
                    $update = true;
                }
                continue;
            }

            $limit = isset($resource->limit) ? $resource->limit : self::UNLIMITED_VALUE;
            if (in_array($propertyName, array(tenant::STORAGE_USAGE, tenant::LOCAL_STORAGE_USAGE)) && $limit > 0) {
                $limit = ceil(UomConvertHelper::convert($limit, $resource->unit, UomConvertHelper::MEASURABLE_B));
            }
            if ($tenant->{$propertyName}->limit != $limit) {
                $tenant->{$propertyName}->limit = $limit;
                $update = true;
            }
        }
        return $update;
    }

    /**
     * @param $login
     * @param $datacenterId
     * @return bool
     * @throws Service_BaseClient_HTTPException
     */
    protected function checkLoginExist($login, $datacenterId)
    {
        $tenant = $this->tenant;
        if ($tenant->type === tenant::TYPE_RESELLER) {
            $this->validateDatacenterId($datacenterId);
            $datacenter = $this->_getDatacenterById($tenant->global, $datacenterId);
        } elseif ($datacenterId) {
            $datacenter = $this->_getDatacenterById($tenant->global, $datacenterId);
        } else {
            $datacenter = $this->tenant->datacenter;
        }

        $cloudApi = $this->createCloudApiInstance($datacenter);

        return $cloudApi->checkLogin($login);
    }

    /**
     * @param $datacenterId
     * @throws RestException
     */
    protected function validateDatacenterId($datacenterId)
    {
        if (empty($datacenterId)) {
            throw new RestException(400, sprintf(
                self::EXCEPTION_FIELD_MUST_BE_PRESENT_MESSAGE,
                self::FIELD_STORAGE_DATACENTER_ID
            ));
        }
    }

    /**
     * @return PartnerCredentials
     */
    protected function getPartnerCredentials()
    {
        Log::getInstance()->event(self::LOG_CALLED_GET_PARTNER_CREDENTIALS);
        $tenant = $this->tenant;
        $datacenter = $tenant->datacenter;
        $credentials = new PartnerCredentials();
        $credentials->baasUrl = $datacenter->url;
        $credentials->partnerLogin = $datacenter->login;
        $credentials->partnerPassword = $datacenter->password;
        return $credentials;
    }

    /**
     * @param bool $tryLink
     */
    protected function _updateProfileLink($tryLink = true)
    {
        $apsApi = $this->getApsApi();
        $tenant = $this->tenant;
        $profile = $apsApi->getProfile($this->_getSubscriptionResources());
        if (is_null($profile)) {
            if (!$tenant->profile->isEmpty()) {
                Log::getInstance()->warning(
                    sprintf(
                        self::LOG_WARNING_BACKUP_PROFILE_HAS_BEEN_REMOVED,
                        $tenant->profile->resourceTypeId,
                        $tenant->aps->id
                    )
                );
            }
            return;
        }

        if ($tenant->profile->isEmpty()) {
            $tenant->profile->set($profile);
            if ($tryLink) {
                $this->getApsc()->linkResource($tenant, tenant::LINK_NAME_PROFILE, $tenant->profile);
            }
            return;
        }

        if ($profile->aps->id !== $tenant->profile->aps->id) {
            Log::getInstance()->warning(
                sprintf(
                    self::LOG_WARNING_BACKUP_PROFILE_HAS_BEEN_CHANGED,
                    $tenant->profile->resourceTypeId,
                    $profile->resourceTypeId,
                    $tenant->aps->id
                )
            );
        }
    }

    /**
     * @param $value
     * @return float
     */
    protected function convertStorageUsageToMb($value)
    {
        return ceil(UomConvertHelper::convert($value, 'b', 'mb'));
    }

    /**
     * @param $datacenter
     * @return int
     */
    protected function _getOrCreateParentGroupId($datacenter = null)
    {
        $tenant = $this->tenant;
        $dc = is_null($datacenter) ? $tenant->datacenter : $datacenter;
        $openApiRetriever = $this->managerFactory->getOpenApiRetrieverForDatacenter(
            $this->getApsc(),
            $tenant->global,
            $dc,
            $this->createMspApiInstance($dc),
            $this->getOpenApi()
        );
        return $openApiRetriever->getGroupIdByPoaAccountId($tenant->accountId);
    }

    /**
     * @return bool
     */
    protected function updateStartProductionTime()
    {
        $tenant = $this->tenant;
        if (is_null($tenant->startProductionTime) && !$tenant->subscription->trial) {
            $tenant->startProductionTime = time();
            return true;
        }
        return false;
    }

    /**
     * @param tenant $tenant
     * @param int $planId
     * @return null | int
     */
    protected function _getStorageForPlan($tenant, $planId)
    {
        $storageRelation = $this->_getStorageIdByPlanId($tenant, $planId);
        if (is_null($storageRelation)) {
            return NULL;
        }
        return $this->_getStorageById($tenant, $storageRelation);
    }

    /**
     * @param $tenant
     * @param $planId
     * @return null | StorageRelation
     */
    protected function _getStorageIdByPlanId($tenant, $planId)
    {
        if (is_array($tenant->storageRelations)) {
            foreach ($tenant->storageRelations as $relation) {
                if ($relation->planId == $planId) {
                    return $relation;
                }
            }
        }

        return NULL;
    }

    /**
     * @param tenant $tenant
     * @param StorageRelation $storageRelation
     * @return null | string
     */
    protected function _getStorageById($tenant, $storageRelation)
    {
        $storages = $tenant->getStorages();
        foreach ($storages as $storage) {
            if ($storage[self::FIELD_STORAGE_ID] == $storageRelation->storageId &&
                $storage[self::FIELD_STORAGE_DATACENTER_ID] == $storageRelation->datacenterId
            ) {
                return $storage;
            }
        }
        return NULL;
    }

    /**
     * @param Service_MspApiClient_ApiInterface|null $mspApi
     */
    protected function _migrateMspGroup(Service_MspApiClient_ApiInterface $mspApi = null)
    {
        if (is_null($mspApi)) {
            $mspApi = $this->getMspApi();
        }
        $tenant = $this->tenant;
        $mspGroupManager = $this->createMspGroupManagerInstance($mspApi, $this->getApsApi());
        $mspGroupManager->migrateGroupForCustomer($tenant->mspGroupId, $tenant, $this->_getOrCreateParentGroupId());
    }

    /**
     * @throws RestException
     */
    protected function throwNoContent()
    {
        throw new RestException(200);
    }

    /**
     * @return array
     * @throws RestException
     * @throws \Swagger\Client\ApiException
     * @throws \Swagger\Client\HttpException
     */
    protected function fetchTenantLimitsAndUsages()
    {
        $backupApplicationManager = new BackupApplication($this->tenant, $this->getMspApi());

        return $backupApplicationManager->fetchLimitsAndUsages();
    }
}