<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

/**
 * Class Model_Resource_Tenant_AcronisPartner
 */
class Model_Resource_Tenant_AcronisPartner extends Model_Resource_Tenant_Partner
{
    const EXCEPTION_YOUR_ACCOUNT_HAS_NO_AAN = 'Your account has no AAN.';
    const EXCEPTION_TARGET_ACCOUNT_HAS_NO_AAN = 'Your account has no AAN.';
    const EXCEPTION_ACCOUNT_HAS_DIFFERENT_AAN = 'The group with ID %s is assigned to account which has another AAN. Please, specify the correct groupId.';
    const EXCEPTION_SUBSCRIPTION_DIFFERENT_MC_TIER_MESSAGE = 'Subscription and account have different minimum commitment tiers. Account has resource "%s". Subscription has resource "%s".';
    const EXCEPTION_MINIMAL_COMMITMENT_IS_NOT_SPECIFIED_MESSAGE = 'Minimal commitment is not specified for account "%s". Please, specify account attribute "%s".';

    const LOG_EVENT_MIGRATE_GROUP_ID = 'Migrate group id "%s"';

    const DISCOUNT_ON_STORAGE_USAGE_FOR_WORKSTATIONS = 0.4;
    const NAME_PREFIX_TO_BE_DELETED = '[TO BE DELETED]';

    public function provision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->checkMinimalCommitmentTiers();
        $this->linkMinimalCommitmentTier(false);

        $this->checkMinimalCommitmentAccountAttributes();
        $this->setAccountAttributeMinimalCommitmentTier();

        $this->checkProfile();
        $this->linkProfile(false);

        $this->_resolveDatacenterAndStorageForPartner();
        $this->linkDatacenter(false);

        $this->updateCounterLimits();

        $tenant = $this->tenant;
        $subscriptionParametersManager = $this->getSubscriptionParametersManager();
        $tenant->wsDiscount = $subscriptionParametersManager->getParameterWSDiscount($tenant->subscriptionId);
        $groupId = $subscriptionParametersManager->getParameterGroupId($tenant->subscriptionId);
        if (is_null($groupId)) {
            $tenant->mspGroupId = $this->_createGroupForPartner();
        } else {
            Log::getInstance()->event(sprintf(self::LOG_EVENT_MIGRATE_GROUP_ID, $groupId));

            $tenant->ismigration = true;
            $this->checkGroupCanBeAssignedOrEx($groupId, $tenant->datacenter->id);

            $tenant->mspGroupId = $groupId;

            $this->_migrateMspGroup($this->getRootMspApi());
            $this->_updateGroupForPartner();
            $this->_retrieveUsageForPartner();
        }
    }

    public function provisionAsync()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateStartProductionTime();

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();

        $tenant = $this->tenant;
        if ($tenant->ismigration) {
            $tenant->ismigration = false;
            return;
        }

        //add provisional links on resources
        $this->linkMinimalCommitmentTier(false);
        $this->linkDatacenter(false);
        $this->linkProfile(false);

        $this->createAdministratorAccount();
    }

    public function retrieve()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkMinimalCommitmentTier($realLink);
        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->_migrateMspGroup();
        $this->_retrieveUsageForPartner();
    }

    public function unprovision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }

        $this->linkDatacenter(false);

        $mspApi = $this->getMspApi();

        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        if (is_null($groupInfo)) {
            return;
        }

        if (!$this->checkGroupHasUsage($groupInfo)) {
            parent::unprovision();
        } else {
            $updateGroupInfo = new Service_MspApiClient_Request_Struct_UpdateGroup();
            if ($groupInfo->getStatus() != Model_TenantManager::STATUS_INACTIVE) {
                $updateGroupInfo->setStatus(Model_TenantManager::STATUS_INACTIVE);
            }
            $mspGroupManager = new Model_MspGroupManager($mspApi, $this->getApsApi());
            $groupName = $mspGroupManager->fetchNameForGroup(
                self::NAME_PREFIX_TO_BE_DELETED . $groupInfo->getName(),
                $groupInfo
            );
            $updateGroupInfo->setName($groupName);
            $mspApi->updateGroup($groupInfo->getId(), $updateGroupInfo, $groupInfo->getVersion());
        }
    }

    public function enable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkMinimalCommitmentTier();
        parent::enable();
    }

    public function upgrade()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->wsDiscount)) {
            $tenant->wsDiscount = false;
        }

        $this->updateStartProductionTime();
        $this->updateCounterLimits();
        $this->update();

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();

        $realLink = !$tenant->subscription->disabled;
        if ($realLink) {
            $this->checkProfile();
        }

        $this->linkMinimalCommitmentTier($realLink);
        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);
    }

    public function onTenantAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkMinimalCommitmentTier();
        parent::onTenantAvailable($event);
    }

    public function onSubscriptionChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;

        $oldMinimalCommitmentTierId = null;
        if (!$tenant->minimalCommitmentTier->isEmpty()) {
            $oldMinimalCommitmentTierId = $tenant->minimalCommitmentTier->aps->id;
        }

        $realLink = !$tenant->subscription->disabled;
        $this->linkMinimalCommitmentTier($realLink);

        $newMinimalCommitmentTierId = null;
        if (!$tenant->minimalCommitmentTier->isEmpty()) {
            $newMinimalCommitmentTierId = $tenant->minimalCommitmentTier->aps->id;
        }

        if ($newMinimalCommitmentTierId && $oldMinimalCommitmentTierId != $newMinimalCommitmentTierId) {
            $this->setAccountAttributeMinimalCommitmentTier();
        }

        parent::onSubscriptionChange($event);
    }

    /**
     * @return int
     */
    private function _createGroupForPartner()
    {
        $tenant = $this->tenant;
        $account = $this->getPbaApi()->getAccount($tenant->accountId);
        $mspGroupManager = $this->getMspGroupManager();
        $parentMspGroupId = $this->_getOrCreateParentGroupId();
        return $mspGroupManager->createGroupForCustomer(
            $tenant,
            $parentMspGroupId,
            $tenant->storageId,
            Service_MspApiClient_Request_Struct_CreateGroup::RESELLER_GROUP_KIND_ID,
            $account->getLocaleID()
        );
    }

    protected function _retrieveUsageForPartner()
    {
        $tenant = $this->tenant;
        $mspApi = $this->getMspApi();
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        if (is_null($groupInfo)) {
            return;
        }
        $usageReportManager = $this->getUsageReportManager();
        $report = $usageReportManager->fetchCurrentReportForGroup($tenant->mspGroupId);

        if (is_null($report)) {
            $this->resetUsage();
            return;
        }

        $tenant->StorageUsage->usage = $report->totalStorage;
        $tenant->ServerCount->usage = $report->servers;
        $tenant->VmCount->usage = $report->vms;
        $tenant->WorkstationCount->usage = $report->workstations;
        $tenant->MobileDeviceCount->usage = $report->mobiles;
        $tenant->MailboxCount->usage = $report->mailboxes;
        $tenant->WebsiteCount->usage = $report->websites;

        $tenant->StorageUsageMbH->usage = $this->convertStorageUsageToMb($tenant->StorageUsage->usage);
        $tenant->ServerCountItemH->usage = $tenant->ServerCount->usage;
        $tenant->VmCountItemH->usage = $tenant->VmCount->usage;
        $tenant->WorkstationCountItemH->usage = $tenant->WorkstationCount->usage;
        $tenant->MobileDeviceCountItemH->usage = $tenant->MobileDeviceCount->usage;
        $tenant->MailboxCountItemH->usage = $tenant->MailboxCount->usage;

        $tenant->AcronisHostedStorage->usage = $report->acronisStorage;
        $tenant->ServiceProviderHostedStorage->usage = $report->partnerStorage;

        $tenant->AcronisHostedStorageMbH->usage = $this->convertStorageUsageToMb($tenant->AcronisHostedStorage->usage);
        $tenant->ServiceProviderHostedStorageMbH->usage = $this->convertStorageUsageToMb($tenant->ServiceProviderHostedStorage->usage);

        if (!$tenant->wsDiscount || $tenant->StorageUsage->usage == 0) {
            return;
        }
        $discountInBytes = floor(self::DISCOUNT_ON_STORAGE_USAGE_FOR_WORKSTATIONS * ($report->acronisWorkstationStorage + $report->partnerWorkstationStorage));
        $tenant->StorageUsage->usage -= $discountInBytes;
        $tenant->StorageUsageMbH->usage = $this->convertStorageUsageToMb($tenant->StorageUsage->usage);

        if ($tenant->AcronisHostedStorage->usage > 0) {
            $discountInBytes = floor(self::DISCOUNT_ON_STORAGE_USAGE_FOR_WORKSTATIONS * $report->acronisWorkstationStorage);
            $tenant->AcronisHostedStorage->usage -= $discountInBytes;
            $tenant->AcronisHostedStorageMbH->usage = $this->convertStorageUsageToMb($tenant->AcronisHostedStorage->usage);
        }
        if ($tenant->ServiceProviderHostedStorage->usage > 0) {
            $discountInBytes = floor(self::DISCOUNT_ON_STORAGE_USAGE_FOR_WORKSTATIONS * $report->partnerWorkstationStorage);
            $tenant->ServiceProviderHostedStorage->usage -= $discountInBytes;
            $tenant->ServiceProviderHostedStorageMbH->usage = $this->convertStorageUsageToMb($tenant->ServiceProviderHostedStorage->usage);
        }
    }

    /**
     * @throws \Rest\RestException
     */
    protected function checkMinimalCommitmentAccountAttributes()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if ($tenant->minimalCommitmentTier->isEmpty()) {
            return;
        }

        $accountAttributeManager = $this->getAccountAttributesManager();
        if ($tenant->minimalCommitmentTier->level == minimalCommitmentTier::LEVEL_FIXED_COMMITMENT) {
            $minimalCommitment = $accountAttributeManager->getAttributeMinimalCommitmentValue($tenant->accountId);
            if (is_null($minimalCommitment) || $minimalCommitment === '') {
                throw new \Rest\RestException(400, Str::format(
                    self::EXCEPTION_MINIMAL_COMMITMENT_IS_NOT_SPECIFIED_MESSAGE,
                    $tenant->accountId,
                    Model_OSA_AccountAttributesManagerInterface::ATTRIBUTE_MINIMAL_COMMITMENT_VALUE
                ));
            }
        }

        $minimalCommitmentResourceTypeId = $accountAttributeManager->getAttributeMinimalCommitmentTier($tenant->accountId);
        if (!empty($minimalCommitmentResourceTypeId) && $tenant->minimalCommitmentTier->resourceTypeId != $minimalCommitmentResourceTypeId) {
            throw new \Rest\RestException(
                409,
                Str::format(
                    self::EXCEPTION_SUBSCRIPTION_DIFFERENT_MC_TIER_MESSAGE,
                    $minimalCommitmentResourceTypeId,
                    $tenant->minimalCommitmentTier->resourceTypeId
                )
            );
        }
    }

    /**
     * @throws \Rest\RestException
     */
    protected function setAccountAttributeMinimalCommitmentTier()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if ($tenant->minimalCommitmentTier->isEmpty()) {
            return;
        }
        $accountId = $tenant->accountId;

        Log::getInstance()->event('getAccountAttributesManager');
        $accountAttributesManager = $this->getAccountAttributesManager();
        $newMinimalCommitmentTierId = $tenant->minimalCommitmentTier->resourceTypeId;
        $currentMinimalCommitmentTierId = $accountAttributesManager->getAttributeMinimalCommitmentTier($accountId);
        if ($currentMinimalCommitmentTierId != $newMinimalCommitmentTierId) {
            Log::getInstance()->event("Set value %s to attribute 'MinComTier'.", $newMinimalCommitmentTierId);
            $accountAttributesManager->setAttributeMinimalCommitmentTier($accountId, $newMinimalCommitmentTierId);
        }
        $newMinimumCommitmentValue = $this->getMinimalCommitmentValueFromTier();
        if (is_null($newMinimumCommitmentValue)) {
            Log::getInstance()->event("Account %s has no minimum commitment.", $accountId);
            return;
        }
        $currentMinimalCommitmentValue = $accountAttributesManager->getAttributeMinimalCommitmentValue($accountId);
        if ($currentMinimalCommitmentValue != $newMinimumCommitmentValue) {
            Log::getInstance()->event("Set value %s to attribute 'MinCom'.", $newMinimalCommitmentTierId);
            $accountAttributesManager->setAttributeMinimalCommitmentValue($accountId, $newMinimumCommitmentValue);
        }
    }

    private function getMinimalCommitmentValueFromTier()
    {
        $tenant = $this->tenant;
        if ($tenant->minimalCommitmentTier->isEmpty()) {
            return null;
        }

        $minimalCommitmentTier = $tenant->minimalCommitmentTier;
        if($minimalCommitmentTier->level <= minimalCommitmentTier::LEVEL_FIXED_COMMITMENT) {
            return null;
        }
        if($minimalCommitmentTier->level == minimalCommitmentTier::LEVEL_NO_COMMITMENT) {
            return 0;
        }

        $accountCurrencyInfo = $this->getPbaApi()->getVendorCurrency($tenant->accountId);
        $currency = $accountCurrencyInfo->getCurrencyID();

        if (!is_array($minimalCommitmentTier->minimalCommitments)) {
            throw new \Rest\RestException(404, Str::format(
                'Minimum commitment tier %s does not have minimum commitment for currency %s.',
                $minimalCommitmentTier->resourceTypeId, $currency
            ));
        }

        $minimalCommitments = array_filter($minimalCommitmentTier->minimalCommitments, function($minimalCommitment) use ($currency) {
           return $minimalCommitment->currency === $currency;
        });
        if (empty($minimalCommitments)) {
            throw new \Rest\RestException(404, Str::format(
                'Minimum commitment tier %s does not have minimum commitment for currency %s.',
                $minimalCommitmentTier->resourceTypeId, $currency
            ));
        }
        if (count($minimalCommitments) > 1) {
            throw new \Rest\RestException(409, Str::format(
                'Minimum commitment tier %s has several minimum commitment for currency %s.',
                $minimalCommitmentTier->resourceTypeId, $currency
            ));
        }
        $minimalCommitment = reset($minimalCommitments);
        return $minimalCommitment->value;
    }

    private function checkGroupHasUsage(Service_MspApiClient_Response_Struct_GroupInfo $groupInfo)
    {
        $groupUsage = $groupInfo->getUsage();
        return (
            $groupUsage->getMailboxCount() != 0
            || $groupUsage->getMobileCount() != 0
            || $groupUsage->getServerCount() != 0
            || $groupUsage->getVmCount() != 0
            || $groupUsage->getWorkstationCount() != 0
            || $groupUsage->getStorageSize() != 0
        );
    }

    protected function checkGroupCanBeAssignedOrEx($groupId, $datacenterId)
    {
        $tenants = $this->getApsApi()->getTenantsByGroupIdAndDatacenterId($groupId, $datacenterId);
        if (count($tenants) == 0) {
            return;
        }

        if (count($tenants) > 1) {
            throw new \Rest\RestException(409, Str::format(Service_APSApi_Api::EXCEPTION_GROUP_ALREADY_ASSIGNED, $groupId, $tenants[0]->subscription->subscriptionId));
        }

        $tenant = $this->tenant;
        $accountAttributesManager = $this->getAccountAttributesManager();
        $selfAAN = $accountAttributesManager->getAttributeAAN($tenant->account->id);

        if (empty($selfAAN)) {
            throw new \Rest\RestException(409, self::EXCEPTION_YOUR_ACCOUNT_HAS_NO_AAN);
        }

        $targetTenant = reset($tenants);
        $targetAAN = $accountAttributesManager->getAttributeAAN($targetTenant->account->id);

        if (empty($targetAAN)) {
            throw new \Rest\RestException(409, self::EXCEPTION_TARGET_ACCOUNT_HAS_NO_AAN);
        }

        if ($selfAAN != $targetAAN) {
            throw new \Rest\RestException(409, Str::format(self::EXCEPTION_ACCOUNT_HAS_DIFFERENT_AAN, $groupId));
        }

        $targetTenant->mspGroupId = null;
        $this->getApsc()->updateResource($targetTenant);
        $subscriptionParametersManager = $this->getSubscriptionParametersManager();
        $subscriptionParametersManager->setParameterExcludedFromReport(
            $targetTenant->subscriptionId,
            Model_OSA_SubscriptionParametersManagerInterface::EXCLUDE_FROM_REPORT_ON
        );
    }

    private function resetUsage()
    {
        $tenant = $this->tenant;
        $tenant->StorageUsage->usage = 0;
        $tenant->StorageUsageMbH->usage = 0;
        $tenant->ServerCount->usage = 0;
        $tenant->ServerCountItemH->usage = 0;
        $tenant->VmCount->usage = 0;
        $tenant->VmCountItemH->usage = 0;
        $tenant->WorkstationCount->usage = 0;
        $tenant->WorkstationCountItemH->usage = 0;
        $tenant->MobileDeviceCount->usage = 0;
        $tenant->MobileDeviceCountItemH->usage = 0;
        $tenant->MailboxCount->usage = 0;
        $tenant->MailboxCountItemH->usage = 0;
        $tenant->WebsiteCount->usage = 0;
        $tenant->AcronisHostedStorage->usage = 0;
        $tenant->AcronisHostedStorageMbH->usage = 0;
        $tenant->ServiceProviderHostedStorage->usage = 0;
        $tenant->ServiceProviderHostedStorageMbH->usage = 0;
    }

    /**
     * @param Service_MspApiClient_Request_Struct_AdministratorInfo $adminInfo
     * @return string
     */
    protected function createLoginForAdministrator(Service_MspApiClient_Request_Struct_AdministratorInfo $adminInfo)
    {
        return self::DEFAULT_ADMIN_USER_NAME_PREFIX . $this->tenant->subscriptionId;
    }
}
