<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

/**
 * Class Model_Resource_Tenant_Customer
 */
class Model_Resource_Tenant_Customer extends Model_Resource_Tenant
{
    /* ======= RESOURCE METHODS ======= */

    public function provision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->checkProfile();
        $this->linkProfile(false);

        $tenant = $this->tenant;
        $tenant->datacenterId = $this->resolveDatacenterId();
        $this->linkDatacenter(false);

        $this->updateCounterLimits();

        if ($tenant->ismigration && !is_null($tenant->mspGroupId)) {
            Log::getInstance()->event(self::LOG_MIGRATION_WAS_STARTED);
            $this->_migrateMspGroup();

            $this->_retrieveUsageForCustomer();
            $this->_updateGroupForCustomer();
        } else {
            $tenant->mspGroupId = $this->_createGroupForCustomer();
        }
        $this->updateTenantId();
    }

    public function provisionAsync()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateStartProductionTime();

        $this->linkDatacenter(false);
        $this->linkProfile(false);

        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    public function retrieve()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->updateCounterLimits();
        $this->_retrieveUsageForCustomer();
    }

    public function configure($new)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;

        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }

        $this->_updateGroupForCustomer();

        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

        $this->updateUsersLimits();
    }

    public function unprovision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if ($tenant->ismigration) {
            Log::getInstance()->event(self::LOG_UNPROVISIONING_TENANT_IS_MIGRATION);
            return;
        }

        if (is_null($tenant->mspGroupId)) {
            Log::getInstance()->warning(self::LOG_WARNING_UNPROVISION_TENANT_DOES_NOT_HAVE_GROUP);
            return;
        }
        Log::getInstance()->event(sprintf(self::LOG_FORMAT_DELETE_MSP_GROUP, $tenant->account->companyName));

        $this->linkDatacenter(false);

        try {
            $this->_deleteGroupForCustomer();
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                throw $e;
            }
            sleep(10);
            $this->_deleteGroupForCustomer();
        }
    }

    public function enable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();
        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->getTenantManager()->enableGroup($tenant->mspGroupId);
    }

    public function disable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->linkProfile(false);
        $this->linkDatacenter(false);
        $this->getTenantManager()->disableGroup($tenant->mspGroupId);
    }

    public function upgrade()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->datacenterId)) {
            $tenant->datacenterId = globals::DEFAULT_DATACENTER_ID;
        }

        $this->updateStartProductionTime();
        $this->updateCounterLimits();
        $this->updateTenantId();
        $this->update();

        $realLink = !$tenant->subscription->disabled;
        if ($realLink) {
            $this->checkProfile();
        }

        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);
        $this->_getOrCreateParentGroupId($tenant->datacenter);  // add parents in map
        $this->_updateGroupForCustomer();
        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);
        $this->_upgradeUsers();

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    /* ======= RESOURCE EVENTS ======= */

    public function onTenantAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();

        if ($this->tenant->ismigration) {
            return;
        }

        $applicationConfig = $this->getApplicationConfig();
        if (!$applicationConfig->service_auto_assignation) {
            return;
        }
        $admin = $this->getApsApi()->getAdmin($this->tenant);
        if (is_null($admin)) {
            return;
        }
        $this->provisionDefaultUser($admin);
    }

    public function onSubscriptionChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $needUpdate = $this->updateStartProductionTime();
        if ($this->updateCounterLimits()) {
            $needUpdate = true;
        }
        if ($needUpdate) {
            $this->update();
        }
        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);
        $this->_updateGroupForCustomer();
        $this->updateUsersLimits();
    }

    public function onUserChanged($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForCustomer();
    }

    public function onAccountChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForCustomer();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @return array
     * @throws Exception
     */
    public function getStorages()
    {
        Log::getInstance()->event(
            self::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        return $this->memoize(function () {
            $tenant = $this->tenant;
            $response = [];
            $storages = $this->getMspApi()->getStorages($tenant->mspGroupId);
            if (is_null($storages)) {
                Log::getInstance()->error(self::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
                throw new Exception(self::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
            }

            foreach ($storages as $storage) {
                /** @var Service_MspApiClient_Response_Struct_Storage $storage */
                if ($storage->getInherited()) {
                    continue;
                }
                $response[] = [
                    self::FIELD_STORAGE_ID => (string)$storage->getId(),
                    self::FIELD_STORAGE_NAME => $storage->getName(),
                ];
            }

            return $response;
        });
    }

    /**
     * @return object
     */
    public function getDatacenter()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkDatacenter($realLink);
        return (object)$tenant->datacenter->toArray();
    }

    /**
     * @param $login
     * @param $datacenterId
     * @return bool
     */
    public function loginExists($login, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        // Is the user's parent the same tenant?
        $user = $this->findUserByLogin($this->tenant->mspGroupId, $login);
        if (empty($user)) {
            // Does the user exist in Cloud
            return $this->checkLoginExist($login, $datacenterId);
        }

        // Is the user already related to an OA user?
        $rql = RqlBuilder::createInstance()
            ->implementing(APSResource::CLOUD_BACKUP_USERS_TYPE)
            ->eq('endUserLogin', $login)
            ->limit(0, 1)
            ->getQuery();
        $resources = $this->getApsc()->impersonate($this->tenant->global)->getResources($rql);
        return !empty($resources);
    }

    /**
     * @return array
     */
    public function getQuotasStatistics()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $quotasStatistics = array();
        foreach (self::$userQuotasPropertiesMap as $tenantQuotaProperty => $userQuotaProperty) {
            $allocatedTotal = 0;
            $tenant = $this->tenant;
            $rql = RqlBuilder::createInstance()
                ->implementing(APSResource::CLOUD_BACKUP_USERS_TYPE)
                ->eq('tenant.aps.id', $tenant->aps->id)
                ->getQuery();
            $users = $this->getApsc()->getResources($rql);
            foreach ($users as $user) {
                if (!isset($user->{$userQuotaProperty})) {
                    continue;
                }
                $counter = $user->{$userQuotaProperty};
                if (!is_object($counter) || !isset($counter->limit)) {
                    continue;
                }
                if ($counter->limit == self::UNLIMITED_VALUE) {
                    $allocatedTotal = self::UNLIMITED_VALUE;
                    break;
                }
                $allocatedTotal += $counter->limit;
            }

            $limit = isset($tenant->{$tenantQuotaProperty}->limit) ? $tenant->{$tenantQuotaProperty}->limit : 0;

            $quotasStatistics[lcfirst($tenantQuotaProperty)] = compact(
                'allocatedTotal',
                'limit'
            );
        }
        return $quotasStatistics;
    }

    /**
     * @return array
     */
    public function updateUsageAndLimits()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->updateCounterLimits();
        $this->_updateProfileLink();

        $tenantUsages = $this->fetchTenantLimitsAndUsages();

        $userQuota = $this->_getProfileUserQuota();
        // ToDo: Implement to retrieve statistics about user via baseUsers::retrieve & baseUsers::update
        $tenant = $this->tenant;
        $mspUsers = $this->getMspApi()->getUsers($tenant->mspGroupId);
        $userUsage = [];
        if (is_array($mspUsers)) {
            foreach ($mspUsers as $userInfo) {
                $item = [];
                $usage = $userInfo->getUsage();
                $privileges = $userInfo->getPrivileges();
                $item[users::STATISTICS_USER_STORAGE_USAGE] = [
                    static::USAGE => $usage->getStorageSize(),
                    static::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->StorageUsage->limit,
                        $privileges->getStorageSize()
                    ),
                ];
                $item[users::STATISTICS_USER_SERVER_COUNT] = [
                    static::USAGE => $usage->getServerCount(),
                    static::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->ServerCount->limit,
                        $privileges->getServerCount()
                    ),
                ];
                $item[users::STATISTICS_USER_VM_COUNT] = [
                    static::USAGE => $usage->getVmCount(),
                    static::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->VmCount->limit,
                        $privileges->getVmCount()
                    ),
                ];
                $item[users::STATISTICS_USER_WORKSTATION_COUNT] = [
                    static::USAGE => $usage->getWorkstationCount(),
                    static::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->WorkstationCount->limit,
                        $privileges->getWorkstationCount()
                    ),
                ];
                $item[users::STATISTICS_USER_MOBILE_DEVICE_COUNT] = [
                    static::USAGE => $usage->getMobileCount(),
                    static::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->MobileDeviceCount->limit,
                        $privileges->getMobileCount()
                    ),
                ];
                $item[users::STATISTICS_USER_WEB_HOSTING_SERVER_COUNT] = [
                    static::USAGE => $usage->getWebHostingServerCount(),
                    static::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->WebHostingServerCount->limit,
                        $privileges->getWebHostingServerCount()
                    ),
                ];
                $item[users::STATISTICS_USER_WEBSITE_COUNT] = [
                    static::USAGE => $usage->getWebsiteCount(),
                    static::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->WebsiteCount->limit,
                        $privileges->getWebsiteCount()
                    ),
                ];
                $userUsage[$userInfo->getId()] = $item;
            }
        }

        return [
            static::TENANT => $tenantUsages,
            static::USER_USAGE => $userUsage,
            static::ERROR => null,
        ];
    }

    public function createAdministrator($parameters, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministrators()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function updateAdministrator($adminId, $parameters, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function deleteAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getUserGuideUrl()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministratorConsole($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function resendAdministratorLink($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function enableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function disableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function adminUrlRedirect()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getDatacenters()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlans()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function setStorageForPlan($planId, $parameters)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function unsetStorageForPlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    /* ======= OTHER METHODS ======= */

    protected function createDefaultUserResource($tenant, $serviceUser)
    {
        $userQuota = Service_APSApi_Api::PROFILE_USER_QUOTA_LIMITS;
        $backupNotification = Enum_NotificationsBackup::DISABLE_NOTIFICATIONS;
        $managementNotification = false;

        $profile = $tenant->profile;
        if (!$profile->isEmpty()) {
            $userQuota = $profile->userQuota;
            $backupNotification = $profile->backupNotification;
            $managementNotification = (bool)$profile->managementNotification;
        }

        $user = \APS\TypeLibrary::newResourceByTypeId(\APS\TypeLibrary::getTypeByServiceId('users'));
        $user->user->set($serviceUser);
        $user->tenant->set($tenant);
        $user->notifyBackup = $backupNotification;
        $user->notificationsManagement = $managementNotification;

        $user->userStorageUsage = new Counter();
        $user->userServerCount = new Counter();
        $user->UserVmCount = new Counter();
        $user->UserWorkstationCount = new Counter();
        $user->UserMobileDeviceCount = new Counter();
        $user->userWebHostingServerCount = new Counter();
        $user->UserWebsiteCount = new Counter();

        if ($userQuota) {
            $user->userStorageUsage->limit = $this->getResourceLimit($tenant->StorageUsage->limit, $userQuota);
            $user->userServerCount->limit = $this->getResourceLimit($tenant->ServerCount->limit, $userQuota);
            $user->UserVmCount->limit = $this->getResourceLimit($tenant->VmCount->limit, $userQuota);
            $user->UserWorkstationCount->limit = $this->getResourceLimit($tenant->WorkstationCount->limit, $userQuota);
            $user->UserMobileDeviceCount->limit = $this->getResourceLimit($tenant->MobileDeviceCount->limit, $userQuota);
            $user->userWebHostingServerCount->limit = $this->getResourceLimit($tenant->WebHostingServerCount->limit, $userQuota);
            $user->UserWebsiteCount->limit = $this->getResourceLimit($tenant->WebsiteCount->limit, $userQuota);
        } else {
            $user->userStorageUsage->limit = self::UNLIMITED_VALUE;
            $user->userServerCount->limit = self::UNLIMITED_VALUE;
            $user->UserVmCount->limit = self::UNLIMITED_VALUE;
            $user->UserWorkstationCount->limit = self::UNLIMITED_VALUE;
            $user->UserMobileDeviceCount->limit = self::UNLIMITED_VALUE;
            $user->userWebHostingServerCount->limit = self::UNLIMITED_VALUE;
            $user->UserWebsiteCount->limit = self::UNLIMITED_VALUE;
        }

        $user->privileged = users::ADMINISTRATOR_PRIVILEGES_YES;
        $user->activateDuringCreation = false;
        return $user;
    }

    /**
     * @param $serviceUser
     */
    protected function provisionDefaultUser($serviceUser)
    {
        $tenant = $this->tenant;

        $user = $this->createDefaultUserResource($tenant, $serviceUser);

        $apsc = $this->getApsc()->impersonate($tenant);
        $apsc->resetSession();
        $apsc->provisionResource($user);
    }

    /**
     * @param $limit
     * @param $userQuota
     * @return int
     */
    protected function getResourceLimit($limit, $userQuota)
    {
        if (is_null($limit)) {
            return 0;
        }
        if ($limit == 0) {
            return 0;
        }
        if ($userQuota === Service_APSApi_Api::PROFILE_USER_QUOTA_ANY ||
            $userQuota === Service_APSApi_Api::PROFILE_USER_QUOTA_OFF
        ) {
            return self::UNLIMITED_VALUE;
        }
        return $limit; // Service_APSApi_Api::PROFILE_USER_QUOTA_LIMITS
    }

    /**
     * @return object
     */
    protected function getApplicationConfig()
    {
        return Settings::getInstance()->getApplicationConfig();
    }

    public function resolveDatacenterId()
    {
        $tenant = $this->tenant;
        // Do nothing if datacenter ID already was set from external side (online store / migration script)
        if (isset($tenant->datacenterId)) {
            return $tenant->datacenterId;
        }

        $datacenters = $tenant->global->datacenters;
        if ($datacenters->count() === 0) {
            throw new \Rest\RestException(401, static::EXCEPTION_NEED_TO_SET_CREDENTIALS_MESSAGE);
        }

        $profile = $tenant->profile;
        $storage = $profile->storagePolicy == profile::STORAGE_POLICY_PARENT
            ? $this->resolveProviderStorage($profile)
            : $this->resolveResellerStorage();

        return $storage->datacenterId;
    }

    public function resolveBackupStorageId()
    {
        $tenant = $this->tenant;

        $profile = $tenant->profile;
        $storage = $profile->storagePolicy == profile::STORAGE_POLICY_PARENT
            ? $this->resolveProviderStorage($profile)
            : $this->resolveResellerStorage();

        if ($tenant->datacenterId != $storage->datacenterId) {
            throw new \Rest\RestException(409, Str::format(
                'Please select storage from datacenter "%s" in the Backup Profile "%s".',
                $tenant->datacenterId, $profile->resourceTypeId
            ));
        };

        return $storage->storageId;
    }

    /**
     * @param $profile
     * @return object
     * @throws \Rest\RestException
     */
    private function resolveProviderStorage($profile)
    {
        $tenant = $this->tenant;
        $storages = is_array($profile->storages) ? $profile->storages : [];
        $storagesCount = count($storages);
        if ($storagesCount === 0) {
            throw new \Rest\RestException(400, self::EXCEPTION_SELECT_STORAGE_IN_BACKUP_PROFILE_MESSAGE);
        }

        if ($storagesCount === 1) {
            return reset($storages);
        }

        $subscriptionParametersManager = $this->getSubscriptionParametersManager();
        $storageOptionId = $subscriptionParametersManager->getParameterStorageId($tenant->subscription->subscriptionId);

        foreach ($storages as $storage) {
            if ($storageOptionId != $this->calculateStorageOptionId($storage->datacenterId, $storage->storageId)) {
                continue;
            }

            return $storage;
        }

        throw new \Rest\RestException(400, self::EXCEPTION_UNKNOWN_STORAGE_MESSAGE);
    }

    /**
     * @param $datacenterId
     * @param $storageId
     * @return int
     */
    private function calculateStorageOptionId($datacenterId, $storageId)
    {
        $optionId = ((int)$datacenterId) << 16;
        $optionId |= (((int)$storageId) & 0xffff);

        return $optionId;
    }

    private function resolveResellerStorage()
    {
        $tenant = $this->tenant;
        $parentTenant = $this->_getParentTenant($tenant);
        if (is_null($parentTenant)) {
            return null;
        }
        $planId = $this->_getServicePlanId();
        $storage = $this->_getStorageForPlan($parentTenant, $planId);

        if (is_null($storage)) {
            throw new Exception(self::EXCEPTION_RESELLER_HAS_TO_SET_STORAGE_FOR_PLAN_MESSAGE);
        }

        return (object)[
            'datacenterId' => $storage[self::FIELD_STORAGE_DATACENTER_ID],
            'storageId' => $storage[self::FIELD_STORAGE_ID],
        ];
    }

    /**
     * @param $tenant
     * @return \APS\ResourceBase
     */
    private function _getParentTenant($tenant)
    {
        $accountId = $this->_getParentAccountId($tenant->accountId);
        $tenants = $this->getApsApi()->getTenantsByAccountId($accountId);
        return count($tenants) ? $tenants[0] : null;
    }

    /**
     * @param int $accountId
     * @return int
     */
    private function _getParentAccountId($accountId)
    {
        $openApi = $this->getOpenApi();
        $accountInfo = $openApi->getAccountInfo($accountId);
        return $accountInfo->getParentAccountId();
    }

    /**
     * @return int
     * @throws Exception | Service_PbaApi_Api_Exception
     */
    private function _getServicePlanId()
    {
        $tenant = $this->tenant;
        $pbaApi = $this->getPbaApi();
        $subscriptionDetails = $pbaApi->getSubscriptionDetails($tenant->subscription->subscriptionId);
        return $subscriptionDetails->getPlanId();
    }

    /**
     * @throws \Acronis\App\UsageReport\Exception\ManagerException
     * @throws \Rest\RestException
     */
    private function _retrieveUsageForCustomer()
    {
        $this->updateCountersUsage($this->tenant->datacenter);
    }

    /**
     * @return void
     */
    private function _updateGroupForCustomer()
    {
        $tenant = $this->tenant;
        $mspGroupManager = $this->getMspGroupManager();
        $mspGroupManager->updateGroupForCustomer($tenant->mspGroupId, $tenant);
        $this->_lockGroup($tenant->mspGroupId);
    }

    private function _lockGroup($groupId)
    {
        try {
            $this->getMspApi()->lockGroup($groupId);
        } catch (Exception $e) {
            Log::getInstance()->error(sprintf(self::ERROR_FAILED_LOCK_GROUP, $groupId));
        }
    }

    /**
     * @return int
     */
    private function _createGroupForCustomer()
    {
        $tenant = $this->tenant;
        $parentMspGroupId = $this->_getOrCreateParentGroupId();
        $mspGroupManager = $this->getMspGroupManager();
        $groupId = $mspGroupManager->createGroupForCustomer($tenant, $parentMspGroupId);
        $this->_lockGroup($groupId);
        return $groupId;
    }

    private function updateUsersLimits()
    {
        $userQuota = $this->_getProfileUserQuota();
        $tenant = $this->tenant;
        foreach ($tenant->users as $user) {
            $this->updateUserApsResource($tenant, $user, $userQuota);

            try {
                $this->updateUserCloudAccount($user);
            } catch (Service_BaseClient_HTTPException $e) {
                if ($e->getCode() !== Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                    throw $e;
                }
                $this->updateUserCloudAccount($user);
            }
        }
    }

    /**
     * @param tenant $tenant
     * @param $user
     * @param $userQuota
     * @throws Exception
     */
    private function updateUserApsResource(tenant $tenant, $user, $userQuota)
    {
        $update = false;
        foreach (Model_Resource_Tenant::$userQuotasPropertiesMap as $tenantCounterName => $userCounterName) {
            if (is_null($user->{$userCounterName})) {
                $user->{$userCounterName} = new Counter();
                $user->{$userCounterName}->limit = 0;
                $user->{$userCounterName}->usage = 0;
            }

            if (empty($tenant->{$tenantCounterName}->limit)) {
                if ($user->{$userCounterName}->limit != 0) {
                    $user->{$userCounterName}->limit = 0;
                    $update = true;
                }
            } elseif (!$userQuota) {
                if ($user->{$userCounterName}->limit != self::UNLIMITED_VALUE) {
                    $user->{$userCounterName}->limit = self::UNLIMITED_VALUE;
                    $update = true;
                }
            }
        }
        if ($update) {
            /** @var \APS\ControllerProxy $apsc */
            $apsc = $tenant->getAPSC();
            $apsc->updateResource($user->resource);
        }
    }

    /**
     * @param $user
     * @throws Service_BaseClient_HTTPException
     */
    private function updateUserCloudAccount($user)
    {
        if (!$user->mspUserId) {
            return;
        }

        $mspApi = $this->getMspApi();
        $userInfo = $mspApi->getUserById($user->mspUserId);

        $oldPrivileges = $userInfo->getPrivileges();
        $newPrivileges = new Service_MspApiClient_Request_Struct_UserPrivileges();

        $limit = $user->userStorageUsage->limit;
        if ($limit != $oldPrivileges->getStorageSize()) {
            $newPrivileges->setStorageSize($limit);
        }
        $limit = $user->userServerCount->limit;
        if ($limit != $oldPrivileges->getServerCount()) {
            $newPrivileges->setServerCount($limit);
        }
        $limit = $user->UserVmCount->limit;
        if ($limit != $oldPrivileges->getVmCount()) {
            $newPrivileges->setVmCount($limit);
        }
        $limit = $user->UserWorkstationCount->limit;
        if ($limit != $oldPrivileges->getWorkstationCount()) {
            $newPrivileges->setWorkstationCount($limit);
        }
        $limit = $user->UserMobileDeviceCount->limit;
        if ($limit != $oldPrivileges->getMobileCount()) {
            $newPrivileges->setMobileCount($limit);
        }
        $limit = $user->userWebHostingServerCount->limit;
        if ($limit != $oldPrivileges->getWebHostingServerCount()) {
            $newPrivileges->setWebHostingServerCount($limit);
        }
        $limit = $user->UserWebsiteCount->limit;
        if ($limit != $oldPrivileges->getWebsiteCount()) {
            $newPrivileges->setWebsiteCount($limit);
        }

        if (!$newPrivileges->isExistsUpdateFields()) {
            return;
        }

        $updateUserInfo = new Service_MspApiClient_Request_Struct_UpdateUserInfo();
        $updateUserInfo->setPrivileges($newPrivileges);

        try {
            $mspApi->updateUser($user->mspUserId, $updateUserInfo, $userInfo->getVersion());
        } catch (Service_BaseClient_HTTPException $e) {
            $skipCodes = [
                Service_MspApiClient_Api::EXCEPTION_HTTP_CODE_NOT_FOUND,
                Service_MspApiClient_Api::EXCEPTION_HTTP_CODE_NOT_AVAILABLE,
            ];
            if (!in_array($e->getCode(), $skipCodes)) {
                throw $e;
            }
            Log::getInstance()->warning(
                'Unable to update user "%s". Error: "%s".',
                $user->mspUserId, $e->getMessage()
            );
        }
    }

    /**
     * @return bool
     */
    private function _getProfileUserQuota()
    {
        $tenant = $this->tenant;
        $profile = $tenant->profile;
        if ($profile->isEmpty()) {
            $userQuota = true;
        } else {
            $userQuota = $profile->userQuota !== Service_APSApi_Api::PROFILE_USER_QUOTA_OFF;
        }
        return $userQuota;
    }

    /**
     * @throws Service_BaseClient_HTTPException
     */
    private function _deleteGroupForCustomer()
    {
        $mspApi = $this->getMspApi();
        $tenant = $this->tenant;
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        if (is_null($groupInfo)) {
            Log::getInstance()->warning(sprintf(self::LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP, $tenant->mspGroupId));
            return;
        }
        $version = $groupInfo->getVersion();
        if ($groupInfo->getStatus() == 1) {
            $version = $this->getTenantManager()->disableGroup($tenant->mspGroupId, $version);
        }
        $mspApi->deleteGroup($tenant->mspGroupId, $version);
    }

    private function _getCounterLimit($userQuota, $tenantCounterLimit, $userCounterLimit)
    {
        if ($userQuota || $tenantCounterLimit) {
            return $userCounterLimit;
        }
        return $tenantCounterLimit;
    }

    private function _upgradeUsers()
    {
        $tenant = $this->tenant;
        $mspApi = $this->getMspApi();
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        $status = $groupInfo->getStatus();
        $version = $groupInfo->getVersion();
        $mspGroupManager = $this->getTenantManager();
        if ($status == Service_MspApiClient_Request_Struct_UpdateGroup::DISABLED_GROUP_STATUS) {
            $version = $mspGroupManager->enableGroup($tenant->mspGroupId, $version);
        }

        $exception = null;
        try {
            $users = $tenant->users;
            if (!count($users)) {
                $users = $this->getApsApi()->getUsersForTenant($tenant);
            }
            foreach ($users as $resource) {
                $this->upgradeUser($resource->aps->id);
            }
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
            $exception = $e;
        }

        if ($status == Service_MspApiClient_Request_Struct_UpdateGroup::DISABLED_GROUP_STATUS) {
            $mspGroupManager->disableGroup($tenant->mspGroupId, $version);
        }

        if (!is_null($exception)) {
            throw $exception;
        }
    }

    /**
     * @param $userId
     */
    private function upgradeUser($userId)
    {
        $user = $this->tenant->createUser();
        $user->fillResource($userId);
        $user->upgrade();
    }

    /**
     * @param $mspGroupId
     * @param $endUserLogin
     * @return null|Service_MspApiClient_Response_Struct_UserItem
     * @throws Exception
     */
    private function findUserByLogin($mspGroupId, $endUserLogin)
    {
        $users = $this->getMspApi()->getUsers($mspGroupId);
        foreach ($users as $user) {
            if ($user->getLogin() === $endUserLogin) {
                return $user;
            }
        }
        return null;
    }
}
