<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

/**
 * Class Model_Resource_Tenant_Customer
 */
class Model_Resource_Tenant_Customer extends Model_Resource_Tenant
{
    /* ======= RESOURCE METHODS ======= */

    public function provision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->checkProfile();
        $this->linkProfile(false);

        $this->_resolveDatacenterAndStorageForCustomer();
        $this->linkDatacenter(false);

        $this->updateCounterLimits();

        $tenant = $this->tenant;
        if ($tenant->ismigration && !is_null($tenant->mspGroupId)) {
            Log::getInstance()->event(self::LOG_MIGRATION_WAS_STARTED);
            $this->_migrateMspGroup();

            $this->_retrieveUsageForCustomer();
            $this->_updateGroupForCustomer();
        } else {
            $tenant->mspGroupId = $this->_createGroupForCustomer();
        }
    }

    public function provisionAsync()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateStartProductionTime();

        $this->linkDatacenter(false);
        $this->linkProfile(false);

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    public function retrieve()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->_retrieveUsageForCustomer();
    }

    public function configure($new)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;

        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }

        $this->_updateGroupForCustomer();
        $this->updateUsersLimits();
    }

    public function unprovision()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if ($tenant->ismigration) {
            Log::getInstance()->event(self::LOG_UNPROVISIONING_TENANT_IS_MIGRATION);
            return;
        }

        if (is_null($tenant->mspGroupId)) {
            Log::getInstance()->warning(self::LOG_WARNING_UNPROVISION_TENANT_DOES_NOT_HAVE_GROUP);
            return;
        }
        Log::getInstance()->event(sprintf(self::LOG_FORMAT_DELETE_MSP_GROUP, $tenant->account->companyName));

        $this->linkDatacenter(false);

        try {
            $this->_deleteGroupForCustomer();
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                throw $e;
            }
            sleep(10);
            $this->_deleteGroupForCustomer();
        }
    }

    public function enable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();
        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->getTenantManager()->enableGroup($tenant->mspGroupId);
    }

    public function disable()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->linkProfile(false);
        $this->linkDatacenter(false);
        $this->getTenantManager()->disableGroup($tenant->mspGroupId);
    }

    public function upgrade()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->datacenterId)) {
            $tenant->datacenterId = globals::DEFAULT_DATACENTER_ID;
        }

        $this->updateStartProductionTime();
        $this->updateCounterLimits();
        $this->update();

        $realLink = !$tenant->subscription->disabled;
        if ($realLink) {
            $this->checkProfile();
        }

        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);
        $this->_getOrCreateParentGroupId($tenant->datacenter);  // add parents in map
        $this->_updateGroupForCustomer();
        $this->_upgradeUsers();

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    /* ======= RESOURCE EVENTS ======= */

    public function onTenantAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();
    }

    public function onSubscriptionChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $needUpdate = $this->updateStartProductionTime();
        if ($this->updateCounterLimits()) {
            $needUpdate = true;
        }
        if ($needUpdate) {
            $this->update();
        }
        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);
        $this->_updateGroupForCustomer();
        $this->updateUsersLimits();
    }

    public function onUserChanged($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForCustomer();
    }

    public function onAccountChange($event)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForCustomer();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @return array
     */
    public function getOptionalCounters()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $privileges = $this->getMspApi()->getAccountPrivileges();
        $counters = array(
            self::FIELD_MOBILE_COUNTER => $privileges[Service_MspApiClient_Api::CACHE_MOBILE_COUNTER],
            self::FIELD_MAILBOX_COUNTER => $privileges[Service_MspApiClient_Api::CACHE_MAILBOX_COUNTER],
            self::FIELD_WEBSITE_COUNTER => $privileges[Service_MspApiClient_Api::CACHE_WEBSITE_COUNTER],
        );
        return $counters;
    }

    /**
     * @return array
     * @throws Exception
     */
    public function getStorages()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $response = array();
        $storages = $this->getMspApi()->getStorages($tenant->mspGroupId);
        if (is_null($storages)) {
            Log::getInstance()->error(self::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
            throw new Exception(self::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
        }

        foreach ($storages as $storage) {
            /** @var Service_MspApiClient_Response_Struct_Storage $storage */
            if ($storage->getInherited()) {
                continue;
            }
            $response[] = array(
                self::FIELD_STORAGE_ID => (string)$storage->getId(),
                self::FIELD_STORAGE_NAME => $storage->getName()
            );
        }

        return $response;
    }

    /**
     * @return object
     */
    public function getDatacenter()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkDatacenter($realLink);
        return (object)$tenant->datacenter->toArray();
    }

    /**
     * @param $login
     * @param $datacenterId
     * @return bool
     */
    public function loginExists($login, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        return $this->checkLoginExist($login, $datacenterId);
    }

    /**
     * @param $language
     */
    public function getAgents($language)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        return $this->getAgentsForAll($language);
    }

    /**
     * @return array
     */
    public function getQuotasStatistics()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $quotasStatistics = array();
        foreach (self::$quotasPropertiesMap as $tenantQuotaProperty => $userQuotaProperty) {
            $allocatedTotal = 0;
            $tenant = $this->tenant;
            $rql = RqlBuilder::createInstance()
                ->implementing(APSResource::CLOUD_BACKUP_USERS_TYPE)
                ->eq('tenant.aps.id', $tenant->aps->id)
                ->getQuery();
            $users = $this->getApsc()->getResources($rql);
            foreach ($users as $user) {
                if (!isset($user->{$userQuotaProperty})) {
                    continue;
                }
                $counter = $user->{$userQuotaProperty};
                if (!is_object($counter) || !isset($counter->limit)) {
                    continue;
                }
                if ($counter->limit == self::UNLIMITED_VALUE) {
                    $allocatedTotal = self::UNLIMITED_VALUE;
                    break;
                }
                $allocatedTotal += $counter->limit;
            }

            $limit = isset($tenant->{$tenantQuotaProperty}->limit) ? $tenant->{$tenantQuotaProperty}->limit : 0;

            $quotasStatistics[lcfirst($tenantQuotaProperty)] = compact(
                'allocatedTotal',
                'limit'
            );
        }
        return $quotasStatistics;
    }

    /**
     * @return array
     */
    public function updateUsageAndLimits()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateCounterLimits();
        $this->_updateProfileLink();

        $error = null;
        try {
            $this->retrieve();
            $this->update();
        } catch (Exception $e) {
            $error = $e->getMessage();
            Log::getInstance()->error($error);
        }

        $userQuota = $this->_getProfileUserQuota();
        // ToDo: Implement to retrieve statistics about user via baseUsers::retrieve & baseUsers::update
        $tenant = $this->tenant;
        $mspUsers = $this->getMspApi()->getUsers($tenant->mspGroupId);
        $userUsage = array();
        if (is_array($mspUsers)) {
            foreach ($mspUsers as $userInfo) {
                $item = array();
                $usage = $userInfo->getUsage();
                $privileges = $userInfo->getPrivileges();
                $item[users::STATISTICS_USER_STORAGE_USAGE] = array(
                    self::USAGE => $usage->getStorageSize(),
                    self::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->StorageUsage->limit,
                        $privileges->getStorageSize()
                    )
                );
                $item[users::STATISTICS_USER_SERVER_COUNT] = array(
                    self::USAGE => $usage->getServerCount(),
                    self::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->ServerCount->limit,
                        $privileges->getServerCount()
                    )
                );
                $item[users::STATISTICS_USER_VM_COUNT] = array(
                    self::USAGE => $usage->getVmCount(),
                    self::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->VmCount->limit,
                        $privileges->getVmCount()
                    )
                );
                $item[users::STATISTICS_USER_WORKSTATION_COUNT] = array(
                    self::USAGE => $usage->getWorkstationCount(),
                    self::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->WorkstationCount->limit,
                        $privileges->getWorkstationCount()
                    )
                );
                $item[users::STATISTICS_USER_MOBILE_DEVICE_COUNT] = array(
                    self::USAGE => $usage->getMobileCount(),
                    self::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->MobileDeviceCount->limit,
                        $privileges->getMobileCount()
                    )
                );
                $item[users::STATISTICS_USER_MAILBOX_COUNT] = array(
                    self::USAGE => $usage->getMailboxCount(),
                    self::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->MailboxCount->limit,
                        $privileges->getMailboxCount()
                    )
                );
                $item[users::STATISTICS_USER_WEBSITE_COUNT] = array(
                    self::USAGE => $usage->getWebsiteCount(),
                    self::LIMIT => $this->_getCounterLimit(
                        $userQuota,
                        $tenant->WebsiteCount->limit,
                        $privileges->getWebsiteCount()
                    )
                );
                $userUsage[$userInfo->getId()] = $item;
            }
        }

        return array(
            self::TENANT => array(
                tenant::STORAGE_USAGE => array(
                    self::USAGE => $tenant->StorageUsage->usage,
                    self::LIMIT => $tenant->StorageUsage->limit
                ),
                tenant::SERVER_COUNT => array(
                    self::USAGE => $tenant->ServerCount->usage,
                    self::LIMIT => $tenant->ServerCount->limit
                ),
                tenant::VM_COUNT => array(
                    self::USAGE => $tenant->VmCount->usage,
                    self::LIMIT => $tenant->VmCount->limit
                ),
                tenant::WORKSTATION_COUNT => array(
                    self::USAGE => $tenant->WorkstationCount->usage,
                    self::LIMIT => $tenant->WorkstationCount->limit
                ),
                tenant::MOBILE_DEVICE_COUNT => array(
                    self::USAGE => $tenant->MobileDeviceCount->usage,
                    self::LIMIT => $tenant->MobileDeviceCount->limit
                ),
                tenant::MAILBOX_COUNT => array(
                    self::USAGE => $tenant->MailboxCount->usage,
                    self::LIMIT => $tenant->MailboxCount->limit
                ),
                tenant::WEBSITE_COUNT => array(
                    self::USAGE => $tenant->WebsiteCount->usage,
                    self::LIMIT => $tenant->WebsiteCount->limit
                )
            ),
            self::USER_USAGE => $userUsage,
            self::ERROR => $error
        );
    }

    public function createAdministrator($parameters, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministrators()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function updateAdministrator($adminId, $parameters, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function deleteAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getUserGuideUrl()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getAdministratorConsole($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function resendAdministratorLink($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function enableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function disableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function adminUrlRedirect()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getDatacenters()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlans()
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function setStorageForPlan($planId, $parameters)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function unsetStorageForPlan($planId)
    {
        Log::getInstance()->event(Str::format(
            self::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    /* ======= OTHER METHODS ======= */

    private function _resolveDatacenterAndStorageForCustomer()
    {
        $tenant = $this->tenant;
        // Do nothing if datacenter ID already was set from external side (online store / migration script)
        if (isset($tenant->datacenterId)) {
            return;
        }

        $datacenters = $tenant->global->datacenters;
        $datacentersCount = count($datacenters);
        if ($datacentersCount === 0) {
            throw new \Rest\RestException(401, self::EXCEPTION_NEED_TO_SET_CREDENTIALS_MESSAGE);
        }

        $profile = $tenant->profile;
        if ($profile->isEmpty()) {
            if ($datacentersCount > 1) {
                throw new \Rest\RestException(400, self::EXCEPTION_DATACENTER_DOES_NOT_SET_MESSAGE);
            }
            $datacenter = $datacenters[0];
            $tenant->datacenterId = $datacenter->id;
            $tenant->storageId = null;
            return;
        }

        if ($profile->storagePolicy == profile::STORAGE_POLICY_PARENT) {
            $this->resolveProviderStorage($profile, $datacenters);
            return;
        }

        $this->resolveResellerStorage();
    }

    /**
     * @param $profile
     * @param $datacenters
     * @throws \Rest\RestException
     */
    private function resolveProviderStorage($profile, $datacenters)
    {
        $tenant = $this->tenant;
        $datacentersCount = count($datacenters);
        $storages = is_array($profile->storages) ? $profile->storages : array();
        $storagesCount = count($storages);
        if ($storagesCount === 0) {
            if ($datacentersCount > 1) {
                throw new \Rest\RestException(400, self::EXCEPTION_SELECT_STORAGE_IN_BACKUP_PROFILE_MESSAGE);
            }
            $datacenter = $datacenters[0];
            $tenant->datacenterId = $datacenter->id;
            $tenant->storageId = null;

            return;
        }

        if ($storagesCount === 1) {
            $storage = $storages[0];
            $tenant->datacenterId = $storage->datacenterId;
            $tenant->storageId = $storage->storageId;

            return;
        }

        $subscriptionParametersManager = $this->getSubscriptionParametersManager();
        $storageOptionId = $subscriptionParametersManager->getParameterStorageId($tenant->subscription->subscriptionId);

        foreach ($storages as $storage) {
            if ($storageOptionId == $this->calculateStorageOptionId($storage->datacenterId, $storage->storageId)) {
                $tenant->datacenterId = $storage->datacenterId;
                $tenant->storageId = $storage->storageId;

                return;
            }
        }
        throw new \Rest\RestException(400, self::EXCEPTION_UNKNOWN_STORAGE_MESSAGE);
    }

    /**
     * @param $datacenterId
     * @param $storageId
     * @return int
     */
    private function calculateStorageOptionId($datacenterId, $storageId)
    {
        $optionId = ((int)$datacenterId) << 16;
        $optionId |= (((int)$storageId) & 0xffff);
        return $optionId;
    }

    private function resolveResellerStorage()
    {
        $tenant = $this->tenant;
        $parentTenant = $this->_getParentTenant($tenant);
        if (is_null($parentTenant)) {
            return NULL;
        }
        $planId = $this->_getServicePlanId();
        $storage = $this->_getStorageForPlan($parentTenant, $planId);

        if (is_null($storage)) {
            throw new Exception(self::EXCEPTION_RESELLER_HAS_TO_SET_STORAGE_FOR_PLAN_MESSAGE);
        }
        $tenant->datacenterId = $storage[self::FIELD_STORAGE_DATACENTER_ID];
        $tenant->storageId = $storage[self::FIELD_STORAGE_ID];
    }

    /**
     * @param $tenant
     * @return \APS\ResourceBase
     */
    private function _getParentTenant($tenant)
    {
        $accountId = $this->_getParentAccountId($tenant->accountId);
        $tenants = $this->getApsApi()->getTenantsByAccountId($accountId);
        return count($tenants) ? $tenants[0] : null;
    }

    /**
     * @param int $accountId
     * @return int
     */
    private function _getParentAccountId($accountId)
    {
        $openApi = $this->getOpenApi();
        $accountInfo = $openApi->getAccountInfo($accountId);
        return $accountInfo->getParentAccountId();
    }

    /**
     * @return int
     * @throws Exception | Service_PbaApi_Api_Exception
     */
    private function _getServicePlanId()
    {
        $tenant = $this->tenant;
        $pbaApi = $this->getPbaApi();
        $subscriptionDetails = $pbaApi->getSubscriptionDetails($tenant->subscription->subscriptionId);
        return $subscriptionDetails->getPlanId();
    }

    private function _retrieveUsageForCustomer()
    {
        $mspApi = $this->getMspApi();
        $tenant = $this->tenant;
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        if (is_null($groupInfo)) {
            return;
        }
        $groupUsage = $groupInfo->getUsage();
        $tenant->StorageUsage->usage = $groupUsage->getStorageSize();
        $tenant->StorageUsageMbH->usage = $this->convertStorageUsageToMb($groupUsage->getStorageSize());
        $tenant->ServerCount->usage = $groupUsage->getServerCount();
        $tenant->ServerCountItemH->usage = $groupUsage->getServerCount();
        $tenant->VmCount->usage = $groupUsage->getVmCount();
        $tenant->VmCountItemH->usage = $groupUsage->getVmCount();
        $tenant->WorkstationCount->usage = $groupUsage->getWorkstationCount();
        $tenant->WorkstationCountItemH->usage = $groupUsage->getWorkstationCount();
        $tenant->MobileDeviceCount->usage = $groupUsage->getMobileCount();
        $tenant->MobileDeviceCountItemH->usage = $groupUsage->getMobileCount();
        $tenant->MailboxCount->usage = $groupUsage->getMailboxCount();
        $tenant->MailboxCountItemH->usage = $groupUsage->getMailboxCount();
        $tenant->WebsiteCount->usage = $groupUsage->getWebsiteCount();

        $tenant->AcronisHostedStorage->usage = 0;
        $tenant->AcronisHostedStorageMbH->usage = 0;
        $tenant->ServiceProviderHostedStorage->usage = 0;
        $tenant->ServiceProviderHostedStorageMbH->usage = 0;

        if ($tenant->StorageUsage->usage == 0) {
            return;
        }

        $storages = $mspApi->getStoragesByGroupId($tenant->mspGroupId);
        foreach ($storages as $storage) {
            if ($storage->getInherited() !== self::INHERITED_STORAGE) {
                continue;
            }
            $tenant->AcronisHostedStorage->usage += intval($storage->getSpaceUsage());
        }
        $tenant->AcronisHostedStorageMbH->usage = $this->convertStorageUsageToMb($tenant->AcronisHostedStorage->usage);
        $tenant->ServiceProviderHostedStorage->usage = $tenant->StorageUsage->usage - $tenant->AcronisHostedStorage->usage;
        $tenant->ServiceProviderHostedStorageMbH->usage = $this->convertStorageUsageToMb($tenant->ServiceProviderHostedStorage->usage);
    }

    /**
     * @return void
     */
    private function _updateGroupForCustomer()
    {
        $tenant = $this->tenant;
        $mspGroupManager = $this->getMspGroupManager();
        $mspGroupManager->updateGroupForCustomer($tenant->mspGroupId, $tenant);
        $this->_lockGroup($tenant->mspGroupId);
    }

    private function _lockGroup($groupId)
    {
        try {
            $this->getMspApi()->lockGroup($groupId);
        } catch (Exception $e) {
            Log::getInstance()->error(sprintf(self::ERROR_FAILED_LOCK_GROUP, $groupId));
        }
    }

    /**
     * @return int
     */
    private function _createGroupForCustomer()
    {
        $tenant = $this->tenant;
        $parentMspGroupId = $this->_getOrCreateParentGroupId();
        $mspGroupManager = $this->getMspGroupManager();
        $groupId = $mspGroupManager->createGroupForCustomer($tenant, $parentMspGroupId, $tenant->storageId);
        $this->_lockGroup($groupId);
        return $groupId;
    }

    private function updateUsersLimits()
    {
        if (!$this->_getProfileUserQuota()) {
            return;
        }
        $tenant = $this->tenant;
        foreach ($tenant->users as $user) {
            $update = false;
            if (empty($tenant->StorageUsage->limit) && $user->userStorageUsage->limit != 0) {
                $user->userStorageUsage->limit = 0;
                $update = true;
            }
            if (empty($tenant->ServerCount->limit) && $user->userServerCount->limit != 0) {
                $user->userServerCount->limit = 0;
                $update = true;
            }
            if (empty($tenant->VmCount->limit) && $user->UserVmCount->limit != 0) {
                $user->UserVmCount->limit = 0;
                $update = true;
            }
            if (empty($tenant->WorkstationCount->limit) && $user->UserWorkstationCount->limit != 0) {
                $user->UserWorkstationCount->limit = 0;
                $update = true;
            }
            if (empty($tenant->MobileDeviceCount->limit) && $user->UserMobileDeviceCount->limit != 0) {
                $user->UserMobileDeviceCount->limit = 0;
                $update = true;
            }
            if (empty($tenant->MailboxCount->limit) && $user->UserMailboxCount->limit != 0) {
                $user->UserMailboxCount->limit = 0;
                $update = true;
            }
            if (empty($tenant->WebsiteCount->limit) && $user->UserWebsiteCount->limit != 0) {
                $user->UserWebsiteCount->limit = 0;
                $update = true;
            }
            if ($update && $user->mspUserId) {
                $userInfo = $this->getMspApi()->getUserById($user->mspUserId);
                $userPrivileges = new Service_MspApiClient_Request_Struct_UserPrivileges();
                $userPrivileges->setStorageSize($user->userStorageUsage->limit);
                $userPrivileges->setServerCount($user->userServerCount->limit);
                $userPrivileges->setVmCount($user->UserVmCount->limit);
                $userPrivileges->setWorkstationCount($user->UserWorkstationCount->limit);
                $userPrivileges->setMobileCount($user->UserMobileDeviceCount->limit);
                $userPrivileges->setMailboxCount($user->UserMailboxCount->limit);
                $userPrivileges->setWebsiteCount($user->UserWebsiteCount->limit);

                $mspApi = $this->getMspApi();
                $status = $userInfo->getStatus();
                $version = $userInfo->getVersion();
                try {
                    if ($status == Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_DISABLE) {
                        $updateUserInfo = new Service_MspApiClient_Request_Struct_UpdateUserInfo();
                        $updateUserInfo->setStatus(Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_ENABLE);
                        $version = $mspApi->updateUser($user->mspUserId, $updateUserInfo, $version);
                    }

                    $updateUserInfo = new Service_MspApiClient_Request_Struct_UpdateUserInfo();
                    $updateUserInfo->setPrivileges($userPrivileges);
                    $version = $mspApi->updateUser($user->mspUserId, $updateUserInfo, $version);

                    if ($status == Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_DISABLE) {
                        $updateUserInfo = new Service_MspApiClient_Request_Struct_UpdateUserInfo();
                        $updateUserInfo->setStatus(Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_DISABLE);
                        $mspApi->updateUser($user->mspUserId, $updateUserInfo, $version);
                    }
                } catch (Exception $e) {
                    $skipCodes = array(
                        Service_MspApiClient_Api::EXCEPTION_HTTP_CODE_NOT_FOUND,
                        Service_MspApiClient_Api::EXCEPTION_HTTP_CODE_NOT_AVAILABLE
                    );
                    if (!in_array($e->getCode(), $skipCodes)) {
                        throw $e;
                    }
                }

                $this->getApsc()->updateResource($user->getBaseResource());
            }
        }
    }

    /**
     * @return bool
     */
    private function _getProfileUserQuota()
    {
        $tenant = $this->tenant;
        $profile = $tenant->profile;
        if ($profile->isEmpty()) {
            $userQuota = true;
        } else {
            $userQuota = $profile->userQuota !== Service_APSApi_Api::PROFILE_USER_QUOTA_OFF;
        }
        return $userQuota;
    }

    /**
     * @throws Service_BaseClient_HTTPException
     */
    private function _deleteGroupForCustomer()
    {
        $mspApi = $this->getMspApi();
        $tenant = $this->tenant;
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        if (is_null($groupInfo)) {
            Log::getInstance()->warning(sprintf(self::LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP, $tenant->mspGroupId));
            return;
        }
        $version = $groupInfo->getVersion();
        if ($groupInfo->getStatus() == 1) {
            $version = $this->getTenantManager()->disableGroup($tenant->mspGroupId, $version);
        }
        $mspApi->deleteGroup($tenant->mspGroupId, $version);
    }

    private function _getCounterLimit($userQuota, $tenantCounterLimit, $userCounterLimit)
    {
        if ($userQuota || $tenantCounterLimit) {
            return $userCounterLimit;
        }
        return $tenantCounterLimit;
    }

    private function _upgradeUsers()
    {
        $tenant = $this->tenant;
        $mspApi = $this->getMspApi();
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        $status = $groupInfo->getStatus();
        $version = $groupInfo->getVersion();
        $mspGroupManager = $this->getTenantManager();
        if ($status == Service_MspApiClient_Request_Struct_UpdateGroup::DISABLED_GROUP_STATUS) {
            $version = $mspGroupManager->enableGroup($tenant->mspGroupId, $version);
        }

        $exception = null;
        try {
            $users = $tenant->users;
            if (!count($users)) {
                $users = $this->getApsApi()->getUsersForTenant($tenant);
            }
            foreach ($users as $resource) {
                $this->upgradeUser($resource->aps->id);
            }
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
            $exception = $e;
        }

        if ($status == Service_MspApiClient_Request_Struct_UpdateGroup::DISABLED_GROUP_STATUS) {
            $mspGroupManager->disableGroup($tenant->mspGroupId, $version);
        }

        if (!is_null($exception)) {
            throw $exception;
        }
    }

    /**
     * @param $userId
     */
    private function upgradeUser($userId)
    {
        $user = $this->tenant->createUser();
        $user->fillResource($userId);
        $user->upgrade();
    }
}
