<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

/**
 * Class Model_Resource_Tenant_Partner
 */
class Model_Resource_Tenant_Partner extends Model_Resource_Tenant
{
    //default notification values for administrator account in baas
    const DEFAULT_GROUP_ADMIN_NOTIFY_BACKUP = 7;
    const DEFAULT_GROUP_ADMIN_NOTIFY_MANAGE = 1;

    /* ======= RESOURCE METHODS ======= */

    public function provision()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->checkProfile();
        $this->linkProfile(false);

        $tenant = $this->tenant;
        $tenant->datacenterId = $this->resolveDatacenterId();
        $this->linkDatacenter(false);

        $this->updateCounterLimits();

        $tenant->mspGroupId = $this->_createGroupForPartner();
        $this->updateTenantId();
    }

    public function provisionAsync()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->updateStartProductionTime();

        $this->linkDatacenter(false);
        $this->linkProfile(false);

        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();

        $admins = $this->getAdministrators();
        if (count($admins) == 0) {
            $this->createAdministratorAccount();
        }
    }

    public function retrieve()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->updateCounterLimits();
        $this->_retrieveUsageForPartner();
    }

    public function configure($new)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;

        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);

        if (is_null($tenant->mspGroupId)) {
            return;
        }

        $this->_updateGroupForPartner();
        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);
    }

    public function unprovision()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if ($tenant->ismigration) {
            Log::getInstance()->event(static::LOG_UNPROVISIONING_TENANT_IS_MIGRATION);
            return;
        }

        if (is_null($tenant->mspGroupId)) {
            Log::getInstance()->warning(static::LOG_WARNING_UNPROVISION_TENANT_DOES_NOT_HAVE_GROUP);
            return;
        }
        Log::getInstance()->event(sprintf(static::LOG_FORMAT_DELETE_MSP_GROUP, $tenant->account->companyName));

        $this->linkDatacenter(false);

        try {
            $this->_deleteGroupForPartner();
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_VERSION_EXPIRED) {
                throw $e;
            }
            sleep(10);
            $this->_deleteGroupForPartner();
        }
    }

    public function enable()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();
        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->getTenantManager()->enableGroup($tenant->mspGroupId);
    }

    public function disable()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->mspGroupId)) {
            return;
        }
        $this->linkProfile(false);
        $this->linkDatacenter(false);
        $this->getTenantManager()->disableGroup($tenant->mspGroupId);
    }

    public function upgrade()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        if (is_null($tenant->datacenterId)) {
            $tenant->datacenterId = globals::DEFAULT_DATACENTER_ID;
        }

        $this->updateStartProductionTime();
        $this->updateCounterLimits();
        $this->updateTenantId();
        $this->update();

        $realLink = !$tenant->subscription->disabled;
        if ($realLink) {
            $this->checkProfile();
        }

        $this->linkProfile($realLink);
        $this->linkDatacenter($realLink);
        $this->_getOrCreateParentGroupId($tenant->datacenter);  // add parents in map
        $this->_updateGroupForPartner();
        $this->getMspGroupManager()->updatePrivilegesForCustomers($this->tenant, $this);

        $this->_subscribeToOnSubscriptionChangeEvent();
        $this->_subscribeToOnUserChangeEvent();
        $this->_subscribeToOnAccountChangeEvent();
    }

    /* ======= RESOURCE EVENTS ======= */

    public function onTenantAvailable($event)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->linkProfile();
        $this->linkDatacenter();
    }

    public function onSubscriptionChange($event)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $needUpdate = $this->updateStartProductionTime();
        if ($this->updateCounterLimits()) {
            $needUpdate = true;
        }
        if ($needUpdate) {
            $this->update();
        }
        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->_updateProfileLink($realLink);
        $this->linkDatacenter($realLink);
        $this->_updateGroupForPartner();
    }

    public function onUserChanged($event)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForPartner();
    }

    public function onAccountChange($event)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $this->_updateGroupForPartner();
    }

    /* ======= RESOURCE OPERATIONS ======= */

    /**
     * @return array
     */
    public function getAdministrators()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        /**
         * @var Service_MspApiClient_Response_Struct_AdministratorItem[] $mspAdmins
         */
        $mspAdmins = $this->getMspApi()->getAdministrators($tenant->mspGroupId);
        $adminUsers = array();
        foreach ($mspAdmins as $mspAdmin) {
            $admin = new stdClass();
            $admin->id = $mspAdmin->getId();
            $admin->firstName = $mspAdmin->getFirstName();
            $admin->lastName = $mspAdmin->getLastName();
            $admin->login = $mspAdmin->getLogin();
            $admin->email = $mspAdmin->getEmail();
            $admin->status = $mspAdmin->getStatus();
            $adminUsers[] = $admin;
        }
        return $adminUsers;
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @return object
     */
    public function getAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $mspAdmin = $this->getMspApi()->getAdministratorById($adminId);

        $admin = new stdClass();
        $admin->id = $mspAdmin->getId();
        $admin->firstName = $mspAdmin->getFirstName();
        $admin->lastName = $mspAdmin->getLastName();
        $admin->login = $mspAdmin->getLogin();
        $admin->email = $mspAdmin->getEmail();
        $admin->status = $mspAdmin->getStatus();
        $admin->notificationsBackup = $mspAdmin->getNotificationsBackup();
        $admin->notificationsManagement = $mspAdmin->getNotificationsManagement();
        return $admin;
    }

    /**
     * @param $adminId
     * @param $parameters
     * @param string $datacenterId
     * @throws \Rest\RestException
     */
    public function updateAdministrator($adminId, $parameters, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $mspApi = $this->getMspApi();
        $adminInfo = $mspApi->getAdministratorById($adminId);
        $adminUpdate = new Service_MspApiClient_Request_Struct_AdministratorUpdate();
        if (isset($parameters->firstName) && $parameters->firstName != $adminInfo->getFirstName()) {
            $adminUpdate->setFirstName($parameters->firstName);
        }
        if (isset($parameters->lastName) && $parameters->lastName != $adminInfo->getLastName()) {
            $adminUpdate->setLastName($parameters->lastName);
        }
        if (isset($parameters->email) && $parameters->email != $adminInfo->getEmail()) {
            $adminUpdate->setEmail($parameters->email);
        }
        if (isset($parameters->notificationsBackup) && $parameters->notificationsBackup != $adminInfo->getNotificationsBackup()) {
            $adminUpdate->setNotificationsBackup(new Enum_NotificationsBackup((int)$parameters->notificationsBackup));
        }
        if (isset($parameters->notificationsManagement) && $parameters->notificationsManagement != $adminInfo->getNotificationsManagement()) {
            $adminUpdate->setNotificationsManagement((int)$parameters->notificationsManagement);
        }

        if ($adminUpdate->isExistsUpdateFields()) {
            $mspApi->updateAdministrator($adminUpdate, $adminId, $adminInfo->getVersion());
        }
        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @throws \Rest\NoContent
     */
    public function resendAdministratorLink($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $admin = $this->getMspApi()->getAdministratorById($adminId);
        $this->getMspApi()->postMailActivate(
            new Service_MspApiClient_Request_Struct_PostMailActivate($admin->getEmail(), $admin->getLogin()
            ));
        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @throws \Rest\NoContent
     */
    public function enableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $admin = $this->getMspApi()->getAdministratorById($adminId);
        $adminEnable = new Service_MspApiClient_Request_Struct_AdministratorUpdate();
        $adminEnable->setStatus(Service_MspApiClient_Request_Struct_AdministratorUpdate::STATUS_ENABLE);
        $this->getMspApi()->updateAdministrator($adminEnable, $adminId, $admin->getVersion());
        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @throws \Rest\RestException
     */
    public function deleteAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $adminInfo = $this->getMspApi()->getAdministratorById($adminId);
        $this->getMspApi()->deleteAdministrator($adminId, $adminInfo->getVersion());
        $this->throwNoContent();
    }

    /**
     * @param $adminId
     * @param string $datacenterId
     * @throws \Rest\RestException
     */
    public function disableAdministrator($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $mspApi = $this->getMspApi();

        $adminInfo = $mspApi->getAdministratorById($adminId);
        if ($adminInfo->getStatus() == Service_MspApiClient_Request_Struct_AdministratorUpdate::STATUS_DISABLE) {
            $this->throwNoContent();
        }

        $adminUpdate = new Service_MspApiClient_Request_Struct_AdministratorUpdate();
        $adminUpdate->setStatus(Service_MspApiClient_Request_Struct_AdministratorUpdate::STATUS_DISABLE);
        $mspApi->updateAdministrator($adminUpdate, $adminId, $adminInfo->getVersion());
        $this->throwNoContent();
    }

    /**
     * @return array
     * @throws Exception
     */
    public function getStorages()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        return $this->memoize(function () {
            $response = [];
            $tenant = $this->tenant;
            $storages = $this->getMspApi()->getStorages($tenant->mspGroupId);
            if (is_null($storages)) {
                Log::getInstance()->error(static::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
                throw new Exception(static::EXCEPTION_CAN_NOT_RECEIVE_STORAGES_MESSAGE);
            }

            foreach ($storages as $storage) {
                /** @var Service_MspApiClient_Response_Struct_Storage $storage */
                if ($storage->getInherited()) {
                    continue;
                }
                $response[] = [
                    static::FIELD_STORAGE_ID => (string)$storage->getId(),
                    static::FIELD_STORAGE_NAME => $storage->getName(),
                ];
            }

            return $response;
        });
    }

    /**
     * @return object
     */
    public function getDatacenter()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $realLink = !$tenant->subscription->disabled;
        $this->linkDatacenter($realLink);
        return (object)$tenant->datacenter->toArray();
    }

    /**
     * @param $login
     * @param $datacenterId
     * @return bool
     */
    public function loginExists($login, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        return $this->checkLoginExist($login, $datacenterId);
    }

    public function updateUsageAndLimits()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->updateCounterLimits();
        $this->_updateProfileLink();

        $tenantUsages = $this->fetchTenantLimitsAndUsages();

        return [
            static::TENANT => $tenantUsages,
            static::USER_USAGE => [],
            static::ERROR => null,
        ];
    }

    public function createAdministrator($parameters, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        if (!isset($parameters->login)) {
            throw new \Rest\RestException(400, Str::format(
                static::EXCEPTION_FIELD_MUST_BE_PRESENT_MESSAGE,
                Service_MspApiClient_Request_Struct_AdministratorInfo::FIELD_LOGIN
            ));
        }
        if (!isset($parameters->email)) {
            throw new \Rest\RestException(400, Str::format(
                static::EXCEPTION_FIELD_MUST_BE_PRESENT_MESSAGE,
                Service_MspApiClient_Request_Struct_AdministratorInfo::FIELD_EMAIL
            ));
        }

        $adminInfo = new Service_MspApiClient_Request_Struct_AdministratorInfo();
        $adminInfo->setType(static::DEFAULT_GROUP_ADMIN_ACCOUNT_PERMS);

        $adminInfo->setLogin($parameters->login);
        $adminInfo->setEmail($parameters->email);

        if (isset($parameters->firstName)) {
            $adminInfo->setFirstName($parameters->firstName);
        }
        if (isset($parameters->lastName)) {
            $adminInfo->setLastName($parameters->lastName);
        }
        if (isset($parameters->notificationsBackup)) {
            $adminInfo->setNotificationsBackup(new Enum_NotificationsBackup($parameters->notificationsBackup));
        }
        if (isset($parameters->notificationsManagement)) {
            $adminInfo->setNotificationsManagement($parameters->notificationsManagement);
        }

        $mspApi = $this->tenant->getMspApi();

        $mspApi->createAdministrator($adminInfo, $this->tenant->mspGroupId);
        try {
            $mspApi->postMailActivate(new Service_MspApiClient_Request_Struct_PostMailActivate(
                $adminInfo->getEmail(), $adminInfo->getLogin()
            ));
        } catch (Exception $e) {
            Log::getInstance()->error($e->getMessage());
        }

        $this->throwNoContent();
    }

    public function getAdministratorConsole($adminId, $datacenterId = '')
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function adminUrlRedirect()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));

        $tenant = $this->tenant;
        $mspApi = $this->getMspApi();
        $brand = $mspApi->getBrand($tenant->mspGroupId);
        $url = $brand->getAccountServerUrl();

        return (object) array(
            static::FIELD_URL => Str::format(static::REDIRECT_URL, $url, $tenant->mspGroupId)
        );
    }

    public function getDatacenters()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlans()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getServicePlan($planId)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function setStorageForPlan($planId, $parameters)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function unsetStorageForPlan($planId)
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getQuotasStatistics()
    {
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $this->throwNotApplicableMethodException();
    }

    public function getUserGuideUrl(){
        Log::getInstance()->event(Str::format(
            static::LOG_METHOD_CALL, $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        ));
        $brand = $this->getMspApi()->getBrand($this->tenant->mspGroupId);
        return $brand->getUserGuideUrl();
    }

    /* ======= OTHER METHODS ======= */

    public function resolveDatacenterId()
    {
        $tenant = $this->tenant;
        // Do nothing if datacenter ID already was set from external side (online store / migration script)
        if (isset($tenant->datacenterId)) {
            return $tenant->datacenterId;
        }

        $datacenters = $tenant->global->datacenters;
        if ($datacenters->count() === 0) {
            throw new \Rest\RestException(401, static::EXCEPTION_NEED_TO_SET_CREDENTIALS_MESSAGE);
        }

        $profileDatacenters = $tenant->profile->datacenters;

        if (count($profileDatacenters) == 0) {
            throw new \Rest\RestException(400, Str::format(
                'Please specify datacenter in the backup profile "%s".',
                $tenant->profile->resourceTypeId
            ));
        }
        if (count($profileDatacenters) == 1) {
            /** @var BackupProfileDatacenter $firstDatacenter */
            $firstDatacenter = reset($profileDatacenters);

            return $firstDatacenter->datacenterId;
        }

        $parametersManager = $this->getSubscriptionParametersManager();
        $datacenterId = $parametersManager->getParameterDatacenterId($tenant->subscriptionId);

        $datacenters = array_filter($profileDatacenters, function ($datacenter) use ($datacenterId) {
            return $datacenter->datacenterId == $datacenterId;
        });

        if (empty($datacenters)) {
            throw new \Rest\RestException(400, Str::format(
                'Please add the datacenter with ID "%s" to the list of datacenters in backup profile "%s".',
                $datacenterId, $tenant->profile->resourceTypeId
            ));
        } else {
            /** @var BackupProfileDatacenter $firstDatacenter */
            $firstDatacenter = reset($datacenters);
            $datacenterId = $firstDatacenter->datacenterId;
        }

        return $datacenterId;
    }

    public function resolveBackupStorageId()
    {
        Log::getInstance()->event(
            static::LOG_METHOD_CALL,
            $this->tenant->aps->id, $this->tenant->accountId, $this->tenant->subscriptionId, __METHOD__
        );

        $this->throwNotApplicableMethodException();
    }

    /**
     * @return int
     */
    private function _createGroupForPartner()
    {
        $parentMspGroupId = $this->_getOrCreateParentGroupId();
        $tenant = $this->tenant;
        $mspGroupManager = $this->getMspGroupManager();
        $groupId = $mspGroupManager->createGroupForCustomer(
            $tenant, $parentMspGroupId, Service_MspApiClient_Request_Struct_CreateGroup::RESELLER_GROUP_KIND_ID
        );
        return $groupId;
    }

    protected function _updateGroupForPartner()
    {
        $tenant = $this->tenant;
        $mspGroupManager = $this->getMspGroupManager();
        $mspGroupManager->updateGroupForCustomer($tenant->mspGroupId, $tenant);
    }

    /**
     * @throws Service_BaseClient_HTTPException
     */
    private function _deleteGroupForPartner()
    {
        $mspApi = $this->getMspApi();
        $tenant = $this->tenant;
        $groupInfo = $mspApi->getGroupById($tenant->mspGroupId);
        if (is_null($groupInfo)) {
            Log::getInstance()->warning(sprintf(static::LOG_WARNING_UNPROVISION_UNABLE_TO_FIND_GROUP, $tenant->mspGroupId));
            return;
        }
        $version = $groupInfo->getVersion();
        if ($groupInfo->getStatus() == 1) {
            $version = $this->getTenantManager()->disableGroup($tenant->mspGroupId, $version);
        }
        $mspApi->deleteGroup($tenant->mspGroupId, $version);
    }

    /**
     * @throws \Acronis\App\UsageReport\Exception\ManagerException
     * @throws \Rest\RestException
     */
    private function _retrieveUsageForPartner()
    {
        $this->updateCountersUsage($this->tenant->datacenter);
    }

    protected function createAdministratorAccount()
    {
        $userInfo = $this->createAdministratorInfo();
        //add administrator account to baas with data from PA staff member
        $userManager = $this->getUserManager();
        $userManager->provisionAdministratorAccount(
            $this->tenant->mspGroupId,
            $this->tenant->subscriptionId,
            $userInfo
        );
    }

    /**
     * @return Service_MspApiClient_Request_Struct_AdministratorInfo
     */
    private function createAdministratorInfo()
    {
        $admInfo = new Service_MspApiClient_Request_Struct_AdministratorInfo();
        $apsApi = new Service_APSApi_Api($this->getApsc()->impersonate($this->tenant));

        $adminInfoFromPA = $apsApi->getFirstStaffMemberInfo($this->tenant);

        $admInfo->setType(static::DEFAULT_GROUP_ADMIN_ACCOUNT_PERMS);
        $admInfo->setFirstName($adminInfoFromPA->givenName);
        $admInfo->setLastName($adminInfoFromPA->familyName);
        $admInfo->setEmail($adminInfoFromPA->email);

        $login = $this->createLoginForAdministrator($admInfo);
        $admInfo->setLogin($login);
        if ($this->tenant->profile->isEmpty()) {
            $admInfo->setNotificationsBackup(new Enum_NotificationsBackup(static::DEFAULT_GROUP_ADMIN_NOTIFY_BACKUP));
            $admInfo->setNotificationsManagement(static::DEFAULT_GROUP_ADMIN_NOTIFY_MANAGE);
        } else {
            $profile = $this->tenant->profile;
            $admInfo->setNotificationsBackup(new Enum_NotificationsBackup((int) $profile->backupNotification));
            $admInfo->setNotificationsManagement($profile->managementNotification);
        }

        return $admInfo;
    }


    /**
     * @param Service_MspApiClient_Request_Struct_AdministratorInfo $adminInfo
     * @return string
     */
    protected function createLoginForAdministrator(Service_MspApiClient_Request_Struct_AdministratorInfo $adminInfo)
    {
        $login = trim($adminInfo->getFirstName() . '_' . $adminInfo->getLastName(), '_');
        $login = preg_replace('/\s/', '_', $login);
        if (empty($login)) {
            $login = reset(explode('@', $adminInfo->getEmail()));
        }
        $i = 1;
        while (true) {
            $newLogin = $login . '_' . $i;
            if (!$this->loginExists($newLogin, $this->tenant->datacenterId)) {
                return $newLogin;
            }
            $i++;
        }
    }
}