<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class Model_UserManager implements Model_UserManagerInterface
{
    /** @var Service_MspApiClient_ApiInterface */
    private $_mspApi;
    /** @var FacadeInterface */
    private $_facade;
    /** @var Service_MspApiClient_Request_Struct_UpdateUserInfo */
    private $_updateUserInfo;

    /**
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param FacadeInterface $facade
     * @param Service_MspApiClient_Request_Struct_UpdateUserInfo $updateUserInfo
     */
    public function __construct(Service_MspApiClient_ApiInterface $mspApi,
        FacadeInterface $facade = NULL,
        Service_MspApiClient_Request_Struct_UpdateUserInfo $updateUserInfo = NULL)
    {
        $this->_mspApi= $mspApi;
        $this->_facade = is_null($facade) ? new Facade() : $facade;
        $this->_updateUserInfo = is_null($updateUserInfo) ? new Service_MspApiClient_Request_Struct_UpdateUserInfo() : $updateUserInfo;
    }

    /**
     * @param string $login
     * @param string $password
     */
    public function activateUser($login, $password)
    {
        $token = $this->_mspApi->mailActivate($login);
        $this->_mspApi->activateAccountWithPassword($token, $password);
    }

    /**
     * @param string $email
     * @param string $password
     * @throws Exception
     */
    public function resetUserPassword($email, $password)
    {
        try
        {
            $token = $this->_mspApi->mailReset($email);
            $this->_mspApi->resetPassword($token, $password);
        }
        catch (Exception $e)
        {
            if ($e->getCode() == Service_BaseClient_RestTransport::HTTP_CODE_FORBIDDEN)
            {
                return;
            }
            throw $e;
        }
    }

    /**
     * @param int $userId
     * @param int $version
     * @return int
     */
    public function enableUser($userId, $version)
    {
        return $this->_setUserStatus($userId, $version, Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_ENABLE);
    }

    /**
     * @param int $userId
     * @param int $version
     * @return int
     */
    public function disableUser($userId, $version)
    {
        return $this->_setUserStatus($userId, $version, Service_MspApiClient_Request_Struct_UpdateUserInfo::STATUS_DISABLE);
    }

    /**
     * @param int $groupId
     * @return array
     */
    public function getBackItems($groupId)
    {
        $backupConsole = $this->_mspApi->getBackupConsole($groupId);
        $link = $backupConsole->getHost();
        $accessToken = $backupConsole->getToken();

        $amsApi = $this->_facade->getAmsApi($link);
        $amsApi->createRemoteConnection($accessToken);
        return $amsApi->getBackupItems();
    }

    /**
     * @param string $email
     * @param int $groupId
     * @return NULL|Service_MspApiClient_Response_Struct_UserInfo
     */
    public function getUserByEmail($email, $groupId)
    {
        $users = $this->_mspApi->getUsers($groupId);
        if (!is_array($users))
        {
            return NULL;
        }
        foreach ($users as $userItem)
        {
            if ($userItem->getEmail() == $email)
            {
                return $this->_mspApi->getUserById($userItem->getId());
            }
        }
        return NULL;
    }

    /**
     * @param string $companyName
     * @return NULL|Service_MspApiClient_Response_Struct_GroupInfo
     */
    public function getGroupByName($companyName)
    {
        $groups = $this->_mspApi->search($companyName);
        if (!is_array($groups))
        {
            return NULL;
        }
        foreach ($groups as $group)
        {
            if ($group->getName() == $companyName)
            {
                return $group;
            }
        }
        return NULL;
    }

    /**
     * @param int $userId
     * @param int $version
     * @param int $status
     * @return int
     */
    private function _setUserStatus($userId, $version, $status)
    {
        $this->_updateUserInfo->setStatus($status);
        return $this->_mspApi->updateUser($userId, $this->_updateUserInfo, $version);
    }

    /**
     * add administrator account to baas with data from PA staff member
     *
     * @param int $mspGroupId
     * @param int $subscriptionId
     * @param Service_MspApiClient_Request_Struct_AdministratorInfo $userInfo
     * @throws Exception
     */
    public function provisionAdministratorAccount($mspGroupId, $subscriptionId, Service_MspApiClient_Request_Struct_AdministratorInfo $userInfo)
    {
        try {
            $this->_mspApi->createAdministrator($userInfo, $mspGroupId);
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() != Service_BaseClient_HTTPException::HTTP_CODE_CONFLICT) {
                throw $e;
            }
            $admin = $this->getCurrentAdministrator($mspGroupId, $userInfo);
            if (is_null($admin)) {
                throw $e;
            }
            if ($admin->getStatus() !== Service_MspApiClient_Response_Struct_UserInfo::STATUS_INACTIVE) {
                Log::getInstance()->event(
                    sprintf(
                        users::LOG_FORMAT_ATTEMPT_TO_CREATE_EXISTING_MSP_USER,
                        $userInfo->getLogin(),
                        $subscriptionId
                    )
                );
                throw new Exception(users::EXCEPTION_USER_ALREADY_EXISTS);
            }
        }
        $this->sendActivateEmailForUser($userInfo);
    }

    /**
     * @param $mspGroupId
     * @param Service_MspApiClient_Request_Struct_AdministratorInfo $userInfo
     * @return null|Service_MspApiClient_Response_Struct_AdministratorItem
     */
    private function getCurrentAdministrator($mspGroupId, Service_MspApiClient_Request_Struct_AdministratorInfo $userInfo)
    {
        $administrators = $this->_mspApi->getAdministrators($mspGroupId);
        if (empty($administrators)) {
            return null;
        }
        foreach ($administrators as $administrator) {
            /** @var Service_MspApiClient_Response_Struct_AdministratorItem $administrator */
            if ($administrator->getLogin() == $userInfo->getLogin()) {
                return $administrator;
            }
        }
        return null;
    }

    /**
     * @param Service_MspApiClient_Request_Struct_AdministratorInfo $userInfo
     */
    private function sendActivateEmailForUser(Service_MspApiClient_Request_Struct_AdministratorInfo $userInfo)
    {
        $this->_mspApi->postMailActivate(new Service_MspApiClient_Request_Struct_PostMailActivate(
            $userInfo->getEmail(),
            $userInfo->getLogin()
        ));
    }
}
