<?php

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */
class Service_MspApiClient_Api extends Service_BaseClient_ApiAbstract implements Service_MspApiClient_ApiInterface
{
    const ACCOUNT_WITH_LOGIN_NOT_FOUND = 'Account with login %s not found';

    const CACHE_ACCOUNT_INFO_PREFIX = 'MSP_API_ACCOUNT_';
    const CACHE_COUNTERS_PREFIX = "COUNTERS_";
    const CACHE_MOBILE_COUNTER = 'MOBILE_COUNTER';
    const CACHE_MAILBOX_COUNTER = 'MAILBOX_COUNTER';
    const CACHE_WEBSITE_COUNTER = 'WEBSITE_COUNTER';
    const CACHE_VERSIONS_PREFIX = 'MSP_API_LOGIC_VERSIONS_';
    const CACHE_ACCOUNT_GROUP_ID_PREFIX = 'MSP_API_ACCOUNT_GROUP_ID_';
    const CACHE_DEFAULT_TTL = 86400; // 1 day

    const INVALID_CREDENTIALS = 'Provided credentials are not valid.';
    const LOGIN_LOCKED = 'The account is locked for %s minutes due to too many attempts to sign in with an incorrect login or password.';
    const LOGIN_EMPTY = 'Please fill-in all connection settings for Acronis Management console in Services > Acronis Backup > Connection Settings';
    const EXCEPTION_CREDENTIALS_DO_NOT_SET_MESSAGE = 'Credentials don\'t set.';
    const LOG_WARNING_CONNECTION_ERROR = 'Connection error:';
    const EXCEPTION_HTTP_INTERNAL_SERVER_ERROR_CODE = 500;
    const EXCEPTION_HTTP_BAD_REQUEST_CODE = 400;
    const EXCEPTION_HTTP_UNAUTHORIZED_CODE = 401;
    const EXCEPTION_HTTP_LOCKED_CODE = 423;
    const EXCEPTION_HTTP_CODE_NOT_FOUND = 404;
    const EXCEPTION_HTTP_CODE_NOT_AVAILABLE = 405;
    const RETRY_AFTER = 'Retry-After';
    const SEVERAL = 'several';
    const API_VERSION_MOBILE = 5;
    const API_VERSION_DELIMITER = '.';

    const LOG_EVENT_CHANGE_GROUP_PARENT = 'Parent of group with ID :group_id was updated. Details: group_name=":group_name", group_new_version=:group_new_version, old_parent_id=:old_parent_id, new_parent_id=:new_parent_id';

    /**
     * @var string
     */
    private $_login;

    /**
     * @var string
     */
    private $_password;

    /**
     * @var string
     */
    private $_url;

    /**
     * @var string
     */
    private $_accountServerUrl = NULL;

    /**
     * @var Session_StoreInterface
     */
    private $_sessionStore;

    /**
     * @var Service_MspApiClient_Request_Struct_Factory
     */
    private $_groupInfoFactory;

    /**
     * @var int
     */
    private $_accountGroupId;

    /**
     * @param Service_BaseClient_IRestClient $restClient
     * @param Session_StoreInterface $sessionStore
     */
    public function __construct(Service_BaseClient_IRestClient $restClient = NULL, Session_StoreInterface $sessionStore = NULL, Service_MspApiClient_Request_Struct_Factory $groupInfoFactory = NULL)
    {
        parent::__construct($restClient);

        if (is_null($sessionStore)) {
            $sessionStore = Session_Store::getInstance();
            $sessionStore->setSessionId(\APS\Request::getController()->getInstanceId());
        }
        $this->_sessionStore = $sessionStore;
        $this->_groupInfoFactory = is_null($groupInfoFactory) ? new Service_MspApiClient_Request_Struct_Factory() : $groupInfoFactory;
    }

    /**
     * @param string $login
     * @param string $baasUrl
     * @param bool $force
     * @return Service_MspApiClient_Response_Struct_Account
     */
    public function getAccountInfo($login, $baasUrl, $force = false)
    {
        //trying to get account info from cache data
        $cacheKey = self::CACHE_ACCOUNT_INFO_PREFIX . md5($login . $baasUrl);
        $this->_sessionStore->startSession();
        $accountInfo = $this->_sessionStore->getValue($cacheKey);
        if (is_null($accountInfo) || $force) {
            //if no account info in session, get actual account info, write to cache and return
            $this->_restClient->setApiUri($baasUrl . MSP_API_URI);
            $accountInfo = $this->_call(new Service_MspApiClient_Request_Accounts($login));

            $this->_sessionStore->setValue($cacheKey, $accountInfo, self::CACHE_DEFAULT_TTL);
        }
        $this->_sessionStore->stopSession();
        return $accountInfo;
    }

    /**
     * @param int $adminId
     * @return string
     */
    public function getImpersonationToken($adminId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_ImpersonateToken($adminId));
    }

    /**
     * @param int $userId
     * @return string
     */
    public function getImpersonationUserToken($userId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_ImpersonateUserToken($userId));
    }

    /**
     * @param string $login
     * @param string $password
     * @param string $url
     * @param bool $force
     * @throws Exception
     */
    public function setCredentials($login, $password, $url, $force = false)
    {
        if (empty($login) || empty($password) || empty($url)) {
            throw new Exception(self::LOGIN_EMPTY);
        }

        $userInfo = $this->getAccountInfo($login, $url, $force);
        if (is_null($userInfo)) {
            throw new Exception(sprintf(self::ACCOUNT_WITH_LOGIN_NOT_FOUND, $login));
        }

        $this->_login = $login;
        $this->_password = $password;
        $this->_url = $url;

        $this->_accountServerUrl = $userInfo->getServerUrl();
        $this->_restClient->setApiUri($this->_accountServerUrl . MSP_API_URI);
    }

    /**
     * @return Service_MspApiClient_Response_Struct_Login
     * @throws Exception | Service_BaseClient_HTTPException
     */
    public function login()
    {
        if (empty($this->_login) || empty($this->_password) || empty($this->_url)) {
            throw new Exception(self::LOGIN_EMPTY);
        }

        $credentials = new Service_MspApiClient_Request_Struct_Credentials($this->_login, $this->_password);
        try {
            $loginInfo = $this->_call(new Service_MspApiClient_Request_Login($credentials));
        } catch (Service_BaseClient_HTTPException $e) {
            Log::getInstance()->warning($e->getMessage());

            $code = $e->getCode();
            if ($code === self::EXCEPTION_HTTP_LOCKED_CODE) {
                $retry = $e->getHeader(self::RETRY_AFTER);
                if (is_null($retry)) {
                    $retry = self::SEVERAL;
                } else {
                    $retry = date('i:s', $retry);
                }
                throw new Exception(sprintf(self::LOGIN_LOCKED, $retry));
            }
            if ($code >= self::EXCEPTION_HTTP_INTERNAL_SERVER_ERROR_CODE) {
                throw $e;
            }

            throw new Exception(self::INVALID_CREDENTIALS);
        }

        if (is_null($loginInfo)) {
            throw new Exception(self::INVALID_CREDENTIALS);
        }

        $this->_restClient->invalidateCookies();
        $this->_restClient->refresh();
        return $loginInfo;
    }

    /**
     * @param Service_MspApiClient_Request_Struct_UserInfo $userInfo
     * @param int $groupId
     * @return Service_MspApiClient_Response_Struct_CreateGroup
     */
    public function createUser(Service_MspApiClient_Request_Struct_UserInfo $userInfo, $groupId)
    {
        if (!$this->_checkMobilePrivilegeSupport()) {
            $userInfo->getPrivileges()->setMobileCount(NULL);
        }
        if (!$this->_checkMailboxPrivilegeSupport()) {
            $userInfo->getPrivileges()->setMailboxCount(NULL);
        }
        if (!$this->_checkWebsitePrivilegeSupport()) {
            $userInfo->getPrivileges()->setWebsiteCount(NULL);
        }
        return $this->_authorizedCall(new Service_MspApiClient_Request_CreateNewUser($userInfo, $groupId));
    }

    /**
     * @param Service_MspApiClient_Request_Struct_AdministratorInfo $userInfo
     * @param int $groupId
     * @return Service_MspApiClient_Response_Struct_CreateGroup
     */
    public function createAdministrator(Service_MspApiClient_Request_Struct_AdministratorInfo $userInfo, $groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_CreateNewAdministrator($userInfo, $groupId));
    }

    /**
     * @param int $adminId
     * @param int $version
     * @return Service_MspApiClient_Response_Version
     */
    public function deleteAdministrator($adminId, $version)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_DeleteAdministrator($adminId, $version));
    }

    /**
     * @param int $adminId
     * @return Service_MspApiClient_Response_Struct_AdministratorInfo
     */
    public function getAdministratorById($adminId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetAdministratorById($adminId));
    }

    /**
     * @param Service_MspApiClient_Request_Struct_AdministratorUpdate $adminInfo
     * @param int $adminId
     * @param int $version
     * @return int
     */
    public function updateAdministrator(Service_MspApiClient_Request_Struct_AdministratorUpdate $adminInfo, $adminId, $version)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_UpdateAdministrator($adminInfo, $adminId, $version));
    }

    /**
     * @param int $groupId
     * @return array
     */
    public function getAdministrators($groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetAdministrators($groupId));
    }

    /**
     * @param int $groupId
     * @return Service_MspApiClient_Response_Struct_UserItem[]
     */
    public function getUsers($groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetUsers($groupId));
    }

    /**
     * @param int $id
     * @return Service_MspApiClient_Response_Struct_UserInfo
     */
    public function getUserById($id)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetUserById($id));
    }

    /**
     * @param int $groupId
     * @return Service_MspApiClient_Response_Struct_GroupInfo
     */
    public function getGroupById($groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetGroupById($groupId));
    }

    /**
     * @param int $groupId
     * @return Service_MspApiClient_Response_Struct_Storage[]
     */
    public function getStoragesByGroupId($groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetStoragesByGroupId($groupId));
    }

    /**
     * @param int $userId
     * @return NULL
     */
    public function activateServiceUser($userId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_ActivateServiceUser($userId));
    }

    /**
     * @param Service_MspApiClient_Request_Struct_CreateGroupInterface $groupInfo
     * @param int|null $parentGroupId
     * @return Service_MspApiClient_Response_Struct_CreateGroup
     */
    public function createGroup(
        Service_MspApiClient_Request_Struct_CreateGroupInterface $groupInfo,
        $parentGroupId = NULL
    )
    {
        if (is_null($parentGroupId)) {
            $parentGroupId = $this->getAccountGroupId();
        }
        if (!$this->_checkMobilePrivilegeSupport()) {
            $groupInfo->getPrivileges()->setMobileCount(NULL);
        }
        if (!$this->_checkMailboxPrivilegeSupport()) {
            $groupInfo->getPrivileges()->setMailboxCount(NULL);
        }
        if (!$this->_checkWebsitePrivilegeSupport()) {
            $groupInfo->getPrivileges()->setWebsiteCount(NULL);
        }
        return $this->_authorizedCall(new Service_MspApiClient_Request_CreateGroup($groupInfo, $parentGroupId));
    }

    /**
     * @param int $userId
     * @param Service_MspApiClient_Request_Struct_UpdateUserInfo $userInfo
     * @param int $version
     * @return int
     * @throws Service_BaseClient_HTTPException
     */
    public function updateUser($userId, Service_MspApiClient_Request_Struct_UpdateUserInfo $userInfo, $version)
    {
        if (!$this->_checkMobilePrivilegeSupport()) {
            $userInfo->getPrivileges()->setMobileCount(null);
        }
        if (!$this->_checkMailboxPrivilegeSupport()) {
            $userInfo->getPrivileges()->setMailboxCount(NULL);
        }
        if (!$this->_checkWebsitePrivilegeSupport()) {
            $userInfo->getPrivileges()->setWebsiteCount(NULL);
        }
        return $this->_authorizedCall(new Service_MspApiClient_Request_UpdateUser($userInfo, $userId, $version));
    }

    /**
     * @param int $userId
     * @param int $version
     * @return NULL
     */
    public function deleteUser($userId, $version)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_DeleteUser($userId, $version));
    }

    /**
     * @param int|null $parentGroupId
     * @return Service_MspApiClient_Response_Struct_GroupItem[]
     */
    public function getGroups($parentGroupId = null)
    {
        if (is_null($parentGroupId)) {
            $parentGroupId = $this->getAccountGroupId();
        }
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetGroups($parentGroupId));
    }

    /**
     * @param string $prefix
     * @return Service_MspApiClient_Response_Struct_SearchItem[]
     */
    public function search($prefix)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_Search($prefix));
    }

    /**
     * @param int $groupId
     * @param int $version
     * @return NULL
     */
    public function deleteGroup($groupId, $version)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_DeleteGroup($groupId, $version));
    }

    /**
     * @param int $groupId
     * @param string $language
     * @return Service_MspApiClient_Response_Struct_BackupConsole
     */
    public function getBackupConsole($groupId, $language = NULL)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_BackupConsole($groupId, $language));
    }

    /**
     * @param string $token
     * @return Service_MspApiClient_Response_Struct_ActivateServiceUser
     */
    public function activate($token)
    {
        return $this->_call(new Service_MspApiClient_Request_Activate($token));
    }

    /**
     * @param string $token
     * @param string $password
     * @return Service_MspApiClient_Response_Struct_ActivateServiceUser
     */
    public function activateAccountWithPassword($token, $password)
    {
        $result = $this->_call(new Service_MspApiClient_Request_AccountsActivate($token, $password));
        return $result;
    }

    /**
     * @param string $login
     * @param string $email
     * @return NULL
     * @throws Service_BaseClient_HTTPException
     */
    public function mailUpdate($login, $email)
    {
        $token = $this->_authorizedCall(new Service_MspApiClient_Request_MailUpdate($login, $email));
        return $this->_authorizedCall(new Service_MspApiClient_Request_AccountsEmailUpdate($token));
    }

    /**
     * @param string $login
     * @return Service_MspApiClient_Response_Struct_ActivateServiceUser
     * @throws Service_BaseClient_HTTPException
     */
    public function mailActivate($login)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_MailActivate($login));
    }

    /**
     * @param Service_MspApiClient_Request_Struct_PostMailActivateInterface $mailActivate
     * @return Service_MspApiClient_Response_Empty
     */
    public function postMailActivate(Service_MspApiClient_Request_Struct_PostMailActivateInterface $mailActivate)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_PostMailActivate($mailActivate));
    }

    /**
     * @param string $mail
     * @return string
     */
    public function mailReset($mail)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_MailReset($mail));
    }

    /**
     * @param string $token
     * @param string $password
     * @return Service_MspApiClient_Response_Struct_Account
     */
    public function resetPassword($token, $password)
    {
        $this->_call(new Service_MspApiClient_Request_Reset($token));
        return $this->_call(new Service_MspApiClient_Request_ResetSend($token, $password));
    }

    /**
     * @param int $groupId
     * @param Service_MspApiClient_Request_Struct_UpdateGroupInterface $groupInfo
     * @param int $version
     * @return int
     */
    public function updateGroup($groupId, Service_MspApiClient_Request_Struct_UpdateGroupInterface $groupInfo, $version)
    {
        if (!$this->_checkMobilePrivilegeSupport()) {
            $groupInfo->getPrivileges()->setMobileCount(NULL);
        }
        if (!$this->_checkMailboxPrivilegeSupport()) {
            $groupInfo->getPrivileges()->setMailboxCount(NULL);
        }
        if (!$this->_checkWebsitePrivilegeSupport()) {
            $groupInfo->getPrivileges()->setWebsiteCount(NULL);
        }
        return $this->_authorizedCall(new Service_MspApiClient_Request_UpdateGroup($groupInfo, $groupId, $version));
    }

    /**
     * @param int|null $groupId
     * @return array
     */
    public function getStorages($groupId = null)
    {
        if (is_null($groupId)) {
            $groupId = $this->getAccountGroupId();
        }
        return $this->_authorizedCall(new Service_MspApiClient_Request_Storages($groupId));
    }

    /**
     * @param int $domainController
     * @param int $externalGroupId
     * @return Service_MspApiClient_Response_MigrationGroup
     * @throws Exception
     */
    public function getMigrationGroup($domainController, $externalGroupId)
    {
        $result = null;
        $apiUri = $this->_restClient->getApiUri();
        try {
            $this->_restClient->setApiUri($this->_accountServerUrl);
            $result = $this->_authorizedCall(new Service_MspApiClient_Request_MigrationGroup($domainController, $externalGroupId));
            $this->_restClient->setApiUri($apiUri);
        } catch (Exception $exception) {
            $this->_restClient->setApiUri($apiUri);
            throw $exception;
        }

        return $result;
    }

    /**
     * @param int $domainController
     * @param int $externalUserId
     * @return Service_MspApiClient_Response_MigrationUser
     * @throws Exception
     */
    public function getMigrationUser($domainController, $externalUserId)
    {
        $result = null;
        $apiUri = $this->_restClient->getApiUri();
        try {
            $this->_restClient->setApiUri($this->_accountServerUrl);
            $result = $this->_authorizedCall(new Service_MspApiClient_Request_MigrationUser($domainController, $externalUserId));
            $this->_restClient->setApiUri($apiUri);
        } catch (Exception $exception) {
            $this->_restClient->setApiUri($apiUri);
            throw $exception;
        }

        return $result;
    }

    /**
     * @param $groupId
     * @return NULL
     */
    public function lockGroup($groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_LockGroup($groupId));
    }

    /**
     * @param $groupId
     * @return NULL
     */
    public function unlockGroup($groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_UnLockGroup($groupId));
    }

    /**
     * @return Service_MspApiClient_Response_Struct_LogicVersions
     */
    public function getLogicVersions()
    {
        $this->_sessionStore->startSession();
        $versions = $this->_sessionStore->getValue(self::CACHE_VERSIONS_PREFIX);
        if (is_null($versions)) {
            $versions = $this->_call(new Service_MspApiClient_Request_LogicVersions());
            $this->_sessionStore->setValue(
                self::CACHE_VERSIONS_PREFIX,
                $versions,
                self::CACHE_DEFAULT_TTL
            );
        }
        $this->_sessionStore->stopSession();
        return $versions;
    }

    /**
     * @return bool | NULL
     */
    public function getApiMajorVersion()
    {
        $apiVersion = $this->getLogicVersions()->getApi();
        $parts = explode(self::API_VERSION_DELIMITER, $apiVersion);
        if ($parts > 0) {
            return $parts[0];
        }
        return NULL;
    }

    /**
     * @return null|string
     */
    public function getAccountServerUrl()
    {
        return $this->_accountServerUrl;
    }

    /**
     * @return int
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     */
    public function getAccountGroupId()
    {
        if (!$this->_accountGroupId) {
            $this->_restoreAccountGroupIdFromCache();
            if (!$this->_accountGroupId) {
                $accountGroupInf = $this->getGroupById('self');
                $this->_accountGroupId = $accountGroupInf->getId();
                $this->_saveAccountGroupIdToCache();
            }
        }
        return $this->_accountGroupId;
    }

    private function _restoreAccountGroupIdFromCache()
    {
        $this->_sessionStore->startSession();
        $this->_accountGroupId = intval($this->_sessionStore->getValue($this->_getAccountGroupIdCacheKey()));
        $this->_sessionStore->stopSession();
    }

    private function _saveAccountGroupIdToCache()
    {
        $this->_sessionStore->startSession();
        $this->_sessionStore->setValue(
            $this->_getAccountGroupIdCacheKey(),
            $this->_accountGroupId,
            self::CACHE_DEFAULT_TTL
        );
        $this->_sessionStore->stopSession();
    }

    private function _getAccountGroupIdCacheKey()
    {
        return self::CACHE_ACCOUNT_GROUP_ID_PREFIX . md5($this->_login . $this->_url);
    }

    /**
     * @param Service_BaseClient_Request_IRequest $request
     * @param string $authHeaders
     * @return mixed
     * @throws Service_BaseClient_HTTPException
     */
    protected function _authorizedCall(Service_BaseClient_Request_IRequest $request, $authHeaders = NULL)
    {
        if (!$this->_restClient->hasCookieFile()) {
            $this->login();
        }

        try {
            return $this->_call($request, $authHeaders);
        } catch (Service_BaseClient_IOException $e) {
            Log::getInstance()->warning(self::LOG_WARNING_CONNECTION_ERROR . $e->getMessage());
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() !== self::EXCEPTION_HTTP_UNAUTHORIZED_CODE && $e->getCode() !== self::EXCEPTION_HTTP_BAD_REQUEST_CODE) {
                throw $e;
            }
        }
        $this->login();
        return $this->_call($request, $authHeaders);
    }

    /**
     * @return array
     */
    public function getAccountPrivileges()
    {
        $this->_sessionStore->startSession();
        $cacheKey = $this->getCountersCacheKey();
        $counters = $this->_sessionStore->getValue($cacheKey);
        if (is_null($counters)) {
            $accountGroupInf = $this->getGroupById('self');
            $counters = array(
                self::CACHE_MOBILE_COUNTER => !is_null($accountGroupInf->getPrivileges()->getMobileCount()),
                self::CACHE_MAILBOX_COUNTER => !is_null($accountGroupInf->getPrivileges()->getMailboxCount()),
                self::CACHE_WEBSITE_COUNTER => !is_null($accountGroupInf->getPrivileges()->getWebsiteCount())
            );
            $this->_sessionStore->setValue(
                $cacheKey,
                $counters,
                self::CACHE_DEFAULT_TTL
            );
        }
        $this->_sessionStore->stopSession();
        return $counters;
    }

    /**
     * @return void
     */
    public function resetCountersCache()
    {
        $this->_sessionStore->startSession();
        $this->_sessionStore->unsetValue($this->getCountersCacheKey());
        $this->_sessionStore->stopSession();
    }

    /**
     * @return string
     */
    private function getCountersCacheKey()
    {
        return self::CACHE_COUNTERS_PREFIX . md5($this->getAccountServerUrl());
    }

    /**
     * @param Service_MspApiClient_Request_Struct_RequestUsageReport $reportAttributes
     * @param $groupId
     * @return mixed
     * @throws Service_BaseClient_HTTPException
     */
    public function requestUsageReport(Service_MspApiClient_Request_Struct_RequestUsageReport $reportAttributes, $groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_RequestUsageReport($reportAttributes, $groupId));
    }

    /**
     * @param Service_MspApiClient_Request_Struct_UsageReport $reportAttributes
     * @param $groupId
     * @return mixed
     * @throws Service_BaseClient_HTTPException
     */
    public function getUsageReport(Service_MspApiClient_Request_Struct_UsageReport $reportAttributes, $groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetUsageReport($reportAttributes, $groupId));
    }

    /**
     * @param int $groupId
     * @return Service_MspApiClient_Response_Struct_Brand
     */
    public function getBrand($groupId)
    {
        return $this->_authorizedCall(new Service_MspApiClient_Request_GetBrand($groupId));
    }

    private function _checkMobilePrivilegeSupport()
    {
        $counters = $this->getAccountPrivileges();
        return $counters[self::CACHE_MOBILE_COUNTER];
    }

    private function _checkMailboxPrivilegeSupport()
    {
        $counters = $this->getAccountPrivileges();
        return $counters[self::CACHE_MAILBOX_COUNTER];
    }

    private function _checkWebsitePrivilegeSupport()
    {
        $counters = $this->getAccountPrivileges();
        return $counters[self::CACHE_WEBSITE_COUNTER];
    }
}
