<?php
/**
 * @Copyright © 2002-2019 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\Services\AgentsManager;

use Arr;
use Service_AmsApiClient_Response_Struct_LinksListAgentsItem as Agent;

class AgentsManager
{
    const LANGUAGE_EN = 'en';
    const LANGUAGE_MULTI = 'multi';

    const ARCHITECTURE_32 = '32';
    const ARCHITECTURE_64 = '64';
    const ARCHITECTURE_NOT_SPECIFIED = '';

    const SYSTEM_WINDOWS_SERVER = 'windows_server';
    const SYSTEM_WINDOWS = 'windows';
    const SYSTEM_LINUX = 'linux';
    const SYSTEM_APPLE = 'apple';
    const SYSTEM_ANDROID = 'android';
    const SYSTEM_IOS = 'ios';
    const SYSTEM_HYPERV = 'hyperv';
    const SYSTEM_VMWARE = 'vmware';
    const SYSTEM_VIRTUOZZO = 'pcs';
    const SYSTEM_MICROSOFT_SQL_SERVER = 'sql';
    const SYSTEM_MICROSOFT_EXCHANGE_SERVER = 'arx';
    const SYSTEM_MICROSOFT_ACTIVE_DIRECTORY = 'ad';
    const SYSTEM_MICROSOFT_OFFICE_365 = 'office';

    const TYPE_CONSUMER = 'consumer';
    const TYPE_ENTERPRISE = 'enterprise';

    const CATEGORY_WORKSTATIONS = 'workstations';
    const CATEGORY_SERVERS = 'servers';
    const CATEGORY_MOBILE_DEVICES = 'mobileDevices';
    const CATEGORY_VIRTUALIZATION_HOSTS = 'virtualizationHosts';
    const CATEGORY_APPLICATIONS = 'applications';

    private $agentRules;

    function __construct()
    {
        $this->agentRules = $this->buildAgentRules();
    }

    public function buildAgentsListForUser($agents, $language)
    {
        $agentsHashmap = $this->buildAgentsHashmap($agents);
        $result = [];

        foreach ($this->agentRules as $agentId => $rule) {
            $agent = $this->findAgent($rule, $agentsHashmap, $language);
            if (is_null($agent)) {
                continue;
            }

            $result[$agentId] = $this->getAgentProperties($rule, $agent);
        }

        return $result;
    }

    /**
     * @param Agent[] $agents
     * @return array
     */
    private function buildAgentsHashmap($agents)
    {
        return Arr::map($agents, function ($agent) {
            return $this->buildHash($agent);
        }, function ($agent) {
            return $agent;
        });
    }

    /**
     * @return AgentRule[]
     */
    private function buildAgentRules()
    {
        $rulesList = require DIR_APP . DS . 'backup_agents.php';

        return array_map(function ($ruleOptions) {
            return new AgentRule($ruleOptions);
        }, $rulesList);
    }

    /**
     * @param AgentRule $rule
     * @param Agent[] $agentsHashmap
     * @param string $language
     * @return Agent|null
     */
    private function findAgent($rule, $agentsHashmap, $language)
    {
        $languages = $language == static::LANGUAGE_EN
            ? [static::LANGUAGE_MULTI, static::LANGUAGE_EN]
            : [static::LANGUAGE_MULTI, $language, static::LANGUAGE_EN];
        $webPriority = [true, null, false];

        foreach ($webPriority as $web) {
            foreach ($languages as $lang) {
                $agentHash = $this->buildHash($rule, $lang, $web);
                if (Arr::has($agentsHashmap, $agentHash)) {
                    return Arr::get($agentsHashmap, $agentHash);
                }
            }
        }

        return null;
    }

    /**
     * @param AgentRule $rule
     * @param Agent $agent
     * @return array
     */
    private function getAgentProperties($rule, $agent)
    {
        return [
            'url' => $agent->getUrl(),
            'category' => $rule->getCategory(),
            'title' => $rule->getTitle(),
        ];
    }

    /**
     * @param Agent $agent
     * @param null|string $language
     * @return string
     */
    private function buildHash($agent, $language = null, $web = null)
    {
        return md5(json_encode([
            is_null($language) ? $agent->getLanguage() : $language,
            $agent->getArchitecture(),
            $agent->getSystem(),
            $agent->getType(),
            is_null($web) ? $agent->getWeb() : $web,
        ]));
    }
}