<?php
/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\UsageReport\Aggregation;

use Acronis\App\Base\ApplicationCounter as Counter;
use Acronis\App\CloudApi\Api;
use Acronis\App\Services\CountersManager\CountersManagerAwareTrait;
use Acronis\App\Services\Logger\AppLoggerAwareTrait;
use Acronis\App\UsageReport\CsvFormatParser;
use Acronis\App\UsageReport\ReportRowWrapperInterface;
use Acronis\App\Utils\MemoizeTrait;
use APS\ControllerProxy;
use Json;
use Str;

class AggregationStrategy implements AggregationStrategyInterface
{
    use AppLoggerAwareTrait;
    use CountersManagerAwareTrait;
    use MemoizeTrait;

    const USAGE_KIND_FOR_PARTNER_AND_RESELLER = 'production';
    const USAGE_STRATEGY_FOR_PARTNER_AND_RESELLER = 'absolute';

    const USAGE_KIND_FOR_CUSTOMER = 'total';
    const USAGE_STRATEGY_FOR_CUSTOMER = 'absolute';

    /** @var ControllerProxy */
    private $apsc;

    /** @var Api */
    private $cloudApi;

    /** @var string */
    private $affinityHelper;

    /**
     * AggregationStrategy constructor.
     *
     * @param ControllerProxy $apsc
     */
    public function __construct($cloudApi, $apsc)
    {
        $this->cloudApi = $cloudApi;
        $this->apsc = $apsc;
        $this->affinityHelper = new AffinityHelper($cloudApi, $apsc->getInstanceId());
    }

    /**
     * @param Accumulator $accumulator
     * @param \Acronis\App\UsageReport\ReportRowWrapperInterface $reportRow
     * @param \tenant $tenant
     * @throws \Exception
     * @return mixed|void
     */
    public function aggregateRow(&$accumulator, $reportRow, $tenant)
    {
        $counters = $this->resolveCounters($reportRow, $tenant->type);

        if (count($counters)) {
            $reportRowJson = Json::encode($reportRow->getRawData());
            $countersList = implode(', ', array_map(function ($counter) {
                return $counter->getName();
            }, $counters));
            $this->getLog()->notice(
                'Row \'%s\' is mapped to counters %s.',
                $reportRowJson, $countersList
            );
        }

        $usageCalculationStrategy = $this->getUsageCalculationStrategy($tenant->type);
        $usageCountingKind = $this->getUsageCountingKind($tenant->type);
        $this->getLog()->notice(
            'Getting usage from "usage.%s.%s".',
            $usageCalculationStrategy, $usageCountingKind
        );

        foreach ($counters as $counter) {
            $usage = $reportRow->getUsage($usageCalculationStrategy, $usageCountingKind);
            $usage = $this->convertUsage($counter, $usage);
            $accumulator->addUsage($counter->getName(), $usage);
        }
    }

    private function getUsageCalculationStrategy($tenantType)
    {
        if ($tenantType === \tenant::TYPE_CUSTOMER) {
            return static::USAGE_STRATEGY_FOR_CUSTOMER;
        }

        return static::USAGE_STRATEGY_FOR_PARTNER_AND_RESELLER;
    }

    private function getUsageCountingKind($tenantType)
    {
        if ($tenantType === \tenant::TYPE_CUSTOMER) {
            return static::USAGE_KIND_FOR_CUSTOMER;
        }

        return static::USAGE_KIND_FOR_PARTNER_AND_RESELLER;
    }

    private function convertUsage($counter, $usage)
    {
        return $this->getCountersManager()->convertValueFromOfferingItemUnits($counter, $usage);
    }

    /**
     * @return Counter[][][]
     */
    private function getCountersHashMap()
    {
        return $this->memoize(function () {
            $counters = $this->getCountersManager()->getAllCounters();
            $countersHashMap = [];
            foreach ($counters as $counter) {
                if (!$counter->hasUsage()) {
                    continue;
                }

                foreach ($counter->getTenantTypes() as $tenantType) {
                    $countersHashMap[$counter->getOfferingItemName()][$tenantType][] = $counter;
                }
            }

            return $countersHashMap;
        });
    }

    /**
     * @param ReportRowWrapperInterface $reportRow
     * @param string $tenantType
     * @throws \Exception
     * @return Counter[]
     */
    private function resolveCounters($reportRow, $tenantType)
    {
        $offeringItemName = $reportRow->getOfferingItemName();
        $countersHashMap = $this->getCountersHashMap();

        if (!isset($countersHashMap[$offeringItemName][$tenantType])) {
            return [];
        }

        $counters = $countersHashMap[$offeringItemName][$tenantType];
        $result = [];
        foreach ($counters as $counter) {
            if (!$this->checkAggregationRules($counter, $reportRow)) {
                continue;
            }

            $result[] = $counter;
        }

        return $result;
    }

    /**
     * @param Counter $counter
     * @param ReportRowWrapperInterface $reportRow
     * @return bool
     * @throws \Exception
     */
    private function checkAggregationRules(Counter $counter, $reportRow)
    {
        $rules = $counter->getAggregationRules();

        $this->getLog()->notice(
            'Checking aggregation rules \'%s\' for counter "%s".',
            $counter->getName(), Json::encode($rules)
        );

        foreach ($rules as $reportRowKey => $allowedValues) {
            if (!is_array($allowedValues)) {
                $allowedValues = [$allowedValues];
            }

            $value = $this->getReportRowValue($reportRowKey, $reportRow);

            if (!in_array($value, $allowedValues)) {
                return false;
            }
        }

        return true;
    }

    /**
     * @param string $reportRowKey
     * @param ReportRowWrapperInterface $reportRow
     * @return string|null|int
     * @throws \Exception
     */
    private function getReportRowValue($reportRowKey, $reportRow)
    {
        switch ($reportRowKey) {
            case CsvFormatParser::SYNTETIC_COLUMN_INFRA_AFFINITY:
                return $this->affinityHelper->resolveAffinity($reportRow);
            case CsvFormatParser::COLUMN_INFRA_BACKEND_TYPE:
                return $reportRow->getInfraBackendType();
        }

        throw new \InvalidArgumentException(Str::format(
            'Unsupported report row key "%s" was provided.',
            $reportRowKey
        ));
    }
}