<?php
/**
 * @Copyright © 2002-2018 Acronis International GmbH. All rights reserved
 */

namespace Acronis\App\UsageReport\Aggregation;

use Acronis\App\Services\Cache\CacheInterface;
use Acronis\App\Services\Logger\AppLoggerAwareTrait;
use Acronis\App\UsageReport\CompositeKey;
use Acronis\App\UsageReport\CsvReportIterator;
use Acronis\App\UsageReport\Exception\AggregatorException;
use Acronis\App\UsageReport\FormatParserInterface;
use Acronis\App\UsageReport\ReportRowWrapperInterface;
use Acronis\App\UsageReport\ReportStorageInterface;
use Arr;
use Str;

class Aggregator
{
    use AppLoggerAwareTrait;

    /** @var string[] */
    private $registeredAccumulators = [];

    /** @var CacheInterface */
    private $cache;

    /** @var ReportStorageInterface */
    private $storage;

    /** @var AggregationStrategyInterface */
    private $aggregationStrategy;

    /** @var \tenant[] */
    private $tenants;

    /** @var string */
    private $date;

    /**
     * Aggregator constructor.
     * @param CacheInterface $cache
     * @param ReportStorageInterface $storage
     * @param AggregationStrategyInterface $aggregationStrategy
     * @param FormatParserInterface $formatParser
     */
    public function __construct(
        CacheInterface $cache,
        ReportStorageInterface $storage,
        AggregationStrategyInterface $aggregationStrategy,
        $tenants
    ) {
        $this->cache = $cache;
        $this->storage = $storage;
        $this->aggregationStrategy = $aggregationStrategy;
        $this->tenants = $tenants;
    }

    /**
     * @param CsvReportIterator $reportIterator
     * @param string $date
     * @param string $datacenterId
     * @throws \Exception
     */
    public function aggregateReportUsage($reportIterator, $date)
    {
        $this->date = $date;
        $this->invalidateStorage();
        $this->aggregationLoop($reportIterator);
        $this->saveAllAccumulatorsToStorage();
    }

    /**
     * @param CsvReportIterator $reportIterator
     * @throws \Exception
     */
    protected function aggregationLoop($reportIterator)
    {
        $accumulator = null;

        $this->getLog()->notice('Started aggregation loop.');

        foreach ($reportIterator as $rowIndex => $reportRow) {
            $tenant = $this->getTenant($reportRow);
            // Skip report rows for which there are no subscriptions
            if (is_null($tenant)) {
                continue;
            }

            $accumulator = $this->resolveAccumulator($reportRow, $accumulator);

            $this->getLog()->notice(
                'Started report row #%s aggregation for subscription "%s".',
                $rowIndex + 1, $tenant->subscriptionId
            );

            $this->aggregationStrategy->aggregateRow($accumulator, $reportRow, $tenant);

            $this->getLog()->notice(
                'Finished report row #%s aggregation.',
                $rowIndex + 1
            );
        }

        // Saving last accumulator
        if (!is_null($accumulator)) {
            $this->saveAccumulatorToCache($accumulator);
        }

        $this->getLog()->notice('Finished aggregation loop.');
    }

    /**
     * @param ReportRowWrapperInterface $reportRow
     * @return \tenant|null
     */
    private function getTenant($reportRow)
    {
        return Arr::get($this->tenants, $reportRow->getTenantId());
    }

    /* ====== Utility methods for working with accumulators ====== */

    /**
     * @param Accumulator $accumulator
     */
    private function registerAccumulator($accumulator)
    {
        $key = (string)new CompositeKey($accumulator->getId(), $this->date);
        $this->registeredAccumulators[$key] = true;
    }

    /**
     * @param Accumulator $accumulator
     * @return bool
     */
    private function isAccumulatorRegistered($accumulator)
    {
        $key = (string)new CompositeKey($accumulator->getId(), $this->date);

        return isset($this->registeredAccumulators[$key]);
    }

    /**
     * @param ReportRowWrapperInterface $reportRow
     * @throws AggregatorException
     * @return mixed
     */
    protected function resolveAccumulatorId($reportRow)
    {
        $tenantId = $reportRow->getTenantId();

        if (!isset($this->tenants[$tenantId])) {
            throw new AggregatorException(Str::format(
                'Tenant with tenantId "%s" was not found in tenants map.',
                $tenantId
            ));
        }

        return $this->tenants[$tenantId]->subscriptionId;
    }

    /**
     * @param ReportRowWrapperInterface $reportRow
     * @param Accumulator $accumulator
     * @return Accumulator
     * @throws AggregatorException
     */
    protected function resolveAccumulator($reportRow, $accumulator)
    {
        $accumulatorId = $this->resolveAccumulatorId($reportRow);

        if (is_null($accumulator)) {
            $accumulator = new Accumulator($accumulatorId);
        }

        if ($accumulator->getId() !== $accumulatorId) {
            // Saving old accumulator before its change to another subscription (accumulatorId)
            $this->saveAccumulatorToCache($accumulator);

            $accumulator = new Accumulator($accumulatorId);
            // Loading accumulator usages from cache if accumulator was previously registered
            $this->loadAccumulatorFromCache($accumulator);
        }

        return $accumulator;
    }

    /* ====== Utility methods for working with cache ====== */

    /**
     * @param Accumulator $accumulator
     */
    private function saveAccumulatorToCache($accumulator)
    {
        $this->registerAccumulator($accumulator);
        $this->cache->set((string)new CompositeKey($accumulator->getId(), $this->date), $accumulator->getUsages());
    }

    /**
     * @param Accumulator $accumulator
     */
    private function loadAccumulatorFromCache($accumulator)
    {
        if ($this->isAccumulatorRegistered($accumulator)) {
            $key = (string)new CompositeKey($accumulator->getId(), $this->date);
            $accumulator->setUsages($this->cache->get($key));
        }
    }

    /* ====== Utility methods for working with storage ====== */

    private function invalidateStorage()
    {
        $this->getLog()->notice('Invalidating report storage.');
        $this->storage->deleteAll();
    }

    private function saveAllAccumulatorsToStorage()
    {
        $this->getLog()->notice('Saving aggregated data to storage.');

        foreach (array_keys($this->registeredAccumulators) as $key) {
            $usages = $this->cache->get($key);

            $this->getLog()->notice(
                'Copying usage "%s" from cache to storage.',
                $key
            );

            $this->storage->set($key, $usages);
        }
    }
}