<?php

/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */
class ProtectedSettings extends AbstractSettings
{
    const FIELD_SALT = 'salt';
    const FIELD_ENCRYPTED_FLAG = 'encrypted';
    const FIELD_KEY = 'key';
    const PROTECTED_SETTINGS_CONFIG_FILE = 'protected_settings.ini';
    const DATACENTER_CREDENTIALS_CONFIG_FILE = 'datacenter_credentials.ini';
    const DATACENTER_CREDENTIALS_CONFIG_NAME = 'Datacenter';
    const PROTECTED_SETTINGS_CONFIG_NAME = "ProtectedSettings";
    const ALGORITHM = 'aes128';
    const MIN_KEY_VALUE = 1000000000000000;
    const MAX_KEY_VALUE = 9999999999999999;
    const KEY_LENGTH = 16;
    const EVENT_PROTECTED_CONFIG_WAS_ENCRYPTED = 'Protected config was encrypted';
    const EVENT_PROTECTED_CONFIG_WAS_DECRYPTED = 'Protected config was decrypted';

    /**
     * Crypt algorithm
     * @var string
     */
    protected $method;

    /**
     * @var string
     */
    protected $salt;

    /**
     * @var int
     */
    protected $key;

    /** @var array */
    private $datacenterConfig;

    /** @var ProtectedSettings */
    private static $instance;

    /**
     * @return ProtectedSettings
     */
    public static function getInstance()
    {
        if (is_null(self::$instance)) {
            self::$instance = new ProtectedSettings();
        }
        return self::$instance;
    }


    private function __construct()
    {
        $this->method = self::ALGORITHM;
    }

    /**
     * @param string $data
     * @return string
     */
    private function encrypt($data)
    {
        $encrypted_data = openssl_encrypt($data, $this->method, $this->salt, false, $this->key);
        Log::getInstance()->event(self::EVENT_PROTECTED_CONFIG_WAS_ENCRYPTED);
        return $encrypted_data;
    }

    /**
     * @param string $data
     * @return string
     */
    private function decrypt($data)
    {
        $decrypted_data = openssl_decrypt($data, $this->method, $this->salt, false, $this->key);
        Log::getInstance()->event(self::EVENT_PROTECTED_CONFIG_WAS_DECRYPTED);
        return $decrypted_data;
    }

    /**
     * @return array
     * @throws Exception
     */
    public function getDatacenterConfig()
    {
        if (is_null($this->datacenterConfig)) {
            $fileName = DIR_CONFIG . self::DATACENTER_CREDENTIALS_CONFIG_FILE;

            try {
                $this->datacenterConfig = $this->readConfig($fileName, self::DATACENTER_CREDENTIALS_CONFIG_NAME);
            } catch (Exception $e) {
                Log::getInstance()->error($e->getMessage());
                throw $e;
            }
        }
        return $this->datacenterConfig;
    }

    /**
     * @param string $fileName
     * @param string $configName
     * @return array
     * @throws Exception
     */
    protected function readConfig($fileName, $configName)
    {
        $this->checkFileAvailable($fileName, $configName);
        $content = file_get_contents($fileName);

        if (strpos($content, self::FIELD_ENCRYPTED_FLAG) === false) {
            $this->initProtectedSettings();
            $content = $this->decrypt($content);
        } else {
            $this->generateProtectedSettings();
            $encryptData = $this->encrypt($content);
            $this->writeConfig($fileName, $configName, $encryptData);
        }

        return parse_ini_string($content, true);
    }

    /**
     * Initialised settings for decrypt
     */
    private function initProtectedSettings()
    {
        $fileName = DIR_CONFIG . self::PROTECTED_SETTINGS_CONFIG_FILE;
        $protectedConfig = parent::readConfig($fileName, self::PROTECTED_SETTINGS_CONFIG_NAME);

        $this->salt = $protectedConfig[self::FIELD_SALT];
        $this->key = $protectedConfig[self::FIELD_KEY];
    }

    /**
     * Generate settings for encrypt
     */
    private function generateProtectedSettings()
    {
        $this->key = substr(mt_rand(self::MIN_KEY_VALUE, self::MAX_KEY_VALUE), 0, self::KEY_LENGTH);
        $this->salt = uniqid(mt_rand(), true);
        $content = "salt={$this->salt}
key={$this->key}";
        $fileName = DIR_CONFIG . self::PROTECTED_SETTINGS_CONFIG_FILE;
        $this->writeConfig($fileName, self::PROTECTED_SETTINGS_CONFIG_NAME, $content);
    }
}
