<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class RqlBuilder
{
    const IMPLEMENTING = 'implementing(%s)';
    const EQ = 'eq(%s,%s)';
    const NE = 'ne(%s,%s)';
    const GT = 'gt(%s,%s)';
    const GE = 'ge(%s,%s)';
    const LT = 'lt(%s,%s)';
    const LE = 'le(%s,%s)';
    const IN = 'in(%s,(%s))';
    const OUT = 'out(%s,(%s))';
    const LIMIT = 'limit(%s,%s)';
    const COMPOSING = 'composing(%s)';
    const LIKE = 'like(%s,%s)';
    const SELECT = 'select(%s)';
    const LINKED_WITH = 'linkedWith(%s)';

    const JOIN_AND = 'and';
    const JOIN_OR = 'or';
    const AND_SEPARATOR = ',';
    const OR_SEPARATOR = '|';

    const VALUE_NULL = 'null()';
    const VALUE_TRUE = 'true()';
    const VALUE_FALSE = 'false()';
    const VALUE_EMPTY = 'empty()';

    const CLOSING_PARENTHESIS = ')';
    const OPENING_PARENTHESIS = '(';

    protected $query;

    public static function createInstance()
    {
        return new static;
    }

    /**
     * @param string $resourceType
     * @param string $operator
     * @return $this
     */
    public function implementing($resourceType, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::IMPLEMENTING, $resourceType), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param array $values
     * @param string $operator
     * @return $this
     */
    public function in($property, $values, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::IN, $property, $this->getImplodeString($values)), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param $values
     * @param string $operator
     * @return $this
     */
    public function out($property, $values, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::OUT, $property, $this->getImplodeString($values)), $operator);
        return $this;
    }

    /**
     * @param $start
     * @param $number
     * @param string $operator
     * @return $this
     */
    public function limit($start, $number, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::LIMIT, $start, $number), $operator);
        return $this;
    }

    /**
     * @param $resourceType
     * @param string $operator
     * @return $this
     */
    public function composing($resourceType, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::COMPOSING, $resourceType), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param $pattern
     * @param string $operator
     * @return $this
     */
    public function like($property, $pattern, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::LIKE, $property, $pattern), $operator);
        return $this;
    }

    /**
     * @param array $attributes
     * @param string $operator
     * @return $this
     */
    public function select($attributes, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::SELECT, $this->getImplodeString($attributes)), $operator);
        return $this;
    }

    /**
     * @param $resourceType
     * @param string $operator
     * @return $this
     */
    public function linkedWith($resourceType, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::LINKED_WITH, $resourceType), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param $value
     * @param string $operator
     * @return $this
     */
    public function eq($property, $value, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::EQ, $property, $this->prepareValue($value)), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param $value
     * @param string $operator
     * @return $this
     */
    public function ne($property, $value, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::NE, $property, $this->prepareValue($value)), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param $value
     * @param string $operator
     * @return $this
     */
    public function gt($property, $value, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::GT, $property, $value), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param $value
     * @param string $operator
     * @return $this
     */
    public function ge($property, $value, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::GE, $property, $value), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param $value
     * @param string $operator
     * @return $this
     */
    public function lt($property, $value, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::LT, $property, $value), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param $value
     * @param string $operator
     * @return $this
     */
    public function le($property, $value, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::LE, $property, $value), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param string $operator
     * @return $this
     */
    public function isNull($property, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::EQ, $property, self::VALUE_NULL), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param string $operator
     * @return $this
     */
    public function isNotNull($property, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::NE, $property, self::VALUE_NULL), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param string $operator
     * @return $this
     */
    public function isTrue($property, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::EQ, $property, self::VALUE_TRUE), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param string $operator
     * @return $this
     */
    public function isFalse($property, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::EQ, $property, self::VALUE_FALSE), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param string $operator
     * @return $this
     */
    public function isEmpty($property, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::EQ, $property, self::VALUE_EMPTY), $operator);
        return $this;
    }

    /**
     * @param $property
     * @param string $operator
     * @return $this
     */
    public function isNotEmpty($property, $operator = self::JOIN_AND)
    {
        $this->addUnit(Str::format(self::NE, $property, self::VALUE_EMPTY), $operator);
        return $this;
    }

    /**
     * @param $operator
     * @param RqlBuilder $builder
     * @return $this
     */
    public function joinQuery($operator, RqlBuilder $builder)
    {
        if ($this->getSeparatorByOperator($operator) == self::OR_SEPARATOR) {
            $this->query = $this->wrapParenthesis(
                $this->wrapParenthesis($this->getQuery()) . self::OR_SEPARATOR . $this->wrapParenthesis($builder->getQuery())
            );
        } else {
            $query = $this->wrapParenthesis($this->getQuery());
            $this->flush();
            $this->addUnit($query);
            $this->query = $this->wrapParenthesis($this->addUnit($this->wrapParenthesis($builder->getQuery()), self::JOIN_AND)->getQuery());
        }

        return $this;
    }

    /**
     * @return mixed
     */
    public function getQuery()
    {
        return empty($this->query) ? '' : $this->query;
    }

    /**
     * Flush of query
     */
    public function flush()
    {
        $this->query = null;
    }

    /**
     * @param $parameters
     * @return string
     */
    protected function getImplodeString($parameters)
    {
        foreach ($parameters as &$value) {
            $value = $this->prepareValue($value);
        }
        return implode(self::AND_SEPARATOR, $parameters);
    }

    /**
     * @param $unit
     * @param $operator
     * @return $this
     */
    protected function addUnit($unit, $operator = '')
    {
        $unitSeparator = $this->getSeparatorByOperator($operator);
        switch (true) {
            case empty($this->query):
                $this->query = $unit;
                break;
            case $unitSeparator == self::OR_SEPARATOR:
                $prevUnitPosition = ($position = strrpos($this->query, self::CLOSING_PARENTHESIS . self::AND_SEPARATOR)) ? $position + 2 : 0;
                $this->query = substr($this->query, 0, $prevUnitPosition) . self::OPENING_PARENTHESIS . substr($this->query, $prevUnitPosition) . self::OR_SEPARATOR . $unit . self::CLOSING_PARENTHESIS;
                break;
            default:
                $this->query = $this->query . $unitSeparator . $unit;
        }

        return $this;
    }

    private function getSeparatorByOperator($operator)
    {
        return $operator == self::JOIN_AND ? self::AND_SEPARATOR : self::OR_SEPARATOR;
    }

    private function wrapParenthesis($query)
    {
        return self::OPENING_PARENTHESIS . $query . self::CLOSING_PARENTHESIS;
    }

    /**
     * @param mixed
     * @return string
     */
    private function prepareValue($value)
    {
        switch (true) {
            case is_bool($value):
                $value = $value ? self::VALUE_TRUE : self::VALUE_FALSE;
                break;
            case is_null($value):
                $value = self::VALUE_NULL;
                break;
        }
        return $value;
    }
}
