<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

require_once("OpenApiFacade.php");
require_once("MigrationHelper.php");

class ApsMigration
{
    const EXCEPTION_RESOURCE_NOT_FOUND_TEMPLATE = "'Not all resource found in subscription with id %s'";
    const EXCEPTION_INCORRECT_SETTINGS_IN_PROFILE = "In the settings of Backup Profile must be set checkbox 'Manage user quotas' on 'Yes', and set checkbox 'Do not allow to set user quotas exceeding subscription limits' on 'No'";
    const EXCEPTION_ID_NOT_FOUND_IN_CONFIG_FOLDER = "Id is not found in config folder";
    const EXCEPTION_DATACENTER_NOT_FOUND = "Datacenter with id %s not found";
    const EXCEPTION_APS_PACKAGE_ID_NOT_FOUND = 'APS Package ID not found';

    const OUTPUT_START_MIGRATION_FOR_GROUP = "Starting migration for group %s... \n";
    const OUTPUT_STOP_MIGRATION_FOR_GROUP = "Migration for group %s was finished \n";
    const OUTPUT_FAIL_MIGRATION_FOR_GROUP = "Migration for group %s was failed \n";
    const OUTPUT_SUCCESS_MESSAGE = "Success\n";
    const OUTPUT_ERROR_MESSAGE_AND_EXPLANATION = "Error %s\n";
    const OUTPUT_GET_ACCOUNT_TOKEN_MESSAGE = "Get account token \n";
    const OUTPUT_CREATING_TENANT_MESSAGE = "Creating tenant\n";
    const OUTPUT_CREATED_TENANT_MESSAGE = "Created tenant: %s\n";
    const OUTPUT_GETTING_USERS_MESSAGE = "Getting users\n";
    const OUTPUT_MIGRATING_USER_MESSAGE = "Migrating user %s\n";
    const OUTPUT_SUCCESSFULLY_MIGRATED_MESSAGE = "Subscription migrated\n";

    const TENANT_COUNTER_STORAGE_USAGE = 'StorageUsage';

    const LOG_START_UNPROVISIONING_RESOURCES = 'Start unporovisioning for new resources';
    const LOG_STOP_UNPROVISIONING_RESOURCES = 'Stop unporovision for new resources';
    const LOG_UNPROVISIONING_RESOURCE = 'Unprovisioning resource with id %s ...';
    const LOG_RESOURCE_UNPROVISIONED = 'Resource with id %s was unprovisioned';

    const DEFAULT_CONTINUE_ORDER_STATUS = 'CP';
    const DEFAULT_TERMINATE_ORDER_STATUS = 'PF';

    const CONFIG_FOLDER_DIR = "../config";
    const CONFIG_FILE_SUFFIX = ".apsc.pem";

    const PROPERTY_CONTROLLER_URI = "controller_uri";
    const PROPERTY_APS_TOKEN = "aps_token";

    const PROVIDER_ACCOUNT_ID = 1;
    const PROVIDER_SUBSCRIPTION_ID = 1;

    /**
     * @var Service_OpenApi_Response_SubscriptionInfo
     */
    private $subscription;

    /**
     * @var MigrationHelper
     */
    private $helper;

    /**
     * BA application ID
     * @var int
     */
    private $instanceId;

    /**
     * @var ApiFactory
     */
    private $apiFactory;

    /**
     * @var Service_OpenApi_Api
     */
    private $openApi;

    /**
     * @var BackupAgentSoapClient
     */
    private $baSoapClient;

    /**
     * @var Service_PbaApi_Api
     */
    private $pbaApi;

    /**
     * @var string[]
     */
    private $continueStatuses;

    /**
     * @var string[]
     */
    private $terminateStatuses;

    /**
     * @var string UUID
     */
    private $apsPackageId;

    /**
     * @param ApiFactory $apiFactory
     * @param string[] $continueStatuses
     * @param string[] $terminateStatuses
     */
    public function __construct(
        $apiFactory,
        array $continueStatuses = array(self::DEFAULT_CONTINUE_ORDER_STATUS),
        array $terminateStatuses = array(self::DEFAULT_TERMINATE_ORDER_STATUS)
    )
    {
        $this->apiFactory = $apiFactory;
        $this->continueStatuses = $continueStatuses;
        $this->terminateStatuses = $terminateStatuses;

        $this->baSoapClient = $apiFactory->getBackupAgentSoapClient(
            BACKUPAGENT_SERVER_DNS, BACKUPAGENT_SERVER_ADMIN_LOGIN,
            BACKUPAGENT_SERVER_ADMIN_PASSWORD, BACKUPAGENT_SERVER_PROTOCOL
        );
        $this->openApi = $apiFactory->getOpenApi(defined('OPEN_API_URI') ? OPEN_API_URI : POA_API_URI);
        $this->pbaApi = $apiFactory->getPbaApi(Settings::getInstance()->getBillingConfig());
    }

    public function migrate($groupNames, $applicationId, $activateUser, $compositeResourceIDs, $dataCenterID)
    {
        foreach ($groupNames as $groupName) {
            try {
                echo sprintf(self::OUTPUT_START_MIGRATION_FOR_GROUP, $groupName);
                $this->migrateGroup($groupName, $applicationId, $activateUser, $compositeResourceIDs, $dataCenterID);
                echo sprintf(self::OUTPUT_STOP_MIGRATION_FOR_GROUP, $groupName);
            } catch (Exception $e) {
                echo sprintf(self::OUTPUT_FAIL_MIGRATION_FOR_GROUP, $groupName);
                echo $e->getMessage() . PHP_EOL;
                Log::getInstance()->error($e->getMessage());
            }
        }
    }

    /**
     * Migrate one particular subscription of the account
     * @param string $groupLogin
     * @param int $applicationId
     * @param boolean $activateUser
     * @param array $compositeResourceIDs
     * @param string $dataCenterID
     * @throws Exception
     */
    private function migrateGroup($groupLogin, $applicationId, $activateUser, $compositeResourceIDs, $dataCenterID)
    {
        $apiHelper = new OpenApiFacade($this->openApi);
        $this->instanceId = $apiHelper->getApplicationInstanceId($groupLogin, $applicationId);
        $subscriptionId = $apiHelper->findSubscriptionId($this->instanceId);
        $this->subscription = $this->openApi->getSubscription($subscriptionId);
        $accountId = $this->subscription->getOwnerId();

        $apsc = $this->createApsControllerInstance($accountId, $subscriptionId);
        $this->apsPackageId = $this->fetchApsPackageId($apsc);
        $datacenter = $this->getDatacenter($apsc, $subscriptionId, $dataCenterID);
        $mspApi = $this->apiFactory->getMspApi($datacenter, $apsc->getInstanceId());

        $this->helper = new MigrationHelper($apsc, $this->openApi, $mspApi, $this->pbaApi, $this->baSoapClient, $this->apsPackageId);

        $subscriptionResources = $this->helper->getSubscriptionResources($subscriptionId);
        $this->checkResources($subscriptionResources);
        $abcStorageResourceId = $this->getStorageResourceId($subscriptionResources);
        $mspGroupId = $this->helper->getMspGroupId($groupLogin);

        $userForMigrate = $this->helper->getUserForMigrate($this->instanceId, $this->subscription->getOwnerId());
        $fakeProvision = !empty($userForMigrate);

        $this->helper->checkTenantMigrated($mspGroupId, $subscriptionId, $fakeProvision);
        echo self::OUTPUT_CREATING_TENANT_MESSAGE;

        $tenant = $this->helper->createTenant(
            $this->subscription->getSubscriptionId(),
            $this->subscription->getOwnerId(),
            $mspGroupId,
            $datacenter,
            $fakeProvision
        );

        $this->helper->updateSubscriptionLimits($tenant, $this->continueStatuses, $this->terminateStatuses, $abcStorageResourceId, $compositeResourceIDs);

        $newResources = array();
        $providerApsc = $this->createApsControllerInstance(self::PROVIDER_ACCOUNT_ID, self::PROVIDER_SUBSCRIPTION_ID);
        $tenant = $this->helper->provisionTenant($tenant, $this->subscription->getSubscriptionId(), $mspGroupId, $fakeProvision, $datacenter, $providerApsc);
        $newResources[] = $tenant;
        echo sprintf(self::OUTPUT_CREATED_TENANT_MESSAGE, $tenant->aps->id);
        try {
            echo self::OUTPUT_GETTING_USERS_MESSAGE;
            $newUsers = array();
            foreach ($userForMigrate as $user) {
                $newUsers[] = $this->createUser($activateUser, $user, $tenant);
            }
            $this->helper->setMigrationIsCompleted($tenant, $newUsers);
        } catch (Exception $e) {
            if (!empty($newResources)) {
                Log::getInstance()->error(self::LOG_START_UNPROVISIONING_RESOURCES);
                foreach ($newResources as $newResource) {
                    Log::getInstance()->event(sprintf(self::LOG_UNPROVISIONING_RESOURCE, $newResource->aps->id));
                    $this->helper->apsc->unprovisionResource($newResource);
                    Log::getInstance()->event(sprintf(self::LOG_RESOURCE_UNPROVISIONED, $newResource->aps->id));
                }
                Log::getInstance()->error(self::LOG_STOP_UNPROVISIONING_RESOURCES);
            }
            throw $e;
        }
        echo self::OUTPUT_SUCCESSFULLY_MIGRATED_MESSAGE;
    }

    /**
     * @param array $subscriptionResources
     * @throws Exception
     */
    private function checkResources($subscriptionResources)
    {
        $checkAcronisBackupService = false;
        $checkTenant = false;
        $checkUser = false;
        $checkStorageUsage = false;

        $globalsType = \APS\TypeLibrary::getTypeByServiceId('globals', $this->apsPackageId);
        $tenantType = \APS\TypeLibrary::getTypeByServiceId('tenant', $this->apsPackageId);
        $profileType = \APS\TypeLibrary::getTypeByServiceId('profile', $this->apsPackageId);
        $usersType = \APS\TypeLibrary::getTypeByServiceId('users', $this->apsPackageId);

        foreach ($subscriptionResources as $resource) {
            if (!isset($resource->apsType)) {
                continue;
            }
            if (!$checkAcronisBackupService
                && !isset($resource->property)
                && $resource->apsType == $globalsType
            ) {
                $checkAcronisBackupService = true;
                continue;
            }
            if (!$checkTenant && !isset($resource->property) && $resource->apsType == $tenantType) {
                $checkTenant = true;
                continue;
            }
            if (!$checkStorageUsage
                && isset($resource->property)
                && $resource->property == self::TENANT_COUNTER_STORAGE_USAGE
                && $resource->apsType == $tenantType
            ) {
                $checkStorageUsage = true;
                continue;
            }
            if (!$checkUser && !isset($resource->property) && $resource->apsType == $usersType) {
                $checkUser = true;
                continue;
            }
            if ($resource->apsType == $profileType) {
                $this->checkBackupProfile($resource);
                continue;
            }
        }

        if (!$checkAcronisBackupService
            || !$checkTenant
            || !$checkStorageUsage
            || !$checkUser
        ) {
            throw new Exception(sprintf(self::EXCEPTION_RESOURCE_NOT_FOUND_TEMPLATE, $this->subscription->getSubscriptionId()));
        }
    }

    /**
     * @param $activateUser
     * @param $user
     * @param $tenant
     * @return mixed | $newUser
     * @throws Exception
     */
    private function createUser($activateUser, $user, $tenant)
    {
        $newUser = null;
        try {
            echo sprintf(self::OUTPUT_MIGRATING_USER_MESSAGE, $user->displayName);
            $newUser = $this->helper->createUser($user, $tenant, $activateUser);
            echo self::OUTPUT_SUCCESS_MESSAGE;
        } catch (Exception $exception) {
            echo sprintf(self::OUTPUT_ERROR_MESSAGE_AND_EXPLANATION, $exception->getMessage());
            throw $exception;
        }
        return $newUser;
    }

    private function getStorageResourceId($subscriptionResources)
    {
        foreach ($subscriptionResources as $resource) {
            if (!isset($resource->apsType)) {
                continue;
            }
            if (isset($resource->property)
                && $resource->property == self::TENANT_COUNTER_STORAGE_USAGE
                && $resource->apsType == \APS\TypeLibrary::getTypeByServiceId('tenant', $this->apsPackageId)
            ) {
                return $resource->id;
            }
        }
        return null;
    }

    private function checkBackupProfile($backupProfile)
    {
        $resource = $this->helper->apsc->getResource($backupProfile->apsId);
        if ($resource->userQuota != profile::USER_QUOTA_SOFT) {
            throw new Exception(self::EXCEPTION_INCORRECT_SETTINGS_IN_PROFILE);
        }
    }

    private function createApsControllerInstance($accountId, $subscriptionId)
    {
        $accountToken = $this->openApi->getAccountToken($accountId, $subscriptionId);
        return $this->apiFactory->getApsController(
            $accountToken[self::PROPERTY_CONTROLLER_URI],
            $accountToken[self::PROPERTY_APS_TOKEN]
        );
    }

    private function getDatacenter($apsc, $subscriptionId, $datacenterId)
    {
        $aps = new APS\ControllerProxy($apsc->getEndPoint());
        $sub = $apsc->getResources(sprintf(MigrationHelper::RQL_IMPLEMENTING_SUBSCRIPTION_EQ_SUBSCRIPTION_ID, $subscriptionId));
        $resources = $sub[0]->resources();
        $globalsSubscriptionResource = null;
        foreach ($resources as $resource) {
            if (!isset($resource->apsType)) {
                continue;
            }
            if (!isset($resource->property) && $resource->apsType == \APS\TypeLibrary::getTypeByServiceId('globals', $this->apsPackageId)) {
                $globalsSubscriptionResource = $resource;
                break;
            }
        }
        if (empty($globalsSubscriptionResource)) {
            throw new Exception(sprintf(MigrationHelper::EXCEPTION_GLOBALS_NOT_FOUND_TEMPLATE, $sub[0]->subscriptionId));
        }
        $global = $aps->getResource($globalsSubscriptionResource->apsId);
        $dc = null;
        foreach ($global->datacenters as $datacenter) {
            if ($datacenter->id == $datacenterId) {
                $dc = $datacenter;
                break;
            }
        }

        if (is_null($dc)) {
            throw new Exception(sprintf(self::EXCEPTION_DATACENTER_NOT_FOUND, $datacenterId));
        }
        return $dc;
    }

    /**
     * @param $apsc
     * @return null
     * @throws Exception
     */
    private function fetchApsPackageId($apsc)
    {
        $packages = json_decode($apsc->getIo()->sendRequest('GET', 'aps/2/packages'));
        $packageId = null;
        foreach ($packages as $package) {
            if ($package->id == APSResource::PACKAGE_ID) {
                $packageId = $package->aps->id;
            }
        }

        if (is_null($packageId)) {
            throw new Exception(self::EXCEPTION_APS_PACKAGE_ID_NOT_FOUND);
        }

        return $packageId;
    }
}
