<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

require_once(__DIR__ . "/ResourceQuotas.php");

class CsvMigration
{
    const DEFAULT_CONTINUE_ORDER_STATUS = 'CP';
    const DEFAULT_TERMINATE_ORDER_STATUS = 'PF';

    const CONFIG_FOLDER_DIR = "../config";
    const CONFIG_FILE_SUFFIX = ".apsc.pem";

    const PROPERTY_CONTROLLER_URI = "controller_uri";
    const PROPERTY_APS_TOKEN = "aps_token";
    const PROPERTY_RESOURCE_ID = 'resource_id';
    const PROPERTY_ADDITIONAL_AMOUNT = 'additional_amount';
    const PROPERTY_SERVICE_USER_ID = 'osa_service_user_id';
    const PROPERTY_MSP_USER_LOGIN = 'abc_login';
    const PROPERTY_STORAGE_QUOTA = 'user_storage_quota';
    const PROPERTY_PRIVILEGED = 'privileged';
    const PROPERTY_ACTIVATE = 'activate';

    const USE_PAY_TOOL = null;
    const USE_DEFAULT_PAY_TOOL = 1;

    const LINK_DATACENTER = 'datacenter';

    const APS_STATUS_READY = 'aps:ready';
    const APS_TYPE_SUBSCRIPTION = 'http://parallels.com/aps/types/pa/subscription/1.0';
    const APS_TYPE_SERVICE_USER = 'http://aps-standard.org/types/core/service-user/1.0';

    const STORAGE_USAGE = 'StorageUsage';
    const MOBILE_DEVICE_COUNT = 'MobileDeviceCount';
    const WORKSTATION_COUNT = 'WorkstationCount';
    const SERVER_COUNT = 'ServerCount';
    const VM_COUNT = 'VmCount';
    const MAILBOX_COUNT = 'MailboxCount';
    const WEBSITE_COUNT = 'WebsiteCount';

    const EX_MSG_WARNING_QUOTA_IS_LESS_THAN_USAGE = "Warning: Quota for %s is less than usage %s < %s.";

    /**
     * @var ApiFactory
     */
    private $apiFactory;

    /**
     * @var Service_OpenApi_Api
     */
    private $openApi;

    /**
     * @var Service_PbaApi_Api
     */
    private $pbaApi;

    /**
     * @var \APS\ControllerProxy
     */
    private $subscriptionApsController;

    /**
     * @var \APS\ControllerProxy
     */
    private $instanceApsController;

    /**
     * @var string[]
     */
    private $continueStatuses;

    /**
     * @var string[]
     */
    private $terminateStatuses;

    /**
     * @var array
     */
    private $instanceId;

    /**
     * @param ApiFactory $apiFactory
     * @param string[] $continueStatuses
     * @param string[] $terminateStatuses
     * @throws Exception
     */
    public function __construct(
        $apiFactory,
        array $continueStatuses = array(self::DEFAULT_CONTINUE_ORDER_STATUS),
        array $terminateStatuses = array(self::DEFAULT_TERMINATE_ORDER_STATUS)
    )
    {
        $this->apiFactory = $apiFactory;
        $this->continueStatuses = $continueStatuses;
        $this->terminateStatuses = $terminateStatuses;

        $this->openApi = $apiFactory->getOpenApi(OPEN_API_URI);
        $this->pbaApi = $apiFactory->getPbaApi(Settings::getInstance()->getBillingConfig());
    }

    /**
     * @param string $fileName
     * @throws Exception
     */
    public function migrate($fileName)
    {
        if (!is_readable($fileName)) {
            throw new Exception("Cannot open the file \"$fileName\".");
        }

        $handle = fopen($fileName, "r");
        if ($handle === false) {
            throw new Exception("Cannot open the file \"$fileName\".");
        }

        echo sprintf("Migration has started \"%s\".", $fileName) . PHP_EOL;
        $line = 0;
        while (($data = fgetcsv($handle)) !== false) {
            $line++;
            if ($line == 1) {
                continue;
            }
            try {
                $instanceId = $this->getIntProperty(0, $data);
                $applicationId = $this->getIntProperty(1, $data);
                $subscriptionId = $this->getIntProperty(2, $data, false);
                $accountId = $this->getIntProperty(3, $data);
                $datacenterId = $this->getIntProperty(4, $data);
                $mspGroupId = $this->getIntProperty(5, $data);
                $users = $this->getJsonProperty(6, $data, false);
                $additionalResources = $this->getJsonProperty(7, $data, false);

                $this->instanceApsController = $this->createApsControllerInstanceForApplicationInstance($instanceId, $applicationId);

                if (empty($subscriptionId)) {
                    $planId = $this->getIntProperty(8, $data);
                    $planPeriodId = $this->getIntProperty(9, $data);
                    $subscriptionId = $this->createSubscription($planId, $planPeriodId, $mspGroupId, $accountId, $instanceId, $applicationId);
                } else {
                    $this->subscriptionApsController = $this->createApsControllerInstanceForSubscription($accountId, $subscriptionId, $instanceId, $applicationId);
                }

                echo sprintf("Migration for subscription \"%s\" has started.", $subscriptionId) . PHP_EOL;
                $this->migrateSubscription($subscriptionId, $accountId, $datacenterId, $mspGroupId, $users, $additionalResources);
                echo sprintf("Migration for subscription \"%s\" has finished.", $subscriptionId) . PHP_EOL;
            } catch (Exception $e) {
                echo "Migration for subscription has failed." . PHP_EOL;
                echo sprintf('Error: %s Line: #%s.', $e->getMessage(), $line) . PHP_EOL;
            }
        }
        echo "Migration has finished." . PHP_EOL;
    }

    /**
     * @param $planId
     * @param $planPeriodId
     * @param $mspGroupId
     * @param $accountId
     * @param $instanceId
     * @param $applicationId
     * @return
     * @throws Exception
     */
    private function createSubscription($planId, $planPeriodId, $mspGroupId, $accountId, $instanceId, $applicationId)
    {
        $tenant = $this->fetchApsTenantByGroupId($mspGroupId);
        if (!is_null($tenant)) {
            throw new Exception(sprintf('Group "%s" already bound to subscription "%s".', $mspGroupId, $tenant->subscriptionId));
        }

        echo sprintf('Subscription creating for customer "%s" has started.', $accountId) . PHP_EOL;

        $accountData = $this->openApi->getAccountInfo($accountId);
        $orderData = $this->placeOrderForCustomer($planId, $planPeriodId, $accountId, $accountData->getParentAccountId());
        $orderId = $orderData[1];

        $subscriptionsList = $this->pbaApi->getSubscriptionsListByOrder($orderId);
        if (empty($subscriptionsList[0])) {
            throw new Exception(sprintf('Order "%s" doesn\'t provisioning any subscriptions.', $orderId));
        }
        $subscriptionId = $subscriptionsList[0]->getSubscriptionId();

        $attemptsCount = 10;
        while ($attemptsCount) {
            sleep(1);
            try {
                $this->subscriptionApsController = $this->createApsControllerInstanceForSubscription($accountId, $subscriptionId, $instanceId, $applicationId);
            } catch (Exception $e) {
                echo 'Trying to get the subscription again...' . PHP_EOL;
                $attemptsCount--;
                if ($attemptsCount === 0) {
                    throw $e;
                }
                continue;
            }
            break;
        }
        echo sprintf('Subscription "%s" has been created.', $subscriptionId) . PHP_EOL;
        return $subscriptionId;
    }

    private function placeOrderForCustomer($planId, $planPeriod, $accountId, $parentAccountId)
    {
        $provisioningItemId = 0;
        $parentItemId = -1;
        $provisioningDataSlotCounter = 0;

        $provisioningItemsCounter = 1;

        $contactDataCounter = 2;
        $login = 'user_' . \APS\generatePassword(6, '');

        $payToolCounter = 3;
        $ipAddress = '127.0.0.1';

        $parameters = [
            $provisioningItemsCounter,
            "{$planId}={$planPeriod}={$provisioningItemId}={$parentItemId}",
            $provisioningDataSlotCounter,
            $contactDataCounter,
            "LoginID={$login}",
            "AccountID={$accountId}",
            $payToolCounter,
            'PayToolID=3',
            'PluginID=0',
            "IPAddressID={$ipAddress}"
        ];

        return $this->pbaApi->placeOrderAndAuthorize(
            $parentAccountId,
            $parameters
        );
    }

    /**
     * @param int $subscriptionId
     * @param int $accountId
     * @param array $additionalResources
     * @throws Exception
     */
    private function upgradeSubscription($subscriptionId, $accountId, $additionalResources)
    {
        if (empty($additionalResources)) {
            echo 'There are no resources to upgrade' . PHP_EOL;
            return;
        }

        $orderResources = array();
        $subscriptionResources = $this->pbaApi->getSubscriptionResourcesList($subscriptionId);
        foreach ($additionalResources as $additionalResource) {
            $id = $this->getIntProperty(self::PROPERTY_RESOURCE_ID, $additionalResource);
            $amount = $this->getIntProperty(self::PROPERTY_ADDITIONAL_AMOUNT, $additionalResource);
            if ($amount < 0) {
                throw new Exception(sprintf('Additional amount for resource "%s" cannot be negative.', $id));
            }

            $subscriptionResource = $this->fetchPbaSubscriptionResourceById($subscriptionResources, $id);
            if (is_null($subscriptionResource)) {
                throw new Exception(sprintf('Service plan does not have a resource rate for resource "%s".', $id));
            }
            $additionalAmount = $subscriptionResource->getAdditionalAmount();
            if ($additionalAmount >= $amount) {
                echo sprintf('Subscription already has enough additional amount (%s>=%s) for resource "%s".', $additionalAmount, $amount, $id) . PHP_EOL;
                continue;
            }

            $orderResource = new Service_PbaApi_Api_Request_Struct_ResourceUpgradeOrder();
            $orderResource->setResourceID($id);
            $orderResource->setNewResourceID($id);
            $orderResource->setAmount($amount - $additionalAmount);
            $orderResources[] = $orderResource;
        }

        if (empty($orderResources)) {
            return;
        }

        $fromIp = $this->fetchEndPointHost();
        $this->placeUpgradeOrder($accountId, $subscriptionId, $fromIp, $orderResources);
    }

    /**
     * @param int $accountId
     * @param int $subscriptionId
     * @param string $fromIp
     * @param array $resources
     * @throws Exception
     * @throws Service_PbaApi_Api_Exception
     */
    private function placeUpgradeOrder($accountId, $subscriptionId, $fromIp, array $resources)
    {
        $attemptsCount = 5;
        while (true) {
            try {
                $result = $this->pbaApi->placeResourceUpgradeOrder(
                    $subscriptionId,
                    self::USE_PAY_TOOL,
                    self::USE_DEFAULT_PAY_TOOL,
                    $accountId,
                    $fromIp,
                    $resources
                );
                break;
            } catch (Exception $e) {
                echo 'Trying to upgrade resources again...' . PHP_EOL;
                $attemptsCount--;
                if ($attemptsCount === 0) {
                    throw $e;
                }
            }
            sleep(3);
        }
        $orderId = $result->getNewOrderId();
        echo sprintf("Placed Upgrade Order #%s for subscription %s.", $orderId, $subscriptionId) . PHP_EOL;
        do {
            sleep(3);
            $order = $this->pbaApi->getOrder($orderId);
        } while (!in_array($order->getOrderStatusID(), array_merge($this->continueStatuses, $this->terminateStatuses)));

        if (in_array($order->getOrderStatusID(), $this->terminateStatuses)) {
            throw new Exception(sprintf('Failed to place upgrade resource order for subscription with ID %s.', $subscriptionId));
        }
    }

    /**
     * @param int $subscriptionId
     * @param int $accountId
     * @param int $datacenterId
     * @param int $mspGroupId
     * @param array $users
     * @param array $additionalResources
     * @throws Exception
     * @throws Service_BaseClient_HTTPException
     * @internal param int $accountId
     */
    private function migrateSubscription($subscriptionId, $accountId, $datacenterId, $mspGroupId, $users, $additionalResources)
    {
        $subscription = $this->fetchApsSubscriptionById($subscriptionId);

        $this->checkSubscriptionIsMigrated($subscription, $mspGroupId);

        $this->upgradeSubscription($subscriptionId, $accountId, $additionalResources);

        $subscriptionResources = $subscription->resources();
        $this->checkSubscriptionResources($subscriptionResources);

        $quotas = $this->fetchTenantQuotas($subscriptionResources);
        $global = $this->fetchApsGlobal($subscriptionResources);
        $profile = $this->fetchApsProfile($subscriptionResources);

        $this->checkBackupProfile($profile);
        $datacenter = $this->fetchApsDatacenterById($global, $datacenterId);
        $account = $subscription->account;

        $mspApi = $this->createMspApiInstance($datacenter->url, $datacenter->login, $datacenter->password);
        $groupInfo = $this->fetchMspGroupInfoById($mspApi, $mspGroupId);

        $this->checkTenantQuotas($quotas, $groupInfo->getUsage());

        $tenant = $this->createTenantInAps($global, $account, $subscription, $profile, $datacenter, $mspGroupId, $quotas);
        try {
            if (!empty($users)) {
                echo "Migrating users." . PHP_EOL;
                foreach ($users as $user) {
                    $serviceUserId = $this->getIntProperty(self::PROPERTY_SERVICE_USER_ID, $user);
                    $mspUserLogin = (string)$this->getProperty(self::PROPERTY_MSP_USER_LOGIN, $user);
                    $storageQuota = $this->getIntProperty(self::PROPERTY_STORAGE_QUOTA, $user);
                    if ($storageQuota < -1 || $storageQuota == 0) {
                        throw new Exception(sprintf('Storage quota for user "%s" is invalid. Specify user storage quota to -1 (unlimited) or greater than 0.', $serviceUserId));
                    }
                    $privileged = (bool)$this->getProperty(self::PROPERTY_PRIVILEGED, $user, false);
                    $activate = (bool)$this->getProperty(self::PROPERTY_ACTIVATE, $user, false);

                    $userInfo = $this->fetchMspUserInfoByLogin($mspApi, $datacenter->url, $mspGroupId, $mspUserLogin);
                    $mspUserId = $userInfo->getId();

                    echo sprintf("Migrating user \"%s\".", $serviceUserId) . PHP_EOL;
                    $this->migrateUser($subscription, $tenant, $profile, $account, $serviceUserId, $mspUserId, $storageQuota, $privileged, $activate);
                    echo sprintf("User \"%s\" has been migrated.", $serviceUserId) . PHP_EOL;
                }
                echo "Users have been migrated." . PHP_EOL;
            } else {
                echo "There are no users to migrate." . PHP_EOL;
            }

            $this->completeMigration($tenant);
        } catch (Exception $e) {
            $this->rollbackMigration($tenant);
            throw $e;
        }
    }

    /**
     * @param array $subscriptionResources
     * @throws Exception
     */
    private function checkSubscriptionResources(array $subscriptionResources)
    {
        $this->checkSubscriptionResourceLimitEqual($subscriptionResources, 1, \APS\TypeLibrary::getTypeByServiceId('globals'));
        $this->checkSubscriptionResourceLimitEqual($subscriptionResources, 1, \APS\TypeLibrary::getTypeByServiceId('tenant'));
        $this->checkSubscriptionResourceLimitEqual($subscriptionResources, -1, \APS\TypeLibrary::getTypeByServiceId('users'));
        $this->checkSubscriptionResourceLimitEqual($subscriptionResources, 1, \APS\TypeLibrary::getTypeByServiceId('profile'));
        $this->checkSubscriptionResourceLimitNotEqual($subscriptionResources, 0, \APS\TypeLibrary::getTypeByServiceId('tenant'), self::STORAGE_USAGE);
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $subscription
     * @param \APS\ResourceBase|\APS\ResourceProxy $tenant
     * @param \APS\ResourceBase|\APS\ResourceProxy $profile
     * @param \APS\ResourceBase|\APS\ResourceProxy $account
     * @param int $serviceUserId
     * @param int $mspUserId
     * @param int $storageQuota
     * @param bool $privileged
     * @param bool $activate
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function migrateUser($subscription, $tenant, $profile, $account, $serviceUserId, $mspUserId, $storageQuota, $privileged, $activate)
    {
        $serviceUser = $this->fetchApsServiceUserById($account, $serviceUserId);

        $this->checkUserIsMigrated($subscription, $serviceUser, $mspUserId);

        $quotas = new ResourceQuotas();
        $quotas->setStorageUsage($storageQuota);
        $quotas->setMobileDeviceCount(-1);
        $quotas->setWorkstationCount(-1);
        $quotas->setServerCount(-1);
        $quotas->setVmCount(-1);
        $quotas->setMailboxCount(-1);
        $quotas->setWebsiteCount(-1);

        return $this->createUserInAps($tenant, $profile, $serviceUser, $mspUserId, $quotas, $privileged, $activate);
    }

    /**
     * @param string $url
     * @param string $login
     * @param string $password
     * @return Service_MspApiClient_Api
     */
    private function createMspApiInstance($url, $login, $password)
    {
        $instanceId = $this->instanceApsController->getInstanceId();
        $apiUri = $url . MSP_API_URI;
        $cookieFile = DIR_COOKIES . $instanceId . '_' . md5($url . $login);

        $restClient = new Service_BaseClient_RestClient();
        $restClient->setApiUri($apiUri);
        $restClient->setCookieFile($cookieFile);
        $store = new Session_StoreStub();
        $mspApi = new Service_MspApiClient_Api($restClient, $store);
        $mspApi->setCredentials($login, $password, $url, false);
        $mspApi->login();
        return $mspApi;
    }

    /**
     * @param int $accountId
     * @param int $subscriptionId
     * @param int $instanceId
     * @param int $applicationId
     * @return \APS\ControllerProxy
     * @throws \Exception
     */
    private function createApsControllerInstanceForSubscription($accountId, $subscriptionId, $instanceId, $applicationId)
    {
        $token = $this->openApi->getAccountToken($accountId, $subscriptionId);
        return $this->apiFactory->getApsController(
            $token[self::PROPERTY_CONTROLLER_URI],
            $token[self::PROPERTY_APS_TOKEN],
            $this->getInstanceResourceId($applicationId, $instanceId)
        );
    }

    /**
     * @param int $instanceId
     * @param int $applicationId
     * @return \APS\ControllerProxy
     * @throws \Exception
     */
    private function createApsControllerInstanceForApplicationInstance($instanceId, $applicationId)
    {
        $token = $this->openApi->getApplicationInstanceToken($instanceId);
        return $this->apiFactory->getApsController(
            $token[self::PROPERTY_CONTROLLER_URI],
            $token[self::PROPERTY_APS_TOKEN],
            $this->getInstanceResourceId($applicationId, $instanceId)
        );
    }

    /**
     * @param $applicationId
     * @param $instanceId
     * @return mixed
     * @throws \Exception
     * @throws \Service_OpenApi_Validator_BaseException
     */
    private function getInstanceResourceId($applicationId, $instanceId)
    {
        if (isset($this->instanceId[$instanceId])) {
            return $this->instanceId[$instanceId];
        }
        $instanceGuid = $this->findInstanceGuidById($applicationId, $instanceId);
        if (is_null($instanceGuid)) {
            throw new Exception(sprintf('Cannot find instance "%s" in application "%s".', $instanceId, $applicationId));
        }
        $this->instanceId[$instanceId] = $instanceGuid;
        return $this->instanceId[$instanceId];
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $global
     * @param \APS\ResourceBase|\APS\ResourceProxy $account
     * @param \APS\ResourceBase|\APS\ResourceProxy $subscription
     * @param \APS\ResourceBase|\APS\ResourceProxy $profile
     * @param \APS\ResourceBase|\APS\ResourceProxy $datacenter
     * @param int $mspGroupId
     * @param ResourceQuotas $quotas
     * @return \APS\ResourceBase
     */
    public function createTenantInAps($global, $account, $subscription, $profile, $datacenter, $mspGroupId, $quotas)
    {
        echo sprintf("Creating tenant for subscription \"%s\"...", $subscription->subscriptionId) . PHP_EOL;

        $tenant = APS\TypeLibrary::newResourceByTypeId(\APS\TypeLibrary::getTypeByServiceId('tenant'));
        $tenant->global->set($global->aps->id);
        $tenant->account->set($account->aps->id);
        $tenant->subscription->set($subscription->aps->id);
        $tenant->profile->set($profile->aps->id);

        $tenant->datacenterId = $datacenter->id;
        $tenant->mspGroupId = $mspGroupId;
        $tenant->ismigration = true;

        $counter = \APS\TypeLibrary::newResourceByTypeId('http://aps-standard.org/types/core/resource/1.0#Counter');
        $counter->limit = 0;
        $counter->usage = 0;

        $tenant->StorageUsage = clone $counter;
        $tenant->StorageUsage->limit = $quotas->getStorageUsage();

        $tenant->MobileDeviceCount = clone $counter;
        $tenant->MobileDeviceCount->limit = $quotas->getMobileDeviceCount();

        $tenant->WorkstationCount = clone $counter;
        $tenant->WorkstationCount->limit = $quotas->getWorkstationCount();

        $tenant->ServerCount = clone $counter;
        $tenant->ServerCount->limit = $quotas->getServerCount();

        $tenant->VmCount = clone $counter;
        $tenant->VmCount->limit = $quotas->getVmCount();

        $tenant->MailboxCount = clone $counter;
        $tenant->MailboxCount->limit = $quotas->getMailboxCount();

        $tenant->WebsiteCount = clone $counter;
        $tenant->WebsiteCount->limit = $quotas->getWebsiteCount();

        $result = $this->subscriptionApsController->provisionResource($tenant);
        while ($result->aps->status !== self::APS_STATUS_READY) {
            sleep(3);
            $result = $this->subscriptionApsController->getResource($result->aps->id);
        }

        echo sprintf("Tenant \"%s\" has been created.", $result->aps->id) . PHP_EOL;
        return $result;
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $tenant
     * @param \APS\ResourceBase|\APS\ResourceProxy $profile
     * @param \APS\ResourceBase|\APS\ResourceProxy $serviceUser
     * @param int $mspUserId
     * @param ResourceQuotas $quotas
     * @param bool $privileged
     * @param bool $activate
     * @return \APS\ResourceBase
     */
    public function createUserInAps($tenant, $profile, $serviceUser, $mspUserId, ResourceQuotas $quotas, $privileged = false, $activate = true)
    {
        echo sprintf("Creating user \"%s\"", $serviceUser->displayName) . PHP_EOL;

        $user = APS\TypeLibrary::newResourceByTypeId(\APS\TypeLibrary::getTypeByServiceId('users'));
        $user->tenant->set($tenant->aps->id);
        $user->user->set($serviceUser->aps->id);

        $user->mspUserId = $mspUserId;
        $user->privileged = $privileged ? users::ADMINISTRATOR_PRIVILEGES_YES : users::ADMINISTRATOR_PRIVILEGES_NO;
        $user->activateDuringCreation = $activate;

        $user->notifyBackup = is_null($profile) ? Enum_NotificationsBackup::DISABLE_NOTIFICATIONS : $profile->backupNotification;
        $user->notificationsManagement = is_null($profile) ? false : (bool)$profile->managementNotification;

        $counter = new Counter();
        $counter->limit = 0;
        $counter->usage = 0;

        $user->userStorageUsage = clone $counter;
        $user->userStorageUsage->limit = $quotas->getStorageUsage();

        $user->UserMobileDeviceCount = clone $counter;
        $user->UserMobileDeviceCount->limit = $quotas->getMobileDeviceCount();

        $user->UserWorkstationCount = clone $counter;
        $user->UserWorkstationCount->limit = $quotas->getWorkstationCount();

        $user->userServerCount = clone $counter;
        $user->userServerCount->limit = $quotas->getServerCount();

        $user->UserVmCount = clone $counter;
        $user->UserVmCount->limit = $quotas->getVmCount();

        $user->UserMailboxCount = clone $counter;
        $user->UserMailboxCount->limit = $privileged ? $quotas->getMailboxCount() : 0;

        $user->UserWebsiteCount = clone $counter;
        $user->UserWebsiteCount->limit = $quotas->getWebsiteCount();

        $result = $this->subscriptionApsController->provisionResource($user);

        while ($result->aps->status !== self::APS_STATUS_READY) {
            sleep(3);
            $result = $this->subscriptionApsController->getResource($result->aps->id);
        }

        echo sprintf("User \"%s\" has been created.", $serviceUser->displayName) . PHP_EOL;
        return $result;
    }

    /**
     * Finalization of migration. After this method provisioning of
     * users and tenant goes in usual way (ismigration flag is set to false)
     * @param \APS\ResourceBase|\APS\ResourceProxy $tenant
     */
    private function completeMigration($tenant)
    {
        $tenant->ismigration = false;
        $this->subscriptionApsController->updateResource($tenant);
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $tenant
     */
    private function rollbackMigration($tenant)
    {
        $this->subscriptionApsController->unprovisionResource($tenant);
    }

    /**
     * @param int $subscriptionId
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsSubscriptionById($subscriptionId)
    {
        $resources = $this->subscriptionApsController->getResources(sprintf(
            "implementing(%s),eq(subscriptionId,%s)",
            self::APS_TYPE_SUBSCRIPTION, $subscriptionId
        ));
        if (!count($resources)) {
            throw new Exception(sprintf('Cannot fetch subscription by ID "%s".', $subscriptionId));
        }
        return $resources[0];
    }

    /**
     * @param array $subscriptionResources
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsGlobal(array $subscriptionResources)
    {
        $id = $this->fetchSubscriptionResourceApsId($subscriptionResources,  \APS\TypeLibrary::getTypeByServiceId('globals'));
        if (is_null($id)) {
            throw new Exception(sprintf('Cannot fetch application by ID "%s".', $id));
        }
        return $this->subscriptionApsController->getResource($id);
    }

    /**
     * @param array $subscriptionResources
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsProfile(array $subscriptionResources)
    {
        $id = $this->fetchSubscriptionResourceApsId($subscriptionResources, \APS\TypeLibrary::getTypeByServiceId('profile'));
        if (is_null($id)) {
            throw new Exception(sprintf('Cannot fetch Backup Profile by ID "%s".', $id));
        }
        return $this->subscriptionApsController->getResource($id);
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $global
     * @param int $datacenterId
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsDatacenterById($global, $datacenterId)
    {
        $path = $global->datacenters->getLink()->href;
        $resources = $this->instanceApsController->getResources(sprintf("eq(id,%s)", $datacenterId), $path);
        if (!count($resources)) {
            throw new Exception(sprintf('Cannot fetch datacenter by ID "%s".', $datacenterId));
        }
        return $resources[0];
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $account
     * @param int $serviceUserId
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsServiceUserById($account, $serviceUserId)
    {
        $path = $account->users->getLink()->href;
        $serviceUsers = $this->subscriptionApsController->getResources(
            sprintf("implementing(%s)", self::APS_TYPE_SERVICE_USER), $path
        );
        foreach ($serviceUsers as $serviceUser) {
            if ($serviceUser->userId == $serviceUserId) {
                return $serviceUser;
            }
        }
        throw new Exception(sprintf('Cannot fetch service user by ID "%s".', $serviceUserId));
    }

    /**
     * @param int $mspGroupId
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsTenantByGroupId($mspGroupId)
    {
        $resources = $this->instanceApsController->getResources(sprintf(
            "implementing(%s),eq(mspGroupId,%s)",
            \APS\TypeLibrary::getTypeByServiceId('tenant'), $mspGroupId
        ));
        if (!count($resources)) {
            return null;
        }
        return $resources[0];
    }

    /**
     * @param int $subscriptionId
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsTenantBySubscriptionId($subscriptionId)
    {
        $resources = $this->instanceApsController->getResources(sprintf(
            "implementing(%s),eq(subscriptionId,%s)",
            \APS\TypeLibrary::getTypeByServiceId('tenant'), $subscriptionId
        ));
        if (!count($resources)) {
            return null;
        }
        return $resources[0];
    }

    /**
     * @param int $mspUserId
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsUserById($mspUserId)
    {
        $resources = $this->instanceApsController->getResources(sprintf(
            "implementing(%s),eq(mspUserId,%s)",
            \APS\TypeLibrary::getTypeByServiceId('users'), $mspUserId
        ));
        if (!count($resources)) {
            return null;
        }
        return $resources[0];
    }

    /**
     * @param int $subscriptionId
     * @param string $serviceUserId
     * @return \APS\ResourceBase
     * @throws Exception
     */
    private function fetchApsUserBySubscriptionIdAndServiceUserId($subscriptionId, $serviceUserId)
    {
        $resources = $this->instanceApsController->getResources(sprintf(
            "implementing(%s),eq(subscriptionId,%s),eq(serviceUserId,%s)",
            \APS\TypeLibrary::getTypeByServiceId('users'), $subscriptionId, $serviceUserId
        ));
        if (!count($resources)) {
            return null;
        }
        return $resources[0];
    }

    /**
     * @param array $subscriptionResources
     * @return ResourceQuotas
     * @throws Exception
     */
    private function fetchTenantQuotas(array $subscriptionResources)
    {
        $quotas = new ResourceQuotas();
        $quotas->setStorageUsage($this->fetchSubscriptionResourceCounterLimit($subscriptionResources, self::STORAGE_USAGE));
        $quotas->setMobileDeviceCount($this->fetchSubscriptionResourceCounterLimit($subscriptionResources, self::MOBILE_DEVICE_COUNT));
        $quotas->setWorkstationCount($this->fetchSubscriptionResourceCounterLimit($subscriptionResources, self::WORKSTATION_COUNT));
        $quotas->setServerCount($this->fetchSubscriptionResourceCounterLimit($subscriptionResources, self::SERVER_COUNT));
        $quotas->setVmCount($this->fetchSubscriptionResourceCounterLimit($subscriptionResources, self::VM_COUNT));
        $quotas->setMailboxCount($this->fetchSubscriptionResourceCounterLimit($subscriptionResources, self::MAILBOX_COUNT));
        $quotas->setWebsiteCount($this->fetchSubscriptionResourceCounterLimit($subscriptionResources, self::WEBSITE_COUNT));
        return $quotas;
    }

    /**
     * @param array $subscriptionResources
     * @param int $limit
     * @param string $apsType
     * @param string $apsProperty
     * @throws Exception
     */
    private function checkSubscriptionResourceLimitEqual(array $subscriptionResources, $limit, $apsType, $apsProperty = null)
    {
        $resources = $this->filterSubscriptionResources($subscriptionResources, $apsType, $apsProperty);
        $count = count($resources);
        if (!$count) {
            throw new Exception(sprintf('Service template does not have resource type for APS type "%s#%s".', $apsType, $apsProperty));
        }
        if ($count > 1) {
            throw new Exception(sprintf('Service template has too many resource types for APS type "%s#%s".', $apsType, $apsProperty));
        }
        $resource = reset($resources);
        $resourceLimit = isset($resource->limit) ? $resource->limit : -1;
        if ($resourceLimit != $limit) {
            throw new Exception("Limit for resource type is wrong");
        }
    }

    /**
     * @param array $subscriptionResources
     * @param int $limit
     * @param string $apsType
     * @param string $apsProperty
     * @throws Exception
     */
    private function checkSubscriptionResourceLimitNotEqual(array $subscriptionResources, $limit, $apsType, $apsProperty = null)
    {
        $resources = $this->filterSubscriptionResources($subscriptionResources, $apsType, $apsProperty);
        $count = count($resources);
        if (!$count) {
            throw new Exception(sprintf('Service template does not have resource type for APS type "%s#%s".', $apsType, $apsProperty));
        }
        if ($count > 1) {
            throw new Exception(sprintf('Service template has too many resource types for APS type "%s#%s".', $apsType, $apsProperty));
        }
        $resource = reset($resources);
        $resourceLimit = isset($resource->limit) ? $resource->limit : -1;
        if ($resourceLimit == $limit) {
            throw new Exception("Limit for resource type is wrong");
        }
    }

    /**
     * @param array $subscriptionResources
     * @param string $property
     * @return int
     * @throws Exception
     */
    private function fetchSubscriptionResourceCounterLimit(array $subscriptionResources, $property)
    {
        $resources = $this->filterSubscriptionResources($subscriptionResources, \APS\TypeLibrary::getTypeByServiceId('tenant'), $property);
        $count = count($resources);
        if (!$count) {
            return 0;
        }
        if ($count > 1) {
            throw new Exception("Invalid resource count");
        }
        $resource = reset($resources);
        return isset($resource->limit) ? $resource->limit : -1;
    }

    /**
     * @param array $subscriptionResources
     * @param string $apsType
     * @return string
     * @throws Exception
     */
    private function fetchSubscriptionResourceApsId(array $subscriptionResources, $apsType)
    {
        $resources = $this->filterSubscriptionResources($subscriptionResources, $apsType);
        $count = count($resources);
        if (!$count) {
            return null;
        }
        if ($count > 1) {
            throw new Exception("Invalid resource count");
        }
        $resource = reset($resources);
        if (!isset($resource->apsId)) {
            throw new Exception('Resource does not have property "apsId".');
        }
        return $resource->apsId;
    }

    /**
     * @param array $subscriptionResources
     * @param string $apsType
     * @param string $apsProperty
     * @return array
     */
    private function filterSubscriptionResources(array $subscriptionResources, $apsType, $apsProperty = null)
    {
        return array_filter($subscriptionResources, function ($resource) use ($apsType, $apsProperty) {
            $typeIsMatched = isset($resource->apsType) && $resource->apsType === $apsType;
            $propertyIsMatched = is_null($apsProperty)
                ? !isset($resource->property)
                : isset($resource->property) && $resource->property === $apsProperty;

            return $typeIsMatched && $propertyIsMatched;
        });
    }

    /**
     * @return string
     * @throws Exception
     */
    private function fetchEndPointHost()
    {
        $ehif = new EndpointHostIpFetcher($this->instanceApsController);
        return $ehif->fetchApsEndpointIpAddressOrFallback();
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $profile
     * @throws Exception
     */
    private function checkBackupProfile($profile)
    {
        if ($profile->userQuota != profile::USER_QUOTA_SOFT) {
            throw new Exception("In the backup profile settings, 'Manage user quotas' must be set to 'Yes' and 'Do not allow to set user quotas exceeding subscription limits' must be set to 'No'.");
        }
    }

    /**
     * @param string $column
     * @param array|object $object
     * @param bool $throwException
     * @return int|null
     * @throws Exception
     */
    private function getIntProperty($column, $object, $throwException = true)
    {
        $value = $this->getProperty($column, $object, $throwException);
        if (trim($value) === '') {
            if ($throwException) {
                throw new Exception(sprintf('Value from column #%s is empty.', $column));
            }
            return null;
        }
        if (!is_numeric($value)) {
            throw new Exception(sprintf('Cannot convert value from column #%s to integer.', $column));
        }
        return intval($value);
    }

    /**
     * @param string $column
     * @param array|object $object
     * @param bool $throwException
     * @return mixed
     * @throws Exception
     */
    private function getJsonProperty($column, $object, $throwException = true)
    {
        $value = $this->getProperty($column, $object, $throwException);
        if (trim($value) === '') {
            if ($throwException) {
                throw new Exception(sprintf('Value from column #%s is empty.', $column));
            }
            return null;
        }
        $jsonObject = json_decode($value);
        if (is_null($jsonObject)) {
            throw new Exception(sprintf('Cannot parse JSON from column #%s.', $column));
        }
        return $jsonObject;
    }

    /**
     * @param string $column
     * @param array|object $object
     * @param bool $throwException
     * @return int|null
     * @throws Exception
     */
    private function getProperty($column, $object, $throwException = true)
    {
        $data = (array)$object;
        if (array_key_exists($column, $data)) {
            return $data[$column];
        }
        if ($throwException) {
            throw new Exception(sprintf('Cannot get value from column #%s.', $column));
        }
        return null;
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $subscription
     * @param int $mspGroupId
     * @throws Exception
     */
    private function checkSubscriptionIsMigrated($subscription, $mspGroupId)
    {
        $tenant = $this->fetchApsTenantBySubscriptionId($subscription->subscriptionId);
        if (!is_null($tenant)) {
            if ($tenant->mspGroupId == $mspGroupId) {
                throw new Exception(sprintf('Subscription "%s" already migrated.', $subscription->subscriptionId));
            }
            throw new Exception(sprintf('Subscription "%s" already bound to group "%s".', $subscription->subscriptionId, $tenant->mspGroupId));
        }

        $tenant = $this->fetchApsTenantByGroupId($mspGroupId);
        if (!is_null($tenant)) {
            if ($tenant->subscriptionId == $subscription->subscriptionId) {
                throw new Exception(sprintf('Subscription "%s" already migrated.', $subscription->subscriptionId));
            }
            throw new Exception(sprintf('Group "%s" already bound to subscription "%s".', $mspGroupId, $tenant->subscriptionId));
        }
    }

    /**
     * @param \APS\ResourceBase|\APS\ResourceProxy $subscription
     * @param \APS\ResourceBase|\APS\ResourceProxy $serviceUser
     * @param int $mspUserId
     * @throws Exception
     */
    private function checkUserIsMigrated($subscription, $serviceUser, $mspUserId)
    {
        $backupUser = $this->fetchApsUserBySubscriptionIdAndServiceUserId($subscription->subscriptionId, $serviceUser->aps->id);
        if (!is_null($backupUser)) {
            if ($backupUser->mspUserId == $mspUserId) {
                throw new Exception(sprintf('User group "%s" already migrated.', $mspUserId));
            }
            throw new Exception(sprintf('Service user "%s" already bound to user group "%s".', $serviceUser->login, $backupUser->mspUserId));
        }

        $backupUser = $this->fetchApsUserById($mspUserId);
        if (!is_null($backupUser)) {
            if ($backupUser->serviceUserId == $serviceUser->aps->id) {
                throw new Exception(sprintf('User group "%s" already migrated.', $mspUserId));
            }
            throw new Exception(sprintf('User group "%s" already bound to service user "$s".', $mspUserId, $backupUser->user->login));
        }
    }

    /**
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param int $mspGroupId
     * @return Service_MspApiClient_Response_Struct_GroupInfo
     * @throws Exception
     */
    private function fetchMspGroupInfoById(Service_MspApiClient_ApiInterface $mspApi, $mspGroupId)
    {
        try {
            $groupInfo = $mspApi->getGroupById($mspGroupId);
            return $groupInfo;
        } catch (Service_BaseClient_HTTPException $e) {
            if ($e->getCode() == 404) {
                throw new Exception(sprintf('Group "%s" does not exist.', $mspGroupId));
            }
            throw $e;
        }
    }

    /**
     * @param Service_MspApiClient_ApiInterface $mspApi
     * @param string $url
     * @param int $mspGroupId
     * @param string $mspUserLogin
     * @return Service_MspApiClient_Response_Struct_UserItem
     * @throws Exception
     */
    private function fetchMspUserInfoByLogin(Service_MspApiClient_ApiInterface $mspApi, $url, $mspGroupId, $mspUserLogin)
    {
        $users = $mspApi->getUsers($mspGroupId);
        foreach ($users as $user) {
            if ($user->getLogin() == $mspUserLogin) {
                return $user;
            }
        }
        throw new Exception(sprintf('Cannot find user by login "%s" in group "%s".', $mspUserLogin, $mspGroupId));
    }

    /**
     * @param ResourceQuotas $quotas
     * @param Service_MspApiClient_Response_Struct_UserUsage $usages
     * @throws Exception
     */
    private function checkTenantQuotas(ResourceQuotas $quotas, Service_MspApiClient_Response_Struct_UserUsage $usages)
    {
        $quota = $quotas->getStorageUsage();
        $usage = $usages->getStorageSize();
        if ($quota != -1 && $quota < $usage) {
            throw new Exception(sprintf(self::EX_MSG_WARNING_QUOTA_IS_LESS_THAN_USAGE, 'Storage', $quota, $usage));
        }
        $quota = $quotas->getMobileDeviceCount();
        $usage = $usages->getMobileCount();
        if ($quota != -1 && $quota < $usage) {
            throw new Exception(sprintf(self::EX_MSG_WARNING_QUOTA_IS_LESS_THAN_USAGE, 'Mobile Devices', $quota,
                $usage));
        }
        $quota = $quotas->getWorkstationCount();
        $usage = $usages->getWorkstationCount();
        if ($quota != -1 && $quota < $usage) {
            throw new Exception(sprintf(self::EX_MSG_WARNING_QUOTA_IS_LESS_THAN_USAGE, 'Workstations', $quota,
                $usage));
        }
        $quota = $quotas->getServerCount();
        $usage = $usages->getServerCount();
        if ($quota != -1 && $quota < $usage) {
            throw new Exception(sprintf(self::EX_MSG_WARNING_QUOTA_IS_LESS_THAN_USAGE, 'Servers', $quota, $usage));
        }
        $quota = $quotas->getVmCount();
        $usage = $usages->getVmCount();
        if ($quota != -1 && $quota < $usage) {
            throw new Exception(sprintf(self::EX_MSG_WARNING_QUOTA_IS_LESS_THAN_USAGE, 'Virtual Machines', $quota, $usage));
        }
        $quota = $quotas->getMailboxCount();
        $usage = $usages->getMailboxCount();
        if ($quota != -1 && $quota < $usage) {
            throw new Exception(sprintf(self::EX_MSG_WARNING_QUOTA_IS_LESS_THAN_USAGE, 'Mailboxes', $quota, $usage));
        }
        $quota = $quotas->getWebsiteCount();
        $usage = $usages->getWebsiteCount();
        if ($quota != -1 && $quota < $usage) {
            throw new Exception(sprintf(self::EX_MSG_WARNING_QUOTA_IS_LESS_THAN_USAGE, 'Websites', $quota,
                $usage));
        }
    }

    /**
     * @param Service_PbaApi_Api_Response_Struct_SubscriptionResourcesList[] $subscriptionResources
     * @param int $resourceId
     * @return Service_PbaApi_Api_Response_Struct_SubscriptionResourcesList|null
     */
    private function fetchPbaSubscriptionResourceById(array $subscriptionResources, $resourceId)
    {
        foreach ($subscriptionResources as $subscriptionResource) {
            if ($subscriptionResource->getResourceId() == $resourceId) {
                return $subscriptionResource;
            }
        }
        return null;
    }

    /**
     * @param $applicationId
     * @param $instanceId
     * @return null
     * @throws \Exception
     */
    private function findInstanceGuidById($applicationId, $instanceId)
    {
        $instances = $this->openApi->getApplicationInstances($applicationId);

        foreach (\APS\ControllerProxy::listInstances() as $instanceGuid) {
            $applicationInfo = \APS\TypeLibrary::retrieveApplicationServices(false, $instanceGuid);
            $globalsId = $applicationInfo->globals->aps->id;

            foreach ($instances as $instance) {
                if ($instance['application_instance_id'] == $instanceId && $instance['application_resource_id'] == $globalsId) {
                    return $instanceGuid;
                }
            }
        }
        return null;
    }
}
