<?php
/**
 * @Copyright © 2002-2017 Acronis International GmbH. All rights reserved
 */

class BillingReportHelper
{
    const SHEET_TO_BILL = 0;
    const SHEET_CLOUD_RESELLER_OVERALL = 1;
    const SHEET_CLOUD_RESELLER_BREAKDOWN = 2;
    const SHEET_SERVICE_PROVIDER_OVERALL = 3;
    const SHEET_SERVICE_PROVIDER_BREAKDOWN = 4;

    const PROPERTY_CONTROLLER_URI = "controller_uri";
    const PROPERTY_APS_TOKEN = "aps_token";

    const APS_RESOURCES_PATH = '/aps/2/resources';
    const APS_TOKEN_HEADER = 'APS-Token: ';

    const PROVIDER_ACCOUNT_ID = 1;
    const PROVIDER_SUBSCRIPTION_ID = 1;

    const CLOUD_RESELLER = "Cloud Reseller";
    const SERVICE_PROVIDER = 'Service Provider';

    const AAN_ACRONIS = 'Acronis';
    const AAN_TRIAL = '00000000';

    const REPORT_CONFIG = 'report.ini';

    const LEVEL_NO_COMMITMENT = -1;
    const LEVEL_FIXED_COMMITMENT = -2;

    const BILL_MODE_ALL = 'ALL';
    const BILL_MODE_LAST = 'LAST';
    const COMMON_BOOK = 'common';

    private static $sheetIndexes = array(
        self::SHEET_TO_BILL,
        self::SHEET_CLOUD_RESELLER_OVERALL,
        self::SHEET_CLOUD_RESELLER_BREAKDOWN,
        self::SHEET_SERVICE_PROVIDER_OVERALL,
        self::SHEET_SERVICE_PROVIDER_BREAKDOWN
    );

    private static $sqlToFetchUsage;

    private static $headStyle = array(
        'borders' => array(
            'allborders' => array(
                'style' => \PHPExcel_Style_Border::BORDER_THIN,
                'color' => array(
                    'rgb' => '8DB4E2'
                )
            )
        ),
        'font' => array(
            'name' => 'Calibri',
            'bold' => true,
            'color' => array(
                'rgb' => 'FFFFFF'
            )
        ),
        'fill' => array(
            'type' => \PHPExcel_Style_Fill::FILL_SOLID,
            'color' => array(
                'rgb' => '4F81BD'
            )
        )
    );

    private static $bodyStyle = array(
        'borders' => array(
            'allborders' => array(
                'style' => \PHPExcel_Style_Border::BORDER_THIN,
                'color' => array(
                    'rgb' => '8DB4E2'
                )
            )
        ),
        'fill' => array(
            'type' => \PHPExcel_Style_Fill::FILL_SOLID,
            'color' => array(
                'rgb' => 'DCE6F1'
            )
        ),
        'font' => array(
            'name' => 'Calibri'
        ),
    );
    const FORMAT_NUMBER = '#,##0';

    const SHEET_NAME_TO_BILL = 'To Bill';
    const SHEET_NAME_OVERALL_CLOUD_RESELLER = 'Overall Cloud Reseller';
    const SHEET_NAME_CLOUD_RESELLER_SKU_BREAKDOWN = 'Cloud Reseller SKU Breakdown';
    const SHEET_NAME_OVERALL_SERVICE_PROVIDER = 'Overall Service Provider';
    const SHEET_NAME_SERVICE_PROVIDER_SKU_BREAKDOWN = 'Service Provider SKU Breakdown';

    /**
     * @var PHPExcel[]
     */
    private $books = array();

    /**
     * @var array
     */
    private $bookRowPositions = array();

    /**
     * @var array
     */
    private $sheetColumns = array();

    /**
     * @var \PDO
     */
    private $pdo;

    /**
     * @var Service_OpenApi_Api
     */
    private $publicApi;

    private $config;
    private $apsToken;
    private $tiers;
    private $datacenters;
    private $startPeriodDate;
    private $endPeriodDate;

    private $distributorDiscountStmt;

    public function __construct()
    {
        $this->config = parse_ini_file(REPORT_DIR_CONFIG . self::REPORT_CONFIG, true);

        $billMode = isset($this->config['reports']['bill_mode']) ?
            strtoupper($this->config['reports']['bill_mode']) :
            self::BILL_MODE_LAST;

        if (!in_array($billMode, array(
            self::BILL_MODE_LAST,
            self::BILL_MODE_ALL,
        ))
        ) {
            throw new \Exception('Unsupported bill mode.');
        }

        $this->billMode = $billMode;

        $publicApiSettings = $this->getPublicApiConnectionSettingsFromConfig();
        $this->publicApi = new Service_OpenApi_Api(new Service_OpenApi_Client($publicApiSettings['host']));

        $this->startPeriodDate = new \DateTime('first day of previous month');
        $this->startPeriodDate->setTime(0, 0, 0);

        $this->endPeriodDate = new \DateTime('last day of previous month');
        $this->endPeriodDate->setTime(23, 59, 59);
    }

    public function generate()
    {
        \PHPExcel_Shared_Font::setTrueTypeFontPath(REPORT_DIR_CONFIG);
        \PHPExcel_Shared_Font::setAutoSizeMethod(\PHPExcel_Shared_Font::AUTOSIZE_METHOD_EXACT);

        Log::getInstance()->notice(
            'Generate a report for the period from %s to %s.',
            $this->startPeriodDate->format('c'),
            $this->endPeriodDate->format('c')
        );

        $this->apsToken = $this->publicApi->getAccountToken(self::PROVIDER_ACCOUNT_ID, self::PROVIDER_SUBSCRIPTION_ID);
        $this->tiers = $this->getMinimalCommitmentTiers();
        if (empty($this->tiers)) {
            Log::getInstance()->warning('No tier resources are fetched.');
        }
        $this->datacenters = $this->getDatacenters();
        if (empty($this->datacenters)) {
            Log::getInstance()->warning('No datacenters are fetched.');
        }

        $this->initColumns();

        $this->openDbConnection();
        $this->createResourcesTable();
        $this->createBreakdownView();

        $this->dumpDataOnCloudResellerOverallSheet();
        $this->dumpDataOnCloudResellerBreakdownSheet();
        $this->dumpDataOnServiceProviderOverallSheet();
        $this->dumpDataOnServiceProviderBreakdownSheet();

        $files = $this->saveBooks();
        $this->sendReport($files);
    }

    /**
     * @return PHPExcel
     */
    private function createBook()
    {
        $book = new PHPExcel();

        $book->getProperties()
            ->setCreator("Acronis (c)")
            ->setLastModifiedBy("Acronis (c)");

        $book->createSheet(self::SHEET_CLOUD_RESELLER_OVERALL);
        $book->createSheet(self::SHEET_CLOUD_RESELLER_BREAKDOWN);
        $book->createSheet(self::SHEET_SERVICE_PROVIDER_OVERALL);
        $book->createSheet(self::SHEET_SERVICE_PROVIDER_BREAKDOWN);

        foreach (self::$sheetIndexes as $index) {
            $table = $this->sheetColumns[$index];
            $sheet = $book->getSheet($index);
            $sheet->setTitle($table['title']);
            $this->fillRow($sheet, 1, $table['columns']);

            $columnsCount = $table['count'];
            for($i = 0; $i < $columnsCount; $i++) {
                $sheet->getColumnDimensionByColumn($i)->setAutoSize(true);
            }

            $sheet->getStyleByColumnAndRow(0, 1, $columnsCount - 1, 1)
                ->applyFromArray(self::$headStyle);
        }
        return $book;
    }

    private function initColumns()
    {
        $this->sheetColumns[self::SHEET_TO_BILL] = $this->initToBillColumns();
        $this->sheetColumns[self::SHEET_CLOUD_RESELLER_OVERALL] = $this->initCloudResellerOverallColumns();
        $this->sheetColumns[self::SHEET_CLOUD_RESELLER_BREAKDOWN] = $this->initCloudResellerBreakdownColumns();
        $this->sheetColumns[self::SHEET_SERVICE_PROVIDER_OVERALL] = $this->initServiceProviderOverallColumns();
        $this->sheetColumns[self::SHEET_SERVICE_PROVIDER_BREAKDOWN] = $this->initServiceProviderBreakdownColumns();
    }

    private function initToBillColumns()
    {
        $columns = array(
            'Company Name (Child)',
            'Acronis (Parent)',
            'Acronis Account Number',
            'Contact E-mail',
            'Type',
            'Discount Tier',
            'Distributor Discount',
            'Currency',
            'Overall Cost before Discount',
            'Overall after Discount',
        );

        $currencyFormatBuilder = function ($values) {
            $currencyColumn = 7; // Index of currency in $columns
            $currency = $values[$currencyColumn];
            return BillingReportHelper::buildCurrencyFormat($currency, false);
        };

        $currencyWithSymbolFormatBuilder = function ($values) {
            $currencyColumn = 7; // Index of currency in $columns
            $currency = $values[$currencyColumn];
            return BillingReportHelper::buildCurrencyFormat($currency);
        };

        $format = array(
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       //Company Name (Child)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       //Acronis (Parent)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       //Acronis Account Number
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       //Contact E-mail
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       //Type
            \PHPExcel_Style_NumberFormat::FORMAT_PERCENTAGE, //Discount Tier
            $currencyFormatBuilder,                          //Distributor Discount
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       //Currency
            $currencyWithSymbolFormatBuilder,                //Overall Cost before Discount
            $currencyWithSymbolFormatBuilder,                //Overall after Discount
        );

        return array(
            'title' => self::SHEET_NAME_TO_BILL,
            'columns' => $columns,
            'format' => $format,
            'count' => count($columns)
        );
    }

    private function initCloudResellerOverallColumns()
    {
        $columns = array(
            'Cloud Reseller (Child)',
            'Acronis Legal Entity (Parent)',
            'Acronis Account Number',
            'Contact E-mail',
            'Base Discount',
            'Tier 1 Support Discount',
            'Distributor Discount',
            'Currency',
            'Overall Cloud Reseller Cost before Discount',
            'Overall Cloud Reseller Cost after Discount',
        );

        $currencyFormatBuilder = function ($values) {
            $currencyColumn = 7; // Index of currency in $columns
            $currency = $values[$currencyColumn];
            return BillingReportHelper::buildCurrencyFormat($currency, false);
        };

        $currencyWithSymbolFormatBuilder = function ($values) {
            $currencyColumn = 7; // Index of currency in $columns
            $currency = $values[$currencyColumn];
            return BillingReportHelper::buildCurrencyFormat($currency);
        };

        $format = array(
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Cloud Reseller (Child)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Acronis Legal Entity (Parent)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Acronis Account Number
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Contact E-mail
            \PHPExcel_Style_NumberFormat::FORMAT_PERCENTAGE, // Base Discount
            \PHPExcel_Style_NumberFormat::FORMAT_PERCENTAGE, // Tier 1 Support Discount
            $currencyFormatBuilder,                          // Distributor Discount
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Currency
            $currencyWithSymbolFormatBuilder,                // Overall Cloud Reseller Cost before Discount
            $currencyWithSymbolFormatBuilder,                // Overall Cloud Reseller Cost after Discount
        );
        return array(
            'title' => self::SHEET_NAME_OVERALL_CLOUD_RESELLER,
            'columns' => $columns,
            'format' => $format,
            'count' => count($columns)
        );
    }

    private function initCloudResellerBreakdownColumns()
    {
        $columns = array(
            'Cloud Reseller (Child)',
            'Acronis Legal Entity (Parent)',
            'Acronis Account Number',
            'Contact E-mail',
            'Product',
            'SKU',
            'Currency',
            'Overall Cost per SKU',
        );

        $currencyFormatBuilder = function ($values) {
            $currencyColumn = 6; // Index of currency in $columns
            $currency = $values[$currencyColumn];
            return BillingReportHelper::buildCurrencyFormat($currency, false);
        };

        $format = array(
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT, // Cloud Reseller (Child)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT, // Acronis Legal Entity (Parent)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT, // Acronis Account Number
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT, // Contact E-mail
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT, // Product
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT, // SKU
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT, // Currency
            $currencyFormatBuilder,                    // Overall Cost per SKU
        );
        return array(
            'title' => self::SHEET_NAME_CLOUD_RESELLER_SKU_BREAKDOWN,
            'columns' => $columns,
            'format' => $format,
            'count' => count($columns)
        );
    }

    private function initServiceProviderOverallColumns()
    {
        $columns = array(
            'Service Provider (Child)',
            'Service Provider AAN',
            'Cloud Reseller (Parent)',
            'Cloud Reseller AAN',
            'OSA Account #',
            'Contact E-mail',
            'Licensing Type',
            'Country',
            'Region',
            'Datacenter',
            'Product',
            'SKU',
            'Currency',
            'Status',
            'Minimum Commitment',
            'Minimum Commitment start month',
            'Start',
            'End',
        );

        $currencyFormatBuilder = function ($values) {
            $currencyColumn = 12; // Index of currency in $columns
            $currency = $values[$currencyColumn];
            return BillingReportHelper::buildCurrencyFormat($currency, false);
        };

        $format = array(
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Service Provider (Child)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Service Provider AAN
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Cloud Reseller (Parent)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Cloud Reseller AAN
            \PHPExcel_Style_NumberFormat::FORMAT_GENERAL, //OSA Account #
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Contact E-mail
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Licensing Type
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Country
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Region
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Datacenter
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Product
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //SKU
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Currency
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Status
            self::FORMAT_NUMBER,                          //Minimum Commitment
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Minimum Commitment start month
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //Start
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,    //End
        );

        $billableResources = $this->getBillableResourcesFromConfig();
        $columns = array_merge(
            $columns,
            array_values($billableResources)
        );
        $format = array_merge(
            $format,
            array_fill(1, count($billableResources), \PHPExcel_Style_NumberFormat::FORMAT_GENERAL)
        );

        $columns[] = 'Discount Tier';
        $columns[] = 'Overall Service Provider Cost';

        $format[] = \PHPExcel_Style_NumberFormat::FORMAT_PERCENTAGE;
        $format[] = $currencyFormatBuilder;

        return array(
            'title' => self::SHEET_NAME_OVERALL_SERVICE_PROVIDER,
            'columns' => $columns,
            'format' => $format,
            'count' => count($columns)
        );
    }

    private function initServiceProviderBreakdownColumns()
    {
        $columns = array(
            'Service Provider (Child)',
            'Service Provider AAN',
            'Cloud Reseller (Parent)',
            'Cloud Reseller AAN',
            'OSA Account #',
            'OSA Subscription #',
            'Service Plan Name',
            'Contact E-mail',
            'Licensing Type',
            'Country',
            'Region',
            'Datacenter',
            'Product',
            'SKU',
            'Currency',
            'Status',
            'Minimum Commitment',
            'Minimum Commitment start month',
            'Start',
            'End',
            'Quantity',
            'WS40 Discount',
            'Discount Tier',
            'Cost per Item before Discount',
            'Cost per Item after Discount',
            'Overall Cost per SKU',
            'Close Date',
        );

        $currencyFormatBuilder = function ($values) {
            $currencyColumn = 14; // Index of currency in $columns
            $currency = $values[$currencyColumn];
            return BillingReportHelper::buildCurrencyFormat($currency, false);
        };

        $format = array(
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Service Provider (Child)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Service Provider AAN
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Cloud Reseller (Parent)
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Cloud Reseller AAN
            \PHPExcel_Style_NumberFormat::FORMAT_GENERAL,    // OSA Account #
            \PHPExcel_Style_NumberFormat::FORMAT_GENERAL,    // OSA Subscription #
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Service Plan Name
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Contact E-mail
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Licensing Type
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Country
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Region
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Datacenter
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Product
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // SKU
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Currency
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Status
            self::FORMAT_NUMBER,                             // Minimum Commitment
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Minimum Commitment start month
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Start
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // End
            \PHPExcel_Style_NumberFormat::FORMAT_GENERAL,    // Quantity
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // WS40 Discount
            \PHPExcel_Style_NumberFormat::FORMAT_PERCENTAGE, // Discount Tier
            \PHPExcel_Style_NumberFormat::FORMAT_GENERAL,    // Cost per Item before Discount
            \PHPExcel_Style_NumberFormat::FORMAT_GENERAL,    // Cost per Item after Discount
            $currencyFormatBuilder,                          // Overall Cost per SKU
            \PHPExcel_Style_NumberFormat::FORMAT_TEXT,       // Close
        );

        if ($this->billMode == self::BILL_MODE_ALL) {
            $columns[] = 'Duration';
            $columns[] = 'Detail start';
            $columns[] = 'Detail end';

            $format[] = \PHPExcel_Style_NumberFormat::FORMAT_NUMBER_00;
            $format[] = \PHPExcel_Style_NumberFormat::FORMAT_TEXT;
            $format[] = \PHPExcel_Style_NumberFormat::FORMAT_TEXT;
        }
        return array(
            'title' => self::SHEET_NAME_SERVICE_PROVIDER_SKU_BREAKDOWN,
            'columns' => $columns,
            'format' => $format,
            'count' => count($columns)
        );
    }

    private function dumpDataOnCloudResellerOverallSheet()
    {
        Log::getInstance()->notice('Execute query for "%s" sheet.', self::SHEET_NAME_OVERALL_CLOUD_RESELLER);
        $stmt = $this->executeSqlQuery($this->readSQLFile('report_cr_overall.sql'));

        Log::getInstance()->notice('Dump data on "%s" sheet.', self::SHEET_NAME_OVERALL_CLOUD_RESELLER);
        while ($row = $stmt->fetch(PDO::FETCH_OBJ)) {

            $distributorDiscount = $this->countDistributorDiscount($row->CRID);
            $overallDiscount = max($row->OverallDiscount - $distributorDiscount, 0.0);

            $reportGroup = $row->CRAAN !== self::AAN_ACRONIS ?  $row->ReportGroup : null;

            $this->dumpData(self::SHEET_CLOUD_RESELLER_OVERALL, array(
                $row->CRName,
                $row->CRVendor,
                $row->CRAAN,
                $row->CREmail,
                $row->CRBDisc,
                $row->CRT1Disc,
                $distributorDiscount,
                $row->CRCurrency,
                $row->Overall,
                $overallDiscount,
            ), $reportGroup);

            if ($row->Overall > 0 && !in_array($row->CRAAN, array(self::AAN_ACRONIS, self::AAN_TRIAL))) {
                $this->dumpData(self::SHEET_TO_BILL, array(
                    $row->CRName,
                    $row->CRVendor,
                    $row->CRAAN,
                    $row->CREmail,
                    self::CLOUD_RESELLER,
                    $row->CRBDisc + $row->CRT1Disc,
                    $distributorDiscount,
                    $row->CRCurrency,
                    $row->Overall,
                    $overallDiscount
                ));
            }
        }
    }

    private function dumpDataOnCloudResellerBreakdownSheet()
    {
        Log::getInstance()->notice('Execute query for "%s" sheet.', self::SHEET_NAME_CLOUD_RESELLER_SKU_BREAKDOWN);
        $stmt = $this->executeSqlQuery($this->readSQLFile('report_cr_breakdown.sql'));

        Log::getInstance()->notice('Dump data on "%s" sheet.', self::SHEET_NAME_CLOUD_RESELLER_SKU_BREAKDOWN);
        while ($row = $stmt->fetch(PDO::FETCH_OBJ)) {
            $reportGroup = $row->CRAAN !== self::AAN_ACRONIS ?  $row->ReportGroup : null;

            $this->dumpData(self::SHEET_CLOUD_RESELLER_BREAKDOWN, array(
                $row->CRName,
                $row->CRVendor,
                $row->CRAAN,
                $row->CREmail,
                $row->ResourceName,
                $row->ResourceSKU,
                $row->SPCurrency,
                $row->Overall,
            ), $reportGroup);
        }
    }

    private function dumpDataOnServiceProviderOverallSheet()
    {
        Log::getInstance()->notice('Execute query for "%s" sheet.', self::SHEET_NAME_OVERALL_SERVICE_PROVIDER);
        $stmt = $this->executeSqlQuery($this->readSQLFile('report_sp_overall.sql'));

        $billableResources = $this->getBillableResourcesFromConfig();

        Log::getInstance()->notice('Dump data on "%s" sheet.', self::SHEET_NAME_OVERALL_SERVICE_PROVIDER);
        while ($row = $stmt->fetch(PDO::FETCH_OBJ)) {
            $datacenterName = $this->getDatacenterName($row->SPDatacenter);
            $minimalCommitment = $this->getMinimalCommitmentValue($row->MinComTier, $row->MinCom, $row->SPCurrency);
            $subscriptions = explode(',', $row->subscriptions);
            $usage = $this->fetchUsageForSubscriptions($subscriptions, $billableResources);

            $values = array(
                $row->SPName,
                $row->SPAAN,
                $row->CRName,
                $row->CRAAN,
                $row->SPID,
                $row->SPEmail,
                $row->LType,
                $row->SPCountry,
                $row->Region,
                $datacenterName,
                $row->Product,
                $row->SKU,
                $row->SPCurrency,
                $row->StatusName,
                $minimalCommitment,
                $row->MinComMonth,
                $row->Start,
                $row->End,
            );

            foreach ($billableResources as $id => $name) {
                $values[] = isset($usage[$id]) ? $usage[$id] : 0;
            }

            $values[] = $row->Discount;
            $values[] = $row->Overall;

            $reportGroup = $row->CRAAN !== self::AAN_ACRONIS ?  $row->ReportGroup : null;

            $this->dumpData(self::SHEET_SERVICE_PROVIDER_OVERALL, $values, $reportGroup);

            if ($row->Overall > 0 && in_array($row->CRAAN, array(self::AAN_ACRONIS, self::AAN_TRIAL))) {
                $this->dumpData(self::SHEET_TO_BILL, array(
                    $row->SPName,
                    $row->CRName,
                    $row->SPAAN,
                    $row->SPEmail,
                    self::SERVICE_PROVIDER,
                    0, // Discount tier
                    0, // Distributor discount
                    $row->SPCurrency,
                    $row->Overall,
                    $row->Overall
                ));
            }
        }
    }

    private function dumpDataOnServiceProviderBreakdownSheet()
    {
        Log::getInstance()->notice('Execute query for "%s" sheet.', self::SHEET_NAME_SERVICE_PROVIDER_SKU_BREAKDOWN);
        switch ($this->billMode) {
            case self::BILL_MODE_ALL:
                $filename = 'report_sp_breakdown_all.sql';
                break;
            case self::BILL_MODE_LAST:
            default:
                $filename = 'report_sp_breakdown.sql';
        }
        $stmt = $this->executeSqlQuery($this->readSQLFile($filename));

        Log::getInstance()->notice('Dump data on "%s" sheet.', self::SHEET_NAME_SERVICE_PROVIDER_SKU_BREAKDOWN);
        while ($row = $stmt->fetch(PDO::FETCH_OBJ)) {
            $datacenterName = $this->getDatacenterName($row->SPDatacenter);
            $minimalCommitment = $this->getMinimalCommitmentValue($row->MinComTier, $row->MinCom, $row->SPCurrency);

            $values = array(
                $row->SPName,
                $row->SPAAN,
                $row->CRName,
                $row->CRAAN,
                $row->SPID,
                $row->SubscriptionID,
                $row->PlanName,
                $row->SPEmail,
                $row->LType,
                $row->SPCountry,
                $row->SPRegion,
                $datacenterName,
                $row->ResourceName,
                $row->ResourceSKU,
                $row->SPCurrency,
                $row->StatusName,
                $minimalCommitment,
                $row->MinComMonth,
                $row->Start,
                $row->End,
                $row->ResourceQuantity,
                $row->WSDiscount,
                $row->Discount,
                $row->ResourcePrice,
                $row->ResourcePriceWithDiscount,
                $row->Amount,
                $row->Close,
            );
            if ($this->billMode == self::BILL_MODE_ALL) {
                $values[] = $row->Duration;
                $values[] = $row->DetSDate;
                $values[] = $row->DetEDate;
            }
            $reportGroup = $row->CRAAN !== self::AAN_ACRONIS ?  $row->ReportGroup : null;
            $this->dumpData(self::SHEET_SERVICE_PROVIDER_BREAKDOWN, $values, $reportGroup);
        }
    }

    /**
     * @param int $sheetIndex
     * @param array $values
     * @param string|null $reportGroup
     */
    private function dumpData($sheetIndex, array $values, $reportGroup = null)
    {
        $this->dumpDataToBook(self::COMMON_BOOK, $sheetIndex, $values);
        if (!is_null($reportGroup)) {
            $this->dumpDataToBook($reportGroup, $sheetIndex, $values);
        }
    }

    /**
     * @param $bookId
     * @param $sheetIndex
     * @param array $values
     */
    private function dumpDataToBook($bookId, $sheetIndex, array $values)
    {
        $book = $this->getBook($bookId);
        $sheet = $book->getSheet($sheetIndex);
        $row = isset($this->bookRowPositions[$bookId][$sheetIndex]) ? $this->bookRowPositions[$bookId][$sheetIndex] : 2;

        $table = $this->sheetColumns[$sheetIndex];
        $format = isset($table['format']) ? $table['format'] : array();

        $this->fillRow($sheet, $row, $values, $format);

        $columnsCount = $table['count'];
        $sheet->getStyleByColumnAndRow(0, $row, $columnsCount - 1, $row)
            ->applyFromArray(self::$bodyStyle);

        $this->bookRowPositions[$bookId][$sheetIndex] = ++$row;
    }

    /**
     * @param $bookId
     * @return PHPExcel
     */
    private function getBook($bookId)
    {
        if (!isset($this->books[$bookId])) {
            $this->books[$bookId] = $this->createBook();
        }
        $book = $this->books[$bookId];
        return $book;
    }

    /**
     * @param PHPExcel_Worksheet $sheet
     * @param integer $row
     * @param array $values
     * @param string $format
     */
    private function fillRow(PHPExcel_Worksheet $sheet, $row, array $values, $format = null)
    {
        $i = 0;
        foreach ($values as $value) {
            $sheet->setCellValueByColumnAndRow($i, $row, $value);
            if (isset($format[$i])) {
                $cellFormat = $format[$i];
                if (is_callable($cellFormat)) {
                    $cellFormat = $cellFormat($values);
                }
                $style = $sheet->getStyleByColumnAndRow($i, $row);
                $style->getNumberFormat()->setFormatCode($cellFormat);
            }
            $i++;
        }
    }

    private function getTiersResourcesNames()
    {
        $resources = array();
        foreach ($this->tiers as $tier) {
            $resources[$tier->resourceTypeId] = $tier->name;
        }
        return $resources;
    }

    private function getPublicApiConnectionSettingsFromConfig()
    {
        return $this->config['poa_api'];
    }

    private function getPbaDbConnectionSettingsFromConfig()
    {
        return $this->config['pba_db'];
    }

    private function getBillableResourcesFromConfig()
    {
        return $this->config['resources'];
    }

    private function getReportSettingFromConfig()
    {
        return $this->config['reports'];
    }

    private function getEmailSettings()
    {
        return $this->config['email'];
    }

    private function openDbConnection()
    {
        $config = $this->getPbaDbConnectionSettingsFromConfig();
        $dsn = 'pgsql:';
        $dsn .= 'host=' . (isset($config['host']) ? $config['host'] : 'localhost');
        $dsn .= ';port=' . (!empty($config['port']) ? $config['port'] : 5432);
        $dsn .= ';dbname=' . (!empty($config['dbname']) ? $config['dbname'] : '');
        $dsn .= ';user=' . (!empty($config['user']) ? $config['user'] : '');
        $dsn .= ';password=' . (!empty($config['password']) ? $config['password'] : '');

        Log::getInstance()->notice("Open connection to database %s.", $dsn);

        $this->pdo = new PDO($dsn);
        $this->pdo->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
        $this->pdo->exec("SET timezone TO 'Europe/Tallinn'");
    }
    
    private function executeSqlQuery($sql) 
    {
        Log::getInstance()->event('Execute SQL: %s', $sql);
        return $this->pdo->query($sql);
    }

    private function prepareSqlQuery($sql)
    {
        Log::getInstance()->event('Prepare SQL: %s', $sql);
        return $this->pdo->prepare($sql);
    }

    private function readSQLFile($filename)
    {
        $sql = @file_get_contents(REPORT_DIR_SQL . $filename);
        if (empty($sql)) {
            throw new Exception(Str::format('Cannot read a file "%s".', $filename));
        }
        return $sql;
    }

    private function createResourcesTable()
    {
        Log::getInstance()->notice("Create a resource table.");
        $this->executeSqlQuery($this->readSQLFile('table_resources.sql'));

        $resources = $this->getTiersResourcesNames() + $this->getBillableResourcesFromConfig();
        Log::getInstance()->notice("Populate the resource table with %s.", json_encode($resources));
        $insetStmt = $this->prepareSqlQuery($this->readSQLFile('table_resources_insert.sql'));
        foreach ($resources as $resourceId => $resourceName) {
            $insetStmt->execute(array(':resourceId' => $resourceId, ':resourceName' => $resourceName));
        }
    }

    private function createBreakdownView()
    {
        Log::getInstance()->notice("Create a breakdown view.");
        switch ($this->billMode) {
            case self::BILL_MODE_ALL:
                $filename = 'view_breakdown_all.sql';
                break;
            case self::BILL_MODE_LAST:
            default:
                $filename = 'view_breakdown.sql';
        }
        $this->executeSqlQuery($this->readSQLFile($filename));
    }

    /**
     * @return string[]
     * @throws Exception
     */
    private function saveBooks()
    {
        $settings = $this->getReportSettingFromConfig();
        $format = $settings['format'];
        $dir = rtrim($settings['dir_name'], DS) . DS;
        $filenameTemplate = $settings['file_name'];

        if (!file_exists($dir) || !is_writable($dir)) {
            throw new Exception(Str::format('A directory "%s" does not exist.', $dir));
        }

        $dir .= 'report_' . date('YmdHis') . DS;
        mkdir($dir);

        $files = array();
        foreach ($this->books as $bookId => $book) {
            $filename = str_replace(
                array('{date}', '{group}'),
                array(date('Ymd'), self::clearFileName($bookId)),
                $filenameTemplate
            );
            $path = $dir . $filename;
            if ($bookId !== self::COMMON_BOOK) {
                $book->removeSheetByIndex(self::SHEET_TO_BILL);
            }
            $this->saveBook($book, $path, $format);
            $files[] = $path;
        }

        return $files;
    }

    /**
     * @param PHPExcel $book
     * @param string $path
     * @param string $format
     */
    private function saveBook($book, $path, $format)
    {
        Log::getInstance()->notice("Write the report into file '%s'.", $path) ;
        $book->setActiveSheetIndex(0);
        $bookWriter = PHPExcel_IOFactory::createWriter($book, $format);
        $bookWriter->save($path);
    }

    /**
     * @return array
     */
    private function getMinimalCommitmentTiers()
    {
        return $this->fetchApsResources(APSResource::CLOUD_BACKUP_MINIMAL_COMMITMENT_TIER);
    }

    /**
     * @return array
     */
    private function getDatacenters()
    {
        return $this->fetchApsResources(APSResource::CLOUD_BACKUP_DATACENTER_TYPE);
    }

    private function fetchApsResources($apsType)
    {
        $url =  $this->apsToken[self::PROPERTY_CONTROLLER_URI] . self::APS_RESOURCES_PATH;
        $url .= '?implementing(' . $apsType . ')';
        $headers = array(self::APS_TOKEN_HEADER . $this->apsToken[self::PROPERTY_APS_TOKEN]);
        $resources = json_decode($this->doHttpGetRequest($url, $headers));
        return $resources;
    }

    private function getMinimalCommitmentsDictionary()
    {
        $dictionary = array();
        foreach ($this->tiers as $tier) {
            if ($tier->level < 0) {
                continue;
            }
            if (!isset($tier->minimalCommitments) || !is_array($tier->minimalCommitments)) {
                continue;
            }
            $minimalCommitments = array();
            foreach ($tier->minimalCommitments as $minimalCommitment) {
                $minimalCommitments[$minimalCommitment->currency] = $minimalCommitment->value;
            }
            $dictionary[$tier->resourceTypeId] = $minimalCommitments;
        }
        return $dictionary;
    }

    private function getDatacentersDictionary()
    {
        $dictionary = array();
        foreach ($this->datacenters as $datacenter) {
            $dictionary[$datacenter->id] = $datacenter->name;
        }
        return $dictionary;
    }

    private function doHttpGetRequest($url, array $headers = array())
    {
        $curl = new Curl();
        $curl->setOption(CURLOPT_URL, $url);
        $curl->setOption(CURLOPT_HTTPHEADER, $headers);
        $curl->setOption(CURLOPT_RETURNTRANSFER, true);
        $curl->setOption(CURLOPT_FOLLOWLOCATION, true);
        $curl->setOption(CURLOPT_SSL_VERIFYPEER, false);
        $curl->setOption(CURLOPT_SSL_VERIFYHOST, false);
        $return = $curl->execute();
        $errorNo = $curl->getErrorNo();
        if ($errorNo) {
            $curl->refresh();
            throw new \Exception($curl->getError(), $errorNo);
        }
        $code = $curl->getInfo(CURLINFO_HTTP_CODE);
        if ($code < 200 || $code >= 300) {
            $curl->refresh();
            throw new \Exception($return, $code);
        }
        $curl->refresh();
        return $return;
    }

    private function sendReport(array $files)
    {
        $settings = $this->getEmailSettings();
        if (!isset($settings['enabled']) || $settings['enabled'] == 0) {
            return;
        }

        $startDate = $this->startPeriodDate->format('Y-m-d');
        $endDate = $this->endPeriodDate->format('Y-m-d');

        Log::getInstance()->notice("Send the report.");
        $message = \Swift_Message::newInstance();

        if (!isset($settings['email_to']) || empty($settings['email_to'])) {
            throw new \Exception('Email field "to" must be present in the settings.');
        }

        $to = explode(';', $settings['email_to']);
        $message->setTo($to);

        if (isset($settings['email_cc']) && !empty($settings['email_cc'])) {
            $cc = explode(';', $settings['email_cc']);
            $message->setCc($cc);
        }
        if (isset($settings['email_from']) && !empty($settings['email_from'])) {
            $from = $settings['email_from'];
            $message->setFrom($from);
        }
        if (isset($settings['email_subject']) && !empty($settings['email_subject'])) {
            $subject = $settings['email_subject'];
            $subject = str_replace(
                array('{start_period}', '{end_period}'),
                array($startDate, $endDate),
                $subject
            );
            $message->setSubject($subject);
        }
        if (isset($settings['email_body']) && !empty($settings['email_body'])) {
            $body = $settings['email_body'];
            $body = str_replace(
                array('{start_period}', '{end_period}'),
                array($startDate, $endDate),
                $body
            );
            $message->setBody($body);
        }

        foreach ($files as $path) {
            $message->attach(\Swift_Attachment::fromPath($path));
        }

        $transport = \Swift_SendmailTransport::newInstance();
        $mailer = \Swift_Mailer::newInstance($transport);
        $mailer->send($message);
    }

    /**
     * @param int $level
     * @return array
     */
    private function getTiersResourcesWithLevel($level)
    {
        $tiers = $this->filterTiersByLevel($level);
        return $this->getTiersResources($tiers);
    }

    /**
     * @param int $level
     * @return array
     */
    private function filterTiersByLevel($level)
    {
        return array_filter($this->tiers, function ($tier) use ($level) {
            return $tier->level == $level;
        });
    }

    /**
     * @param array $tiers
     * @return array
     */
    private function getTiersResources($tiers)
    {
        return array_map(function ($tier) {
            return $tier->resourceTypeId;
        }, $tiers);
    }

    /**
     * @param $datacenterId
     * @return string
     */
    private function getDatacenterName($datacenterId)
    {
        $datacenters = $this->getDatacentersDictionary();
        $datacenter = isset($datacenters[$datacenterId]) ? $datacenters[$datacenterId] : $datacenterId;
        return $datacenter;
    }

    /**
     * @param $tierId
     * @param $minimalCommitment
     * @param $currencyId
     * @return int|null
     */
    private function getMinimalCommitmentValue($tierId, $minimalCommitment, $currencyId)
    {
        $minimalCommitmentsDictionary = $this->getMinimalCommitmentsDictionary();
        $noCommitmentTiersResources = $this->getTiersResourcesWithLevel(self::LEVEL_NO_COMMITMENT);
        $fixedCommitmentTiersResources = $this->getTiersResourcesWithLevel(self::LEVEL_FIXED_COMMITMENT);

        if (in_array($tierId, $noCommitmentTiersResources)) {
            return 0;
        }
        if (in_array($tierId, $fixedCommitmentTiersResources)) {
            return $minimalCommitment;
        }
        if (isset($minimalCommitmentsDictionary[$tierId][$currencyId])) {
            return $minimalCommitmentsDictionary[$tierId][$currencyId];
        }
        return null;
    }

    private function fetchUsageForSubscriptions($subscriptions, $billableResources)
    {
        $sql = $this->getSqlToFetchUsage();
        $sql = str_replace('$$SUBSCRIPTIONS$$', implode(',', $subscriptions), $sql);
        $usageStmt = $this->executeSqlQuery($sql);
        $resources = $usageStmt->fetchAll();
        $usage = array();
        foreach ($billableResources as $id => $name) {
            $usage[$id] = 0;
        }

        foreach ($resources as $resource) {
            $id = $resource['id'];
            if (!isset($usage[$id])) {
                continue;
            }
            $usage[$id] = $resource['usage'];
        }
        return $usage;
    }

    /**
     * @return bool|string
     */
    private function getSqlToFetchUsage()
    {
        if (!self::$sqlToFetchUsage) {
            switch ($this->billMode) {
                case self::BILL_MODE_ALL:
                    $filename = 'select_resources_usage.sql';
                    break;
                case self::BILL_MODE_LAST:
                default:
                    $filename = 'select_resources_usage.sql';
            }
            self::$sqlToFetchUsage = $this->readSQLFile($filename);
        }
        return self::$sqlToFetchUsage;
    }

    /**
     * @param $accountId
     * @return float
     */
    private function countDistributorDiscount($accountId)
    {
        $distributorDiscountStmt = $this->getDistributorDiscountStmt();
        $distributorDiscountStmt->execute(array(':CR_VENDOR_ID' => $accountId));
        $result = $distributorDiscountStmt->fetch(PDO::FETCH_OBJ);
        $distributorDiscount = $result ? $result->Discount : 0.0;
        return $distributorDiscount;
    }

    /**
     * @return PDOStatement
     */
    private function getDistributorDiscountStmt()
    {
        if (!$this->distributorDiscountStmt) {
            $this->distributorDiscountStmt = $this->prepareSqlQuery($this->readSQLFile('select_distributor_discount.sql'));
        }
        return $this->distributorDiscountStmt;
    }

    public static function getCurrencyPrecision($currency)
    {
        if (in_array($currency, array('AOA', 'CNY', 'HKD', 'MOP', 'TWD'))) {
            return 1;
        }
        if (in_array($currency, array('BHD', 'JOD', 'KWD', 'LYD', 'OMR', 'TND'))) {
            return 3;
        }
        if (in_array($currency, array(
            'AMD', 'BIF', 'BYR', 'CLP', 'COP', 'DJF', 'GNF', 'HUF', 'IDR', 'IQD', 'IRR',
            'ISK', 'JPY', 'KHR', 'KMF', 'KPW', 'KRW', 'LAK', 'LBP', 'MMK', 'PYG', 'RWF',
            'SLL', 'STD', 'UGX', 'VND', 'VUV', 'XAF', 'XOF', 'XPF', 'YER', 'ZMK'
        ))) {
            return 0;
        }
        return 2;
    }

    /**
     * @param string $currency
     * @param bool $showName
     * @return string
     */
    public static function buildCurrencyFormat($currency, $showName = true)
    {
        $precision = BillingReportHelper::getCurrencyPrecision($currency);
        $format = $showName ? "[$$currency] " . self::FORMAT_NUMBER : self::FORMAT_NUMBER;
        if ($precision) {
            $format .= '.' . str_repeat('0', $precision);
        }
        return $format;
    }

    private static function clearFileName($filename)
    {
        return preg_replace('/\x00-\x1f\x7f"\*\/:<>\?\\\\|\+,\.;=\[\]/', '', $filename);
    }
}