<?php
/**
 * @Copyright © 2002-2015 Acronis International GmbH. All rights reserved
 */

class BackupAgentSoapClient extends SoapClient
{
    /**
     * Override the constructor to inject the correct username and password
     *
     * @access public
     * @params string $wsdl
     * @params array $options
     * @params array $credentials   The credentials used to identify ourselves
     */
    public function __construct($wsdl, $options, $credentials = array()) {
        $this->username = $credentials['username'];
        $this->password = $credentials['password'];

        return parent::__construct($wsdl, $options);
    }

    /**
     * Override the default __doRequest function to use WSSE and WSA
     *
     * @access public
     * @param String $request The SOAP request
     * @param String $location The URL to request
     * @param String $action The SOAP action
     * @param String $version The SOAP version
     * @param null $one_way
     * @return string
     */
    public function __doRequest($request, $location, $action, $version, $one_way = NULL) {
        $dom = new DOMDocument();
        $dom->loadXML($request);

        $objWSA = new WSASoap($dom);
        $objWSA->addAction($action);
        $objWSA->addTo($location);
        $objWSA->addMessageID();
        $objWSA->addReplyTo();
        $dom = $objWSA->getDoc();
        $objWSSE = new WSSESoap($dom);

        /* Sign all headers to include signing the WS-Addressing headers */
        $objWSSE->signAllHeaders = TRUE;
        $objWSSE->addTimestamp();
        $objWSSE->addUserToken($this->username, htmlspecialchars($this->password));

        /* create new XMLSec Key using RSA SHA-1 and type is private key */
        $objKey = new XMLSecurityKey(XMLSecurityKey::RSA_SHA1, array('type'=>'private'));

        /* load the private key from file - last arg is bool if key in file (TRUE) or is string (FALSE) */
        $objKey->loadKey(__DIR__.'/private.key', TRUE);

        /* Sign the message - also signs appropraite WS-Security items */
        $objWSSE->signSoapDoc($objKey);

        /* Add certificate (BinarySecurityToken) to the message and attach pointer to Signature */
        $token = $objWSSE->addBinaryToken(file_get_contents(__DIR__.'/public.crt'));
        $objWSSE->attachTokentoSig($token);
        $request = $objWSSE->saveXML();
        $return = parent::__doRequest($request, $location, $action, $version);
        return $return;
    }

    /**
     * Magic call method
     *
     * @params string $function_name
     * @params array $arguments
     * @access public
     */
    public function __call($function_name, $arguments) {
        $return = $this->decode(parent::__call($function_name, $arguments))->to_array();
        return $return;
    }

    /**
     * Convert SoapVars to their specified objects
     *
     * @param mixed Object to unmangle
     * @return mixed Unmangled object
     * @access public
     */
    private function decode($obj) {
        if (is_object($obj) && get_class($obj) == 'SoapVar') {
            $obj = self::decode($obj->enc_value);
        } elseif (is_object($obj)) {
            foreach ($obj as $key => $value) {
                if (is_array($value) || is_object($value)) {
                    $obj->$key = self::decode($value);
                }
            }
        } elseif (is_array($obj)) {
            foreach ($obj as $key => $value) {
                if (is_array($value) || is_object($value)) {
                    $obj[$key] = self::decode($value);
                }

            }
        }

        return $obj;
    }

    /**
     * Get an instance of the BackupAgentSoapClient
     *
     * @access public
     * @param $ba_wsdl
     * @param array $credentials
     * @return BackupAgentSoapClient
     * @throws Exception
     */
    public static function Get($ba_wsdl, $credentials = array()) {
        $class_map = require 'class_map.php';
        try {
            return new BackupAgentSoapClient($ba_wsdl, array('classmap' => $class_map, 'trace' => true), $credentials);
        } catch (SoapFault $sf) {
            throw new Exception('BackupAgent service is down');
        }
    }
}
